
(******* Quelques definitions du prelude CAML **************)

exception failure of string;

fun failwith s = raise(failure s)
;

(* 1- les paires *)

fun fst (x,y) = x
and snd (x,y) = y
;

fun do_list f =
  let fun do_rec [] = ()
        | do_rec (a::L) = (f a; do_rec L)
  in do_rec
  end
;

fun map f =
  let fun map_rec [] = []
        | map_rec (a::L) = f a :: map_rec L
  in map_rec
  end
;

fun it_list f =
  let fun it_rec (a, []) = a
        | it_rec (a, b::L) = it_rec (f (a, b), L)
  in it_rec
  end
;

fun it_list2 f =
  let fun it_rec  (a, [], [])    = a
        | it_rec  (a, a1::L1, a2::L2) = it_rec (f (a,a1,a2), L1, L2)
        | it_rec  _ = failwith "it_list2"
  in it_rec
  end
;

fun is_there p = (* exists replaced by is_there (JG) *)
  let fun is_there_rec []  = false
        | is_there_rec (a::L) = (p a) orelse (is_there_rec L)
  in is_there_rec
  end
;

fun for_all p =
  let fun for_all_rec []  = true
        | for_all_rec (a::L) = (p a) andalso (for_all_rec L)
  in for_all_rec
  end
;

fun rev_append   ([], L)  = L 

  | rev_append (x::L1, L2) = rev_append (L1, x::L2)
;

fun try_find f =
  let fun try_find_rec  []    = failwith "try_find"
        | try_find_rec (a::L) = (f a) handle failure _ => try_find_rec L
  in try_find_rec
  end
;

fun partition p =
  let fun part_rec   []   = ([],[])
        | part_rec (a::L) =
            let val (pos,neg) = part_rec L in
              if p a then  ((a::pos), neg) else (pos, (a::neg))
            end
  in part_rec
  end
;

(* 3- Les ensembles et les listes d'association *)

fun mem a =
  let fun mem_rec [] = false
        | mem_rec (b::L) = (a=b) orelse mem_rec L
  in mem_rec
  end
;

fun set_union L1 L2 =
  let fun set_union_rec [] = L2
        | set_union_rec (a::L) =
            if mem a L2 then set_union_rec L else a :: set_union_rec L
  in set_union_rec L1
  end
;

fun mem_assoc a =
  let fun mem_rec [] = false
        | mem_rec ((b,_)::L) = (a=b) orelse mem_rec L
  in mem_rec
  end
;

fun assoc a =
  let fun assoc_rec [] = failwith "find"
        | assoc_rec ((b,d)::L) = if a=b then d else assoc_rec L
  in assoc_rec
  end
;

(* 4- Les sorties *)

fun print_string s = #put stdout s;
fun print_num (n:num) = print (stdout,pack n);
fun print_newline () = #put stdout "\n";
fun message s = (#put stdout s;#put stdout "\n");

(* formerly (JG) :
val print_string = String.print;
val print_num = Integer.print;
fun print_newline () = String.print "\n";
fun message s = (String.print s; String.print "\n");
*)

(* 5- Les ensembles *)

fun set_union L1 =
  let fun set_union_rec [] = L1
        | set_union_rec (a::L) = if mem a L1 then set_union_rec L else a :: set_union_rec L
  in set_union_rec
  end
;
(****************** Term manipulations *****************)

datatype term =
    Var of num
  | Term of string * term list;;

fun vars (Var n) = [n]
  | vars (Term(_,L)) = vars_of_list L
and vars_of_list [] = []
  | vars_of_list (t::r) = set_union (vars t) (vars_of_list r)
;

datatype binding = Bind of num * term
;

fun binding (n, []) = failwith "unbound"
  | binding (n, Bind(m,t)::rest) = if n=m then t else binding(n, rest)
;

fun substitute subst =
  let fun subst_rec (Term(oper,sons)) = Term(oper, map subst_rec sons)
        | subst_rec (t as (Var n))    = (binding (n, subst)) handle failure _ => t
  in subst_rec
  end
;

fun change f =
  let fun change_rec (h::t, n) = if n=1 then f h :: t
                                       else h :: change_rec (t, n-1)
        | change_rec _         = failwith "change"
  in change_rec
  end
;

(* Term replacement replace M u N => M[u<-N] *)
fun replace M u N = 

  let fun reprec (_, []) = N
        | reprec (Term(oper,sons), (n::u)) =
             Term(oper, change (fn P => reprec(P,u)) (sons, n))
        | reprec _ = failwith "replace"
  in reprec(M,u)
  end
;

(* matching = - : (term -> term -> subst) *)
fun matching term1 term2 =
  let fun match_rec (subst, Var v, M) =
             ((if M = binding (v, subst) then subst else failwith "matching")
              handle failure "unbound" =>
              Bind(v,M) :: subst)
        | match_rec (subst, Term(op1,sons1), Term(op2,sons2)) =
	  if op1 = op2 then it_list2 match_rec (subst, sons1, sons2)
                       else failwith "matching"
        | match_rec _ = failwith "matching"
  in match_rec ([], term1,term2)
  end
;

(* A naive unification algorithm *)

fun compsubst (subst1, subst2) = 

  (map (fn (Bind(v,t)) => Bind(v, substitute subst1 t)) subst2) @ subst1
;

fun occurs n =
  let fun occur_rec (Var m) = (m=n)
        | occur_rec (Term(_,sons)) = is_there occur_rec sons
  in occur_rec
  end
;

fun unify ((term1 as (Var n1)), term2) =
      if term1 = term2 then []
      else if occurs n1 term2 then failwith "unify"
      else [Bind(n1,term2)]
  | unify (term1, Var n2) =
      if occurs n2 term1 then failwith "unify"
      else [Bind(n2,term1)]
  | unify (Term(op1,sons1), Term(op2,sons2)) =
      if op1 = op2 then 

	it_list2 (fn (s,t1,t2) => compsubst (unify(substitute s t1,
                                                   substitute s t2), s))
                 ([], sons1, sons2)
      else failwith "unify"
;

(* We need to print terms with variables independently from input terms
  obtained by parsing. We give arbitrary names v1,v2,... to their variables. *)

val INFIXES = ["+","*"];

fun pretty_term (Var n) =
      (print_string "v"; print_num n)
  | pretty_term (Term (oper,sons)) =
      if mem oper INFIXES then
        case sons of
            [s1,s2] =>
              (pretty_close s1; print_string oper; pretty_close s2)
          | _ =>
              failwith "pretty_term : infix arity <> 2"
      else
       (print_string oper;
        case sons of
	     []   => ()
          | t::lt =>(print_string "(";
                     pretty_term t;
                     do_list (fn t => (print_string ","; pretty_term t)) lt;
                     print_string ")"))
and pretty_close (M as Term(oper, _)) =
      if mem oper INFIXES then
        (print_string "("; pretty_term M; print_string ")")
      else pretty_term M
  | pretty_close M = pretty_term M
;

(****************** Equation manipulations *************)

(* standardizes an equation so its variables are 1,2,... *)

datatype equation = Equation of term * term
;

datatype rule = Rule of num * num * term * term
;

fun mk_rule (n, M, N) =
  let val all_vars = set_union (vars M) (vars N)
      val (k,subst) =
        it_list (fn ((i,sigma),v) => (i+1, (Bind(v,Var(i)))::sigma))
                ((1,[]), all_vars)
  in Rule(n, (k-1), substitute subst M, substitute subst N)
  end
;

(* checks that rules are numbered in sequence and returns their number *)
fun check_rules rules =
  it_list (fn (n, Rule(k,_,_,_)) =>
             if k=n+1 then k
                      else failwith "Rule numbers not in sequence")
          (0, rules)
;

fun pretty_rule (Rule(k,n,M,N)) =
 (print_num k; print_string " : ";
  pretty_term M; print_string " = "; pretty_term N;
  print_newline())
;

val pretty_rules = do_list pretty_rule
;

(****************** Rewriting **************************)

(* Top-level rewriting. Let eq:L=R be an equation, M be a term such that L<=M.
   With sigma = matching L M, we define the image of M by eq as sigma(R) *)
fun reduce L M =
  substitute (matching L M)
;

(* A more efficient version of can (rewrite1 (L,R)) for R arbitrary *)
fun reducible L =
  let fun redrec M =
    (matching L M; true)
    handle failure _ =>
      case M of Term(_,sons) => is_there redrec sons
              |        _     => false
  in redrec
  end
;

(* mreduce : rules -> term -> term *)
fun mreduce rules M =
  let fun redex (Rule(_,_,L,R)) = reduce L M R in try_find redex rules end
;

(* One step of rewriting in leftmost-outermost strategy, with multiple rules *)
(* fails if no redex is found *)
(* mrewrite1 : rules -> term -> term *)
fun mrewrite1 rules =
  let fun rewrec M =
    (mreduce rules M) handle failure _ =>
      let fun tryrec [] = failwith "mrewrite1"
            | tryrec (son::rest) =
                (rewrec son :: rest) handle failure _ => son :: tryrec rest
      in case M of
          Term(f, sons) => Term(f, tryrec sons)
        | _ => failwith "mrewrite1"
      end
  in rewrec
  end
;

(* Iterating rewrite1. Returns a normal form. May loop forever *)
(* mrewrite_all : rules -> term -> term *)
fun mrewrite_all rules M =
  let fun rew_loop M =
    rew_loop(mrewrite1 rules M)  handle  failure _ => M
  in rew_loop M
  end
;

(*
pretty_term (mrewrite_all Group_rules M where M,_=<<A*(I(B)*B)>>);;
==> A*U
*)


(************************ Recursive Path Ordering ****************************)

datatype ordering = Greater | Equal | NotGE;

fun ge_ord order pair = case order pair of NotGE => false | _ => true
and gt_ord order pair = case order pair of Greater => true | _ => false
and eq_ord order pair = case order pair of Equal => true | _ => false
;

fun rem_eq equiv =
  let fun remrec x [] = failwith "rem_eq"
        | remrec x (y::l) = if equiv (x,y) then l else y :: remrec x l
  in remrec
  end
;

fun diff_eq equiv (x,y) =
  let fun diffrec (p as ([],_)) = p
        | diffrec ((h::t), y) =
            diffrec (t,rem_eq equiv h y) handle failure _ =>
              let val (x',y') = diffrec (t,y) in (h::x',y') end
  in if len x > len y then diffrec(y,x) else diffrec(x,y) (* len instead of length (JG) *)
  end
;

(* multiset extension of order *)
fun mult_ext order (Term(_,sons1), Term(_,sons2)) =
      (case diff_eq (eq_ord order) (sons1,sons2) of
           ([],[]) => Equal
         | (l1,l2) =>
             if for_all (fn N => is_there (fn M => order (M,N) = Greater) l1) l2
             then Greater else NotGE)
  | mult_ext order _ = failwith "mult_ext"
;

(* lexicographic extension of order *)
fun lex_ext order ((M as Term(_,sons1)), (N as Term(_,sons2))) =
      let fun lexrec ([] , []) = Equal
            | lexrec ([] , _ ) = NotGE
            | lexrec ( _ , []) = Greater
            | lexrec (x1::l1, x2::l2) =
                case order (x1,x2) of
                  Greater => if for_all (fn N' => gt_ord order (M,N')) l2 

                             then Greater else NotGE
                | Equal => lexrec (l1,l2)
                | NotGE => if is_there (fn M' => ge_ord order (M',N)) l1 

                           then Greater else NotGE
      in lexrec (sons1, sons2)
      end
  | lex_ext order _ = failwith "lex_ext"
;;

(* recursive path ordering *)

fun rpo op_order ext =
  let fun rporec (M,N) =
    if M=N then Equal else 

      case M of
          Var m => NotGE
        | Term(op1,sons1) =>
            case N of
                Var n =>
                  if occurs n M then Greater else NotGE
              | Term(op2,sons2) =>
                  case (op_order (op1, op2)) of
                      Greater =>
                        if for_all (fn N' => gt_ord rporec (M,N')) sons2
                        then Greater else NotGE
                    | Equal =>
                        ext rporec (M,N)
                    | NotGE =>
                        if is_there (fn M' => ge_ord rporec (M',N)) sons1
                        then Greater else NotGE
  in rporec
  end
;

(****************** Critical pairs *********************)

(* All (u,sig) such that N/u (&var) unifies with M,
   with principal unifier sig *)

fun super M =
  let fun suprec (N as Term(_,sons)) =
      let fun collate ((pairs,n), son) =
                (pairs @ map (fn (u,sigma) => (n::u,sigma)) (suprec son), n+1)
          val insides =
                fst (it_list collate (([],1), sons))
      in ([], unify(M,N)) :: insides  handle failure _ => insides
      end
    | suprec _ = []
  in suprec end
;

(* Ex :
let (M,_) = <<F(A,B)>> 

and (N,_) = <<H(F(A,x),F(x,y))>> in super M N;;
==> [[1],[2,Term ("B",[])];                      x <- B
     [2],[2,Term ("A",[]); 1,Term ("B",[])]]     x <- A  y <- B
*)

(* All (u,sigma), u&[], such that N/u unifies with M *)
(* super_strict : term -> term -> (num list & subst) list *)

fun super_strict M (Term(_,sons)) =
        let fun collate ((pairs,n), son) =
          (pairs @ map (fn (u,sigma) => (n::u,sigma)) (super M son), n+1)
        in fst (it_list collate (([],1), sons)) end
  | super_strict _ _ = []
;

(* Critical pairs of L1=R1 with L2=R2 *)
(* critical_pairs : term_pair -> term_pair -> term_pair list *)
fun critical_pairs (Equation(L1,R1)) (Equation(L2,R2)) =
  let fun mk_pair (u,sigma) =
     Equation(substitute sigma (replace L2 u R1), substitute sigma R2) in
  (map mk_pair (super L1 L2))
  end
;

(* Strict critical pairs of L1=R1 with L2=R2 *)
(* strict_critical_pairs : term_pair -> term_pair -> term_pair list *)
fun strict_critical_pairs (Equation(L1,R1)) (Equation(L2,R2)) =
  let fun mk_pair (u,sigma) =
    Equation(substitute sigma (replace L2 u R1), substitute sigma R2) in
  (map mk_pair (super_strict L1 L2))
  end
;

(* All critical pairs of eq1 with eq2 *)
fun mutual_critical_pairs eq1 eq2 =
  (strict_critical_pairs eq1 eq2) @ (critical_pairs eq2 eq1);

(* Renaming of variables *)

fun rename n (Equation(t1,t2)) =
  let fun ren_rec (Var k) = Var(k+n)
        | ren_rec (Term(oper,sons)) = Term(oper, map ren_rec sons)
  in Equation(ren_rec t1, ren_rec t2)
  end
;

(************************ Completion ******************************)

fun deletion_message (Rule(k,_,_,_)) =
  (print_string "Rule ";print_num k; message " deleted")
;

(* Generate failure message *)
fun non_orientable (Equation(M,N)) =
  (pretty_term M; print_string " = "; pretty_term N; print_newline())
;

fun fetch_rule k =
  let fun fetch_rec [] = failwith "rule deleted"
        | fetch_rec (Rule(n,v,L,R)::rest) =
            if k = n then (v, Equation(L,R)) else fetch_rec rest
  in
    fetch_rec
  end
;

(* Improved Knuth-Bendix completion procedure *)
(* kb_completion : (term_pair -> bool) -> num -> rules -> term_pair list -> (num & num) ->  
term_pair list -> rules *)
fun kb_completion greater =
  let fun kbrec (n, rules) =
    let val normal_form = mrewrite_all rules
        fun process (failures, k, l, eqs) =
         (case eqs of
           [] =>
              if k<l then next_criticals (failures,k+1,l) else
              if l<n then next_criticals (failures,1,l+1) else
               (case failures of
                  [] => rules (* successful completion *)
                | _  => (message "Non-orientable equations :";
                         do_list non_orientable failures;
                         failwith "kb_completion"))
        | Equation(M,N)::eqs =>
              let val M' = normal_form M
                  val N' = normal_form N
                  fun enter_rule(left,right) =
                    let val new_rule = mk_rule(n+1, left, right) in
                     (pretty_rule new_rule;
                      let fun left_reducible (Rule(_,_,L,_)) = reducible left L
                          val (redl,irredl) = partition left_reducible rules
                      in (do_list deletion_message redl;
                          let fun right_reduce (Rule(m,_,L,R)) = 

                              mk_rule(m,L,(mrewrite_all (new_rule::rules) R))
                              val irreds = map right_reduce irredl
                              val eqs' = map (fn (Rule(_,_,L,R)) => Equation(L,R)) redl
                          in kbrec (n+1, new_rule::irreds) ([], k, l, (eqs @ eqs' @ failures))
                          end)
                      end)
                    end
              in if M'=N' then process (failures, k, l, eqs) else
                 if greater(M',N') then enter_rule(M',N') else
                 if greater(N',M') then enter_rule(N',M') else
                   process ((Equation(M',N')::failures), k, l, eqs)
              end)
       and next_criticals (failures, k, l) =
            (let val (v,el) = fetch_rule l rules in
               if k=l then
                 process (failures, k, l, strict_critical_pairs el (rename v el))
               else
                (let val (_,ek) = fetch_rule k rules in
                    process (failures, k, l, mutual_critical_pairs el (rename v ek))
                 end
                 handle failure "rule deleted" (*rule k deleted*) =>
                   next_criticals (failures,k+1,l))
             end
             handle failure "rule deleted" (*rule l deleted*) =>
                next_criticals (failures, 1,l+1))
      in process
      end
  in kbrec
  end
;

fun kb_complete greater complete_rules rules =
    let val n = check_rules complete_rules
        val eqs = map (fn (Rule(_,_,L,R)) => Equation(L,R)) rules
        val completed_rules =
               kb_completion greater (n, complete_rules) ([],n,n,eqs)
    in (message "Canonical set found :";
        pretty_rules (rev completed_rules);
        ())
    end
;

val Group_rules = [
  Rule(1, 1, Term("*", [Term("U",[]), Var 1]), Var 1),
  Rule(2, 1, Term("*", [Term("I",[Var 1]), Var 1]), Term("U",[])),
  Rule(3, 3, Term("*", [Term("*", [Var 1, Var 2]), Var 3]),
             Term("*", [Var 1, Term("*", [Var 2, Var 3])]))];

val Geom_rules = [
 Rule(1,1,Term ("*",[(Term ("U",[])), (Var 1)]),(Var 1)),
 Rule(2,1,Term ("*",[(Term ("I",[(Var 1)])), (Var 1)]),(Term ("U",[]))),
 Rule(3,3,Term ("*",[(Term ("*",[(Var 1), (Var 2)])), (Var 3)]),
  (Term ("*",[(Var 1), (Term ("*",[(Var 2), (Var 3)]))]))),
 Rule(4,0,Term ("*",[(Term ("A",[])), (Term ("B",[]))]),
  (Term ("*",[(Term ("B",[])), (Term ("A",[]))]))),
 Rule(5,0,Term ("*",[(Term ("C",[])), (Term ("C",[]))]),(Term ("U",[]))),
 Rule(6,0,
   Term
   ("*",
    [(Term ("C",[])),
     (Term ("*",[(Term ("A",[])), (Term ("I",[(Term ("C",[]))]))]))]),
  (Term ("I",[(Term ("A",[]))]))),
 Rule(7,0,
  Term
   ("*",
    [(Term ("C",[])),
     (Term ("*",[(Term ("B",[])), (Term ("I",[(Term ("C",[]))]))]))]),
  (Term ("B",[])))
];

fun Group_rank "U" = 0
  | Group_rank "*" = 1
  | Group_rank "I" = 2
  | Group_rank "B" = 3
  | Group_rank "C" = 4
  | Group_rank "A" = 5
;

fun Group_precedence (op1, op2) =
  let val r1 = Group_rank op1
      val r2 = Group_rank op2
  in
    if r1 = r2 then Equal else
    if r1 > r2 then Greater else NotGE
  end
;

val Group_order = rpo Group_precedence lex_ext 

;

fun greater pair =
  case Group_order pair of Greater => true | _ => false
;

fun go rules =
  (kb_complete greater [] rules; ())
;
fun main _ = go Group_rules;
