/* Compiler to C, C code generation (second version, mostly obsolete).
   Copyright (C) 1992-1997 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#if TRANSLATOR
#include "expr.h"
#include "refcount.h"
#include "mem.h"
#include "translate.h"
#include "optimize.h"
#include "codegen.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#include "evlproto.h"
#include "errproto.h"
#include "typproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rcgen.h"
#ifdef HAS_PROTOTYPES
#include "scgen.h"
#endif

#define __SEG__ CODEGEN

EXTERN int errno;

PRIVATE void varNamesSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  varNames *vn;

  vn = (varNames *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE: /* save */
      {
	COUNTER i;
	CHAR c;

	for (i=0; i<V_COUNT; i++)
	  {
	    outInt(vn->maxn[i]);
	    if (errno) return;
	  }
	c = vn->has_calls;
	outChars(&c,1);
	if (errno) return;
	c = vn->has_jumps;
	outChars(&c,1);
	if (errno) return;
      }
      /*fallthrough*/
    case X_MARK: /* mark */
    case X_MARKDATA:
      (*doext->action)(vn->names);
      (*doext->action)(vn->zapx);
      (*doext->action)(vn->zapm);
      break;
    case X_FREE:
      MARKFREE(extdata); /* free the C structure */
      break;
    }
}

PRIVATE ExprPtr varNamesLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  varNames *vn;
  ExprPtr e;
  COUNTER i;
  CHAR c;
  ENTER;

  vn = (varNames *)MALLOC(sizeof(varNames));
  for (i=0; i<V_COUNT; i++)
    {
      vn->maxn[i] = inInt();
      if (errno)
	{
err:
	  MARKFREE(vn);
	  return NIL;
	}
    }
  inChars(&c,1);
  if (errno) goto err;
  vn->has_calls = c;
  inChars(&c,1);
  if (errno) goto err;
  vn->has_jumps = c;
  vn->names = (*load)(sharedref);
  if (errno) goto err;
  vn->zapx = (*load)(sharedref);
  if (errno) goto err;
  vn->zapm = (*load)(sharedref);
  if (errno) goto err;
  e = (shared?Kexternal:MakeExternal)(desc,(char *)vn);
  KEEPONLY(e);
  return e;
}

PUBLIC extDesc vnInfoDesc = {
  (CHAR *)"variable-name-collection",
  'v','n','#',' ', NULL,
  NULL, /* not shared */
  NULL, /* not shared */
  NULL, /* print as default */
  varNamesSweep,
  varNamesLoad
  };

PRIVATE ExprPtr varNamesNew()
{
  varNames *vn;
  ExprPtr e;
  COUNTER i;

  vn = (varNames *)MALLOC(sizeof(varNames));
  vn->names = NewMap();
  for (i=0; i<V_COUNT; i++)
    {
      vn->maxn[i] = 0;
    }
  vn->has_calls = false;
  vn->has_jumps = false;
  vn->zapx = NewMap();
  vn->zapm = NewMap();
  e = MakeExternal(&vnInfoDesc,(char *)vn);
  return KEEP(e);
}

PRIVATE void varFreeListSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  varFreeList *vf;
  COUNTER i;

  vf = (varFreeList *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      {
	COUNTER i;

	for (i=0; i<V_COUNT; i++)
	  {
	    outInt(vf->n[i]);
	    if (errno) return;
	  }
      }
      /*fallthrough*/
    case X_MARK:
    case X_MARKDATA:
      (*doext->action)(vf->pendinguses);
      for (i=0; i<V_COUNT; i++)
	(*doext->action)(vf->free[i]);
      break;
    case X_FREE:
      MARKFREE(extdata); /* free the C structure */
      break;
    }
}

PRIVATE ExprPtr varFreeListLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  varFreeList *vf;
  ExprPtr e;
  COUNTER i;
  ENTER;

  vf = (varFreeList *)MALLOC(sizeof(varFreeList));
  for (i=0; i<V_COUNT; i++)
    {
      vf->n[i] = inInt();
      if (errno)
	{
err:
	  MARKFREE(vf);
	  return NIL;
	}
    }
  vf->pendinguses = (*load)(sharedref);
  if (errno) goto err;
  for (i=0; i<V_COUNT; i++)
    {
      vf->free[i] = (*load)(sharedref);
      if (errno) goto err;
    }
  e = (shared?Kexternal:MakeExternal)(desc,(char *)vf);
  KEEPONLY(e);
  return e;
}

PUBLIC extDesc vfInfoDesc = {
  (CHAR *)"variable-free-list",
  'v','f','#',' ', NULL,
  NULL, /* not shared */
  NULL, /* not shared */
  NULL, /* print as default */
  varFreeListSweep,
  varFreeListLoad
  };

PRIVATE ExprPtr varFreeListNew()
{
  varFreeList *vf;
  ExprPtr e;
  COUNTER i;

  vf = (varFreeList *)MALLOC(sizeof(varFreeList));
  vf->pendinguses = NewMap();
  for (i=0; i<V_COUNT; i++)
    {
      vf->free[i] = NewMap();
      vf->n[i] = 0;
    }
  e = MakeExternal(&vfInfoDesc,(char *)vf);
  return KEEP(e);
}

PRIVATE ExprPtr varFreeListCopy(vf)
IN CONST ExprPtr vf;
{
  ExprPtr vfn;
  varFreeList *v,*vv;
  COUNTER i;

  vfn = varFreeListNew();
  v = VF(vf);
  vv = VF(vfn);
  SETQ(vv->pendinguses,v->pendinguses);
  for (i=0; i<V_COUNT; i++)
    {
      SETQ(vv->free[i],v->free[i]);
      vv->n[i] = v->n[i];
    }
  return vfn;
}

PRIVATE char *varnames[] = {
  "VAR(%ld)",
  "_b%ld",
  "_i%ld",
  "_s%ld",
  "_x%ld",
  "_f%ld",
  "%ld",
  "_ls%ld",
  "(%ld)",
  "_sms%ld",
  "_v%ld"  /* don't change the 'v' character in front! */
};

PRIVATE void varFreeListMerge(vf,vff)
INOUT ExprPtr vf;
IN CONST ExprPtr vff;
{
  COUNTER i;
  COUNTER n,nn;
  char buf[32];
  char *fmt;
  ExprPtr var,*loc;

  setMapDomRestrTo(&VF(vf)->pendinguses,VF(vff)->pendinguses);
  HITBARRIER(vf);
  for (i=0; i<V_COUNT; i++)
    { /* take union of free lists and maximum of counters n[] */
      loc = &VF(vf)->free[i];
      setMapMerge(loc,VF(vff)->free[i]);
      HITBARRIER(vf);
      n = VF(vf)->n[i];
      nn = VF(vff)->n[i];
      VF(vf)->n[i] = (n>nn)?n:nn;
    }
}

PRIVATE COUNTER storageClass(storage,def)
IN COUNTER storage;
IN CONST ExprPtr def;
{
  if (storage & SCF_BOXED)
    return V_EXPRPTR;
  if (storage & SCF_BOOL)
    return V_BOOL;
  if (storage & SCF_INT)
    return V_INT;
  if (storage & SCF_STRING)
    return V_STRING;
  if (storage & SCF_REAL)
    return V_REAL;
  if (storage & SCF_CODE)
    return V_CODE;
  if (storage & SCF_LOCAL) /* local data: virtual couples, as returned by mapget,
			      are composite and unboxed;
			      but so are local refs, iterator variables, and so on. */
    {
      if (def->kind!=(CHAR)E_CONS)
	kinderror("storageClass",def->kind);
      def = CAR(def);
      if (def->kind!=(CHAR)E_COUNTED)
	kinderror("storageClass",def->kind);
      switch (COUNTEDN(def))
	{
	case Z_MAPGET:
	  return V_MAPGET;
	case Z_START_LISTSWEEP:
	  /*return V_LISTSWEEP; not used for GC protection reasons */
	  return V_EXPRPTR;
	case Z_START_MAPSWEEP:
	  return V_MAPSWEEP;
	case Z_START_SUBMAPSWEEP:
	  /*return V_SUBMAPSWEEP; not used for GC protection reasons */
	  return V_EXPRPTR;
	case Z_BOX_LOCAL_REF:
	  return V_EXPRPTR;
	default:
	  kinderror("storageClass",COUNTEDN(def));
	  /*NOTREACHED*/
	  break;
	}
    }
  return V_VOID; /* void */
}

PRIVATE ExprPtr makeVarName(x,deco,vn,vf)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr vn,vf;
{
  ppInfo *pp;
  char buf[32];
  ExprPtr e;
  COUNTER stor;
  varFreeList *vfn;
  INT n;

  vfn = VF(vf);
  pp = getPpInfo(x,deco);
  if (pp->uses==NewMap())
    return NIL;
  stor = storageClass(pp->storage,pp->def);
  e = ChooseInDom(vfn->free[stor]);
  if (e!=NIL)
    {
      vfn->free[stor] = MapRemove(e,vfn->free[stor]);
      return e;
    }
  n = ++vfn->n[stor];
  sprintf(buf,varnames(stor),n);
  if (VN(vn)->maxn[stor]<n)
    VN(vn)->maxn[stor] = n;
  e = Kstr((CHAR *)buf);
  return KEEP(e);
}

PRIVATE ExprPtr makeParName(x,deco,np)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
INOUT COUNTER *np;
{
  char buf[32];
  ExprPtr e;

  sprintf(buf,"p%d",++*np);
  e = Kstr((CHAR *)buf);
  return KEEP(e);
}

PRIVATE void addVarName(x,name,vn)
IN CONST ExprPtr x;
IN CONST ExprPtr name;
INOUT ExprPtr vn;
{
  ExprPtr e;

  e = MapPut(x,name,VN(vn)->names);
  SETQ(VN(vn)->names,e);
  HITBARRIER(vn);
}

ARGSWEEPFUNC(do_updatependinguses)
{
  ExprPtr vn,vf,counter,uses,name,deco;
  COUNTER stor;
  Boolean found;

  vn = MAPDATA(struct upu *)->vn;
  vf = MAPDATA(struct upu *)->vf;
  deco = MAPDATA(struct upu *)->deco;
  counter = MapGet(x,VF(vf)->pendinguses,&found);
  if (found && --IVAL(counter)==0)
    {
      uses = MapRemove(x,VF(vf)->pendinguses);
      SETQ(VF(vf)->pendinguses,uses);
      HITBARRIER(vf);
      name = MapGetSure(x,VN(vn)->names);
      stor = storageClass(getPpInfo(x,deco)->storage,x);
      uses = SetAdd(name,VF(vf)->free[stor]);
      SETQ(VF(vf)->free[stor],uses);
      HITBARRIER(vf);
    }
  return false;
}

PRIVATE updatePendingUses(e,deco,vn,vf)
IN CONST ExprPtr e;
IN CONST ExprPtr deco;
INOUT ExprPtr vn,vf;
{
  struct upu u;

  u.vn = vn;
  u.vf = vf;
  u.deco = deco;
  if (e->kind==(CHAR)E_COUNTED)
    (void) argSweep(COUNTEDVAL(e),do_updatependinguses,&u);
  else (void) do_updatependinguses(e,&u);
}

PRIVATE void createPendingUses(x,deco,vf)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr vf;
{
  ExprPtr uses,card;
  INT n;

  uses = getPpInfo(x,deco)->uses;
  n = MapCard(uses);
  MAKEINT(n,card);
  (void) KEEP(card);
  uses = MapPut(x,card,VF(vf)->pendinguses);
  SETQ(VF(vf)->pendinguses,uses);
  HITBARRIER(vf);
}

PRIVATE void addZap(x,vn,vf,oldx,oldm)
IN CONST ExprPtr x;
INOUT ExprPtr vn,vf;
IN CONST MapPtr oldx,oldm;
{
  ExprPtr diff;
  ENTER;

  diff = SetDiff(VF(vf)->free[V_EXPRPTR],oldx);
  if (diff!=NewMap())
    {
      setMapPut(x,diff,&VN(vn)->zapx);
      HITBARRIER(vn);
    }
  diff = SetDiff(VF(vf)->free[V_MAPGET],oldm);
  if (diff!=NewMap())
    {
      setMapPut(x,diff,&VN(vn)->zapm);
      HITBARRIER(vn);
    }
  QUIT;
}

PRIVATE ExprPtr CGlobalName(i)
IN INT i;
{
  char buf[32];
  ExprPtr name;

  sprintf(buf,"CGlobals[%ld]",i);
  name = Kstr((CHAR *)buf);
  return KEEP(name);
}

PRIVATE ExprPtr CConstName(i)
IN INT i;
{
  char buf[32];
  ExprPtr name;

  sprintf(buf,"CGlobals[%ld]",i);
  name = Kstr((CHAR *)buf);
  return KEEP(name);
}

PRIVATE MapPtr getVarNames1 P((IN CONST ExprPtr x,IN CONST ExprPtr deco,
			       INOUT ExprPtr vn,INOUT ExprPtr vf,
			       IN CONST MapPtr frontier,IN CONST ExprPtr last));

DEF_PRIVATE_ACTION(getvarnames)
{
  struct getvarnames *data;

  data = (struct getvarnames *)env;
  return getVarNames1(data->x,data->deco,data->vn,data->vf,data->frontier,data->last);
}

PRIVATE MapPtr getVarNames1k(x,deco,vn,vf,frontier,last)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr vn,vf;
IN CONST MapPtr frontier;
IN CONST ExprPtr last;
{
  struct getvarnames *data = (struct getvarnames *)alloctmp(sizeof(struct getvarnames));

  data->x = x;
  data->deco = deco;
  data->vn = vn;
  data->vf = vf;
  data->frontier = frontier;
  data->last = last;
  return CallCC(ACTION_NAME(getvarnames),(ExprPtr)data,true);
}

PRIVATE MapPtr getVarNames1(x,deco,vn,vf,frontier,last)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr vn,vf;
IN CONST MapPtr frontier;
IN CONST ExprPtr last;
{
  ExprPtr oldExprNames,oldMapGetNames;
  Boolean found;
  ENTER;

  if (STKFULL(found))
    return getVarNames1k(x,deco,vn,vf,frontier,last);
again:
  KEEPONLY(frontier);
  (void) MapGet(x,VN(vn)->names,&found);
  if (found)
    return frontier;
  oldExprNames = VF(vf)->free[V_EXPRPTR];
  (void) KEEP(oldExprNames);
  oldMapGetNames = VF(vf)->free[V_MAPGET];
  (void) KEEP(oldMapGetNames);
  switch (x->kind)
    {
    case E_NIL:
      break;
    case E_PROG:
      {
	ExprPtr ll,name;
	COUNTER i;

	for (i=0, ll=CAR(x); ll!=NIL; ll=CDR(ll))
	  {
	    name = makeParName(ll,deco,&i);
	    addVarName(x,name,vn);
	  }
      }
      last = x;
      x = CDR(x);
      goto again;
    case E_CONS:
      {
	ExprPtr name,e;

	e = CAR(x);
	updatePendingUses(e,deco,vn,vf);
	if (e->kind==(CHAR)E_COUNTED)
	  switch (COUNTEDN(e))
	    {
	    case Z_CODE:
	      name = Kstr(((Code *)EXTDATA(COUNTEDVAL(e)))->name);
	      (void) KEEP(name);
	      break;
	    case Z_C_CONST:
	      /*!!! name = CConstName(IVAL(CAR(COUNTEDVAL(e))),  ); */
	    case Z_C_GLOBAL:
	      name = CGlobalName(IVAL(CAR(COUNTEDVAL(e))));
	      break;
	    case Z_APPLY:
	    case Z_CALL:
	      VN(vn)->has_calls = true;
	      /*fallthrough*/
	    default:
	      name = makeVarName(x,deco,vn,vf);
	      createPendingUses(x,deco,vf);
	      break;
	    }
	else
	  name = MapGetSure(e,VN(vn)->names);
	addVarName(x,name,vn);
      }
      addZap(x,vn,vf,oldExprNames,oldMapGetNames);
      last = x;
      x = CDR(x);
      goto again;
    case E_IF:
      {
	ExprPtr test;

	test = CAR(CAR(x));
	updatePendingUses(test,deco,vn,vf);
	if (test->kind==(CHAR)E_COUNTED)
	  {
	    if (COUNTEDN(test)==Z_START_HANDLE)
	      VN(vn)->has_jumps = true;
	  }
      }
      addZap(x,vn,vf,oldExprNames,oldMapGetNames);
      last = CAR(x);
      frontier = getVarNames1(CDR(last),deco,vn,varFreeListCopy(vf),frontier,last);
      last = x;
      x = CDR(x);
      goto again;
    case E_GOTO:
      {
	struct upu u;

	u.vn = vn;
	u.vf = vf;
	u.deco = deco;
	(void) do_updatependinguses(MapGetSure(last,COUNTEDVAL(CAR(x))),&u);
      }
      {
	ExprPtr vff;
	Boolean found;

	vff = MapGet(x,frontier,&found);
	if (found)
	  varFreeListMerge(vff,vf);
	else
	  frontier = MapPut(x,vf,frontier);
      }
      break;
    default:
      kinderror("getVarNames1",x->kind);
      /*NOTREACHED*/
      break;
    }
  KEEPONLY(frontier);
  return frontier;
}

PUBLIC ExprPtr getVarNames(c)
IN CONST ExprPtr c;
{
  Code *code;
  ExprPtr vn,vf,x,last,name;
  MapPtr frontier,before;

  code = (Code *)EXTDATA(c);
  vn = varNamesNew();
  vf = varFreeListNew();
  frontier = NewMap();
  before = NewMap();
  x = code->prgm;
  last = NIL;
  while (true)
    {
      frontier = getVarNames1(x,code->deco,vn,vf,frontier,last);
      if (frontier==NewMap())
	break;
      x = getHighestMergeNode(frontier,code->deco,&before);
      vf = MapGetSure(x,frontier);
      frontier = SetRemove(x,frontier);
      name = makeVarName(x,code->deco,vn,vf);
      createPendingUses(x,code->deco,vf);
      switch (storageClass(getPpInfo(x,code->deco)->storage,x))
	{ /* indicate that some values can be zapped (used in outputCode1(),
	     case of E_GOTOs. */
	case V_MAPGET: setMapPut(x,UNIT,&VN(vn)->zapm); HITBARRIER(vn); break;
	case V_EXPRPTR: setMapPut(x,UNIT,&VN(vn)->zapx); HITBARRIER(vn); break;
	}
      last = x;
      x = CDR(x);
    }
  return vn;
}

/***** Splitting codes into run-time and init-time codes. *****/

ARGMAPFUNC(do_rewrite)
{
  return MapGetSure(x,MAPDATA(MapPtr));
}

PRIVATE ExprPtr rewriteExpr(e,done)
IN CONST ExprPtr e;
IN CONST MapPtr done;
{
  ENTER;

  if (e->kind==(CHAR)E_COUNTED)
    e = zApply(COUNTEDN(e),argMap(COUNTEDVAL(e),do_rewrite,done));
  else
    e = do_rewrite(e,done);
  KEEPONLY(e);
  return e;
}

ARGMAPFUNC(do_irewrite)
{
  return getPpInfo(x,MAPDATA(ExprPtr))->def;
}

PRIVATE ExprPtr rewriteIExpr(e,ideco)
IN CONST ExprPtr e;
IN CONST ExprPtr ideco;
{
  ENTER;

  if (e->kind==(CHAR)E_COUNTED)
    e = zApply(COUNTEDN(e),argMap(COUNTEDVAL(e),do_irewrite,ideco));
  else
    e = do_irewrite(e,ideco);
  KEEPONLY(e);
  return e;
}

MAPSWEEPFUNC(do_existsnoninituse)
{
  ExprPtr deco;

  deco = MAPDATA(ExprPtr);
  return (getPpInfo(x,deco)->storage & SCF_NOINIT)!=0;
}

PRIVATE Boolean existsNonInitUse(pp,deco)
IN CONST ppInfo *pp;
IN CONST ExprPtr deco;
{
  return MapSweep(pp->uses,do_existsnoninituse,deco);
}

PRIVATE ExprPtr genCName(funcname,x)
IN CONST CHAR *funcname;
IN CONST ExprPtr x;
{
  ExprPtr s;
  CHAR buf[128];

  sprintf(buf,"_CoNsT_%s_%lx",funcname,(long)x);
  s = Kstr(buf);
  return KEEP(s);
}

PRIVATE INT getKnownName(val)
IN CONST ExprPtr val;
{
  if (val==trueValue)
    return I_TRUEVALUE;
  if (val==falseValue)
    return I_FALSEVALUE;
  if (val==NIL)
    return I_NIL;
  if (val==NewMap())
    return I_EMPTY;
  if (val==tuple0)
    return I_TUPLE0;
  if (val==record0)
    return I_RECORD0;
  if (val==array0)
    return I_ARRAY0;
  if (val==zzero)
    return I_ZZERO;
  if (val==one)
    return I_ONE;
  if (val==mone)
    return I_MONE;
  if (val==two)
    return I_TWO;
  if (val==half)
    return I_HALF;
  if (val==zpi)
    return I_ZPI;
#if FLOAT_FORMAT==IEEE754
  if (val==pinf)
    return I_PINF;
  if (val==minf)
    return I_MINF;
#endif
  if (val==izero)
    return I_IZERO;
  if (val==ione)
    return I_IONE;
  if (val==itwo)
    return I_ITWO;
  return -1;
}

PRIVATE MapPtr splitCode1 P((IN CONST ExprPtr x,IN CONST ExprPtr last,
			     INOUT ExprPtr *runlastp,INOUT ExprPtr *initlastp,
			     IN CONST ExprPtr deco,IN OUT ExprPtr ideco,
			     IN CONST MapPtr ieqns,IN CONST CHAR *funcname,
			     IN CONST MapPtr done));

DEF_PRIVATE_ACTION(splitcode)
{
  struct splitcode *data;

  data = (struct splitcode *)env;
  return splitCode1(data->x,data->last,&data->runlast,&data->initlast,data->deco,data->ideco,data->ieqns,data->funcname,data->done);
}

PRIVATE MapPtr splitCode1k(x,last,runlastp,initlastp,deco,ideco,ieqns,funcname,done)
IN CONST ExprPtr x;
IN CONST ExprPtr last;
INOUT ExprPtr *runlastp,*initlastp;
IN CONST ExprPtr deco;
IN OUT ExprPtr ideco;
IN CONST MapPtr ieqns;
IN CONST CHAR *funcname;
IN CONST MapPtr done;
{
  struct splitcode *data = (struct splitcode *)alloctmp(sizeof(struct splitcode));
  ExprPtr res;

  data->x = x;
  data->last = last;
  data->runlast = *runlastp;
  data->initlast = *initlastp;
  data->deco = deco;
  data->ideco = ideco;
  data->ieqns = ieqns;
  data->funcname = funcname;
  data->done = done;
  res = CallCC(ACTION_NAME(splitcode),(ExprPtr)data,true);
  *runlastp = data->runlast;
  *initlastp = data->initlast;
  return res;
}

PRIVATE MapPtr splitCode1(x,last,runlastp,initlastp,deco,ideco,ieqns,funcname,done)
IN CONST ExprPtr x;
IN CONST ExprPtr last;
INOUT ExprPtr *runlastp,*initlastp;
IN CONST ExprPtr deco;
IN OUT ExprPtr ideco;
IN CONST MapPtr ieqns;
IN CONST CHAR *funcname;
IN CONST MapPtr done;
{
  ppInfo *pp;
  ExprPtr e,y;
  Boolean found;

  if (STKFULL(found))
    return splitCode1k(x,last,runlastp,initlastp,deco,ideco,ieqns,funcname,done);
again:
  switch (x->kind)
    {
    case E_NIL:
      break;
    case E_CONS:
      (void) MapGet(x,done,&found);
      if (found)
	return done;
      pp = getPpInfo(x,deco);
      if ((pp->storage & SCF_NOINIT) ||
	  ((pp->storage & SCF_UNBOXED) &&
	   (pp->storage & (SCF_BOOL | SCF_INT | SCF_STRING | SCF_CODE))))
	{
	  e = rewriteExpr(CAR(x),done);
	  *runlastp = tackAssign(e,*runlastp);
	}
      else
	{
	  if ((pp->storage & SCF_BOXED) &&
	      pp->valset->kind==(CHAR)E_MAPLET)
	    {
	      INT i;

	      i = getKnownName(MAPLETX(pp->valset));
	      if (i>=0)
		{
		  *runlastp = tackApplyBin(Z_C_GLOBAL,KINT(i),
					   KINT(pp->storage),*runlastp);
		  goto next;
		}
	    }
	  e = rewriteIExpr(CAR(x),ideco);
	  y = MapGet(e,ieqns,&found);
	  if (!found)
	    {
	      ppInfo *ppp;

	      *initlastp = y = tackAssign(e,*initlastp);
	      ieqns = MapPut(e,y,ieqns);
	      ppp = getPpInfo(y,ideco); /* create decoration */
	      SETQ(ppp->def,y);
	      SETQ(ppp->eqns,ieqns);
	      SETQ(ppp->fragile,pp->fragile);
	      SETQ(ppp->valset,pp->valset);
	      ppp->storage = pp->storage;
	      memcpy(ppp->tags,pp->tags,NTAGS);
	    }
	  if (existsNonInitUse(pp,deco))
	    { /* if it is used at run-time, generate a global C name for it. */
	      ExprPtr name;

	      name = genCName(funcname,y); /*!!!*/
	      e = zApplyBin(Z_SET_C_GLOBAL,name,y);
	      (void) MapGet(e,ieqns,&found);
	      if (!found) /* we have not yet created a global for this value */
		{
		  ppInfo *ppp;

		  *initlastp = y = tackAssign(e,*initlastp);
		  ieqns = MapPut(e,y,ieqns);
		  ppp = getPpInfo(y,ideco); /* create decoration */
		  SETQ(ppp->def,y);
		  SETQ(ppp->eqns,ieqns);
		  (void) newUse(y,NIL,ideco);
		}
	      *runlastp = tackApplyBin(Z_C_GLOBAL,name,KINT(pp->storage),*runlastp);
	    }
	}
next:
      done = MapPut(x,*runlastp,done);
      last = x;
      x = CDR(x);
      goto again;
    case E_IF:
      (void) MapGet(x,done,&found);
      if (found)
	return done;
      {
	ExprPtr thenlast;

	thenlast = compileIf(rewriteExpr(CAR(CAR(x)),done),*runlastp,runlastp);
	/* split then part */
	done = splitCode1(CDR(CAR(x)),CAR(x),&thenlast,initlastp,deco,ideco,ieqns,funcname,done);
	/* then split else part */
	last = x;
	x = CDR(x);
	goto again;
      }
    case E_GOTO:
      {
	ExprPtr l,r;

	l = do_rewrite(last,done);
	r = do_rewrite(MapGetSure(last,COUNTEDVAL(CAR(x))),done);
	e = MapGet(x,done,&found);
	if (found)
	  *runlastp = mergeContinuations(r,l,e);
	else
	  {
	    *runlastp = mergeContinuations(r,l,NIL);
	    goto next;
	  }
      }
      break;
    default:
      kinderror("splitCode1",x->kind);
      /*NOTREACHED*/
      break;
    }
  return done;
}

PRIVATE MapPtr copyTranslation(arglist,trans)
IN CONST ExprPtr arglist;
OUT struct translation *trans;
{
  ExprPtr list,ll;
  MapPtr done;
  ENTER;

  for (list=NIL, ll=arglist; ll!=NIL; ll=CDR(ll))
    {
      list = MakePair(E_CONS,CAR(ll),list);
      KEEPONLY(list);
    }
  list = Nreverse1(list,NIL);
  KEEPONLY(list);
  done = NewMap();
  for (ll=arglist; ll!=NIL; ll=CDR(ll), list=CDR(list))
    done = MapPut(ll,list,done);
  KEEPONLY(done);
  trans->prgm = MakePair(E_PROG,list,NIL);
  (void) KEEP(trans->prgm);
  return done;
}

PRIVATE ExprPtr splitCode(c,initc)
IN CONST ExprPtr c;
INOUT ExprPtr initc;
{
  Code *code,*initcode;
  struct translation *fun = (struct translation *)alloctmp(sizeof(struct translation));
  ExprPtr runlast,initlast;
  MapPtr done;
  ExprPtr runc;
  Boolean found;
  ExprPtr deco;
  ENTER;

  code = (Code *)EXTDATA(c);
  done = copyTranslation(CAR(code->prgm),fun);
  initcode = (Code *)EXTDATA(initc);
  initlast = initcode->last;
  runlast = fun->prgm;
  done = splitCode1(CDR(code->prgm),code->prgm,&runlast,&initlast,code->deco,
		    initcode->deco,getPpInfo(initlast,initcode->deco)->eqns,
		    initcode->name,done);
  runlast = MapGet(code->last,done,&found);
  if (!found)
    runlast = NIL;
  runc = makeCode(fun,runlast);
  deco = ppDecoInit();
  ppDecorate(fun->prgm,runlast,deco,NewMap(),NewMap());
  fillCode(runc,deco);
  ((Code *)EXTDATA(runc))->flags = code->flags;
  KEEPONLY(runc);
  return runc;
}

/***** Function that output C code. *****/

PRIVATE void flushChars(of)
INOUT outFiles *of;
{
  COUNTER i;
  CHAR *s,c;

  if (of->nChars!=0)
    {
      for (i=0, s=of->chars; i<of->nChars; i++)
	{
	  c = *s++;
	  if (c>=0x20 && c<0x7f)
	    {
	      (void) hfputc('\'',of->messages);
	      switch (c)
		{
		case '\'': case '\\':
		  (void) hfputc('\\',of->messages);
		}
	      (void) hfputc('\'',of->messages);
	    }
	  else
	    (void) hfprintf(of->messages,"%d",(unsigned int)c);
	  (void) hfputc(',',of->messages);
	}
      of->nChars = 0;
      (void) hfputc('\n',of->messages);
    }
}

PRIVATE void outputString(e,of)
IN CONST ExprPtr e;
INOUT outFiles *of;
{
  INT len,pos,n;

  pos = 0;
  len = SLEN(e);
  while (of->nChars+len>=CHARS_PER_LINE)
    {
      n = CHARS_PER_LINE-of->nChars;
      memcpy(of->chars+of->nChars,STEXT(e)+pos,n);
      of->nChars = CHARS_PER_LINE;
      pos += n;
      len -= n;
      flushChars(of);
    }
  memcpy(of->chars+of->nChars,STEXT(e)+pos,len);
  of->nChars += len;
  of->offset += SLEN(e);
}

PRIVATE void outputName(name,hf)
IN CONST ExprPtr name;
INOUT hFILE *hf;
{
  REGISTER INT i,n;

  for (n=SLEN(name), i=0; i<n; i++)
    (void) hfputc(STEXT(name)[i],hf);
}

PRIVATE void outputAssign1(name,of)
IN CONST ExprPtr name;
INOUT outFiles *of;
{
  if (name!=NIL && (STEXT(name)[0]!='_' || STEXT(name)[1]!='v'))
    {
      (void) hfputs("  ",of->runCode);
      outputName(name,of->runCode);
      (void) hfputs(" = ",of->runCode);
    }
  else (void) hfputs("  ",of->runCode);
}

PRIVATE void outputNameOf(x,names,of)
IN CONST ExprPtr x;
IN CONST MapPtr names;
INOUT outFiles *of;
{
  ExprPtr name;

  name = MapGetSure(x,names);
  outputName(name,of->runCode);
}

PRIVATE void outputAssignVar(name,var,names,of)
IN CONST ExprPtr name;
IN CONST ExprPtr var;
IN CONST MapPtr names;
INOUT outFiles *of;
{
  ExprPtr varname;
  Boolean found;

  varname = MapGet(var,names,&found);
  if (found && name!=varname && name!=NIL &&
      (STEXT(name)[0]!='_' || STEXT(name)[1]!='v'))
    { /* if this is a real new name and the values are not void,
	 then output assignment. */
      outputAssign1(name,of);
      outputName(varname,of->runCode);
      (void) hfputs(";\n",of->runCode);
    }
}

PRIVATE void outputExtDef(name,x,hf)
IN CONST ExprPtr name;
IN CONST EXTENDED *x;
INOUT hFILE *hf;
{
  COUNTER i,n;
  char c;
  char *data;
  COUNTER size;
  Boolean comma;

  comma = false;
  if (sizeof(EXTENDED) % sizeof(unsigned long)==0)
    {
      n = sizeof(EXTENDED) / sizeof(unsigned long);
      c = 'L';
      size = sizeof(unsigned long);
    }
  else if (sizeof(EXTENDED) % sizeof(unsigned short)==0)
    {
      n = sizeof(EXTENDED) / sizeof(unsigned short);
      c = 'S';
      size = sizeof(unsigned short);
    }
  else
    {
      n = sizeof(EXTENDED);
      c = 'C';
      size = 1;
    }
  (void) hfputc('(',hf);
  for (i=0, data = (char *)x; i<n; i++, data+=size)
    {
      (void) hfprintf(hf,"SET_X%c(,c);
      outputName(name,hf);
      (void) hfprintf(hf,",%d,0x%lx)",i,
		      ((c=='L')?(*(unsigned long *)data):
		       ((c=='S')?(*(unsigned short *)data):
			(*data))));
      if (comma)
	(void) hfputc(',',hf);
      else
	comma = true;
    }
  (void) hfputc(')',hf);
}

PRIVATE void outputVector(name,arg,names,of,tuple,share)
IN CONST ExprPtr name;
IN CONST ExprPtr arg;
IN CONST MapPtr names;
INOUT outFiles *of;
IN Boolean tuple,share;
{
  COUNTER i,n;

  i = tuple?0:1;
  n = AVAL(arg)->n;
  (void) hfprintf(of->runCode,"  BEGIN_VEC(%d);\n",n-i);
  for (; i<n; i++)
    {
      (void) hfputs("  PUSH_VEC(",of->runCode);
      outputNameOf(AVAL(arg)->expr[i++],names,of);
      (void) hfputs(");\n",of->runCode);
    }
  (void) hfputs(share?"  END_VEC(":"  END_VEC_SHARE(",of->runCode);
  outputName(name,of->runCode);
  (void) hfputs(tuple?",E_TUPLE_GNRL,":",E_RECORD_GNRL,",of->runCode);
  i = 0;
  if (tuple)
    (void) hfputs("NIL",of->runCode);
  else outputNameOf(AVAL(arg)->expr[i++],names,of);
  (void) hfprintf(of->runCode,",%d);\n",n-i);
}

PRIVATE void outputApply(arg,names,of)
IN CONST ExprPtr arg;
IN CONST MapPtr names;
INOUT outFiles *of;
{
  /* there are two kinds of function values:
     - function variables (parameters, locals, globals),
       which must be called as (*f)(),
     - function constants (name of Code structs), which
       must be called as f(). */
  ExprPtr f;
  char c;

  if (arg->kind==(CHAR)E_TUPLE_2)
    f = CAR(arg);
  else
    f = AVAL(arg)->expr[0];
  if (f->kind==(CHAR)E_CONS && CAR(f)->kind==(CHAR)E_COUNTED &&
      COUNTEDN(CAR(f))==Z_CODE)
    { /* immediate call [f()] */
      outputNameOf(f,names,of);
    }
  else
    { /* indirect call */
      (void) hfputs("(*",of->runCode);
      outputNameOf(f,names,of);
      (void) hfputc(')',of->runCode);
    }
  c = '(';
  if (arg->kind==(CHAR)E_TUPLE_2)
    {
      while (arg = CDR(arg), arg->kind==(CHAR)E_TUPLE_2)
	{
	  (void) hfputc(c,of->runCode);
	  c = ',';
	  outputNameOf(CAR(arg),names,of);
	}
      (void) hfputc(c,of->runCode);
      outputNameOf(CDR(arg),names,of);
    }
  else
    {
      COUNTER i,n;

      for (i=1, n=AVAL(arg)->n; i<n; i++)
	{
	  (void) hfputc(c,of->runCode);
	  c = ',';
	  outputNameOf(AVAL(arg)->expr[i],names,of);
	}
    }
  (void) hfputc(')',of->runCode);
}

PRIVATE void outputAssignExpr(name,e,names,modulename,of)
IN CONST ExprPtr name;
IN CONST ExprPtr e;
IN CONST MapPtr names;
IN CONST CHAR *modulename;
INOUT outFiles *of;
{
  ExprPtr arg;
  char *fun;

  arg = COUNTEDVAL(e);
  switch (COUNTEDN(e))
    { /* deal with special cases first */
    case Z_SET_C_GLOBAL:
      outputAssignVar(name,e,names,of);
      return;
    case Z_UNBOXED_REAL:
      {
	EXTENDED x;

	x = (EXTENDED)XVAL(arg);
	switch (classifyX(&x))
	  {
	  case C_QNAN: case C_SNAN: case -C_SNAN:
	  case C_INFINITE: case -C_INFINITE:
	    outputExtDef(name,&x,of->runCode);
	    break;
	  default:
	    (void) hfputs("  DEF_EXTENDED(",of->runCode);
	    outputName(name,of->runCode);
	    (void) hfputc(',',of->runCode);
	    (void) hfprintf(of->runCode,"%15E",x);
	    (void) hfputc(',',of->runCode);
	    outputExtDef(name,&x,of->runCode);
	    (void) hfputc(')',of->runCode);
	    break;
	  }
      }
      break;
    case Z_BOX_UNSHARED_TUPLE_1:
      fun = "  MKSING(E_TUPLE_1,";
using:
      (void) hfprintf(of->runCode,fun);
      outputNameOf(arg,names,of);
      goto uend;
    case Z_BOX_UNSHARED_RECORD_1:
      fun = "  MKPAIR(E_RECORD_1,";
ucons:
      (void) hfprintf(of->runCode,fun);
      outputNameOf(CAR(arg),names,of);
      (void) hfputc(',',of->runCode);
      outputNameOf(CDR(arg),names,of);
uend:
      (void) hfputc(',',of->runCode);
      outputName(name,of->runCode);
      (void) hfputc(')',of->runCode);
      break;
    case Z_CODE: /* codes and globals are given the right name by
		    getVarNames1() already, so don't do anything. */
    case Z_C_GLOBAL:
    case Z_ASSERT_EQUAL: /* this is just an assertion, don't produce code for it */
      return;
    case Z_GETTHEEXPR:
      outputName(name,of->runCode);
      (void) hfputc(" = THEEXPR; THEEXPR = NOEXPR",of->runCode);
      break;
    case Z_QUEUE_TO_LIST: fun = "  Q2L("; goto ucons;
    case Z_BOX_UNSHARED_CONS: fun = "  MKPAIR(E_CONS,"; goto ucons;
    case Z_BOX_UNSHARED_TUPLE_2: fun = "  MKPAIR(E_TUPLE_2,"; goto ucons;
    case Z_BOX_TUPLE_GNRL:
      outputVector(name,arg,names,of,true,true);
      break;
    case Z_BOX_UNSHARED_TUPLE_GNRL:
      outputVector(name,arg,names,of,true,false);
      break;
    case Z_BOX_RECORD_GNRL:
      outputVector(name,arg,names,of,false,true);
      break;
    case Z_BOX_UNSHARED_RECORD_GNRL:
      outputVector(name,arg,names,of,false,false);
      break;
    case Z_CONSENV:
      (void) hfputs("  MKENV(",of->runCode);
      outputNameOf(CAR(arg),names,of);
      (void) hfprintf(of->runCode,",%d,",IVAL(CDR(arg)));
      outputName(name,of->runCode);
      (void) hfputc(')',of->runCode);
      break;
    case Z_FAKEENV:
      (void) hfputs("  FAKEENV(",of->runCode);
      outputNameOf(arg,names,of);
      (void) hfputc(')',of->runCode);
      break;
    case Z_LET:
      (void) hfputs("ELET(",of->runCode);
      outputNameOf(CAR(arg),names,of);
      (void) hfputc(',',of->runCode);
      arg = CDR(arg);
      outputNameOf(CAR(arg),names,of);
      (void) hfprintf(of->runCode,",%ld)",IVAL(CDR(arg)));
      break;
    case Z_MAPGET: fun = "  MAPGET("; goto ucons;
    case Z_BOX_QUEUE: fun = MKQ(; goto using;
    case Z_BOX_1_QUEUE: fun = "MK1Q("; goto using;
    case Z_START_LISTSWEEP: fun = "START_LISTSWEEP("; goto using;
    case Z_START_MAPSWEEP: fun = "START_MAPSWEEP("; goto using;
    case Z_START_SUBMAPSWEEP: fun = "START_SUBMAPSWEEP("; goto using;
    case Z_CONCAT_QUEUE: fun = "CONCATQ("; goto ucons;
    default:
      outputAssign1(name,of);
      switch (COUNTEDN(e))
	{
	case Z_NIL:
	  fun = "NIL";
cst:
	  (void) hfputs(fun,of->runCode);
	  break;
	case Z_UNBOXED_INT:
	  (void) hfprintf(of->runCode,"%ld",IVAL(arg));
	  break;
	case Z_UNBOX_INT:
	  fun = "IVAL(";
unary:
	  (void) hfputs(fun,of->runCode);
	  outputNameOf(arg,names,of);
paren:
	  (void) hfputc(')',of->runCode);
	  break;
	case Z_BOX_INT: fun = "Kint("; goto unary;
	case Z_UNBOX_REAL: fun = "(EXTENDED)XVAL("; goto unary;
	case Z_BOX_REAL: fun = "Kext("; goto unary;
	case Z_UNBOX_TUPLE_NTH:
	  fun = "AVAL(";
indexing:
	  (void) hfputs(fun,of->runCode);
	  outputNameOf(CAR(arg),names,of);
	  (void) hfprintf(of->runCode,")->expr[%ld]",IVAL(CDR(arg)));
	  break;
	case Z_BOX_COMPLEX:
	  fun = "Kcomplex(";
binary:
	  (void) hfputs(fun,of->runCode);
	  outputNameOf(CAR(arg),names,of);
	  (void) hfputc(',',of->runCode);
	  outputNameOf(CDR(arg),names,of);
	  goto paren;
	case Z_UNBOXED_STRING:
	  (void) hfprintf(of->runCode,"_S_%s+%ld",modulename,of->offset);
	  outputString(arg,of);
	  break;
	case Z_UNBOX_STRING_LEN: fun = "SLEN("; goto unary;
	case Z_UNBOX_STRING_TEXT: fun = "STEXT("; goto unary;
	case Z_BOX_STRING: fun = "Kstring("; goto binary;
	case Z_EMPTY: fun = "empty_map"; goto cst;
	case Z_TUPLE_0: fun = "tuple0"; goto cst;
	case Z_RECORD_0: fun = "record0"; goto cst;
	case Z_BOX_TYPE_APPL: fun = "KtypeAppl("; goto binary;
	case Z_BOX_REF: fun = "Kref("; goto unary;
	case Z_BOX_TUPLE_1: fun = "Ktuple1("; goto unary;
	case Z_BOX_RECORD_1: fun = "Krecord1("; goto binary;
	case Z_BOX_CLOSURE: fun = "MakeCompiled(E_COMPILED_CLOSURE,"; goto binary;
	case Z_BOX_CONS: fun = "Kons("; goto binary;
	case Z_BOX_TUPLE_2: fun = "Ktuple2("; goto binary;
	case Z_BOX_SINGLETON: fun = "MapletKons("; goto binary;
	case Z_OVERWRITE: fun = "MapOverwrite("; goto binary;
	case Z_ENEXT: fun = "ENEXT("; goto unary;
	case Z_VAR: fun = "EVARS("; goto indexing;
	case Z_RAISE: fun = "(void) Raise("; goto unary;
	case Z_MAPPUT:
	  fun = "MapAdd(";
ternary:
	  (void) hfputs(fun,of->runCode);
	  outputNameOf(CAR(arg),names,of);
	  (void) hfputc(',',of->runCode);
	  arg = CDR(arg);
	  outputNameOf(CAR(arg),names,of);
	  (void) hfputc(',',of->runCode);
	  outputNameOf(CDR(arg),names,of);
	  goto paren;
	case Z_EQUAL:
	  fun = " == ";
cmp:
	  outputNameOf(CAR(arg),names,of);
	  (void) hfputs(fun,of->runCode);
	  outputNameOf(CDR(arg),names,of);
	  break;
	case Z_NOT: fun = "!("; goto unary;
	case Z_CAR: fun = "CAR("; goto unary;
	case Z_CDR: fun = "CDR("; goto unary;
	case Z_UNBOX_REF: fun = "REFVAL("; goto unary;
	case Z_HAS_TAG:
	  outputNameOf(CAR(arg),names,of);
	  (void) hfprintf(of->runCode,"->kind==(CHAR)%ld",IVAL(CDR(arg)));
	  break;
	case Z_GET_VECTOR_LEN:
	  (void) hfputs("AVAL(",of->runCode);
	  outputNameOf(arg,names,of);
	  (void) hfputs("->n",of->runCode);
	  break;
	case Z_TYPE_MATCH: fun = "TypeMatch("; goto binary;
	case Z_MAP_CARD_DEC:
	  (void) hfputs("MapCardDec(",of->runCode);
	  outputNameOf(CAR(arg),names,of);
	  (void) hfprintf(of->runCode,",%ld)",IVAL(CDR(arg)));
	  break;
	case Z_INT_LEQ: fun = "<="; goto cmp;
	case Z_GET_SHAPE: fun = "ASHAPE("; goto unary;
	case Z_GET_MAPGET_Y: fun = "MAPGETY("; goto unary;
	case Z_GET_MAPGET_FOUND: fun = "MAPGETF("; goto unary;
	case Z_TUPLE_SUB_NTH:
	  (void) hfputs("AVAL(",of->runCode);
	  outputNameOf(CAR(arg),names,of);
	  (void) hfputs(")->expr[",of->runCode);
	  outputNameOf(CDR(arg),names,of);
	  (void) hfputs("-1]",of->runCode);
	  break;
	case Z_MAPPUTBEHIND: fun = "MapSlip("; goto ternary;
	case Z_MAPREMOVE: fun = "MapKill("; goto binary;
	case Z_UNBOXED_FALSE: fun = "false"; goto cst;
	case Z_UNBOXED_TRUE: fun = "true"; goto cst;
	case Z_REF_SET: fun = "SETREFVAL("; goto binary;
	case Z_BOX_NUMTYPE: fun = "KnumType("; goto unary;
	case Z_BOX_TYPEFX: fun = "KtypeFx("; goto binary;
	case Z_APPLY:
	case Z_CALL:
	  outputApply(arg,names,of);
	  break;
	case Z_UNBOX_CLOSURE_FN: fun = "ACTION("; goto unary;
	case Z_UNBOX_CLOSURE_ENV: fun = "QENV("; goto unary;
	case Z_APPEND: fun = "Append0("; goto binary;
	case Z_MAPINTER:
	case Z_MAPDOMRESTRTO: fun = "MapDRT("; goto binary;
	case Z_MAPDOMRESTRBY: fun = "MapDRB("; goto binary;
	case Z_UNBOX_LISTSWEEP: fun = "UNBOX_LISTSWEEP("; goto unary;
	case Z_UNBOX_MAPSWEEP: fun = "UNBOX_MAPSWEEP("; goto unary;
	case Z_UNBOX_SUBMAPSWEEP: fun = "UNBOX_SUBMAPSWEEP("; goto unary;
	case Z_BOX_LOCAL_REF:
	case Z_UNBOX_LOCAL_REF:
	  outputNameOf(arg,names,of);
	  break;
	case Z_LOCAL_REF_SET:
	  outputNameOf(CAR(arg),names,of);
	  (void) hfputs(" = ",of->runCode);
	  outputNameOf(CDR(arg),names,of);
	  break;
	case Z_REVERSE1: fun = "Reverse1("; goto binary;
	case Z_BOX_UNSHARED_EXCEPTION: fun = "MakeException("; goto unary;
	case Z_BOX_SHARED_EXCEPTION: fun = "Kref(E_EXCEPTION,"; goto unary;
	case Z_BOX_BOOL: fun = "Kbool("; goto unary;
	case Z_UNBOX_BOOL: fun = "BVAL("; goto unary;
	case Z_END_HANDLE:
	  (void) hfputs("END_HANDLE()",of->runCode);
	  break;
	case Z_RMU1: fun = "RMU1("; goto ternary;
	case Z_UNSHARED_RMU1: fun = "UNSHARED_RMU1("; goto ternary;
	case Z_RMUX: fun = "RMUX("; goto binary;
	case Z_UNSHARED_RMUX: fun = "UNSHARED_RMUX("; goto binary;
	case Z_MAKE_PROMISE: fun = "MakePromise("; goto unary;
	case Z_BOX_PACK: fun = "MakePack(E_PACK,"; goto binary;
	case Z_CMPSTR: fun = "CompareStrings("; goto binary;
	case Z_INC_LISTSWEEP:
	case Z_INC_MAPSWEEP:
	case Z_INC_SUBMAPSWEEP:
	case Z_START_HANDLE:
	case Z_FAKE_UNBOX_TUPLE_LEN:
	case Z_FAKE_MAPGET:
	case Z_FAKE_NOT_MAPGET:
	case Z_FAKE_MAPGET_FOUND:
	default:
	  kinderror("outputAssignExpr",COUNTEDN(e));
	  /*NOTREACHED*/
	  break;
	}
    }
  (void) hfputs(";\n",of->runCode);
}

MAPSWEEPFUNC(do_zap)
{
  CHAR buf[32];
  struct outzap *data;

  data = MAPDATA(struct outzap *);
  memcpy(buf,STEXT(x),SLEN(x));
  buf[SLEN(x)] = 0;
  (void) hfprintf(data->of->runCode,data->msg,buf);
  return false;
}

PRIVATE void outputZap(x,vn,of)
IN CONST ExprPtr x;
IN CONST ExprPtr vn;
IN CONST outFiles *of;
{
  MapPtr zap;
  struct outzap data;
  Boolean found;

  data.of = of;
  zap = MapGet(x,VN(vn)->zapx,&found);
  if (found && zap!=NewMap())
    {
      data.msg = (CHAR *)"  ZAP(%s);\n";
      (void) MapSweep(zap,do_zap,&data);
    }
  zap = MapGet(x,VN(vn)->zapm,&found);
  if (found && zap!=NewMap())
    {
      data.msg = (CHAR *)"  ZAPM(%s);\n";
      (void) MapSweep(zap,do_zap,&data);
    }
}

PRIVATE MapPtr outputCode1 P((IN CONST ExprPtr x,IN CONST ExprPtr last,
			      IN CONST ExprPtr c,IN CONST ExprPtr vn,
			      IN CONST CHAR *modulename,
			      IN CONST outFiles *of,IN CONST MapPtr done));

DEF_PRIVATE_ACTION(outputcode)
{
  struct outputcode *data;

  data = (struct outputcode *)env;
  return outputCode1(data->x,data->last,data->c,data->vn,data->modulename,data->of,data->done);
}

PRIVATE MapPtr outputCode1k(x,last,c,vn,modulename,of,done)
IN CONST ExprPtr x,last;
IN CONST ExprPtr c;
IN CONST ExprPtr vn;
IN CONST CHAR *modulename;
IN CONST outFiles *of;
IN CONST MapPtr done;
{
  struct outputcode *data = (struct outputcode *)alloctmp(sizeof(struct outputcode));

  data->x = x;
  data->last = last;
  data->c = c;
  data->vn = vn;
  data->modulename = modulename;
  data->of = of;
  data->done = done;
  return CallCC(ACTION_NAME(outputcode),(ExprPtr)data,true);
}

PRIVATE MapPtr outputCode1(x,last,c,vn,modulename,of,done)
IN CONST ExprPtr x,last;
IN CONST ExprPtr c;
IN CONST ExprPtr vn;
IN CONST CHAR *modulename;
IN CONST outFiles *of;
IN CONST MapPtr done;
{
  Code *code;
  ExprPtr name,e;
  MapPtr names;
  Boolean found;
  ENTER;

  if (STKFULL(found))
    return outputCode1k(x,last,c,vn,modulename,of,done);
  code = ((Code *)EXTDATA(c));
  names = VN(vn)->names;
again:
  switch (x->kind)
    {
    case E_NIL:
      (void) hfputs("  _QUIT return ",of->runCode);
      if (code->last==last)
	{
	  if (code->storage!=SCF_TOP &&
	      (name = MapGet(last,names,&found), found))
	    outputName(name,of->runCode);
	  (void) hfputs(";\n",of->runCode);
	}
      break;
    case E_IF:
      e = CAR(CAR(x));
      if (e->kind==(CHAR)E_COUNTED)
	switch (COUNTEDN(e))
	  {
	  case Z_INC_LISTSWEEP:
	    (void) hfputs("  INC_LISTSWEEP(",of->runCode);
inc:
	    name = MapGetSure(COUNTEDVAL(e),names);
	    outputName(name,of->runCode);
	    (void) hfputs(",q",of->runCode);
handle:
	    outputName(name,of->runCode);
	    (void) hfputs(")\n",of->runCode);
	    done = outputCode1(CDR(CAR(x)),CAR(x),c,names,modulename,of,done);
	    (void) hfputs("q",of->runCode);
	    outputName(name,of->runCode);
	    (void) hfputs(":\n",of->runCode);
	    last = x;
	    x = CDR(x);
	    goto again;
	  case Z_INC_MAPSWEEP:
	    (void) hfputs("  INC_MAPSWEEP(",of->runCode);
	    goto inc;
	  case Z_INC_SUBMAPSWEEP:
	    (void) hfputs("  INC_SUBMAPSWEEP(",of->runCode);
	    goto inc;
	  case Z_START_HANDLE:
	    (void) hfputs("  START_HANDLE(q",of->runCode);
	    name = MapGetSure(COUNTEDVAL(e),names);
	    goto handle;
	  default:
	    kinderror("outputCode1",COUNTEDN(e));
	    break;
	  }
      else
	{
	  (void) hfputs("  if (",of->runCode);
	  name = MapGetSure(e,names);
	  outputName(name,of->runCode);
	  (void) hfputs(") {\n",of->runCode);
	  outputZap(x,vn,of);
	  done = outputCode1(CDR(CAR(x)),CAR(x),c,names,modulename,of,done);
	  (void) hfputs("  }\n",of->runCode);
	  outputZap(x,vn,of);
	  last = x;
	  x = CDR(x);
	  goto again;
	}
      break;
    case E_GOTO:
      e = COUNTEDVAL(CAR(x));
      name = MapGetSure(x,names);
      {
	ExprPtr newname;
	struct outzap data;

	newname = MapGetSure(last,e);
	outputAssignVar(name,newname,names,of);
	if (newname!=name)
	  { /* zap the old name if it was different and marked as zappable
	       in getVarNames(). */
	    data.of = of;
	    (void) MapGet(x,VN(vn)->zapx,&found);
	    if (found)
	      {
		data.msg = (CHAR *)"  ZAP(%s);\n";
		do_zap(name,UNIT,(CHAR *)&data);
	      }
	    (void) MapGet(x,VN(vn)->zapm,&found);
	    if (found)
	      {
		data.msg = (CHAR *)"  ZAPM(%s);\n";
		do_zap(name,UNIT,(CHAR *)&data);
	      }
	  }
      }
      (void) MapGet(x,done,&found);
      if (found)
	{
	  (void) hfputs("  goto LAB",of->runCode);
	  outputName(name,of->runCode);
	  (void) hfputs(";\n",of->runCode);
	}
      else
	{
	  done = SetAdd(x,done);
	  (void) hfputs("LAB",of->runCode);
	  outputName(name,of->runCode);
	  (void) hfputs(":\n",of->runCode);
	}
      break;
    case E_CONS:
      name = MapGet(x,names,&found);
      if (!found)
	break;
      e = CAR(x);
      if (e->kind!=(CHAR)E_COUNTED)
	outputAssignVar(name,e,names,of);
      else outputAssignExpr(name,e,names,modulename,of);
      outputZap(x,vn,of);
      last = x;
      x = CDR(x);
      goto again;
    default:
      kinderror("outputCode1",x->kind);
      break;
    }
  KEEPONLY(done);
  return done;
}

PRIVATE char *typename[] = {
  "ExprPtr ",
  "Boolean ",
  "INT ",
  "UBString ",
  "EXTENDED ",
  "compiledAction ",
  "Boolean ",
  "void "
};

PRIVATE void outputProto(name,arglist,res,of,private,init)
IN CONST CHAR *name;
IN CONST ExprPtr arglist;
IN INT res;
IN CONST outFiles *of;
IN Boolean private,init;
{
  hFILE *file;
  CHAR *scope;
  COUNTER stor,i;

  file = init?of->initProto:(private?of->runProto:of->initProto);
  scope = (CHAR *)(private?"PRIVATE ":"IMPORT ");
  (void) hfputs(scope,file);
  stor = storageClass(res,NIL);
  (void) hfputs(typename[stor],file);
  if (init)
    (void) hfputs("_I_",file);
  (void) hfputs(name,file);
  (void) hfputs(" P((",file);
  if (arglist==NIL)
    (void) hfputs("void",file);
  else for (i=0; arglist!=NIL; arglist=CDR(arglist))
    {
      if (i!=0)
	(void) hfputc(',',file);
      stor = storageClass(IVAL(CAR(arglist)),NIL);
      (void) hfputs(typename[stor],file);
      (void) hfprintf(file,"p%d",++i);
    }
  (void) hfputs("));",file);
}

PRIVATE void outputHeader(name,arglist,res,of,private,init)
IN CONST CHAR *name;
IN CONST ExprPtr arglist;
IN INT res;
IN CONST outFiles *of;
IN Boolean private,init;
{
  CHAR *scope;
  COUNTER stor,i;
  hFILE *file;
  ExprPtr ll;

  scope = (CHAR *)(private?"PRIVATE ":"PUBLIC ");
  file = init?of->initCode:of->runCode;
  (void) hfputs(scope,file);
  stor = storageClass(res,NIL);
  (void) hfputs(typename[stor],file);
  if (init)
    (void) hfputs("_I_",file);
  (void) hfputs("\n#ifdef HAS_PROTOTYPES\n  (",file);
  /* output header for the case we have prototypes */
  if (arglist==NIL)
    (void) hfputs("void",file);
  else for (i=0, ll=arglist; ll!=NIL; ll=CDR(ll))
    {
      if (i!=0)
	(void) hfputc(',',file);
      stor = storageClass(IVAL(CAR(ll)),NIL);
      (void) hfputs(typename[stor],file);
      (void) hfprintf(file,"p%d",++i);
    }
  (void) hfputs(")\n#else  (",file);
  /* then output header for the case we don't have prototypes */
  for (i=0, ll=arglist; ll!=NIL; ll=CDR(ll))
    {
      if (i!=0)
	(void) hfputc(',',file);
      (void) hfprintf(file,"p%d",++i);
    }
  (void) hfputs(")\n",file);
  for (i=0, ll=arglist; ll!=NIL; ll=CDR(ll))
    {
      stor = storageClass(IVAL(CAR(ll)),NIL);
      (void) hfputs(typename[stor],file);
      (void) hfprintf(file,"p%d",++i);
      (void) hfputs(";\n",file);
    }
  /* then open brace for beginning of function */
  (void) hfputs("#endif\n{\n",file);
}

PRIVATE void outputFunStart(name,arglist,res,of,private,init)
IN CONST CHAR *name;
IN CONST ExprPtr arglist;
IN INT res;
IN CONST outFiles *of;
IN Boolean private,init;
{
  outputProto(name,arglist,res,of,private,init);
  outputHeader(name,arglist,res,of,private,init);
}

PRIVATE void outputBeamer(name,arglist,res,of)
IN CONST CHAR *name;
IN CONST ExprPtr arglist;
IN INT res;
IN CONST outFiles *of;
{
  COUNTER i,stor;
  ExprPtr ll;
  CHAR c;

  if (arglist==NIL && res==V_VOID)
    {
      (void) hfprintf(of->runProto,"DECL_BEAMER(%s)\n",name);
      (void) hfprintf(of->runCode,"DEF_BEAMER(%s)\n",name);
      (void) hfputs(name,of->runCode);
      (void) hfputs("();\n",of->runCode);
    }
  else
    {
      (void) hfprintf(of->runProto,"struct _Beam_%s {\n",name);
      for (i=0, ll=arglist; ll!=NIL; ll=CDR(ll))
	{
	  stor = storageClass(IVAL(CAR(ll)),NIL);
	  (void) hfputs("  ",of->runProto);
	  (void) hfputs(typename[stor],of->runProto);
	  (void) hfprintf(of->runProto,"p%d;\n",++i);
	}
      if (res!=V_VOID)
	(void) hfprintf(of->runProto,"  %s res;\n",typename[stor]);
      (void) hfputc('}',of->runProto);
      (void) hfputs(";\n",of->runProto);
      (void) hfprintf(of->runProto,"DECL_BEAMER(%s)\n",name);
      (void) hfprintf(of->runCode,"DEF_BEAMER(%s)\n",name);
      (void) hfprintf(of->runCode,"START_BEAMER(%s)\n",name);
      if (res!=V_VOID)
	(void) hfputs("b->res = ",of->runCode);
      (void) hfputs(name,of->runCode);
      c = '(';
      if (arglist==NIL)
	(void) hfputc(c,of->runCode);
      else for (i=0, ll=arglist; ll!=NIL; ll=CDR(ll))
	{
	  (void) hfputc(c,of->runCode);
	  c = ',';
	  (void) hfprintf(of->runCode,"b->p%d");
	}
      (void) hfputs(");\n",of->runCode);
    }
  (void) hfputs("END_BEAMER()\n\n",of->runCode);
}

PRIVATE void outputBeamerJump(name,arglist,res,of)
IN CONST CHAR *name;
IN CONST ExprPtr arglist;
IN INT res;
IN CONST outFiles *of;
{
  COUNTER i;
  ExprPtr ll;

  (void) hfputs("{\n",of->runCode);
  if (arglist==NIL && res==V_VOID)
    {
      (void) hfprintf(of->runCode,"  BEAM(%s,NIL)\n",name);
    }
  else
    {
      (void) hfprintf(of->runCode,"  PREP_BEAM(%s)\n",name);
      for (i=1, ll=arglist; ll!=NIL; ll=CDR(ll), i++)
	{
	  (void) hfprintf(of->runCode,"  b->p%d = p%d;\n",i,i);
	}
      (void) hfprintf(of->runCode,"  BEAM(%s,b)\n",name);
      if (res!=V_VOID)
	(void) hfputs("  return b->res;\n",of->runCode);
    }
  (void) hfputs("}\n",of->runCode);
}

PRIVATE void outputLocalTypedVars(vn,of,stor,regp,op)
IN CONST ExprPtr vn;
IN CONST outFiles *of;
IN COUNTER stor;
IN Boolean regp;
IN CONST CHAR *op;
{
  INT i,n;
  Boolean comma;
  char *fun;

  if ((n = VN(vn)->maxn[stor])!=0)
    {
      if (VN(vn)->has_jumps)
	fun = "  VOLATILE ";
      /* if we call SETJUMP() somewhere in the function, then
	 we are not guaranteed that local variables will have the correct
	 values upon return from a LONGJUMP(), unless we declare them volatile,
	 what we do here. */
      else if (regp)
	fun = "  REGISTER ";
      else
	fun = "  ";
      (void) hfputs(fun,of->runCode);
      (void) hfputs(typename[stor],of->runCode);
      comma = false;
      for (i=1; i<=n; i++)
	{
	  if (comma)
	    (void) hfputc(',',of->runCode);
	  comma = true;
	  if (op!=NULL)
	    {
	      (void) hfputs(op,of->runCode);
	      (void) hfputc('(',of->runCode);
	    }
	  (void) hfprintf(of->runCode,varnames(stor),i);
	  if (op!=NULL)
	    (void) hfputc(')',of->runCode);
	}
      (void) hfputc(';',of->runCode);
      (void) hfputc('\n',of->runCode);
    }
}

PRIVATE void outputLocalVars(vn,of)
IN CONST ExprPtr vn;
IN CONST outFiles *of;
{
  varNames *vnn;
  INT ne,nm,ng,n;

  vnn = (varNames *)EXTDATA(vn);
  ne = vnn->maxn[V_EXPRPTR];
  ng = vnn->maxn[V_MAPGET];
  nm = vnn->maxn[V_MAPSWEEP];
  n = ne+ng+nm;
  (void) hfprintf(of->runCode,"#define _N_EXPR %ld\n#define _N_MAPGET %ld\n#define _N__X %ld+%ld+(MAP_CHAIN_LENGTH+1)*%ld\n#define _QUIT ",ne,ng,ne,ng,nm);
  /* declare number of entries in the local array of protected variables;
     ne entries for ExprPtrs, ng entries for MapGets,
     nm*(MAP_CHAIN_LENGTH+1) entries for MapSweeps
     (those are hand-coded stacks). */
  if (n!=0)
    (void) hfputs("STOP_PROTECTION();",of->runCode);
  (void) hfputc('\n',of->runCode);
  (void) hfputc('\n',of->runCode);
  (void) hfputs("{\n",of->runCode);
  outputLocalTypedVars(vn,of,V_BOOL,true,NULL);
  outputLocalTypedVars(vn,of,V_INT,true,NULL);
  outputLocalTypedVars(vn,of,V_STRING,true,NULL);
  outputLocalTypedVars(vn,of,V_REAL,true,NULL);
  outputLocalTypedVars(vn,of,V_CODE,false,NULL);
  outputLocalTypedVars(vn,of,V_MAPGET,false,"MAPGETF");
  outputLocalTypedVars(vn,of,V_MAPSWEEP,true,"_MPTR");
  if (n!=0)
    (void) hfputs("  DECL_PROTECTION()\n",of->runCode);
  (void) hfputc('\n',of->runCode);
}

/*
	Format of code bodies:
#define _N_EXPR <#entries for V_EXPRPTR>
#define _N_MAPGET <#entries for V_MAPGET>
#define _N__X <#entries for V_EXPRPTR+V_MAPGET+V_MAPSWEEP>
  <qualif> <type> <list-of-variables>, where <qualif> is REGISTER (VOLATILE if
			some part of the code does SETJUMP), and <type> is
			INT, Bool, etc. but not ExprPtr.
#define _QUIT [STOP_PROTECTION();] // decl in brackets if _N__X!=0 only
  DECL_PROTECTION() // if _N__X!=0 only

  TEST_STACK() // if function has calls only; test for stack overflow and
  // go to beamer function if it has overflown.
  <code>
  // code never reaches here, since it must have output a return or a Raise().
  PREP_BEAM(funname) // if function has calls only
  <store input parameters in local structure b>
  BEAM(funname) // jump into hyperspace
  <get back result and return>
#undef _QUIT
#undef _N__X
#undef _N_MAPGET
#undef _N_EXPR
*/

PUBLIC void outputCode(c,of,modulename)
IN CONST ExprPtr c;
IN CONST outFiles *of;
IN CONST CHAR *modulename;
{
  ExprPtr deco,done,vn;
  Code *code;
  ENTER;

  code = (Code *)EXTDATA(c);
  deco = code->deco;
  done = NewMap();
  /* output prototype for beamer function [from "Beam me up, Scotty", Star Trek]. */
  outputBeamer(code->name,CAR(code->prgm),code->storage,of);
  /* output prototype for code */
  outputFunStart(code->name,CAR(code->prgm),code->storage,of,(code->flags & CODE_PRIVATE)!=0,false);
  vn = getVarNames(c);
  outputLocalVars(vn,of);
  /* now, if function calls other functions, check the stack. */
  if (VN(vn)->has_calls)
    (void) hfputs("  TEST_STACK()\n",of->runCode);
  /* output code, at last. */
  (void) outputCode1(CDR(code->prgm),code->prgm,c,vn,modulename,of,NewMap());
  /* then, if we has to check the stack, emit here a call to the beamer function,
     which is responsible for changing stacks when the current one is full. */
  (void) hfputs("}\n",of->runCode);
  if (VN(vn)->has_calls)
    outputBeamerJump(code->name,CAR(code->prgm),code->storage,of);
  (void) hfputs("#undef _QUIT\n#undef _N__X\n#undef _N_MAPGET\n#undef _N_EXPR\n}\n\n",of->runCode);
  QUIT;
}
/*!!!*/

PUBLIC void closeCFiles(of)
INOUT outFiles *of;
{
  if (of->initProto!=NULL)
    {
      (void) hfclose(of->initProto);
      of->initProto = NULL;
    }
  if (of->initCode!=NULL)
    {
      (void) hfclose(of->initCode);
      of->initCode = NULL;
    }
  if (of->runProto!=NULL)
    {
      (void) hfclose(of->runProto);
      of->runProto = NULL;
    }
  if (of->runCode!=NULL)
    {
      (void) hfclose(of->runCode);
      of->runCode = NULL;
    }
  if (of->publicProto!=NULL)
    {
      (void) hfclose(of->publicProto);
      of->publicProto = NULL;
    }
  if (of->messages!=NULL)
    {
      flushChars(of);
      (void) hfputc('0',of->messages);
      (void) hfputc('\n',of->messages);
      (void) hfclose(of->messages);
      of->messages = NULL;
    }
}

PUBLIC Boolean openCFiles(modulename,of)
IN CONST CHAR *modulename;
OUT outFiles *of;
{
#define BUF_SIZE 64
  char buf[BUF_SIZE],*s;

  strcpy(buf,(char *)modulename);
  s = buf+strlen((char *)modulename);
  strcpy(s,"_i.h");
  of->initProto = hfopen(buf,"w");
  strcpy(s,"_i.c");
  of->initCode = hfopen(buf,"w");
  strcpy(s,"_c.h");
  of->runProto = hfopen(buf,"w");
  strcpy(s,"_c.c");
  of->runCode = hfopen(buf,"w");
  strcpy(s,"_p.h");
  of->publicProto = hfopen(buf,"w");
  strcpy(s,"_s");
  of->messages = hfopen(buf,"w");
  of->nChars = 0;
  of->offset = 0;
  strcpy(s,"_s.r");
  if (of->initProto==NULL || of->initCode==NULL ||
      of->runProto==NULL || of->runCode==NULL ||
      of->publicProto==NULL || of->messages==NULL)
    {
      closeCFiles(of);
      return false;
    }
  strcpy(s,"_c.c");
  (void) hfprintf(of->runProto,(CHAR *)"/* Prototypes for %s, generated by the HimML compiler */\n\n",buf);
  strcpy(s,"_c.h");
  (void) hfprintf(of->runCode,(CHAR *)"/* Code for %s, generated by the HimML compiler */\n\n#include \"codesupp.h\"\n#include \"%s\"\n\n#define __SEG__ %s%s\n\n",buf,buf,modulename,"_CODE");

  strcpy(s,"_i.c");
  (void) hfprintf(of->initProto,(CHAR *)"/* Prototypes for %s, generated by the HimML compiler */\n\n",buf);
  strcpy(s,"_i.h");
  (void) hfprintf(of->initCode,(CHAR *)"/* Code for %s, generated by the HimML compiler */\n\n#include \"codesupp.h\"\n#include \"%s\"\n\n#define __SEG__ %s%s\n\n",buf,buf,modulename,"_INIT");

  strcpy(s,"_p.h");
  (void) hfprintf(of->runCode,(CHAR *)"#include \"%s\"\n\n",buf);
  (void) hfprintf(of->initCode,(CHAR *)"#include \"%s\"\n\n",buf);
  (void) hfprintf(of->publicProto,(CHAR *)"/* Public prototypes for %s, generated by the HimML compiler */\n\n",modulename);

  strcpy(s,"_s");
  (void) hfprintf(of->initCode,(CHAR *)"#if OS==MAC_OS\nPUBLIC CHAR *_S_%s;\n#else\nPUBLIC CHAR _S_%s[] = { /* constant strings */\n#include \"%s\"\n};\n#endif\n\n",modulename,modulename,buf);
  (void) hfprintf(of->publicProto,(CHAR *)"#if OS==MAC_OS\nEXTERN CHAR *_S_%s;\n#else\nEXTERN CHAR _S_%s[];\n#endif\n\n",modulename,modulename);

  /* output prototype for initialization function */
  outputFunStart(modulename,NIL,V_VOID,of,true,true);
  (void) hfprintf(of->initCode,"#if OS==MAC_OS\n  { Handle h = GetResourceByName('byt#',\"\\p%s_s\");\n    MoveHHi(h); HLock(h); _S_%s = *h;\n  }\n#endif\n",modulename,modulename);

  return true;
#undef BUF_SIZE
}

#endif
