/* Bytecode interpreter (unused).
   Copyright (C) 1992-1997 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#if TRANSLATOR
#include "expr.h"
#include "refcount.h"
#include "mem.h"
#include "interp.h"
#include "codesupp.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#ifdef HAS_STDARG
#include <stdarg.h>
#else
#ifdef HAS_VARARGS
#include <varargs.h>
#endif
#endif
#include "errproto.h"
#include "typproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rinterp.h"
#ifdef HAS_PROTOTYPES
#include "sinterp.h"
#endif

#define __SEG__ EVAL

#include "zinterp.c"

DEF_PRIVATE_ACTION(interp)
{
  ExprPtr *params = (ExprPtr *)env;

  switch (((bFunPtr)params[0])->arity)
    {
#include "zinterp.c"
    default:
      kinderror(message(541),((bFunPtr)params[0])->arity);
      /*NOTREACHED*/
      return NIL;
    }
}

#ifdef HAS_STDARG
PUBLIC ExprPtr Interpret(bFunPtr myself,...)
{
  va_list args;
#else
#ifdef HAS_VARARGS
/*VARARGS*/
PUBLIC ExprPtr Interpret(myself,va_alist)
bFunPtr myself;
va_dcl
{
  va_list args;
#else
PUBLIC ExprPtr Interpret(myself,dummy)
bFunPtr myself;
char *dummy;
{
  char **args;

#define va_arg(a,b) (*((b *) ((a +=  ((sizeof(b)+sizeof(int)-1) & ~(sizeof(int)-1))   ) -  ((sizeof(b)+sizeof(int)-1) & ~(sizeof(int)-1)) )))
#endif
#endif
  REGISTER OPCODE *pc;
  gcStackView *frame;
  ExprPtr **ptrs; /* pointers to data in frame->p */
  char *bools;
  INT *ints;
  CHAR **strings;
  EXTENDED *reals;
  char **codes;
  byteCodeChunk *chunk;
  char *rd;

#ifdef HAS_STDARG
  va_start(args,msg);
#else
#ifdef HAS_VARARGS
  va_start(args);
#else
  args = dummy;
#endif
#endif
  /* First entry in args is a byteCode * */
  chunk = myself->bc.chunk;
  pc = ((OPCODE *)(((char *)chunk)+chunk->instrOff))+myself->bc.start;
  rd = ((char *)chunk)+chunk->rawDataOff;
again:
  switch (*pc++)
    {
    case K_RETURN_EXPR: return frame->p[*pc]; /* return [var-index] */
    case K_RETURN_BOOL: return (ExprPtr)bools[*pc]; /* return [var-index] */
    case K_RETURN_INT: return (ExprPtr)ints[*pc]; /* return [var-index] */
    case K_RETURN_STRING: return (ExprPtr)strings[*pc]; /* return [var-index] */
    case K_RETURN_REAL: return (ExprPtr)&reals[*pc]; /* return [var-index] */
      /* we can only return an address, not a real, because they do not
	 fit in a single pointer in general; this is ok because reals
	 is alloctmp(), and provided we store the result as soon as
	 the call to the real-valued function returns, the real result
	 is put away before reals is freed. */
    case K_RETURN_CODE: return (ExprPtr)codes[*pc]; /* return [var-index] */
    case K_RETURN_VOID: return NIL; /* return */
    case K_PARAM_EXPR: frame->p[*pc++] = va_arg(args,ExprPtr); goto again;
    case K_PARAM_BOOL: bools[*pc++] = (char)va_arg(args,ExprPtr); goto again;
    case K_PARAM_INT: ints[*pc++] = (INT)va_arg(args,ExprPtr); goto again;
    case K_PARAM_STRING: strings[*pc++] = (CHAR *)va_arg(args,ExprPtr); goto again;
    case K_PARAM_CODE: codes[*pc++] = (char *)va_arg(args,ExprPtr); goto again;
    case K_PARAM_REAL: reals[*pc++] = *(EXTENDED *)va_arg(args,ExprPtr); goto again;
      /* param_* [argno] [var-index] */
    case K_END_PARAM:
#ifdef HAS_STDARG
      va_end(args);
#else
#ifdef HAS_VARARG
      va_end(args);
#endif
#endif
      break;
    case K_CONST_EXPR: frame->p[pc[1]] = chunk->consts[*pc]; pc += 2; goto again;
      /* const_expr [var-index] [const-index] */
    case K_CONST_BOOL: bools[pc[1]] = rd[*pc]; pc += 2; goto again;
    case K_CONST_INT: ints[pc[1]] = *(INT *)&rd[*pc]; pc += 2; goto again;
    case K_CONST_STRING: strings[pc[1]] = (CHAR *)&rd[*pc]; pc += 2; goto again;
    case K_CONST_CODE: codes[pc[1]] = *(char **)&rd[*pc]; pc += 2; goto again;
    case K_CONST_REAL: reals[pc[1]] = *(EXTENDED *)&rd[*pc]; pc += 2; goto again;
      /* const_* [var-index] [offset-in-rawdata] */
    case K_ASSIGN_EXPR: frame->p[*pc] = frame->p[pc[1]]; pc += 2; goto again;
    case K_ASSIGN_BOOL: bools[*pc] = bools[pc[1]]; pc += 2; goto again;
    case K_ASSIGN_INT: ints[*pc] = ints[pc[1]]; pc += 2; goto again;
    case K_ASSIGN_STRING: strings[*pc] = strings[pc[1]]; pc += 2; goto again;
    case K_ASSIGN_CODE: codes[*pc] = codes[pc[1]]; pc += 2; goto again;
    case K_ASSIGN_REAL: reals[*pc] = reals[pc[1]]; pc += 2; goto again;
      /* assign_* [dst-index] [src-index] */
    case K_GOTO: pc += *pc; goto again; /* goto [offset] */
    case K_IF: if (bools[*pc]) pc += 2; else pc += pc[1]; goto again;
      /* if [var-index] [offset-on-false] code... */
    case K_UNLESS: if (bools[*pc]) pc += pc[1]; else pc += 2; goto again;
      /* if [var-index] [offset-on-true] code... */
    case K_INC_LISTSWEEP:
      if ((frame->p[*pc]=CDR(frame->p[*pc]))==NIL) pc += pc[1]; else pc+=2;
      goto again;
      /* inc_listsweep [var-index] [offset-on-end] code... */
    case K_INC_MAPSWEEP:
      {
	REGISTER ExprPtr *_mptr = ptrs[*pc];
	REGISTER ExprPtr *_mtop = &frame->p[pc[1]];
	REGISTER ExprPtr _m_val;

	do {
	  if (_mptr==_mtop)
	    {
	      pc += pc[2];
	      goto again; /* exit */
	    }
	  *_mptr-- = NOEXPR; /* allow this to be reclaimed by GC */
	  _m_val = *_mptr; /* pop from stack */
	} while (_m_val==empty_map);
	while (_m_val->kind!=(CHAR)E_MAPLET)
	  {
	    REGISTER ExprPtr _m_aux;

	    _m_aux = MAPLEFT(_m_val);
	    if (_m_aux==empty_map)
	      _m_val = MAPRIGHT(_m_val);
	    else
	      {
		*_mptr++ = MAPRIGHT(_m_val); /* push onto stack */
		_m_val = _m_aux;
	      }
	  }
	ptrs[*pc] = _mptr;
	_mtop[MAP_CHAIN_LENGTH] = _m_val;
	pc += 3;
	goto again;
      }
      /* inc_mapsweep [index-for-mptr-in-ptrs]
	 [index-for-mtop-in-frame->p] [offset-on-end] code... */
    case K_INC_SUBMAPSWEEP:
      {
	REGISTER ExprPtr _m_l,_m_e,_m_val,m = frame->p[*pc];

	_m_l = CAR(m);
	_m_val = CDR(m);
	while (true)
	  {
	    if (_m_l==NIL)
	      {
		SETCAR(m,NIL); /* free that at least */
		pc += pc[1];
		goto again;
	      }
	    _m_e = CAR(_m_l);
	    if (_m_e->kind==(CHAR)E_OR)
	      {
		_m_e->kind = (CHAR)E_AND;
		_m_val = MapPutTrie(CAR(_m_e),CDR(_m_e),_m_val,(CODE)(1L<<NLOWBITS));
		SETCDR(m,_m_val);
		break;
	      }
	    _m_e->kind = (CHAR)E_OR;
	    _m_val = MapRemoveTrie(CAR(_m_e),_m_val,(CODE)(1L<<NLOWBITS));
	    SETCDR(m,_m_val);
	    _m_l = CDR(_m_l);
	  }
	pc += 2;
	goto again;
      }
      /* inc_submapsweep [var-index] [offset-on-end] code... */
    case K_START_HANDLE:
      {
	excHandler *_xh = (excHandler *)alloctmp(sizeof(excHandler));
#if PROFILING & PROFILING_PROPER_TIMES
	ExprPtr _cProf = currentProf;
#endif

	if (SETJUMP(_xh->jb)!=0) /* if abnormal exit */
	  {
	    setHandlerStack(excStack,_xh->next);
#if PROFILING & PROFILING_PROPER_TIMES
	    syncProfProper(_cProf);
#endif
	    pc += pc[0];
	    goto again;
	  }
	/* normal exit, evaluate body of handle clause */
	_xh->next = excStack;
	_xh->tag = X_HANDLER;
	_xh->code = _xh->data = NIL;
	excStack = _xh;
	pc += 1;
	goto again;
      }
      /* start_handle [offset-on-raise] code... */
    case K_TEST_STACK:
      if (STKFULL(codes))
	{
	  char **params = (char **)alloctmp((myself->arity+1)*sizeof(char *));
	  char **prm;

	  params[0] = (char *)myself;
	  prm = params+1;
	  while (true)
	    {
	      switch (*pc++)
		{
		case K_PARAM_EXPR:
		case K_PARAM_BOOL:
		case K_PARAM_INT:
		case K_PARAM_STRING:
	      case K_PARAM_CODE:
		case K_PARAM_REAL:
		  *prm++ = va_arg(args,ExprPtr);
		  pc++;
		  break;
		  /* param_* [var-index] */
		case K_END_PARAM:
#ifdef HAS_STDARG
		  va_end(args);
#else
#ifdef HAS_VARARG
		  va_end(args);
#endif
#endif
		  return CallCC(ACTION_NAME(interp),(ExprPtr)params,true);
		case K_ALLOC_EXPRS: /* ignore */
		  pc += 7;
		  break;
		case K_ALLOC_NOEXPR: /* ignore */
		  pc += 5;
		  break;
		default:
		  kinderror(message(540),pc[-1]);
		  /*NOTREACHED*/
		  break;
		}
	    }
	}
      /*NOTREACHED*/
      goto again;
    case K_ALLOC_EXPRS:
      {
	REGISTER COUNTER nexprs = *pc;
	COUNTER exprsz = VARSIZE(gcStackView,p,nexprs);
	COUNTER ptrsz = exprsz + (pc[1]*sizeof(ExprPtr *));
	COUNTER strsz = ptrsz + (pc[2]*sizeof(CHAR *));
	COUNTER codesz = MODALIGN(strsz + (pc[3]*sizeof(char *)) + (ALIGNMENT-1));
	COUNTER realsz = MODALIGN(codesz + (pc[4]*sizeof(EXTENDED)) + (ALIGNMENT-1));
	COUNTER intsz = realsz + (pc[5]*sizeof(INT));
	COUNTER sz = intsz + pc[6];

	frame = (gcStackView *)alloctmp(sz);
	frame->n = nexprs;
	frame->next = theHeapFrames;
	{
	  REGISTER ExprPtr *_pp,**_ppend;
	  for (_pp = (ExprPtr *)_x.p, _ppend = _pp+nexprs;
	       _pp<_ppend; *_pp++ = NOEXPR);
	}
	theHeapFrames = (gcStackView *)&_x;
	ptrs = (ExprPtr **)(((char *)frame)+exprsz);
	strings = (CHAR **)(((char *)frame)+ptrsz);
	codes = (char **)(((char *)frame)+strsz);
	reals = (EXTENDED *)(((char *)frame)+codesz);
	ints = (INT *)(((char *)frame)+realsz);
	bools = ((char *)frame)+intsz;
	pc += 7;
	goto again;
      }
      /* alloc_exprs [nexprs] [nptrs] [nstrings] [ncodes] [nreals] [nints] [nbools] */
    case K_ALLOC_NOEXPR:
      {
	REGISTER char *locals;
	COUNTER strsz = (pc[0]*sizeof(CHAR *));
	COUNTER codesz = MODALIGN(strsz + (pc[1]*sizeof(char *)) + (ALIGNMENT-1));
	COUNTER realsz = MODALIGN(codesz + (pc[2]*sizeof(EXTENDED)) + (ALIGNMENT-1));
	COUNTER intsz = realsz + (pc[3]*sizeof(INT));
	COUNTER sz = intsz + pc[4];

	locals = (char *)alloctmp(sz);
	strings = (CHAR **)locals;
	codes = (char **)(locals+strsz);
	reals = (EXTENDED *)(locals+codesz);
	ints = (INT *)(locals+realsz);
	bools = locals+intsz;
	pc += 5;
	goto again;
      }
      /* alloc_noexpr [nstrings] [ncodes] [nreals] [nints] [nbools] */
    case K_STOP_PROTECTION: theHeapFrames = frame->next; goto again;
      /* stop_protection */
    }
}

#endif
