/* Optimizer, both for the bytecode compiler and the compiler to C (unused).
   Copyright (C) 1992-1997 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#ifdef TRANSLATOR
#include "expr.h"
#include "ufind.h"
#include "optimize.h"
#include "errproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#ifdef HAS_PROTOTYPES
#include "sopt.h"
#endif

#define __SEG__ OPTIMIZE

EXTERN EXTENDED fone;
EXTERN ExprPtr highCost;
EXTERN ExprPtr matchException,mapgetException;

PRIVATE ExprPtr getContents(di,i)
IN CONST ExprPtr di;
IN COUNTER i;
{
  return AVAL(di)->expr[i];
}

PRIVATE void setContents(di,i,val)
INOUT ExprPtr di;
IN COUNTER i;
IN CONST ExprPtr val;
{
  SETQ(AVAL(di)->expr[i],val);
  HITBARRIER(di);
}

	/* definite equations: */

PRIVATE void dIcost(di,x)
IN CONST ExprPtr di;
OUT REAL *x;
{
  *x = (EXTENDED)XVAL(getContents(di,DEFINFO_COST));
}

PRIVATE void setdIcost(di,x)
INOUT ExprPtr di;
IN CONST REAL *x;
{
  ExprPtr xx;

  xx = Kreal(x);
  setContents(di,DEFINFO_COST,xx);
}

PRIVATE COUNTER dItype(di)
IN CONST ExprPtr di;
{
  return IVAL(getContents(di,DEFINFO_TYPE));
}

PRIVATE void setdItype(di,type)
INOUT ExprPtr di;
IN COUNTER type;
{
  ExprPtr xx;

  xx = Kint((INT)type);
  setContents(di,DEFINFO_TYPE,xx);
}

#define dIfunbodies(di) getContents(di,DEFINFO_FUN_BODIES)
#define setdIfunbodies(di,val) setContents(di,DEFINFO_FUN_BODIES,val)
#define dIpreferred(di) getContents(di,DEFINFO_PREFERRED)
#define setdIpreferred(di,val) setContents(di,DEFINFO_PREFERRED,val)
#define dIvar(di) getContents(di,DEFINFO_VAR)
#define setdIvar(di,val) setContents(di,DEFINFO_VAR,val)

PRIVATE ExprPtr newDefInfo(cost,preferred,type,funbodies,var)
IN CONST REAL *cost;
IN CONST ExprPtr preferred;
IN COUNTER type;
IN CONST ExprPtr funbodies;
IN CONST ExprPtr var;
{
  ExprPtr di;
  ENTER;

  di = MakeArray(N_DEFINFO,NIL);
  (void) KEEP(di);
  setdIcost(di,cost);
  setdIpreferred(di,preferred);
  setdItype(di,type);
  setdIfunbodies(di,funbodies);
  setdIvar(di,var);
  KEEPONLY(di);
  return di;
}

PRIVATE COUNTER interType(type1,type2)
IN COUNTER type1,type2;
{
  if (type1==type2 || type2==BOXED_ANY)
    return type1;
  if (type1==BOXED_ANY)
    return type2;
  switch (type1)
    {
      case BOXED_VOID: /* type of H0 */
	if (type2==type1)
	  return type1;
	break;
      case BOXED_NUM:
	switch (type2)
	  {
	    case BOXED_REAL: case BOXED_NUM: case BOXED_COMPLEX:
	      return type2;
	  }
	break;
      case BOXED_REAL: case BOXED_COMPLEX:
	if (type2==BOXED_NUM)
	  return type1;
	break;
      case UNBOXED_NUM:
	switch (type2)
	  {
	    case UNBOXED_REAL: case UNBOXED_NUM: case UNBOXED_COMPLEX:
	      return type2;
	  }
	break;
      case UNBOXED_REAL: case UNBOXED_COMPLEX:
	if (type2==UNBOXED_NUM)
	  return type1;
	break;
      case BOXED_LIST:
	switch (type2)
	  {
	    case BOXED_NIL: case BOXED_CONS: case BOXED_LIST:
	      return type2;
	  }
	break;
      case BOXED_NIL: case BOXED_CONS:
	if (type2==BOXED_LIST)
	  return type1;
	break;
      default:
        if (type1>=BOXED_TUPLE_START+1 && type1<=BOXED_TUPLE_START+15 &&
	    type2>=BOXED_TUPLE_START+1 && type2<=BOXED_TUPLE_START+15)
	  {
	    type1 = (type1-BOXED_TUPLE_START) & (type2-BOXED_TUPLE_START);
	    if (type1!=0)
	      return BOXED_TUPLE_START+type1;
	  }
	break;
    }
  berror(message(641));
}

PRIVATE ExprPtr interVar(x1,x2)
IN CONST ExprPtr x1,x2;
{
  if (x1==NIL)
    return x2;
  return x1;
}

PRIVATE ExprPtr mergeDefInfo(x1,x2,di1,di2,data)
IN CONST colorPtr x1,x2;
IN CONST defInfoPtr di1,di2;
IN char *data;
{
  REAL cost1,cost2;
  COUNTER type;
  ExprPtr funbodies,var;
  ExprPtr newdi;

  type = interType(dItype(di1),dItype(di2));
  var = interVar(dIvar(di1),dIvar(di2));
  funbodies = MapMerge(dIfunbodies(di1),dIfunbodies(di2));
  dIcost(di1,&cost1); dIcost(di2,&cost2);
  if (cost2<cost1)
    newdi = newDefInfo(&cost2,dIpreferred(di2),type,funbodies,var);
  else
    newdi = newDefInfo(&cost1,dIpreferred(di1),type,funbodies,var);
  return newdi;
}

struct newcongdefinfo {
  REAL cost;
  COUNTER type;
  MapPtr funbodies;
  ExprPtr var;
};

PRIVATE defInfoPtr newCongDefInfo(x,data)
IN CONST colorPtr x;
IN struct newcongdefinfo *data;
{
  ExprPtr di;

  di = newDefInfo(&data->cost,x,data->type,data->funbodies,data->var);
  return di;
}

	/* possible equations: */

PRIVATE COUNTER pItype(pi)
IN CONST ExprPtr pi;
{
  return IVAL(getContents(pi,POSINFO_TYPE));
}

PRIVATE void setpItype(pi,type)
INOUT ExprPtr pi;
IN COUNTER type;
{
  ExprPtr xx;

  xx = Kint((INT)type);
  setContents(pi,POSINFO_TYPE,xx);
}

PRIVATE ExprPtr newPosInfo(type)
IN COUNTER type;
{
  ExprPtr pi;
  ENTER;

  pi = MakeArray(N_POSINFO,NIL);
  (void) KEEP(pi);
  setpItype(pi,type);
  KEEPONLY(pi);
  return pi;
}

PRIVATE COUNTER unionType(type1,type2)
IN COUNTER type1,type2;
{
  if (type1==BOXED_ANY || type2==BOXED_ANY)
    return BOXED_ANY;
  if (type1==type2)
    return type1;
  switch (type1)
    {
      case BOXED_VOID: /* type of H0 */
	if (type2==type1)
	  return type1;
	break;
      case BOXED_REAL: case BOXED_NUM: case BOXED_COMPLEX:
	switch (type2)
	  {
	    case BOXED_REAL: case BOXED_NUM: case BOXED_COMPLEX:
	      return BOXED_NUM;
	  }
	break;
      case UNBOXED_REAL: case UNBOXED_NUM: case UNBOXED_COMPLEX:
	switch (type2)
	  {
	    case UNBOXED_REAL: case UNBOXED_NUM: case UNBOXED_COMPLEX:
	      return UNBOXED_NUM;
	  }
	break;
      case BOXED_NIL: case BOXED_CONS: case BOXED_LIST:
	switch (type2)
	  {
	    case BOXED_NIL: case BOXED_CONS: case BOXED_LIST:
	      return BOXED_LIST;
	  }
	break;
      default:
        if (type1>=BOXED_TUPLE_START+1 && type1<=BOXED_TUPLE_START+15 &&
	    type2>=BOXED_TUPLE_START+1 && type2<=BOXED_TUPLE_START+15)
	  {
	    return BOXED_TUPLE_START+
		   ((type1-BOXED_TUPLE_START) | (type2-BOXED_TUPLE_START));
	  }
	break;
    }
  berror(message(641));
}

PRIVATE ExprPtr mergePosInfo(x1,x2,di1,di2,data)
IN CONST colorPtr x1,x2;
IN CONST posInfoPtr di1,di2;
IN char *data;
{
  COUNTER type;
  ExprPtr newpi;

  type = unionType(pItype(di1),pItype(di2));
  newpi = newPosInfo(type);
  return newpi;
}

struct newcongposinfo {
  COUNTER type;
};

PRIVATE defInfoPtr newCongPosInfo(x,data)
IN CONST colorPtr x;
IN struct newcongposinfo *data;
{
  ExprPtr pi;

  pi = newPosInfo(data->type);
  return pi;
}
