/* Translator from interpreted form to SSA. Part of the bytecode compiler
   and of the compiler to C. (Not used yet.)
   Copyright (C) 1992-1997 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#if TRANSLATOR
#include "expr.h"
#include "refcount.h"
#include "mem.h"
#include "translate.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#include "errproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "iter.h"
#include "rtrl.h"

typedef ExprPtr (*bindCompFun) P((IN CONST ExprPtr bindings,
				  IN CONST compEnv *env,
				  IN CONST ExprPtr last,
				  INOUT ExprPtr *errlastp));

#ifdef HAS_PROTOTYPES
#include "strl.h"
#endif

#define __SEG__ OPTIMIZE

EXTERN ExprPtr emptyException;
EXTERN int errno;
EXTERN ExprPtr none,some;

PUBLIC void initTranslation(arity,trans)
IN COUNTER arity;
OUT struct translation *trans;
{
  ExprPtr list,storage;
  ENTER;

  storage = KINT(SCF_COMPOSITE | SCF_BOXED | SCF_NOINIT);
  /* build formal parameter list */
  for (list=NIL; arity--!=0;)
    {
      list = MakePair(E_CONS,storage,list);
    }
  trans->prgm = MakePair(E_PROG,list,NIL);
  KEEPONLY(trans->prgm);
}

PRIVATE ExprPtr translateStmt P((IN CONST ExprPtr e,IN CONST compEnv *env,
				 INOUT ExprPtr last));

PRIVATE ExprPtr dumpLetStores P((IN CONST MapPtr set,IN CONST ExprPtr env,
				 INOUT ExprPtr last));

PRIVATE ExprPtr nthArg(n,trans)
IN COUNTER n;
IN CONST struct translation *trans;
{
  ExprPtr arg;

  arg = CAR(trans->prgm);
  while (n--!=0)
    arg = CDR(arg);
  return arg;
}

PRIVATE ExprPtr lastArg(trans)
IN CONST struct translation *trans;
{
  ExprPtr arg;

  arg = CAR(trans->prgm);
  while (CDR(arg)!=NIL)
    arg = CDR(arg);
  return arg;
}

PUBLIC ExprPtr zApply(code,arg)
IN CONST INT code;
IN CONST ExprPtr arg;
{
  return KCOUNTED(arg,code);
}

PUBLIC ExprPtr zApplyBin(code,arg1,arg2)
IN CONST INT code;
IN CONST ExprPtr arg1,arg2;
{
  ExprPtr arg;

  arg = Ktuple2(arg1,arg2);
  (void) KEEP(arg);
  return zApply(code,arg);
}

PUBLIC ExprPtr zApplyTer(code,arg1,arg2,arg3)
IN CONST INT code;
IN CONST ExprPtr arg1,arg2,arg3;
{
  ExprPtr arg;

  arg = Ktuple2(arg2,arg3);
  (void) KEEP(arg);
  return zApplyBin(code,arg1,arg);
}

PUBLIC ExprPtr tackAssign(expr,last)
IN CONST ExprPtr expr;
INOUT ExprPtr last;
{
  ExprPtr instr;

  if (last!=NIL)
    {
      instr = MakePair(E_CONS,expr,NIL); /* x := expr */
      SETCDR(last,instr); /* put at continuation */
    }
  else instr = NIL;
  return instr; /* return x */
}

PUBLIC ExprPtr tackApply(code,arg,last)
IN CONST INT code;
IN CONST ExprPtr arg;
INOUT ExprPtr last;
{
  return (last==NIL)?NIL:tackAssign(zApply(code,arg),last);
}

PUBLIC ExprPtr tackApplyBin(code,arg1,arg2,last)
IN CONST INT code;
IN CONST ExprPtr arg1,arg2;
INOUT ExprPtr last;
{
  return (last==NIL)?NIL:tackAssign(zApplyBin(code,arg1,arg2),last);
}

PUBLIC ExprPtr tackApplyTer(code,arg1,arg2,arg3,last)
IN CONST INT code;
IN CONST ExprPtr arg1,arg2,arg3;
INOUT ExprPtr last;
{
  return (last==NIL)?NIL:tackAssign(zApplyTer(code,arg1,arg2,arg3),last);
}

PRIVATE ExprPtr tackCGlobal(idx,storage,last)
IN INT idx;
IN INT storage;
INOUT ExprPtr last;
{
  ENTER;

  last = tackApplyBin(Z_C_GLOBAL,KINT(idx),KINT(storage),last);
  QUIT;
  return last;
}

PRIVATE ExprPtr compileString(e,last)
IN CONST ExprPtr e;
INOUT ExprPtr last;
{
  ExprPtr n;

  n = tackApply(Z_UNBOXED_INT,KINT(SLEN(e)),last);
  /* create an unboxed integer for the length */
  last = tackApply(Z_UNBOXED_STRING,e,n);
  /* create an unboxed string */
  last = tackApplyBin(Z_BOX_STRING,n,last,last); /* build the boxed string */
  return last;
}

PRIVATE ExprPtr compileInt(e,last)
IN CONST ExprPtr e;
INOUT ExprPtr last;
{
  last = tackApply(Z_UNBOXED_INT,e,last); /* create an unboxed integer */
  last = tackApply(Z_BOX_INT,last,last); /* and box it */
  return last;
}

#define compileLabel compileString

PRIVATE ExprPtr compileShape(shape,last)
IN CONST ExprPtr shape;
INOUT ExprPtr last;
{
again:
  switch (shape->kind)
    {
    case E_EMPTY:
      last = tackApply(Z_EMPTY,NIL,last);
      break;
    case E_MAPLET:
      {
	ExprPtr lbl,idx;

	lbl = compileString(MAPLETX(shape),last);
	idx = compileInt(MAPLETY(shape),lbl);
	last = tackApplyBin(Z_BOX_SINGLETON,lbl,idx,idx);
	break;
      }
    case E_MAP:
      {
	ExprPtr aux;

	if (MAPLEFT(shape)==NewMap())
	  {
	    shape = MAPRIGHT(shape);
	    goto again;
	  }
	if (MAPRIGHT(shape)==NewMap())
	  {
	    shape = MAPLEFT(shape);
	    goto again;
	  }
	aux = compileShape(MAPLEFT(shape),last);
	last = compileShape(MAPRIGHT(shape),aux);
	last = tackApplyBin(Z_OVERWRITE,aux,last,last);
	break;
      }
    }
  return last;
}

PRIVATE ExprPtr compileClosure(code,env,last)
IN CONST ExprPtr code;
IN CONST ExprPtr env;
INOUT ExprPtr last;
{
  last = tackApplyBin(Z_BOX_CLOSURE,code,env,last);
  /* box the closure */
  return last;
}

PRIVATE ExprPtr compileNil(env,last)
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  last = tackApply(Z_NIL,NIL,last);
  return last;
}

PUBLIC ExprPtr buildGnrlArg(n,vec)
IN COUNTER n;
IN CONST ExprPtr *vec;
{
  ExprPtr e;

  e = KtupleGnrl(n,vec);
  (void) KEEP(e);
  return e;
}

PRIVATE ExprPtr compileConstantCouple(tag,e,last)
IN INT tag;
IN CONST ExprPtr e;
INOUT ExprPtr last;
{
  ExprPtr car,cdr;

  car = compileConstant(CAR(e),last,true);
  cdr = compileConstant(CDR(e),car,true);
  last = tackApplyBin(tag,car,cdr,cdr);
  return last;
}

PRIVATE ExprPtr compileConstantVector(tag,shape,e,last)
IN INT tag;
IN CONST ExprPtr shape;
IN CONST ExprPtr e;
INOUT ExprPtr last;
{
  COUNTER i,n;
  ExprPtr arg;

  n = AVAL(e)->n;
  if (shape!=NIL)
    n++;
  {
    ExprPtr *elp;
    ExprPtr *elts;

    elts = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
    elp = elts;
    if (shape!=NIL)
      last = *elp++ = compileConstant(shape,last,true);
    n = AVAL(e)->n;
    for (i=0; i<n; i++)
      *elp++ = last = compileConstant(AVAL(e)->expr[i],last,true);
    arg = buildGnrlArg(n,elts);
  }
  last = tackApply(tag,arg,last);
  return last;
}

DEF_PRIVATE_ACTION(compileconstant)
{
  struct compileconstant *data;

  data = (struct compileconstant *)env;
  return compileConstant(data->e,data->last,data->boxedp);
}

PRIVATE ExprPtr compileConstant1(e,last,boxedp)
IN CONST ExprPtr e;
INOUT ExprPtr last;
IN CONST Boolean boxedp;
{
  struct compileconstant *data = (struct compileconstant *)alloctmp(sizeof(struct compileconstant));

  data->e = e;
  data->last = last;
  data->boxedp = boxedp;
  return CallCC(ACTION_NAME(compileconstant),(ExprPtr)data,true);
}

PUBLIC ExprPtr compileConstant(e,last,boxedp)
IN CONST ExprPtr e;
INOUT ExprPtr last;
IN CONST Boolean boxedp;
{
  if (STKFULL(e))
    return compileConstant1(e,last,boxedp);
  switch (e->kind)
    {
    case E_NIL:
      last = tackApply(Z_NIL,NIL,last);
      break;
    case E_EMPTY:
      last = tackApply(Z_EMPTY,NIL,last);
      break;
    case E_INT:
      last = compileInt(e,last);
      break;
    case E_STRING:
      if (boxedp)
	last = compileString(e,last);
      else
	last = tackApply(Z_UNBOXED_STRING,e,last);
      break;
    case E_REAL:
      last = tackApply(Z_UNBOXED_REAL,e,last); /* create an unboxed real */
      if (boxedp)
	last = tackApply(Z_BOX_REAL,last,last); /* and box it */
      break;
    case E_TUPLE_0:
      last = tackApply(Z_TUPLE_0,NIL,last); /* build 0-tuple */
      break;
    case E_RECORD_0:
      last = tackApply(Z_RECORD_0,NIL,last); /* build empty record */
      break;
    case E_CONS:
      last = compileConstantCouple((INT)Z_BOX_CONS,e,last);
      break;
    case E_TUPLE_1:
      last = compileConstant(REFVAL(e),last,true);
      last = tackApply(Z_BOX_TUPLE_1,last,last);
      break;
    case E_TUPLE_2:
      last = compileConstantCouple((INT)Z_BOX_TUPLE_2,e,last);
      break;
    case E_TUPLE_GNRL:
      last = compileConstantVector((INT)Z_BOX_TUPLE_GNRL,NIL,e,last);
      break;
    case E_RECORD_1:
      last = compileConstantCouple((INT)Z_BOX_RECORD_1,e,last);
      break;
    case E_RECORD_GNRL:
      last = compileConstantVector((INT)Z_BOX_RECORD_GNRL,ASHAPE(e),e,last);
      break;
    case E_QUOTE:
      last = compileConstant(REFVAL(e),last,boxedp);
      break;
    case E_MAPLET:
      last = compileConstantCouple((INT)Z_BOX_SINGLETON,e,last);
      break;
    case E_MAP:
      last = compileConstantCouple((INT)Z_OVERWRITE,e,last);
      break;
    case E_TYPE_APPL:
      last = compileConstantCouple((INT)Z_BOX_TYPE_APPL,e,last);
      break;
    case E_NUMTYPE:
      last = compileConstant(REFVAL(e),last,true);
      last = tackApply(Z_BOX_NUMTYPE,last,last);
      break;
    case E_TYPEFX:
      last = compileConstantCouple((INT)Z_BOX_TYPEFX,e,last);
      break;
    default:
      kinderror(message(474),e->kind);
      break;
    }
  return last;
}

PRIVATE void codeSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  Code *pp;

  pp = (Code *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      {
	int len;

	len = strlen((char *)pp->name);
	outInt(len);
	if (errno) return;
	outChars(pp->name,len);
	if (errno) return;
	outInt(pp->fxtypes);
	if (errno) return;
	outChars(pp->tags,NTAGS);
	if (errno) return;
	outInt(pp->size);
	if (errno) return;
	outInt(pp->flags);
	if (errno) return;
	outInt(pp->storage);
	if (errno) return;
      }
      /*fallthrough*/
    case X_MARKDATA:
      (*doext->action)(pp->last); /* save last, increment refcount of last */
      /*fallthrough*/
    case X_MARK:
      (*doext->action)(pp->prgm);
      (*doext->action)(pp->lambda);
      /* no need to mark last, it is accessible through pp->prgm */
      (*doext->action)(pp->deco);
      break;
    case X_FREE:
      if (pp->name!=NULL)
	{
	  MARKFREE(pp->name); /* free the C name */
	}
      MARKFREE(extdata); /* free the C structure */
      break;
    }
}

PRIVATE ExprPtr codeLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  Code *c;
  ExprPtr e;
  int len;
  ENTER;

  c = (Code *)MALLOC(sizeof(Code));
  len = inInt();
  if (errno)
    {
err1:
      MARKFREE(c);
      return NIL;
    }
  c->name = (CHAR *)MALLOC(len+1);
  inChars(c->name,len);
  if (errno)
    {
err2:
      MARKFREE(c->name);
      goto err1;
    }
  c->name[len] = 0;
  c->fxtypes = inInt();
  if (errno) goto err2;
  inChars(c->tags,NTAGS);
  if (errno) goto err2;
  c->size = inInt();
  if (errno) goto err2;
  c->flags = inInt();
  if (errno) goto err2;
  c->storage = inInt();
  if (errno) goto err2;
  c->last = (*load)(sharedref);
  if (errno) goto err2;
  c->prgm = (*load)(sharedref);
  if (errno) goto err2;
  c->lambda = (*load)(sharedref);
  if (errno) goto err2;
  c->deco = (*load)(sharedref);
  if (errno) goto err2;
  e = (shared?Kexternal:MakeExternal)(desc,(char *)c);
  KEEPONLY(e);
  return e;
}

PUBLIC extDesc codeDesc = {
  (CHAR *)"code",
  'c','o','d','e', NULL,
  NULL, /* not shared */
  NULL, /* not shared */
  NULL, /* print as default */
  codeSweep,
  codeLoad
  };

PUBLIC ExprPtr makeCode(fun,res)
IN CONST struct translation *fun;
IN CONST ExprPtr res;
{
  ExprPtr e;
  Code *c;
  COUNTER i;

  c = (Code *)MALLOC(sizeof(Code));
  c->name = NULL;
  c->prgm = fun->prgm;
  c->last = res;
  c->storage = 0;
  c->fxtypes = 0;
  for (i=0; i<NTAGS; i++)
    c->tags[i] = 0;
  c->size = 0;
  c->flags = 0;
  c->lambda = NewMap();
  c->deco = NIL;
  e = MakeExternal(&codeDesc,(char *)c);
  (void) KEEP(e);
  return e;
}

PUBLIC ExprPtr transToCode(fun,res,last)
IN CONST struct translation *fun;
IN CONST ExprPtr res;
INOUT ExprPtr last;
{
  ExprPtr e;

  e = makeCode(fun,res);
  last = tackApply(Z_CODE,e,last);
  return last;
}

PRIVATE ExprPtr translateDataCon(flags,lbl,env,last)
IN INT flags;
IN CONST ExprPtr lbl;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  struct translation *fun = (struct translation *)alloctmp(sizeof(struct translation));
  /* compile fn x => cons(lbl,x) */
  ExprPtr x,res,code,envv;
  ENTER;

  initTranslation(2,fun); /* 2 args, the input variable, the closure env */
  /* the closure env is not really an environment, but the datatype or exception
     constructor. */
  x = nthArg(0,fun); /* input variable */
  envv = nthArg(1,fun); /* the constructor */
  res = tackApplyBin((flags & 1)?Z_BOX_UNSHARED_CONS:Z_BOX_CONS,envv,x,fun->prgm);
  code = transToCode(fun,res,last);
  /* build C code to do computation; this function should be small enough
     to be inlined in all practical cases. */
  envv = translateStmt(lbl,env,code); /* build constructor (fake environment) */
  last = compileClosure(code,envv,envv);
  QUIT;
  return last;
}

PRIVATE ExprPtr translateUnary(code,e,env,last)
IN INT code;
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  last = translateStmt(e,env,last); /* evaluate body */
  last = tackApply(code,last,last); /* and assign expression to a new variable */
  return last;
}

PRIVATE ExprPtr translateBinary(code,e1,e2,env,last)
IN INT code;
IN CONST ExprPtr e1,e2;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr x1;

  x1 = translateStmt(e1,env,last); /* evaluate first component */
  last = translateStmt(e2,env,x1); /* evaluate second component */
  last = tackApplyBin(code,x1,last,last);
  /* and assign new expression to a new variable */
  return last;
}

PRIVATE ExprPtr translateRevBinary(code,e1,e2,env,last)
IN INT code;
IN CONST ExprPtr e1,e2;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr x1;

  x1 = translateStmt(e1,env,last); /* evaluate first component */
  last = translateStmt(e2,env,x1); /* evaluate second component */
  last = tackApplyBin(code,last,x1,last);
  /* and assign new expression to a new variable */
  return last;
}

PRIVATE ExprPtr translateRecord1(code,e,env,last)
IN INT code;
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr lbl;

  lbl = compileLabel(CAR(e),last); /* compile shape (name of label) */
  last = translateStmt(CDR(e),env,lbl); /* compile value of field */
  last = tackApplyBin(code,lbl,last,last);
  return last;
}

PRIVATE ExprPtr translateTupleGnrl(code,e,env,last)
IN INT code;
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  COUNTER i,n;
  ENTER;

  n = AVAL(e)->n;
  {
    ExprPtr *vec;

    vec = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
    for (i=0; i<n; i++)
      {
	last = translateStmt(AVAL(e)->expr[i],env,last);
	vec[i] = last;
      }
    e = buildGnrlArg(n,vec);
  }
  last = tackApply(code,e,last);
  QUIT;
  return last;
}

PRIVATE ExprPtr translateRecordGnrl(code,e,env,last)
IN INT code;
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  COUNTER i,n;
  ExprPtr shape;
  ENTER;

  n = AVAL(e)->n;
  {
    ExprPtr *vec;

    vec = (ExprPtr *)alloctmp((n+1)*sizeof(ExprPtr));
    shape = ASHAPE(e);
    last = compileShape(ASHAPE(shape),last);
    vec[0] = last;
    /* argument will be a vector of (shape,field1,...,fieldn) */
    for (i=0; i<n; i++)
      {
	last = translateStmt(AVAL(e)->expr[i],env,last);
	vec[IVAL(AVAL(shape)->expr[i])] = last;
      }
    e = buildGnrlArg(n+1,vec);
  }
  last = tackApply(code,e,last);
  QUIT;
  return last;
}

PRIVATE ExprPtr compileKthEnv(level,env,lastp)
IN REGISTER COUNTER level;
IN CONST compEnv *env;
INOUT ExprPtr *lastp;
{
  ExprPtr ee;

  if (env->level==level)
    ee = *lastp = tackCGlobal(I_GLOBENV,(SCF_COMPOSITE | SCF_BOXED | SCF_NOINIT),*lastp);
  else
    {
      for (ee = env->env; level!=0 && ISCONS(ee); level--, ee=CDR(ee));
      if (ISCONS(ee))
	ee = CAR(ee);
      else
	{
	  while (level--!=0)
	    {
	      ee = *lastp = tackApply(Z_ENEXT,ee,*lastp);
	    }
	}
    }
  return ee;
}

PRIVATE ExprPtr compileEnvIndex(val,i,last)
IN CONST ExprPtr val;
IN COUNTER i;
INOUT ExprPtr last;
{
  ENTER;

  if (val->kind==(CHAR)E_CONS && COUNTEDN(CAR(val))==Z_FAKEENV)
    {
      /* a var that we fetch from a fakeenv is better fetched from GLOBENV */
      val = last = tackCGlobal(I_GLOBENV,(SCF_COMPOSITE | SCF_BOXED | SCF_NOINIT),last);
    }
  val = tackApplyBin(Z_VAR,val,KINT(i),last);
  QUIT;
  return val;
}

/* env->env is a list of variables, where each variable
   is a level (a vector of values at run-time);
   the end of the list is a variable abbreviating the
   rest of the environment. */
PRIVATE ExprPtr compileLookup(i,level,env,last)
IN COUNTER i,level;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr ee;

  ee = compileKthEnv(level,env,&last);
  return compileEnvIndex(ee,i,last);
}

PRIVATE ExprPtr getEnv(env)
IN CONST compEnv *env;
{
  return ISCONS(env->env)?CAR(env->env):env->env;
}

PUBLIC ExprPtr compileIf(test,last,elselastp)
IN CONST ExprPtr test;
INOUT ExprPtr last;
OUT ExprPtr *elselastp;
{
  ExprPtr ifthen;

  if (last!=NIL)
    {
      ENTER;

      ifthen = MakePair(E_CONS,test,NIL);
      (void) KEEP(ifthen);
      *elselastp = MakePair(E_IF,ifthen,NIL);
      SETCDR(last,*elselastp); /* put at continuation */
      QUIT;
    }
  else
    {
      ifthen = *elselastp = NIL;
    }
  return ifthen; /* return normal (then) continuation */
}

PRIVATE ExprPtr compileLoop(tag,test,last,elselastp)
IN INT tag;
IN CONST ExprPtr test;
INOUT ExprPtr last;
OUT ExprPtr *elselastp;
{
  ExprPtr ifthen;

  if (last!=NIL)
    {
      ENTER;

      test = KCOUNTED(test,tag);
      ifthen = compileIf(test,last,elselastp);
      QUIT;
    }
  else
    {
      ifthen = *elselastp = NIL;
    }
  return ifthen; /* return normal (then) continuation */
}

PUBLIC ExprPtr mergeContinuations(res1,last1,last2)
IN ExprPtr res1;
INOUT ExprPtr last1,last2;
/* make last1 point to last2, if last2 is a merge point,
   or create a merge point, and make last1 and last2 point
   to it. res1 is the result variable for the computation
   ending at last.
   Return the merge point.
*/
{
  ExprPtr info,old;
  ENTER;

  if (last1!=NIL)
    {
      if (last2->kind!=(CHAR)E_GOTO)
	{
	  info = MakeCounted(E_COUNTED,NewMap(),0L);
	  (void) KEEP(info);
	  last2 = MakePair(E_GOTO,info,NIL);
	}
      else
	info = CAR(last2);
      SETCDR(last1,last2);
      old = REFVAL(info);
      setrefMapPut(last1,res1,info); /* add result variable */
      if (REFVAL(info)!=old)
	{
	  SETCOUNTEDN(info,1+COUNTEDN(info)); /* increment count */
	}
    }
  QUIT;
  return last2;
}

PRIVATE ExprPtr tackMerge(last)
INOUT ExprPtr last;
{
  return mergeContinuations(last,last,NIL);
}

PRIVATE ExprPtr compileRaise(exc,env,last)
IN CONST ExprPtr exc;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  if (env->handlers!=NIL)
    { /* compile a goto when the label for the handler is known */
      /* this is not strictly necessary to do it here,
	 as the optimizer does it, too. */
      (void) mergeContinuations(exc,last,CAR(env->handlers));
      return NIL;
    }
  last = tackApply(Z_RAISE,exc,last);
  return last;
}

PRIVATE ExprPtr compileRaiseImm(idx,env,last)
IN CONST INT idx;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  last = tackCGlobal(idx,(SCF_COMPOSITE | SCF_BOXED | SCF_NOINIT),last);
  last = compileRaise(last,env,last);
  return last;
}

PRIVATE ExprPtr compileCheckNotEqual(val,immval,last,errlastp)
IN CONST ExprPtr val,immval;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp; /* will be a merge node, if non NIL */
{
  ExprPtr isimmval,thenlast;

  isimmval = tackApplyBin(Z_EQUAL,val,immval,last);
  thenlast = compileIf(isimmval,isimmval,&last);
  *errlastp = mergeContinuations(val,thenlast,*errlastp); /* does not match */
  /* val is a dummy, here. */
  return last;
}

PRIVATE ExprPtr compileCheckNotEqualImm(val,immval,last,errlastp)
IN CONST ExprPtr val,immval;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp; /* will be a merge node, if non NIL */
{
  last = compileConstant(immval,last,true);
  last = compileCheckNotEqual(val,last,last,errlastp);
  return last;
}

PRIVATE ExprPtr compileCheckEqual(val,immval,last,errlastp)
IN CONST ExprPtr val,immval;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp; /* will be a merge node, if non NIL */
{
  ExprPtr isimmval,elselast;

  isimmval = tackApplyBin(Z_EQUAL,val,immval,last);
  last = compileIf(isimmval,isimmval,&elselast);
  *errlastp = mergeContinuations(val,elselast,*errlastp); /* does not match */
  /* the value val is a dummy. */
  return last;
}

PRIVATE ExprPtr compileCheckEqualImm(val,immval,last,errlastp)
IN CONST ExprPtr val,immval;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp; /* will be a merge node, if non NIL */
{
  last = compileConstant(immval,last,true);
  last = compileCheckEqual(val,last,last,errlastp);
  return last;
}

PRIVATE ExprPtr compileMergeBin(last1,last2)
INOUT ExprPtr last1,last2;
{
  if (last1==NIL)
    return last2;
  if (last2==NIL)
    return last1;
  return mergeContinuations(last1,last1,mergeContinuations(last2,last2,NIL));
}

PRIVATE INT startTag(kind)
IN INT kind;
{
  switch (kind)
    {
    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
      return Z_START_LISTSWEEP;
    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
      return Z_START_MAPSWEEP;
    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor; never empty */
      return Z_START_SUBMAPSWEEP;
    }
}

PRIVATE INT unboxTag(kind)
IN INT kind;
{
  switch (kind)
    {
    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
      return Z_UNBOX_LISTSWEEP;
    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
      return Z_UNBOX_MAPSWEEP;
    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor; never empty */
      return Z_UNBOX_SUBMAPSWEEP;
    }
}

PRIVATE INT incTag(kind)
IN INT kind;
{
  switch (kind)
    {
    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
      return Z_INC_LISTSWEEP;
    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
      return Z_INC_MAPSWEEP;
    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor; never empty */
      return Z_INC_SUBMAPSWEEP;
    }
}

PRIVATE ExprPtr getCarCdr(tag,last,carp,cdrp)
IN INT tag;
IN CONST ExprPtr last;
OUT ExprPtr *carp,*cdrp;
{
  switch (tag)
    {
    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
      last = *carp = tackApply(Z_CAR,last,last);
      *cdrp = NIL; /* NIL is dummy */
      break;
    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
      *carp = tackApply(Z_CAR,last,last);
      last = *cdrp = tackApply(Z_CDR,last,*carp);
      break;
    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor; never empty */
      *carp = last;
      *cdrp = NIL; /* NIL is dummy */
      break;
    }
  return last;
}

PRIVATE ExprPtr compileRefSet(tag,r,x,last)
IN COUNTER tag;
IN CONST ExprPtr r,x;
INOUT ExprPtr last;
{
  last = tackApplyBin(tag,r,x,last);
  last = tackApply(Z_TUPLE_0,NIL,last);
  return last;
}

PRIVATE ExprPtr compileSweep(tag,list,neutral,combine,combine1,body,
			     data,env,last,nonelastp,errlastp)
IN INT tag;
IN CONST ExprPtr list;
IN compileNeutralFunc neutral;
IN compileCombineFunc combine;
IN compileCombine1Func combine1;
IN compileBodyFunc body;
IN CHAR *data;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *nonelastp,*errlastp;
{
  ExprPtr elselast,outlast,nonelast,car,cdr,val,listr,resr,start;

  /* first, test whether list is NIL */
  outlast = NIL;
  switch (tag)
    {
    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
      last = compileCheckNotEqualImm(list,NIL,last,&nonelast);
      break;
    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
      last = compileCheckNotEqualImm(list,NewMap(),last,&nonelast);
      break;
    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor; never empty */
      last = compileCheckNotEqualImm(list,NewMap(),last,&elselast);
      /* if list is empty, just compute body on empty set */
      {
	/* compute func on empty list */
	val = (*body)(list,NIL,data,env,elselast,&nonelast,errlastp); /* NIL is dummy */
	/* if the predicate was false, then return neutral (at nonelast) */
	/* otherwise, return result */
	if (val!=NIL)
	  val = (*combine1)(env,val,errlastp);
	outlast = mergeContinuations(val,val,outlast);
      }
      break;
    }
  if (nonelast!=NIL)
    {
      nonelast = (*neutral)(env,nonelast,nonelastp,errlastp);
      outlast = mergeContinuations(nonelast,nonelast,outlast);
    }
  /* compute function on first element of list */
  listr = tackApply(startTag(tag),list,last);
  start = tackMerge(listr); /* insert a label, i.e, a (not yet useful) merge node. */
  list = tackApply(unboxTag(tag),listr,start); /* let list = !listr */
  last = getCarCdr(tag,list,&car,&cdr);
  val = (*body)(car,cdr,data,env,last,&nonelast,errlastp);
  /* now, is we are on the nonelast branch (predicate false), then loop
     until we find something satisfying the predicate. */
  if (nonelast!=NIL)
    {
      nonelast = compileLoop(incTag(tag),listr,nonelast,&elselast);
      /* if list is not empty, then loop back */
      (void) mergeContinuations(elselast,elselast,start); /* first elselast is dummy */
      /* if list is empty, return neutral element */
      nonelast = (*neutral)(env,nonelast,nonelastp,errlastp);
      outlast = mergeContinuations(nonelast,nonelast,outlast);
    }
  /* if the predicate is finally true, then join the normal thread of execution (in val) */
  if (val!=NIL)
    val = (*combine1)(env,val,errlastp);
  if (val!=NIL) /* if body fails, the whole iteration fails */
    {
      /* now, compile a while loop */
      resr = tackApply(Z_BOX_LOCAL_REF,val,val);
      start = tackMerge(resr); /* insert a label, i.e, a (not yet useful) merge node. */
      list = tackApply(unboxTag(tag),listr,start); /* let list = !listr */
      last = getCarCdr(tag,list,&car,&cdr);
      last = (*body)(car,cdr,data,env,last,&nonelast,errlastp);
      /* if predicate is true, combine last with !resr */
      if (last!=NIL)
	{
	  val = tackApply(Z_UNBOX_LOCAL_REF,resr,last);
	  last = (*combine)(last,env,val,errlastp);
	  last = compileRefSet(Z_LOCAL_REF_SET,resr,last,last); /* store new result */
	}
      last = compileMergeBin(last,nonelast);
      last = compileLoop(incTag(tag),listr,last,&elselast); /* increment list */
      (void) mergeContinuations(elselast,elselast,start); /* and loop back; first elselast is dummy */
      /* otherwise, get result */
      last = tackApply(Z_UNBOX_LOCAL_REF,resr,last);
      outlast = mergeContinuations(last,last,outlast);
    }
  /* and finish: */
  return outlast;
}

#ifdef HAS_PROTOTYPES
PRIVATE ExprPtr compileHasTag(IN CONST CHAR tag,
			      IN CONST ExprPtr val,
			      INOUT ExprPtr last,
			      OUT ExprPtr *elselastp)
#else
PRIVATE ExprPtr compileHasTag(tag,val,last,elselastp)
IN CONST CHAR tag;
IN CONST ExprPtr val;
INOUT ExprPtr last;
OUT ExprPtr *elselastp;
#endif
{
  ENTER;

  val = tackApplyBin(Z_HAS_TAG,val,KINT(tag),last); /* test val->kind==(CHAR)tag */
  last = compileIf(val,val,elselastp);
  QUIT;
  return last;
}

PRIVATE ExprPtr compilePattern P((IN CONST ExprPtr pat,IN CONST ExprPtr val,
				  IN CONST compEnv *env,INOUT ExprPtr last,
				  INOUT ExprPtr *choicelastp,
				  INOUT ExprPtr *errlastp));

PRIVATE ExprPtr compileVectorMatch P((IN CONST ExprPtr pat,IN CONST ExprPtr val,
				      IN COUNTER idx,IN CONST compEnv *env,
				      INOUT ExprPtr last,INOUT ExprPtr *choicelastp,
				      INOUT ExprPtr *errlastp));

DEF_PRIVATE_ACTION(compilevectormatch)
{
  struct compilevectormatch *data;

  data = (struct compilevectormatch *)env;
  return compileVectorMatch(data->pat,data->val,data->idx,data->env,data->last,
			    &data->choicelast,&data->errlast);
}

PRIVATE ExprPtr compileVectorMatch1(pat,val,idx,env,last,choicelastp,errlastp)
IN CONST ExprPtr pat,val;
IN COUNTER idx;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *choicelastp,*errlastp;
{
  struct compilevectormatch *data = (struct compilevectormatch *)
    alloctmp(sizeof(struct compilevectormatch));
  ExprPtr res;

  data->pat = pat;
  data->val = val;
  data->idx = idx;
  data->env = env;
  data->last = last;
  data->choicelast = *choicelastp;
  data->errlast = *errlastp;
  res = CallCC(ACTION_NAME(compilevectormatch),(ExprPtr)data,true);
  *choicelastp = data->choicelast;
  *errlastp = data->errlast;
  return res;
}

PRIVATE ExprPtr compileVectorMatch(pat,val,idx,env,last,choicelastp,errlastp)
IN CONST ExprPtr pat,val; /* pat is a pattern matching GNRL_TUPLE data */
IN COUNTER idx;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *choicelastp,*errlastp; /* will be a merge node, if non NIL */
{
  if (STKFULL(pat))
    return compileVectorMatch1(pat,val,idx,env,last,choicelastp,errlastp);
again:
  switch (pat->kind)
    {
    case E_PAT_TUPLE_GNRL_CONS:
    case E_PAT_RECORD_GNRL_CONS:
      last = tackApplyBin(Z_UNBOX_TUPLE_NTH,val,KINT(idx+1),last);
      last = compilePattern(CAR(pat),last,env,last,choicelastp,errlastp);
      pat = CDR(pat);
      idx++;
      goto again;
    case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_REVRECORD_GNRL_CONS:
      last = compileVectorMatch(CDR(pat),val,idx+1,env,last,choicelastp,errlastp);
      last = tackApplyBin(Z_UNBOX_TUPLE_NTH,val,KINT(idx+1),last);
      last = compilePattern(CAR(pat),last,env,last,choicelastp,errlastp);
      break;
    case E_PATCHOICE:
      {
	ExprPtr faillast;
	ExprPtr saveletdelay;
	Boolean newenv;

	faillast = NIL;
	last = compileVectorMatch(CAR(pat),val,idx,env,last,choicelastp,&faillast);
	/* now, if match fails, we try to match the remaining patterns */
	/* a choice is (primary pattern,(secondary pattern,secondary body)) */
	pat = CDR(pat);
	saveletdelay = REFVAL(env->letdelayref);
	(void) KEEP(saveletdelay);
	faillast = compilePattern(CAR(pat),val,env,faillast,choicelastp,errlastp);
	newenv = (env->env!=NIL && CAR(env->env)==NIL); /* happens in polyfun declarations,
							   where we want to postpone environment
							   creation as much as possible. */
	if (newenv)
	  {
	    faillast = tackAssign(env->letconsenv,faillast);
	    SETCAR(env->env,faillast);
	  }
	faillast = dumpLetStores(REFVAL(env->letdelayref),faillast,faillast);
	faillast = translateStmt(CDR(pat),env,faillast);
	/* faillast is now a return statement */
	*choicelastp = mergeContinuations(faillast,faillast,*choicelastp);
	SETREFVAL(env->letdelayref,saveletdelay);
	if (newenv)
	  {
	    SETCAR(env->env,NIL);
	  }
	break;
      }
    case E_DEBUGNAME:
      pat = DNPAT(pat);
      goto again;
    case E_PAT_ANY: /* ellipsis */
      break;
    default: /* should be an end-of-tuple (or end-of-vector) */
      {
	ExprPtr n;

	last = tackApply(Z_GET_VECTOR_LEN,val,last);
	n = tackApply(Z_UNBOXED_INT,KINT(idx),last); /* create an unboxed integer */
	last = compileCheckEqual(last,n,n,errlastp); /* check we are at the end */
	break;
      }
    }
  return last;
}

PRIVATE ExprPtr compileTupleMatch(pat,val,env,last,choicelastp,errlastp)
IN CONST ExprPtr pat,val; /* pat is a pattern matching GNRL_TUPLE data */
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *choicelastp,*errlastp; /* will be a merge node, if non NIL */
{
  ExprPtr aux,elselast,car,cdr;

  last = compileHasTag((CHAR)E_TUPLE_2,val,last,&elselast);
  /* if val->kind==(CHAR)E_TUPLE_2: */
  car = tackApply(Z_CAR,val,last); /* compute CAR(val) */
  aux = CDR(pat);
  if (aux==patany) /* if has an ellipsis */
    last = compilePattern(CAR(pat),car,env,car,choicelastp,errlastp); /* match CAR(val) */
  else
    if (pat->kind==(CHAR)E_PAT_TUPLE_GNRL_CONS)
      {
	last = compilePattern(CAR(pat),car,env,car,choicelastp,errlastp); /* match CAR(val) */
	cdr = tackApply(Z_CDR,val,last); /* compute CDR(val) */
	last = compilePattern(CAR(aux),cdr,env,cdr,choicelastp,errlastp); /* match it */
      }
    else
      {
	cdr = tackApply(Z_CDR,val,last); /* compute CDR(val) */
	last = compilePattern(CAR(aux),cdr,env,cdr,choicelastp,errlastp); /* match it */
	last = compilePattern(CAR(pat),car,env,last,choicelastp,errlastp); /* match CAR(val) */
      }
  /* if val->kind!=(CHAR)E_TUPLE_2: */
  elselast = compileVectorMatch(pat,val,0,env,elselast,choicelastp,errlastp);
  last = compileMergeBin(last,elselast);
  return last;
}

/* !!! serious bug (in compileTupleMatch, for instance):
   in both branches of the conditional, we have to match values against
   patterns, so that there will be several definition points for the
   same HimML variables; this conflicts with letdelayref, which can only
   remember one definition point per variable...
*/

PRIVATE ExprPtr compileRecordMatch(pat,val,env,last,choicelastp,errlastp)
IN CONST ExprPtr pat,val;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *choicelastp,*errlastp; /* will be a merge node, if non NIL */
{
  /* pat is a list of E_PAT_RECORD_GNRL_CONS or E_PAT_REVRECORD_GNRL_CONS.
     It is terminated by E_RECORD_0 (fixed size records), or by
     E_PAT_ANY (ellipsis).
  */
  ExprPtr elselast,cdr;
  ENTER;

  /* val may be a record_gnrl, or a record_1; in the last case,
     CDR(pat) must be an ellipsis. */
  last = compileHasTag((CHAR)E_RECORD_1,val,last,&elselast);
  /* if val->kind==(CHAR)E_RECORD_1, then check that CDR(pat)==patany and
     that CAR(pat) matches CDR(val) */
  if (CDR(pat)==patany)
    {
      /* case val->kind==(CHAR)E_RECORD_1: */
      cdr = tackApply(Z_CDR,val,last); /* compute cdr of value */
      last = compilePattern(CAR(pat),cdr,env,last,choicelastp,errlastp);
      /* otherwise, match as a vector: */
      elselast = compileVectorMatch(pat,val,0,env,elselast,choicelastp,errlastp);
      last = compileMergeBin(last,elselast);
    }
  else
    {
      /* in the case val->kind==(CHAR)E_RECORD_1, matching fails: */
      *errlastp = mergeContinuations(val,last,*errlastp); /* val is a dummy. */
      /* otherwise, match as a vector: */
      last = compileVectorMatch(pat,val,0,env,elselast,choicelastp,errlastp);
    }
  QUIT;
  return last;
}

PRIVATE ExprPtr compileCheckTag(patbits,val,env,last,errlastp)
IN CONST ExprPtr patbits;
IN CONST ExprPtr val;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{
  COUNTER i;
  ExprPtr thenlast,elselast;

  elselast = last;
  last = NIL;
  for (i=1; i<E_UNUSED; i++)
    {
      if (PATCHECKBITS(patbits)[i>>3] & (1 << (i & 7)))
	{ /* there must be at least one bit here */
	  thenlast = compileHasTag((CHAR)i,val,elselast,&elselast);
	  last = mergeContinuations(val,thenlast,last); /* val is a dummy. */
	}
    }
  *errlastp = mergeContinuations(val,elselast,*errlastp); /* val is a dummy. */
  return last;
}

PRIVATE ExprPtr compileCheckPackPat(pat,val,env,last,errlastp)
IN CONST ExprPtr pat,val;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp; /* will be a merge node, if non NIL */
{
  ExprPtr packtypeval,packtypepat,thenlast;

  packtypeval = tackApply(Z_CDR,val,last); /* PACKTYPE is same as CDR */
  packtypepat = compileConstant(PACKTYPE(pat),packtypeval,true);
  last = tackApplyBin(Z_TYPE_MATCH,packtypeval,packtypepat,packtypepat);
  thenlast = compileIf(last,last,&last);
  *errlastp = mergeContinuations(val,thenlast,*errlastp); /* val is a dummy */
  return last;
}

PRIVATE ExprPtr compileCheckCardEq(n,val,env,last,errlastp)
IN COUNTER n;
IN CONST ExprPtr val;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp; /* will be a merge node, if non NIL */
{
  ExprPtr elselast,zero;

  switch (n)
    {
    case 0:
      last = compileCheckEqualImm(val,NewMap(),last,errlastp);
      break;
    case 1:
      last = compileHasTag((CHAR)E_MAPLET,val,last,&elselast);
      *errlastp = mergeContinuations(val,elselast,*errlastp); /* val is a dummy. */
      break;
    default:
      last = tackApplyBin(Z_MAP_CARD_DEC,val,KINT(n),last);
      /* check MapCardEq(val,COUNTEDN(pat)), i.e, that last==0 */
      zero = tackApply(Z_UNBOXED_INT,izero,last);
      last = compileCheckEqual(last,zero,zero,errlastp);
      break;
    }
  return last;
}

PRIVATE ExprPtr compileCheckCardGeq(n,val,env,last,errlastp)
IN COUNTER n;
IN CONST ExprPtr val;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp; /* will be a merge node, if non NIL */
{
  ExprPtr elselast,zero;

  switch (n)
    {
    case 0:
      break;
    case 1:
      last = compileCheckNotEqualImm(val,NewMap(),last,errlastp);
      break;
    case 2:
      last = compileHasTag((CHAR)E_MAP,val,last,&elselast);
      *errlastp = mergeContinuations(val,elselast,*errlastp); /* val is a dummy. */
      break;
    default:
      last = tackApplyBin(Z_MAP_CARD_DEC,val,KINT(n),last);
      /* check MapCardGeq(val,COUNTEDN(pat)), i.e, that last<=0 */
      zero = tackApply(Z_UNBOXED_INT,izero,last);
      last = tackApplyBin(Z_INT_LEQ,last,zero,zero);
      last = compileIf(last,last,&elselast);
      *errlastp = mergeContinuations(COUNTEDVAL(val),elselast,*errlastp);
      /* COUNTEDVAL(val), i.e, the original val, is a dummy. */
      break;
    }
  return last;
}

PRIVATE ExprPtr compileSelectGnrl1(shape,label,val,env,last)
IN CONST ExprPtr shape,label;
IN CONST ExprPtr val;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr idx;

  if (label->kind==(CHAR)E_INT)
    /* if label is precomputed */
    idx = tackApply(Z_UNBOXED_INT,KINT(IVAL(label)-1),last);
  else
    {
      label = compileLabel(label,shape);
      idx = tackApplyBin(Z_MAPGET,label,shape,label);
      /* the result of a Z_MAPGET call is a virtual couple (y,found),
	 as in y = MapGet(x,map,&found) in C; Z_GET_MAPGET_Y gets
	 the y value from it, and Z_GET_MAPGET_FOUND the unboxed boolean
	 found. */
      idx = tackApply(Z_GET_MAPGET_Y,idx,idx);
      idx = tackApply(Z_UNBOX_INT,idx,idx);
    }
  val = tackApplyBin(Z_TUPLE_SUB_NTH,val,idx,idx); /* AVAL(val)->expr[index-1] */
  return val;
}

PRIVATE ExprPtr compileSelectGnrl(label,val,env,last)
IN CONST ExprPtr label;
IN CONST ExprPtr val;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr shape;

  shape = (label->kind==(CHAR)E_INT)?NIL:tackApply(Z_GET_SHAPE,val,last);
  /* if label is a precomputed position, we do not need to compute the shape of val */
  last = compileSelectGnrl1(shape,label,val,env,last);
  return last;
}

PRIVATE ExprPtr compileRecordMapAddMatch(pat,val,env,last,choicelastp,errlastp)
IN CONST ExprPtr pat,val;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *choicelastp,*errlastp; /* will be a merge node, if non NIL */
{
  /* pat is a E_PAT_RECORD_MAPADD list, terminated by an ellipsis
     (coded as E_PAT_ANY), with possible E_PATCHOICEs inside.
     It cannot be terminated by an end-of-record, since in that
     case the compiler would be able to determine the type and
     assign indexes to all the fields, so that we would have an
     E_PAT_RECORD_GNRL_CONS or E_PAT_REVRECORD_GNRL_CONS pattern.
     This list contains couples (label,pattern) to be matched.
  */
  ExprPtr thenlast,shape;

  thenlast = NIL;
  if (CDR(pat)==patany) /* pattern |[a = <p>,...]|, may match a E_RECORD_1 */
    {
      ExprPtr cdr;

      thenlast = compileHasTag((CHAR)E_RECORD_1,val,last,&last);
      /* if val is a E_RECORD_1, match <p> with the field CDR(val) of val: */
      cdr = tackApply(Z_CDR,val,thenlast);
      thenlast = compilePattern(CDR(CAR(pat)),cdr,env,cdr,choicelastp,errlastp);
      /* otherwise, find the field in the E_RECORD_GNRL val, and match it,
	 as though CDR(pat)!=patany */
      thenlast = mergeContinuations(val,thenlast,NIL); /* val is a dummy. */
    }
  /* val may be a record_gnrl, or a record_1; in the last case,
     CDR(pat) must be an ellipsis, so now we assume val is a record_gnrl. */
  shape = tackApply(Z_GET_SHAPE,val,last);
again:
  switch (pat->kind)
    {
    case E_PAT_RECORD_MAPADD:
      last = compileSelectGnrl1(shape,CAR(CAR(pat)),val,env,last);
      /* select the right entry in val */
      last = compilePattern(CDR(CAR(pat)),last,env,last,choicelastp,errlastp);
      /* and match it */
      pat = CDR(pat);
      /* then match the rest */
      goto again;
    case E_PATCHOICE:
      {
	ExprPtr elselast;
	ExprPtr saveletdelay;
	Boolean newenv;

	elselast = NIL;
	last = compileRecordMapAddMatch(CAR(pat),val,env,last,choicelastp,&elselast);
	/* now, if match fails, we try to match the remaining patterns */
	/* a choice is (primary pattern,(secondary pattern,secondary body)) */
	pat = CDR(pat);
	saveletdelay = REFVAL(env->letdelayref);
	(void) KEEP(saveletdelay);
	elselast = compileRecordMapAddMatch(CAR(pat),val,env,elselast,choicelastp,errlastp);
	newenv = (env->env!=NIL && CAR(env->env)==NIL); /* happens in polyfun declarations,
							   where we want to postpone environment
							   creation as much as possible. */
	if (newenv)
	  {
	    elselast = tackAssign(env->letconsenv,elselast);
	    SETCAR(env->env,elselast);
	  }
	elselast = dumpLetStores(REFVAL(env->letdelayref),elselast,elselast);
	elselast = translateStmt(CDR(pat),env,elselast);
	/* elselast is now a return statement */
	*choicelastp = mergeContinuations(elselast,elselast,*choicelastp);
	SETREFVAL(env->letdelayref,saveletdelay);
	if (newenv)
	  {
	    SETCAR(env->env,NIL);
	  }
	break;
      }
    case E_DEBUGNAME:
      pat = DNPAT(pat);
      goto again;
    default: /* should be an ellipsis */
      break;
    }
  if (thenlast!=NIL)
    last = mergeContinuations(last,last,thenlast); /* the first last is a dummy. */
  return last;
}

PRIVATE ExprPtr compileMapAddPatternCst(cstx,pat,val,env,last,choicelastp,errlastp)
IN CONST ExprPtr cstx;
IN CONST ExprPtr pat;
IN CONST ExprPtr val;
INOUT compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *choicelastp,*errlastp;
{
  ExprPtr elselast,y,found;

  cstx = compileConstant(cstx,last,true);
  y = tackApplyBin(Z_MAPGET,cstx,val,cstx);
  found = tackApply(Z_GET_MAPGET_FOUND,y,y);
  last = compileIf(found,found,&elselast);
  *errlastp = mergeContinuations(val,elselast,*errlastp); /* val is a dummy. */
  last = tackApply(Z_GET_MAPGET_Y,y,last);
  last = compilePattern(CDR(PATMAPADDXY(pat)),last,env,last,choicelastp,errlastp);
  val = tackApplyBin(Z_MAPREMOVE,cstx,val,last);
  pat = PATMAPADDREST(pat);
  last = compilePattern(pat,val,env,last,choicelastp,errlastp);
  return last;
}

/*
PRIVATE Boolean hasPatChoice(pat)
IN CONST ExprPtr pat;
{
again:
  switch (pat->kind)
    {
    case E_CONS:
    case E_UNSHARED_CONS:
    case E_TUPLE_2:
    case E_UNSHARED_TUPLE_2: 
    case E_PAT_TUPLE_GNRL_CONS:
    case E_PAT_RECORD_GNRL_CONS:
    case E_PAT_AS:
    case E_PAT_MAPADD:
    case E_APPL:
      if (hasPatChoice(CAR(pat)))
	return true;
      pat = CDR(pat);
      goto again;
    case E_PAT_REVCONS:
    case E_PAT_REVTUPLE_2:
    case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_REVRECORD_GNRL_CONS:
      if (hasPatChoice(CDR(pat)))
	return true;
      pat = CAR(pat);
      goto again;
    case E_PAT_CONS:
    case E_TUPLE_1:
    case E_UNSHARED_TUPLE_1:
    case E_REF:
      pat = REFVAL(pat);
      goto again;
    case E_RECORD_1:
    case E_UNSHARED_RECORD_1:
      pat = CDR(pat);
      goto again;
    case E_PAT_CHECK:
      pat = PATCHECKVAL(pat);
      goto again;
    case E_PAT_PACK:
      pat = PACKVAL(pat);
      goto again;
    case E_PAT_CARD_EQ:
    case E_PAT_CARD_GEQ:
      pat = COUNTEDVAL(pat);
      goto again;
    case E_PATCHOICE:
      return true;
    case E_DEBUGNAME:
      pat = DNPAT(pat);
      goto again;
    case E_PAT_RECORD_MAPADD:
      if (hasPatChoice(CDR(CAR(pat))))
	return true;
      pat = CDR(pat);
      goto again;
    }
  return false;
}
*/

PRIVATE ExprPtr compileNewMap(env,last)
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  last = tackApply(Z_EMPTY,NIL,last);
  return last;
}

PRIVATE ExprPtr compileNeutralNone(env,last,nonelastp,errlastp)
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *nonelastp,*errlastp;
{
  return last;
}

#define compileNeutralExists compileNeutralNone
PRIVATE ExprPtr compileCombine1Exists(env,last,errlastp)
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{
  ExprPtr elselast;

  last = tackApply(Z_UNBOX_BOOL,last,last);
  last = compileIf(last,last,&elselast);
  *errlastp = mergeContinuations(last,last,*errlastp); /* first last is dummy */
  return elselast;
}
PRIVATE ExprPtr compileCombineExists(e1,env,last,errlastp)
IN CONST ExprPtr e1;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{
  ExprPtr elselast;

  last = tackApply(Z_UNBOX_BOOL,e1,last);
  last = compileIf(last,last,&elselast);
  *errlastp = mergeContinuations(last,last,*errlastp); /* first last is dummy */
  return elselast;
}
#define compilePromoteExists compileIdentity

#define compileNeutralAll compileNeutralNone
PRIVATE ExprPtr compileCombine1All(env,last,errlastp)
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{
  ExprPtr elselast;

  last = tackApply(Z_UNBOX_BOOL,last,last);
  last = compileIf(last,last,&elselast);
  *errlastp = mergeContinuations(elselast,elselast,*errlastp); /* first last is dummy */
  return last;
}
PRIVATE ExprPtr compileCombineAll(e1,env,last,errlastp)
IN CONST ExprPtr e1;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{
  ExprPtr elselast;

  last = tackApply(Z_UNBOX_BOOL,e1,last);
  last = compileIf(last,last,&elselast);
  *errlastp = mergeContinuations(elselast,elselast,*errlastp); /* first last is dummy */
  return last;
}
#define compilePromoteAll compileIdentity

PRIVATE ExprPtr compileNeutralSome(env,last,nonelastp,errlastp)
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *nonelastp,*errlastp;
{
  last = compileString(none,last);
  return last;
}
PRIVATE ExprPtr compileCombine1Some(env,last,errlastp)
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{
  *errlastp = mergeContinuations(last,last,*errlastp);
  return NIL;
}
PRIVATE ExprPtr compileCombineSome(e1,env,last,errlastp)
IN CONST ExprPtr e1;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{
  /* produces code that is never really coded, since combine1
     exits from the loop first. */
  return NIL;
}
PRIVATE ExprPtr compilePromoteSome(env,last)
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr res,somestr;

  res = last;
  somestr = compileString(some,last);
  last = tackApplyBin(Z_BOX_CONS,somestr,res,somestr);
  return last;
}

PRIVATE ExprPtr compilePatMapAddBody(x,y,_data,env,last,nonelastp,errlastp)
IN CONST ExprPtr x;
IN CONST ExprPtr y;
IN CHAR *_data;
IN CONST compEnv *env;
INOUT ExprPtr last;
OUT ExprPtr *nonelastp;
INOUT ExprPtr *errlastp;
{
  struct patmapadd *data;
  ExprPtr maplet,errlast,val;
  ENTER;
  /* assuming pat = {x => y} ++ rest,
     compile
     if match (x,xx) && match (y,yy) (* may be reversed if E_PAT_REVCONS *)
       then match (rest,mapremove(xx,data))
     else false
     */

  data = (struct patmapadd *)_data;
  maplet = PATMAPADDXY(data->pat);
  errlast = *nonelastp = NIL;
  switch (maplet->kind)
    {
    case E_CONS:
      /*case E_UNSHARED_CONS:*/
      last = compilePattern(CAR(maplet),x,env,last,&data->choicelast,&errlast);
      last = compilePattern(CDR(maplet),y,env,last,&data->choicelast,&errlast);
      break;
    case E_PAT_REVCONS:
      last = compilePattern(CDR(maplet),y,env,last,&data->choicelast,&errlast);
      last = compilePattern(CAR(maplet),x,env,last,&data->choicelast,&errlast);
      break;
    }
  val = tackApplyBin(Z_MAPREMOVE,x,data->map,last);
  last = compilePattern(PATMAPADDREST(data->pat),val,env,val,&data->choicelast,&errlast);
  if (last!=NIL)
    *errlastp = mergeContinuations(last,last,*errlastp); /* first last is dummy */
  return errlast;
}

PRIVATE Boolean cstPattern(pat,val)
IN CONST ExprPtr pat;
OUT ExprPtr *val;
{
  switch (pat->kind)
    {
    case E_NIL: case E_TUPLE_0: case E_RECORD_0: case E_EMPTY:
    case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
      *val = pat;
      return true;
    case E_QUOTE:
      *val = REFVAL(pat);
      return true;
    default:
      return false;
    }
}

PRIVATE ExprPtr compileMapAddPatternGnrl(pat,val,env,last,choicelastp,errlastp)
IN CONST ExprPtr pat,val;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *choicelastp,*errlastp; /* will be a merge node, if non NIL */
{
  /* now, pat = {x => y} ++ rest: */
  ExprPtr nonelast,errlast;
  struct patmapadd *data = (struct patmapadd *)alloctmp(sizeof(struct patmapadd));

  nonelast = errlast = NIL;
  data->pat = pat;
  data->map = val;
  data->choicelast = *choicelastp;
  last = compileSweep(E_PAT_REVRECORD_GNRL_CONS, /* in map */
		      val,compileNeutralExists,compileCombineExists,
		      compileCombine1Exists,compilePatMapAddBody,
		      (CHAR *)data,env,last,&nonelast,&errlast);
  *choicelastp = data->choicelast;
  if (last!=NIL)
    *errlastp = mergeContinuations(last,last,*errlastp); /* first last is dummy */
  return errlast;
}

MAPSWEEPFUNC(do_dumpletstores)
{
  struct dumpletstores *data;

  data = MAPDATA(struct dumpletstores *);
  data->last = tackApplyTer(Z_LET,data->env,y,x,data->last);
  return false;
}

PRIVATE ExprPtr dumpLetStores(set,env,last)
IN CONST MapPtr set;
IN CONST ExprPtr env;
INOUT ExprPtr last;
{
  struct dumpletstores *data = (struct dumpletstores *)alloctmp(sizeof(struct dumpletstores));

  data->last = last;
  data->env = env;
  (void) MapSweep(set,do_dumpletstores,data);
  return data->last;
}

DEF_PRIVATE_ACTION(compilepattern)
{
  struct compilepattern *data;

  data = (struct compilepattern *)env;
  return compilePattern(data->pat,data->val,data->env,data->last,&data->choicelast,&data->errlast);
}

PRIVATE ExprPtr compilePattern1(pat,val,env,last,choicelastp,errlastp)
IN CONST ExprPtr pat,val;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *choicelastp,*errlastp;
{
  struct compilepattern *data = (struct compilepattern *)alloctmp(sizeof(struct compilepattern));
  ExprPtr res;

  data->pat = pat;
  data->val = val;
  data->env = env;
  data->last = last;
  data->choicelast = *choicelastp;
  data->errlast = *errlastp;
  res = CallCC(ACTION_NAME(compilepattern),(ExprPtr)data,true);
  *choicelastp = data->choicelast;
  *errlastp = data->errlast;
  return res;
}

PRIVATE ExprPtr compilePattern(pat,val,env,last,choicelastp,errlastp)
IN CONST ExprPtr pat,val;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *choicelastp,*errlastp; /* will be a merge node, if non NIL */
{
  if (STKFULL(pat))
    return compilePattern1(pat,val,env,last,choicelastp,errlastp);
again:
  switch (pat->kind)
    {
    case E_DEBUGNAME:
      pat = DNPAT(pat);
      goto again;
    case E_VARINTRO:
      {
	ExprPtr rank;

	rank = KINT((INT)VARRANK(pat));
	/* now, remember for later the binding to do */
	setrefMapPut(rank,val,env->letdelayref);
	/*VP->expr[VARRANK(pat)] = val;*/
	/* now, we also record in env->funs the fact that val was a function, if it was.
	   This means that we test whether val is a Z_BOX_CLOSURE assignment node, with
	   the environment equal to the current environment (see compilation of E_POLYFUN
	   nodes in translateStmt()). */
	if (val->kind==(CHAR)E_CONS &&
	    CAR(val)->kind==(CHAR)E_COUNTED && COUNTEDN(CAR(val))==Z_BOX_CLOSURE &&
	    CDR(COUNTEDVAL(CAR(val)))==getEnv(env))
	  setcarMapPut(rank,CAR(COUNTEDVAL(CAR(val))),env->funs);
	/* record the Z_CODE part */
      }
      break;
    case E_CONS:
    case E_UNSHARED_CONS:
      last = compileCheckNotEqualImm(val,NIL,last,errlastp); /* check val!=NIL */
      /*fallthrough*/
    case E_MAPLET:
dec_cons:
      {
	ExprPtr car,cdr;

	/* val is a non-nil list, and we can decompose it */
	car = tackApply(Z_CAR,val,last); /* get the car */
	last = compilePattern(CAR(pat),car,env,car,choicelastp,errlastp); /* and match it */
	cdr = tackApply(Z_CDR,val,last); /* get the cdr */
	last = compilePattern(CDR(pat),cdr,env,car,choicelastp,errlastp); /* and match it */
	break;
      }
    case E_PAT_REVCONS:
      last = compileCheckNotEqualImm(val,NIL,last,errlastp); /* check val!=NIL */
dec_rev_cons:
      {
	ExprPtr car,cdr;

	/* val is a non-nil list, and we can decompose it */
	cdr = tackApply(Z_CDR,val,last); /* get the cdr */
	last = compilePattern(CDR(pat),cdr,env,cdr,choicelastp,errlastp); /* and match it */
	car = tackApply(Z_CAR,val,last); /* get the car */
	last = compilePattern(CAR(pat),car,env,car,choicelastp,errlastp); /* and match it */
	break;
      }
    case E_APPL: /* this is only used for exceptions applied to arguments */
      {
	ExprPtr thenlast,exc;

	thenlast = compileHasTag((CHAR)E_CONS,val,last,&last);
	*errlastp = mergeContinuations(last,last,*errlastp); /* first last is dummy */
	exc = CAR(pat);
	/* now the name of the exception as recorded in the pattern
	   is given by a E_VARREF that points to an entry in the environment
	   where we find a compiled closure with the exception in the
	   QENV part. */
	last = compileLookup(VARRANK(exc),VARLEVEL(exc),env,last);
	last = tackApply(Z_UNBOX_CLOSURE_ENV,last,last);
	thenlast = tackApply(Z_CAR,val,last);
	last = compileCheckEqual(thenlast,last,thenlast,errlastp);
	last = tackApply(Z_CDR,val,last);
	last = compilePattern(CDR(pat),last,env,last,choicelastp,errlastp);
	break;
      }
    case E_PAT_CONS:
      last = compileCheckNotEqualImm(val,NIL,last,errlastp); /* check val!=NIL */
      {
	ExprPtr car,cdr;

	/* val is a non-nil list, and we can decompose it */
	car = tackApply(Z_CAR,val,last); /* get the car */
	cdr = tackApply(Z_CDR,val,car); /* get the cdr */
	val = tackApplyBin(Z_BOX_TUPLE_2,car,cdr,cdr); /* build couple */
	last = compilePattern(REFVAL(pat),val,env,val,choicelastp,errlastp); /* and match it */
	break;
      }
    case E_NIL:
      last = compileCheckEqualImm(val,NIL,last,errlastp);
      break;
    case E_TUPLE_0:
      /*last = compileCheckEqualImm(val,tuple0,last,errlastp);*/
      /* because of type checking, tags always match */
      break;
    case E_TUPLE_1:
    case E_UNSHARED_TUPLE_1:
      last = tackApply(Z_UNBOX_REF,val,last); /* dereference ref value */
      last = compilePattern(REFVAL(pat),last,env,last,choicelastp,errlastp);
      break;
    case E_TUPLE_2:
    case E_UNSHARED_TUPLE_2:
      goto dec_cons;
    case E_PAT_TUPLE_GNRL_CONS:
    case E_PAT_REVTUPLE_GNRL_CONS:
      last = compileTupleMatch(pat,val,env,last,choicelastp,errlastp);
      break;
    case E_PAT_REVTUPLE_2:
      goto dec_rev_cons;
    case E_RECORD_0:
      /*last = compileCheckEqualImm(val,record0,last,errlastp);*/
      /* because of type checking, tags always match */
      break;
    case E_RECORD_1:
    case E_UNSHARED_RECORD_1:
      pat = CDR(pat);
      last = val = tackApply(Z_CDR,val,last); /* get the cdr */
      goto again;
    case E_PAT_RECORD_GNRL_CONS: /* records with known indexes */
    case E_PAT_REVRECORD_GNRL_CONS: /* records with known indexes */
      last = compileRecordMatch(pat,val,env,last,choicelastp,errlastp);
      break;
    case E_EMPTY: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
      last = compileCheckEqualImm(val,pat,last,errlastp);
      break;
    case E_REF:
      pat = REFVAL(pat);
      last = val = tackApply(Z_UNBOX_REF,val,last);
      goto again;
    case E_QUOTE:
      last = compileCheckEqualImm(val,REFVAL(pat),last,errlastp);
      break;
    case E_VARREF:
      last = compileLookup(VARRANK(pat),VARLEVEL(pat),env,last);
      last = compileCheckEqual(val,last,last,errlastp);
      break;
    case E_PAT_ANY:
      break;
    case E_PAT_AS:
      last = compilePattern(PATAS2(pat),val,env,last,choicelastp,errlastp);
      pat = PATAS1(pat);
      goto again;
    case E_PAT_CHECK:
      last = compileCheckTag(pat,val,env,last,errlastp);
      pat = PATCHECKVAL(pat);
      goto again;
    case E_PAT_PACK:
      last = compileCheckPackPat(pat,val,env,last,errlastp);
dec_car:
      pat = CAR(pat); /* PACKVAL is same as CAR */
      val = tackApply(Z_CAR,val,last); /* PACKVAL is same as CAR */
      goto again;
    case E_PAT_CARD_EQ:
      last = compileCheckCardEq(COUNTEDN(pat),val,env,last,errlastp);
      pat = COUNTEDVAL(pat);
      goto again;
    case E_PAT_CARD_GEQ:
      last = compileCheckCardGeq(COUNTEDN(pat),val,env,last,errlastp);
      pat = COUNTEDVAL(pat);
      goto again;
    case E_PATCHOICE:
      {
	ExprPtr elselast;
	ExprPtr saveletdelay;
	Boolean newenv;

	elselast = NIL;
	last = compilePattern(CAR(pat),val,env,last,choicelastp,&elselast);
	/* a choice is (primary pattern,(secondary pattern,secondary body)) */
	pat = CDR(pat);
	saveletdelay = REFVAL(env->letdelayref);
	(void) KEEP(saveletdelay);
	elselast = compilePattern(CAR(pat),val,env,elselast,choicelastp,errlastp);
	newenv = (env->env!=NIL && CAR(env->env)==NIL); /* happens in polyfun declarations,
							   where we want to postpone environment
							   creation as much as possible. */
	if (newenv)
	  {
	    elselast = tackAssign(env->letconsenv,elselast);
	    SETCAR(env->env,elselast);
	  }
	elselast = dumpLetStores(REFVAL(env->letdelayref),elselast,elselast);
	elselast = translateStmt(CDR(pat),env,elselast);
	*choicelastp = mergeContinuations(elselast,elselast,*choicelastp);
	SETREFVAL(env->letdelayref,saveletdelay);
	if (newenv)
	  {
	    SETCAR(env->env,NIL);
	  }
      }
      break;
    case E_PAT_RECORD_MAPADD:
      last = compileRecordMapAddMatch(pat,val,env,last,choicelastp,errlastp);
      break;
    case E_PAT_MAPADD:
      {
	Boolean cst;
	ExprPtr patval;

	last = compileCheckNotEqualImm(val,NewMap(),last,errlastp);
	cst = cstPattern(CAR(PATMAPADDXY(pat)),env->constpatmemoref,&patval);
	if (cst)
	  last = compileMapAddPatternCst(patval,pat,val,env,last,choicelastp,errlastp);
	else last = compileMapAddPatternGnrl(pat,val,env,last,choicelastp,errlastp);
	break;
      }
      default:
	kinderror(message(648),pat->kind);
	break;
    }
  return last;
}

PRIVATE ExprPtr compileBindings P((IN CONST ExprPtr bindings,
				   IN CONST compEnv *env,
				   INOUT ExprPtr last,
				   INOUT ExprPtr *errlastp));

PRIVATE ExprPtr compileLet1(bindings,body,env,last,errlastp,comp)
IN CONST ExprPtr bindings;
IN CONST ExprPtr body;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp; /* will be a merge node, if non NIL */
IN CONST bindCompFun comp;
{
  COUNTER n;
  ExprPtr ee,aux;
  compEnv *newenv = (compEnv *)alloctmp(sizeof(compEnv));

  ee = compileKthEnv(0,env,&last);
  newenv->env = env->env;
  newenv->env = env->level;
  newenv->funs = env->funs;
  newenv->handlers = env->handlers;
  n = IVAL(CAR(bindings));
  if (n<0) /* then this is an imported module (an interface file, normally),
	      and the evaluator then first allocates a fake copy of the toplevel
	      environment GLOBENV below the current one. */
    {
      ExprPtr ll,funs;
      COUNTER l;

      ee = last = tackApply(Z_FAKEENV,ee,last);
      newenv->env = MakePair(E_CONS,last,newenv->env); /* extend environment */
      (void) KEEP(newenv->env);
      newenv->level = newenv->level+1;
      funs = NewMap();
      for (ll=env->funs, l=0; ISCONS(ll); l++, ll=CDR(ll))
	if (l==env->level)
	  { /* copy what is known from the toplevel environment */
	    funs = CAR(ll);
	    break;
	  }
      MAKECONS(funs,newenv->fun,aux);
      (void) KEEP(aux);
      newenv->funs = aux;
      n = -1-n;
    }
  last = tackApplyBin(Z_CONSENV,ee,KINT((INT)n),last);
  /* build new environment on top of previous ones */
  /* Z_CONSENV must also fill the new environment with NOEXPR at run-time */
  newenv->env = MakePair(E_CONS,last,newenv->env); /* extend environment */
  (void) KEEP(newenv->env);
  newenv->level = newenv->level+1;
  MAKECONS(NewMap(),newenv->funs,aux);
  (void) KEEP(aux);
  newenv->funs = aux;
  last = compileBindings(CDR(bindings),newenv,last,errlastp);
  last = (*comp)(body,newenv,last,errlastp);
  return last;
}

DEF_PRIVATE_ACTION(compilebinding)
{
  struct compilebinding *data;

  data = (struct compilebinding *)env;
  return compileBinding(data->binding,data->env,data->last,&data->errlast);
}

PRIVATE ExprPtr compileBinding1(binding,env,last,errlastp)
IN CONST ExprPtr binding;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp; /* will be a merge node, if non NIL */
{
  struct compilebinding *data = (struct compilebinding *)alloctmp(sizeof(struct compilebinding));
  ExprPtr res;

  data->binding = binding;
  data->env = env;
  data->last = last;
  data->errlast = *errlastp;
  res = CallCC(ACTION_NAME(compilebinding),(ExprPtr)data,true);
  *errlastp = data->errlast;
  return res;
}

PRIVATE ExprPtr compileBinding(binding,env,last,errlastp)
IN CONST ExprPtr binding;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp; /* will be a merge node, if non NIL */
{
  ExprPtr choicelast;
  ENTER;

  if (binding->kind==(CHAR)E_LET) /* local bindings */
    {
      if (STKFULL(binding))
	return compileBinding1(binding,env,last,errlastp);
      last = compileLet1(LETBIND(binding),LETBODY(binding),env,last,errlastp,compileBindings);
    }
  else
    {
      last = translateStmt(CDR(binding),env,last); /* evaluate value */
      choicelast = NIL;
      env->letdelayref = MakeSing(E_REF,NewMap());
      (void) KEEP(env->letdelayref);
      last = compilePattern(CAR(binding),last,env,last,&choicelast,errlastp);
      /* compile decomposition of pattern; choicelast should remain NIL. */
      last = dumpLetStores(REFVAL(env->letdelayref),getEnv(env),last);
      /* there is no letconsenv trickery here, as we assume that when compiling
	 bindings (typically in a Z_LET), the current env has already been
	 allocated by Z_CONSENV. */
    }
  QUIT;
  return last;
}

PRIVATE ExprPtr compileBindings(bindings,env,last,errlastp)
IN CONST ExprPtr bindings;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp; /* will be a merge node, if non NIL */
{
  while (bindings!=NIL)
    {
      last = compileBinding(CAR(bindings),env,last,errlastp);
      bindings = CDR(bindings);
    }
  return last;
}

PRIVATE ExprPtr compileStmt(stmt,env,last,errlastp)
IN CONST ExprPtr stmt;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp; /* will be a merge node, if non NIL */
{
  return translateStmt(stmt,env,last);
}

PRIVATE ExprPtr compileLet(bindings,body,env,last)
IN CONST ExprPtr bindings;
IN CONST ExprPtr body;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr errlast;

  errlast = NIL;
  last = compileLet1(bindings,body,env,last,&errlast,compileStmt);
  if (errlast!=NIL)
    (void) compileRaiseImm(I_BINDEXC,env,errlast); /* raise Bind */
    /* we assume that raise never returns (check with the debugger) */
  return last;
}

PRIVATE ExprPtr compileAssertGlobEnv(env,last)
IN CONST compEnv *env;
INOUT ExprPtr last;
{ /* for a function at level level, asserts that
     E_NEXT(E_NEXT(...[level times] (env)...)==GLOBENV,
     where env is the environment passed as a parameter to the closure. */
  ExprPtr ee;

  ee = compileKthEnv(env->level,env,&last);
  last = tackCGlobal(I_GLOBENV,(SCF_COMPOSITE | SCF_BOXED | SCF_NOINIT),last);
  last = tackApplyBin(Z_ASSERT_EQUAL,ee,last,last);
  return last;
}

PRIVATE ExprPtr compilePolyFun1(x,match,nlocals,envvar,funs,level,last)
IN CONST ExprPtr x;
IN CONST ExprPtr match;
IN COUNTER nlocals;
IN CONST ExprPtr envvar;
IN CONST ExprPtr funs;
IN COUNTER level;
INOUT ExprPtr last;
{
  compEnv *cenv = (compEnv *)alloctmp(sizeof(compEnv));
  ExprPtr env,choicelast,errlast,ee;
  ENTER;

  cenv->env = envvar;
  cenv->level = level;
  ee = compileKthEnv(0,cenv,&last);
  cenv->letconsenv = zApplyBin(Z_CONSENV,ee,KINT((INT)nlocals));
  env = NIL; /* make dummy environment */
  cenv->env = MakePair(E_CONS,env,cenv->env); /* extend environment */
  (void) KEEP(cenv->env);
  /* cons local environment */
  cenv->level = level+1;
  MAKECONS(NewMap(),funs,cenv->funs);
  (void) KEEP(cenv->funs);
  cenv->handlers = NIL;
  cenv->constpatmemoref = MakeSing(E_REF,NewMap());
  (void) KEEP(cenv->constpatmemoref);
  cenv->letdelayref = MakeSing(E_REF,NewMap());
  (void) KEEP(cenv->letdelayref);
  choicelast = errlast = NIL;
  last = compilePattern(CAR(match),x,cenv,env,&choicelast,&errlast);
  /* compile decomposition of pattern */
  if (errlast!=NIL)
    (void) compileRaiseImm(I_MATCHEXC,cenv,errlast); /* raise Match */
  /* we assume that raise never returns (check with the debugger) */
  last = env = tackAssign(cenv->letconsenv,last);
  SETCAR(cenv->env,env);
  last = dumpLetStores(REFVAL(cenv->letdelayref),env,last);
  last = translateStmt(CDR(match),cenv,last);
  if (choicelast!=NIL)
    last = mergeContinuations(last,last,choicelast);
  /* evaluate body */
  return last;
}

PRIVATE ExprPtr compilePolyFun(match,nlocals,env,outlast)
IN CONST ExprPtr match;
IN COUNTER nlocals;
IN CONST compEnv *env;
INOUT ExprPtr outlast;
{
  struct translation *fun = (struct translation *)alloctmp(sizeof(struct translation));
  ExprPtr last,val;
  ENTER;

  initTranslation(2,fun); /* 2 args: the real arg and the closure environment */
  last = fun->prgm;
  last = compileAssertGlobEnv(env,last);
  last = compilePolyFun1(nthArg(0,fun),match,nlocals,nthArg(1,fun),env->funs,env->level,last);
  val = transToCode(fun,last,outlast);
  KEEPONLY(val);
  return val;
}

PRIVATE ExprPtr compileMemoFn(func,rem,env,last)
IN CONST ExprPtr func,rem;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  struct translation *fun = (struct translation *)alloctmp(sizeof(struct translation));
  ExprPtr x,r,e,y,s,found,elselast,outlast;
  ENTER;

  /* memofn x => f x ==
     let val r = ref {} in
     fn x => if x inset !r
     		then ? (!r) x
	     else let val y = f x
	          in
		     r := !r ++ {x => y}; y
		  end
      end
      (* this function is called the stub *)
      */
  last = compileConstant(rem,last,true); /* rem should be the empty map */
  rem = tackApply(Z_BOX_REF,last,last);
  /* now, compile memoizing stub */
  initTranslation(2,fun); /* 2 args: the real arg and the closure environment;
			    this env is cons(memoizing ref,normal env) */
  last = fun->prgm;
  last = compileAssertGlobEnv(env,last);
  outlast = NIL;
  x = nthArg(0,fun);
  e = nthArg(1,fun);
  r = tackApply(Z_CAR,e,last); /* get memoizing ref */
  s = tackApply(Z_UNBOX_REF,r,r);
  y = tackApplyBin(Z_MAPGET,x,s,s);
  found = tackApply(Z_GET_MAPGET_FOUND,y,y);
  last = compileIf(found,found,&elselast);
  last = tackApply(Z_GET_MAPGET_Y,y,last);
  outlast = mergeContinuations(last,last,outlast);
  /* otherwise: */
  last = tackApply(Z_CDR,e,elselast);
  /* now, inline the call to f (there can be only one anyway, and
     it is a E_POLYFUN). */
  y = compilePolyFun1(nthArg(0,fun),COUNTEDVAL(func),COUNTEDN(func),last,env->funs,env->level,last);
  last = tackApplyBin(Z_BOX_SINGLETON,x,y,y);
  last = tackApplyBin(Z_OVERWRITE,s,last,last);
  last = compileRefSet(Z_REF_SET,r,last,last);
  outlast = mergeContinuations(y,last,outlast);
  last = transToCode(fun,outlast,rem);
  /* and now, compile the closure */
  e = tackApplyBin(Z_BOX_UNSHARED_CONS,rem,getEnv(env),last);
  last = compileClosure(last,e,e);
  QUIT;
  return last;
}

PRIVATE ExprPtr compileApply(fun,arg,env,last)
IN CONST ExprPtr fun,arg;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr ffun,fenv;

  /* identify special cases as in Eval() */
reapply:
  switch (fun->kind)
    {
    case E_POLYFUN: /* case <expr> of ...; this could be done by inlining in
		     the optimization phase, but it is easier to do here. */
      arg = translateStmt(arg,env,last);
      last = compilePolyFun1(arg,COUNTEDVAL(fun),COUNTEDN(fun),env->env,env->funs,env->level,arg);
      break;
    case E_MEMO_FN: /* a memo function as immediate value in the code does not
		       memoize (or it would memoize into a closure that would
		       be immediately garbageable) */
      fun = MEMOFUN(fun);
      goto reapply;
    case E_DELAY: /* same as for memo functions */
      fun = DCODE(fun);
      goto reapply;
    case E_DATACON:
      arg = translateStmt(arg,env,last);
      last = tackApplyBin((COUNTEDN(fun) & 1)?Z_BOX_UNSHARED_CONS:Z_BOX_CONS,
			  translateStmt(COUNTEDVAL(fun),env,arg),arg,arg);
    case E_VARREF:
      { /* we identify the important special case of the application of a well-known
	   function with a known identifier. */
	ExprPtr funs,rank,code;
	COUNTER i;
	Boolean found;

	for (i=0, funs=env->funs; i<VARLEVEL(fun); i++)
	  funs = CDR(funs);
	rank = KINT((INT)VARRANK(fun));
	code = MapGet(rank,CAR(funs),&found);
	if (found)
	  { /* then code is directly a Z_CODE expression */
	    last = arg = translateStmt(arg,env,fun);
	    fenv = compileKthEnv(VARLEVEL(fun),env,&last);
	    ffun = tackAssign(code,last);
	    break;
	  }
      }
      /*fallthrough*/
    default:
      fun = translateStmt(fun,env,last);
      arg = translateStmt(arg,env,fun);
      ffun = tackApply(Z_UNBOX_CLOSURE_FN,fun,arg);
      fenv = tackApply(Z_UNBOX_CLOSURE_ENV,fun,ffun);
      last = tackApplyTer(Z_APPLY,ffun,arg,fenv,fenv);
      break;
    }
  return last;
}

PRIVATE ExprPtr compileSet(ref,val,env,last)
IN CONST ExprPtr ref,val;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ref = translateStmt(ref,env,last);
  val = translateStmt(val,env,ref);
  last = compileRefSet(Z_REF_SET,ref,val,val);
  last = tackApply(Z_TUPLE_0,NIL,last);
  return last;
}

PRIVATE ExprPtr compileMapGet(map,x,env,last)
IN CONST ExprPtr map,x;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr found,elselast,val;

  map = translateStmt(map,env,last);
  x = translateStmt(x,env,map);
  val = tackApplyBin(Z_MAPGET,x,map,x);
  found = tackApply(Z_GET_MAPGET_FOUND,val,val);
  last = compileIf(found,found,&elselast);
  (void) compileRaiseImm(I_MAPGETEXC,env,elselast); /* raise MapGet */
  last = tackApply(Z_GET_MAPGET_Y,val,last);
  return last;
}

PRIVATE ExprPtr compileTupleSelect(index,val,env,last)
IN CONST ExprPtr index,val;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr elselast,thenlast;

  thenlast = NIL;
  elselast = val = translateStmt(val,env,last);
  if (IVAL(index)<=2)
    {
      /* if index>2, then val must be an E_TUPLE_GNRL, because of typing. */
      if (index==ione)
	{ /* if index!=ione, then val cannot be E_TUPLE_1, because of the typing rules. */
	  last = compileHasTag((CHAR)E_TUPLE_1,val,val,&elselast);
	  /* if val is an E_TUPLE_1, just get its REFVAL: */
	  last = tackApply(Z_UNBOX_REF,val,last);
	  thenlast = mergeContinuations(last,last,thenlast);
	}
      /* if val is not an E_TUPLE_1, check whether it is an E_TUPLE_2: */
      last = compileHasTag((CHAR)E_TUPLE_2,val,elselast,&elselast);
      /* if val is an E_TUPLE_2, get its CAR if index==ione, and its CDR otherwise: */
      last = tackApply((index==ione)?Z_CAR:Z_CDR,val,last);
      thenlast = mergeContinuations(last,last,thenlast);
    }
  /* otherwise, val is an E_TUPLE_GNRL: */
  last = tackApplyBin(Z_UNBOX_TUPLE_NTH,val,KINT(IVAL(index)),elselast);
  return last;
}

PRIVATE ExprPtr compileRecordSelect(label,val,env,last)
IN CONST ExprPtr label,val;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr elselast;

  val = translateStmt(val,env,last);
  if (label==ione || label->kind!=(CHAR)E_INT)
    { /* in all other cases, label is an integer >=2, so that val can
	 never be an E_RECORD_1. */
      last = compileHasTag((CHAR)E_RECORD_1,val,val,&elselast);
      /* if val is an E_RECORD_1, then the selection is its CDR: */
      last = tackApply(Z_CDR,val,last);
      /* otherwise, select as on a E_TUPLE_GNRL: */
      elselast = compileSelectGnrl(label,val,env,elselast);
      last = compileMergeBin(last,elselast);
    }
  else
    last = compileSelectGnrl(label,val,env,val);
  return last;
}

PRIVATE ExprPtr translateIf(test,thene,elsee,env,last)
IN CONST ExprPtr test,thene,elsee;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr elselast;

  test = translateStmt(test,env,last);
  test = tackApply(Z_UNBOX_BOOL,test,test);
  last = compileIf(test,test,&elselast);
  last = translateStmt(thene,env,last);
  elselast = translateStmt(elsee,env,elselast);
  last = compileMergeBin(last,elselast);
  return last;
}

PRIVATE ExprPtr compileAnd(e1,e2,env,last)
IN CONST ExprPtr e1,e2;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr elselast,test;

  test = translateStmt(e1,env,last);
  test = tackApply(Z_UNBOX_BOOL,test,test);
  last = compileIf(test,test,&elselast);
  elselast = tackAssign(test,elselast);
  last = translateStmt(e2,env,last);
  last = mergeContinuations(test,elselast,mergeContinuations(last,last,NIL));
  return last;
}

PRIVATE ExprPtr compileOr(e1,e2,env,last)
IN CONST ExprPtr e1,e2;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr elselast,test;

  test = translateStmt(e1,env,last);
  test = tackApply(Z_UNBOX_BOOL,test,test);
  last = compileIf(test,test,&elselast);
  elselast = translateStmt(e2,env,elselast);
  last = tackAssign(test,last);
  last = mergeContinuations(test,last,mergeContinuations(elselast,elselast,NIL));
  return last;
}

PRIVATE ExprPtr compileIdentity(env,last)
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  return last;
}
PRIVATE ExprPtr compileIdentityErr(env,last,errlastp)
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{
  return last;
}

#define compileNeutralProg compileNeutralNone
#define compileCombine1Prog compileIdentityErr
PRIVATE ExprPtr compileCombineProg(e1,env,last,errlastp)
IN CONST ExprPtr e1;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{
  return last;
}
#define compilePromoteProg compileIdentity

PRIVATE ExprPtr compileNeutralList(env,last,nonelastp,errlastp)
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *nonelastp,*errlastp;
{ /* builds an empty queue */
  last = tackApply(Z_BOX_QUEUE,NIL,last);
  /* does an:
     {
     ExprPtr e;

     EALLOC(e);
     e->kind = (CHAR )E_CONS;
     CAR(e) = e;
     CDR(e) = NIL;
     (void) KEEP(e);
     }
     i.e, builds a header whose cdr is the list NIL, and the car
     is the last cons cell in the list (here, the header itself).
     Moreover, the cell for the queue header really should never
     be allocated (boxed).
     */
  return last;
}
#define compileCombine1List compileIdentityErr
PRIVATE ExprPtr compileCombineList(e1,env,last,errlastp)
IN CONST ExprPtr e1;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{ /* builds reversed unshared list, so that it is easier
     to share when the computation is over. If we don't
     want to share, then we should exchange the two
     arguments to Z_CONCAT_QUEUE and forget about the
     reversion at the end of the computation. */
  last = tackApplyBin(Z_CONCAT_QUEUE,e1,last,e1);
  /* Z_CONCAT_QUEUE(q1,q2) does:
     if (CAR(q2)!=q2)
       {
	 SETCDR(CAR(q1),CDR(q2));
         SETCAR(q1,CAR(q2));
       }
     res = q1;
     */
  return last;
}
PRIVATE ExprPtr compilePromoteList(f,env)
IN bcFile *f;
IN CONST compEnv *env;
{ /* builds a queue with only the element last in it */
  bcPut(BC_NEW1QUEUE,f);
}
PRIVATE ExprPtr compileQueueToList(f)
IN bcFile *f;
{
  bcPut(BC_QUEUELIST,f);
}

PRIVATE ExprPtr compileNeutralSet(env,last,nonelastp,errlastp)
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *nonelastp,*errlastp;
{
  last = compileNewMap(env,last);
  return last;
}
#define compileCombine1Set compileIdentityErr
PRIVATE ExprPtr compileCombineSet(e1,env,last,errlastp)
IN CONST ExprPtr e1;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{
  last = tackApplyBin(Z_OVERWRITE,last,e1,last);
  return last;
}
PRIVATE ExprPtr compilePromoteSet(env,last)
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr unit;

  unit = tackApply(Z_TUPLE_0,NIL,last);
  last = tackApplyBin(Z_BOX_SINGLETON,last,unit,unit);
  return last;
}

#define compileNeutralAppend compileNeutralList
#define compileCombine1Append compileCombine1List
#define compileCombineAppend compileCombineList
#define compilePromoteAppend compileIdentity

#define compileNeutralUnion compileNeutralSet
#define compileCombine1Union compileIdentityErr
#define compileCombineUnion compileCombineSet
#define compilePromoteUnion compileIdentity

#define compileNeutralUnionUnder compileNeutralSet
#define compileCombine1UnionUnder compileIdentityErr
PRIVATE ExprPtr compileCombineUnionUnder(e1,env,last,errlastp)
IN CONST ExprPtr e1;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{
  last = tackApplyBin(Z_OVERWRITE,e1,last,last);
  return last;
}
#define compilePromoteUnionUnder compilePromoteUnion

PRIVATE ExprPtr compileNeutralInter(env,last,nonelastp,errlastp)
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *nonelastp,*errlastp;
{
  *nonelastp = mergeContinuations(last,last,*nonelastp); /* first last is dummy */
  return NIL; /* we can do this because we never compute distributed intersections
	       by intersecting all sets with the common neutral, but actually
	       return the neutral only if there are no sets to intersect. */
}
#define compileCombine1Inter compileIdentityErr
PRIVATE ExprPtr compileCombineInter(e1,env,last,errlastp)
IN CONST ExprPtr e1;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{
  last = tackApplyBin(Z_MAPINTER,e1,last,last);
  /* we do not need to check that e1 is NIL or not, as in eval.c,
     since we never intersect with something representing the whole
     universe of sets (we do not use the neutral element when there
     is at least one set to intersect). */
  return last;
}
#define compilePromoteInter compileIdentity

#define compileNeutralMap compileNeutralSet
#define compileCombine1Map compileIdentityErr
#define compileCombineMap compileCombineSet
PRIVATE ExprPtr compilePromoteMap(env,last)
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr car,cdr;

  car = tackApply(Z_CAR,last,last);
  cdr = tackApply(Z_CDR,last,car);
  last = tackApplyBin(Z_BOX_SINGLETON,car,cdr,cdr);
  return last;
}

#define compileNeutralMapUnder compileNeutralSet
#define compileCombine1MapUnder compileIdentityErr
#define compileCombineMapUnder compileCombineUnionUnder
#define compilePromoteMapUnder compilePromoteMap

PRIVATE ExprPtr compileSimpleOf P((IN struct compilesimpleof *data,
				   IN CONST compEnv *env,INOUT ExprPtr last,
				   INOUT ExprPtr *nonelastp,INOUT ExprPtr *errlastp));

PRIVATE ExprPtr compileOneOf(x,y,_data,env,last,nonelastp,errlastp)
IN CONST ExprPtr x,y;
IN CHAR *_data;
IN CONST compEnv *env;
INOUT ExprPtr last;
OUT ExprPtr *nonelastp; /* last node of path when nothing was done (binding failed,
			     predicate was false. */
INOUT ExprPtr *errlastp; /* last node of path exiting out of the whole iteration */
{
  struct compilesimpleof *data;
  ExprPtr choicelast;

  data = (struct compilesimpleof *)_data;
  choicelast = *nonelastp = NIL;
  if (data->mapsweep)
    {
      last = compilePattern(CAR(data->pat),x,env,last,&choicelast,nonelastp);
      last = compilePattern(CDR(data->pat),y,env,last,&choicelast,nonelastp);
    }
  else
    {
      last = compilePattern(data->pat,x,env,last,&choicelast,nonelastp);
    }
  /* choicelast should now be NIL again */
  if (data->domains==NIL)
    {
      last = dumpLetStores(REFVAL(env->letdelayref),getEnv(env),last);
      /* there is no letconsenv trickery here, as we assume that when compiling
	 bindings, the current env has already been allocated by Z_CONSENV. */
      last = compileBindings(data->locals,env,last,nonelastp);
      if (data->p!=trueValue) /* predicate=constant true is a frequent particular case */
	{
	  last = translateStmt(data->p,env,last);
	  last = tackApply(Z_UNBOX_BOOL,last,last);
	  last = compileIf(last,last,nonelastp);
	}
      last = translateStmt(data->f,env,last);
      last = (*data->promote)(env,last);
    }
  else
    last = compileSimpleOf(data,env,last,nonelastp,errlastp);
  return last;
}

DEF_PRIVATE_ACTION(compilesimpleofk)
{
  struct compilesimpleofk *data;

  data = (struct compilesimpleofk *)env;
  return compileSimpleOf(data->data,data->env,data->last,&data->nonelast,&data->errlast);
}

PRIVATE ExprPtr compileSimpleOf1(data,env,last,nonelastp,errlastp)
IN struct compilesimpleof *data;
IN CONST compEnv *env;
INOUT ExprPtr last;
OUT ExprPtr *nonelastp;
INOUT ExprPtr *errlastp;
{
  struct compilesimpleofk *d = (struct compilesimpleofk *)alloctmp(sizeof(struct compilesimpleofk));
  ExprPtr res;

  d->data = data;
  d->env = env;
  d->last = last;
  d->nonelast = *nonelastp;
  d->errlast = *errlastp;
  res = CallCC(ACTION_NAME(compilesimpleofk),(ExprPtr)d,true);
  *nonelastp = d->nonelast;
  *errlastp = d->errlast;
  return res;
}

PRIVATE ExprPtr compileSimpleOf(data,env,last,nonelastp,errlastp)
IN struct compilesimpleof *data;
IN CONST compEnv *env;
INOUT ExprPtr last;
OUT ExprPtr *nonelastp;
INOUT ExprPtr *errlastp;
{
  ExprPtr domains,dompats,savepat,list;
  Boolean savemapsweep;
  ENTER;

  if (STKFULL(savemapsweep))
    return compileSimpleOf1(data,env,last,nonelastp,errlastp);
  domains = data->domains; /* list (with special kinds) of variables
			      containing the different domains */
  dompats = data->dompats;
  *nonelastp = NIL;
  if (domains==NIL)
    last = (*data->neutral)(env,last,nonelastp,errlastp);
  else
    {
      list = CAR(domains);
      data->domains = CDR(domains);
      savepat = data->pat;
      savemapsweep = data->mapsweep;
      data->dompats = CDR(dompats);
      data->pat = CAR(CAR(dompats));
      data->mapsweep = (dompats->kind==(CHAR)E_PAT_REVRECORD_GNRL_CONS);
      last = compileSweep((INT)dompats->kind,
			  list,data->neutral,data->combine,data->combine1,
			  compileOneOf,(CHAR *)data,env,
			  last,nonelastp,errlastp);
      /* restore data */
      data->mapsweep = savemapsweep;
      data->pat = savepat;
      data->domains = domains;
      data->dompats = dompats;
    }
  QUIT;
  return last;
}

PRIVATE ExprPtr compileMapfP(data,domrefs,env,last,errlastp)
IN struct compilesimpleof *data;
IN CONST ExprPtr domrefs;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *errlastp;
{
  ExprPtr dompats,listrs,listr,choicelast,list;

  choicelast = NIL;
  SETCAR(env->letdelayref,NewMap());
  for (dompats = data->dompats, listrs = domrefs; listrs!=NIL;
       dompats = CDR(dompats), listrs = CDR(listrs))
    {
      listr = CAR(listrs);
      switch (dompats->kind)
	{
	case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
	  list = tackApply(Z_UNBOX_LISTSWEEP,listr,last);
	  last = tackApply(Z_CAR,list,list);
	  last = compilePattern(CAR(dompats),last,env,last,&choicelast,errlastp);
	  break;
	case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
	  {
	    ExprPtr x,y;

	    list = tackApply(Z_UNBOX_MAPSWEEP,listr,last);
	    x = tackApply(Z_CAR,list,list);
	    last = compilePattern(CAR(CAR(dompats)),x,env,x,&choicelast,errlastp);
	    y = tackApply(Z_CDR,list,last);
	    last = compilePattern(CDR(CAR(dompats)),y,env,y,&choicelast,errlastp);
	  }
	  break;
	case E_PAT_RECORD_MAPADD: /* sub map domain descriptor; never empty */
	  last = tackApply(Z_UNBOX_SUBMAPSWEEP,listr,last);
	  last = compilePattern(CAR(dompats),last,env,last,&choicelast,errlastp);
	  break;
	}
    }
  /* now, choicelast should still be NIL */
  last = compileBindings(data->locals,env,last,errlastp);
  last = dumpLetStores(REFVAL(env->letdelayref),getEnv(env),last);
  /* there is no letconsenv trickery here, as we assume that when compiling
     bindings, the current env has already been allocated by Z_CONSENV. */
  if (data->p!=trueValue) /* predicate=constant true is a frequent particular case */
    {
      last = translateStmt(data->p,env,last);
      last = tackApply(Z_UNBOX_BOOL,last,last);
      last = compileIf(last,last,errlastp);
    }
  last = translateStmt(data->f,env,last);
  last = (*data->promote)(env,last);
  return last;
}

PRIVATE ExprPtr compileMapInc(data,domrefs,env,last,endlastp)
IN struct compilesimpleof *data;
IN CONST ExprPtr domrefs;
IN CONST compEnv *env;
INOUT ExprPtr last;
OUT ExprPtr *endlastp;
{
  ExprPtr domains,dompats,listrs,listr,elselast,errlast,raiselast;

  errlast = raiselast = NIL;
  for (domains = data->domains, dompats = data->dompats, listrs = domrefs;
       domains!=NIL;
       domains = CDR(domains), dompats = CDR(dompats), listrs = CDR(listrs))
    {
      listr = CAR(listrs);
      elselast = compileLoop(incTag((INT)dompats->kind),listr,last,&last);
      if (domains==data->domains) /* if first list is empty */
	errlast = elselast;
      else /* otherwise, raise ParSweep */
	raiselast = mergeContinuations(elselast,elselast,raiselast);
    }
  if (errlast!=NIL) /* if first list ends up empty, then check that all other
		       are incremented to empty, too */
    for (domains = data->domains, dompats = data->dompats, listrs = domrefs;
	 dompats = CDR(dompats), domains = CDR(domains), listrs = CDR(listrs), domains!=NIL; )
      {
	listr = CAR(listrs);
	errlast = compileLoop(incTag((INT)dompats->kind),listr,errlast,&elselast);
	(void) mergeContinuations(elselast,elselast,raiselast); /* first elselast is dummy */
      }
  *endlastp = errlast;
  return last;
}

PRIVATE ExprPtr compileSimpleMap(data,env,last,nonelastp,errlastp)
IN struct compilesimpleof *data;
IN CONST compEnv *env;
INOUT ExprPtr last;
OUT ExprPtr *nonelastp;
INOUT ExprPtr *errlastp;
{
  ExprPtr list,errlast,elselast,outlast,domrefs,start,res,resr;
  ENTER;

  list = CAR(data->domains);
  /* first, deal with the case where all domains are empty */
  elselast = outlast = errlast = *nonelastp = NIL;
  switch (data->dompats->kind)
    {
    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
      last = compileCheckNotEqualImm(list,NIL,last,&elselast);
      break;
    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
      last = compileCheckNotEqualImm(list,NewMap(),last,&elselast);
      break;
    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor; never empty */
      break;
    }
  /* if first domain is empty, check that all the other domains are empty,
     too; if this fails, raise ParSweep */
  if (elselast!=NIL)
    {
      ExprPtr domains,dompats,buglast;

      for (domains = data->domains, dompats = data->dompats;
	   dompats = CDR(dompats), domains = CDR(domains), domains!=NIL; )
	{
	  if (elselast==NIL)
	    break;
	  list = CAR(domains);
	  switch (dompats->kind)
	    {
	    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
	      elselast = compileCheckEqualImm(list,NIL,elselast,&buglast);
	      errlast = mergeContinuations(buglast,buglast,errlast); /* first buglast is dummy */
	      break;
	    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
	      elselast = compileCheckEqualImm(list,NewMap(),elselast,&elselast);
	      errlast = mergeContinuations(buglast,buglast,errlast); /* first buglast is dummy */
	      break;
	    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor; never empty */
	      errlast = mergeContinuations(elselast,elselast,errlast); /* first elselast is dummy */
	      elselast = NIL;
	      break;
	    }
	}
      if (elselast!=NIL)
	{
	  elselast = (*data->neutral)(env,elselast,nonelastp,errlastp);
	  outlast = mergeContinuations(elselast,elselast,outlast);
	}
    }
  /* now, compute function on first element */
  /* first, initialize all local references to parts of the objects to sweep through */
  {
    ExprPtr domains,dompats,e;

    domrefs = NIL;
    for (domains = data->domains, dompats = data->dompats; domains!=NIL;
	 domains = CDR(domains), dompats = CDR(dompats))
      {
	list = CAR(domains);
	last = tackApply(startTag((INT)dompats->kind),list,last);
	MAKECONS(last,domrefs,e);
	(void) KEEP(e);
	domrefs = e;
      }
    domrefs = Nreverse(domrefs);
  }
  start = tackMerge(last); /* insert label */
  /* second, evaluate predicate and function, if necessary, on data,
   until some element verifying the predicate is found*/
  {
    ExprPtr nonelast,endlast;

    nonelast = NIL;
    last = compileMapfP(data,domrefs,env,start,&nonelast);
    nonelast = compileMapInc(data,domrefs,env,nonelast,&endlast);
    (void) mergeContinuations(nonelast,nonelast,start); /* first nonelast is dummy */
    if (endlast!=NIL) /* if finished before having found an element, then return neutral */
      {
	endlast = (*data->neutral)(env,endlast,nonelastp,errlastp);
	outlast = mergeContinuations(endlast,endlast,outlast);
      }
  }
  if (last!=NIL)
    last = (*data->combine1)(env,last,errlastp);
  if (last!=NIL)
    {
      ExprPtr nonelast,endlast;

      resr = tackApply(Z_BOX_LOCAL_REF,last,last);
      start = tackMerge(resr); /* insert label */
      last = compileMapfP(data,domrefs,env,start,&nonelast);
      if (last!=NIL)
	{
	  res = tackApply(Z_UNBOX_LOCAL_REF,resr,last);
	  last = (*data->combine)(last,env,res,errlastp);
	  last = compileRefSet(Z_LOCAL_REF_SET,resr,last,last); /* store new result */
	}
      last = compileMergeBin(last,nonelast);
      if (last!=NIL)
	{
	  last = compileMapInc(data,domrefs,env,last,&endlast);
	  (void) mergeContinuations(nonelast,nonelast,start); /* first nonelast is dummy */
	  if (endlast!=NIL) /* if finished, then return !resr */
	    {
	      endlast = tackApply(Z_UNBOX_LOCAL_REF,resr,last);
	      outlast = mergeContinuations(endlast,endlast,outlast);
	    }
	}
    }
  if (errlast!=NIL)
    (void) compileRaiseImm(I_MAPEXC,env,errlast); /* raise Map */
  QUIT;
  return outlast;
}

PRIVATE ExprPtr compileDomainsSimple1 P((IN CONST ExprPtr domains,IN CONST compEnv *env,
					 INOUT ExprPtr *lastp,INOUT ExprPtr *locals));

DEF_PRIVATE_ACTION(compiledomainssimplek)
{
  struct compiledomainssimplek *data;

  data = (struct compiledomainssimplek *)env;
  return compileDomainsSimple1(data->domains,data->env,&data->last,&data->locals);
}

PRIVATE ExprPtr compileDomainsSimple1k(domains,env,lastp,locals)
IN CONST ExprPtr domains;
IN CONST compEnv *env;
INOUT ExprPtr *lastp;
INOUT ExprPtr *locals;
{
  struct compiledomainssimplek *data = (struct compiledomainssimplek *)
    alloctmp(sizeof(struct compiledomainssimplek));
  ExprPtr res;

  data->domains = domains;
  data->env = env;
  data->last = *lastp;
  data->locals = *locals;
  res = CallCC(ACTION_NAME(compiledomainssimplek),(ExprPtr)data,true);
  *lastp = data->last;
  *locals = data->locals;
  return res;
}

PRIVATE ExprPtr compileDomainsSimple1(domains,env,lastp,locals)
IN CONST ExprPtr domains;
IN CONST compEnv *env;
INOUT ExprPtr *lastp;
INOUT ExprPtr *locals;
{
  ExprPtr list,rest,res;

  if (STKFULL(res))
    return compileDomainsSimple1k(domains,env,lastp,locals);
  if (domains==NIL)
    {
      *locals = NIL;
      return NIL;
    }
  switch (domains->kind)
    {
    case E_PAT_RECORD_GNRL_CONS:
    case E_PAT_REVRECORD_GNRL_CONS:
    case E_PAT_RECORD_MAPADD:
      /* in map, in list, sub map domain descriptors */
      *lastp = list = translateStmt(CDR(CAR(domains)),env,*lastp);
      rest = compileDomainsSimple1(CDR(domains),env,lastp,locals);
      MAKEPAIR(domains->kind,list,rest,res);
      break;
    case E_TUPLE_1: /* local declarations */
      *locals = REFVAL(domains);
      return NIL;
    }
  return KEEP(res);
}

PRIVATE ExprPtr compileDomainsSimple(domains,env,lastp,locals)
IN CONST ExprPtr domains;
IN CONST compEnv *env;
INOUT ExprPtr *lastp;
OUT ExprPtr *locals;
{
  ExprPtr doms;
  ENTER;

  doms = compileDomainsSimple1(domains,env,lastp,locals);
  KEEPONLY(doms); /* no need to keep *locals, because it is in domains */
  return doms;
}

PRIVATE void fillData(data,ik)
INOUT struct compilesimpleof *data;
IN IterKind ik;
{
  switch (ik)
    {
    case C_PROG:
      data->neutral = compileNeutralProg;
      data->combine = compileCombineProg;
      data->combine1 = compileCombine1Prog;
      data->promote = compilePromoteProg;
      break;
    case C_LIST:
      data->neutral = compileNeutralList;
      data->combine = compileCombineList;
      data->combine1 = compileCombine1List;
      data->promote = compilePromoteList;
      break;
    case C_SET:
      data->neutral = compileNeutralSet;
      data->combine = compileCombineSet;
      data->combine1 = compileCombine1Set;
      data->promote = compilePromoteSet;
      break;
    case C_APPEND:
      data->neutral = compileNeutralAppend;
      data->combine = compileCombineAppend;
      data->combine1 = compileCombine1Append;
      data->promote = compilePromoteAppend;
      break;
    case C_UNION:
      data->neutral = compileNeutralUnion;
      data->combine = compileCombineUnion;
      data->combine1 = compileCombine1Union;
      data->promote = compilePromoteUnion;
      break;
    case C_UNION_UNDER:
      data->neutral = compileNeutralUnionUnder;
      data->combine = compileCombineUnionUnder;
      data->combine1 = compileCombine1UnionUnder;
      data->promote = compilePromoteUnionUnder;
      break;
    case C_INTER:
      data->neutral = compileNeutralInter;
      data->combine = compileCombineInter;
      data->combine1 = compileCombine1Inter;
      data->promote = compilePromoteInter;
      break;
    case C_MAP:
      data->neutral = compileNeutralMap;
      data->combine = compileCombineMap;
      data->combine1 = compileCombine1Map;
      data->promote = compilePromoteMap;
      break;
    case C_MAP_UNDER:
      data->neutral = compileNeutralMapUnder;
      data->combine = compileCombineMapUnder;
      data->combine1 = compileCombine1MapUnder;
      data->promote = compilePromoteMapUnder;
      break;
    case C_ALL:
      data->neutral = compileNeutralAll;
      data->combine = compileCombineAll;
      data->combine1 = compileCombine1All;
      data->promote = compilePromoteAll;
      break;
    case C_EXISTS:
      data->neutral = compileNeutralExists;
      data->combine = compileCombineExists;
      data->combine1 = compileCombine1Exists;
      data->promote = compilePromoteExists;
      break;
    case C_SOME:
      data->neutral = compileNeutralSome;
      data->combine = compileCombineSome;
      data->combine1 = compileCombine1Some;
      data->promote = compilePromoteSome;
      break;
    }
}

PRIVATE ExprPtr compileOf1(f,p,domains,ik,env,last,nonelastp,errlastp)
IN CONST ExprPtr f,p,domains;
IN IterKind ik;
IN CONST compEnv *env;
INOUT ExprPtr last;
OUT ExprPtr *nonelastp;
INOUT ExprPtr *errlastp;
{
  struct compilesimpleof *data = (struct compilesimpleof *)alloctmp(sizeof(struct compilesimpleof));
  ENTER;

  data->dompats = domains;
  data->domains = compileDomainsSimple(data->dompats,env,&last,&data->locals);
  data->f = f;
  data->p = p;
  fillData(data,ik);
  last = compileSimpleOf(data,env,last,nonelastp,errlastp);
  return last;
}

PRIVATE ExprPtr compileMap1(f,p,domains,ik,env,last,nonelastp,errlastp)
IN CONST ExprPtr f,p,domains;
IN IterKind ik;
IN CONST compEnv *env;
INOUT ExprPtr last;
OUT ExprPtr *nonelastp;
INOUT ExprPtr *errlastp;
{
  struct compilesimpleof *data = (struct compilesimpleof *)alloctmp(sizeof(struct compilesimpleof));
  ENTER;

  data->dompats = domains;
  data->domains = compileDomainsSimple(data->dompats,env,&last,&data->locals);
  data->f = f;
  data->p = p;
  fillData(data,ik);
  last = compileSimpleMap(data,env,last,nonelastp,errlastp);
  return last;
}

PRIVATE ExprPtr compileOf(f,p,domains,ik,env,last,nonelastp,errlastp)
IN CONST ExprPtr f,p;
IN CONST ExprPtr domains;
IN IterKind ik;
IN CONST compEnv *env;
INOUT ExprPtr last;
OUT ExprPtr *nonelastp;
INOUT ExprPtr *errlastp;
{
  ExprPtr localenv,ee;
  compEnv *cenv = (compEnv *)alloctmp(sizeof(compEnv));
  INT nlocals;

  nlocals = IVAL(CAR(domains));
  cenv->env = env->env;
  ee = compileKthEnv(0,env,&last);
  localenv = tackApplyBin(Z_CONSENV,ee,KINT(nlocals),last);
  /* cons local environment */
  cenv->env = MakePair(E_CONS,localenv,cenv->env); /* extend environment */
  (void) KEEP(cenv->env);
  cenv->level = env->level+1;
  MAKECONS(NewMap(),env->funs,cenv->funs);
  (void) KEEP(cenv->funs);
  cenv->letdelayref = MakeSing(E_REF,NewMap());
  (void) KEEP(cenv->letdelayref);
  cenv->handlers = env->handlers;
  cenv->constpatmemoref = env->constpatmemoref;
  last = compileOf1(f,p,CDR(domains),ik,cenv,localenv,nonelastp,errlastp);
  return last;
}

PRIVATE ExprPtr compileMap(f,p,domains,ik,env,last,nonelastp,errlastp)
IN CONST ExprPtr f,p;
IN CONST ExprPtr domains;
IN IterKind ik;
IN CONST compEnv *env;
INOUT ExprPtr last;
OUT ExprPtr *nonelastp;
INOUT ExprPtr *errlastp;
{
  ExprPtr localenv,ee;
  compEnv *cenv = (compEnv *)alloctmp(sizeof(compEnv));
  INT nlocals;

  nlocals = IVAL(CAR(domains));
  cenv->env = env->env;
  ee = compileKthEnv(0,env,&last);
  localenv = tackApplyBin(Z_CONSENV,ee,KINT(nlocals),last);
  /* cons local environment */
  cenv->env = MakePair(E_CONS,localenv,cenv->env); /* extend environment */
  (void) KEEP(cenv->env);
  cenv->level = env->level+1;
  MAKECONS(NewMap(),env->funs,cenv->funs);
  (void) KEEP(cenv->funs);
  cenv->letdelayref = MakeSing(E_REF,NewMap());
  (void) KEEP(cenv->letdelayref);
  cenv->handlers = env->handlers;
  cenv->constpatmemoref = env->constpatmemoref;
  last = compileMap1(f,p,CDR(domains),ik,cenv,localenv,nonelastp,errlastp);
  return last;
}

PRIVATE ExprPtr compileOfNoFuss(e,ik,env,last)
IN CONST ExprPtr e;
IN IterKind ik;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),ik,env,last,&nonelast,&errlast);
  return last;
}

PRIVATE ExprPtr compileMapNoFuss(e,ik,env,last)
IN CONST ExprPtr e;
IN IterKind ik;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),ik,env,last,&nonelast,&errlast);
  return last;
}

PRIVATE ExprPtr compileInterOf(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_INTER,env,last,&nonelast,&errlast);
  if (nonelast!=NIL)
    (void) compileRaiseImm(I_EMPTYEXC,env,nonelast); /* raise Empty */
  return last;
}

PRIVATE ExprPtr compileInterMap(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_INTER,env,last,&nonelast,&errlast);
  if (nonelast!=NIL)
    (void) compileRaiseImm(I_EMPTYEXC,env,nonelast); /* raise Empty */
  return last;
}

PRIVATE ExprPtr compileAllOf(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_ALL,env,last,&nonelast,&errlast);
  /* nonelast should be NIL */
  last = tackApply(Z_UNBOXED_TRUE,NIL,last);
  errlast = tackApply(Z_UNBOXED_FALSE,NIL,errlast);
  last = compileMergeBin(last,errlast);
  last = tackApply(Z_BOX_BOOL,last,last);
  return last;
}

PRIVATE ExprPtr compileAllMap(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_ALL,env,last,&nonelast,&errlast);
  /* nonelast should be NIL */
  last = tackApply(Z_UNBOXED_TRUE,NIL,last);
  errlast = tackApply(Z_UNBOXED_FALSE,NIL,errlast);
  last = compileMergeBin(last,errlast);
  last = tackApply(Z_BOX_BOOL,last,last);
  return last;
}

PRIVATE ExprPtr compileExistsOf(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_EXISTS,env,last,&nonelast,&errlast);
  /* nonelast should be NIL */
  last = tackApply(Z_UNBOXED_FALSE,NIL,last);
  errlast = tackApply(Z_UNBOXED_TRUE,NIL,errlast);
  last = compileMergeBin(last,errlast);
  last = tackApply(Z_BOX_BOOL,last,last);
  return last;
}

PRIVATE ExprPtr compileExistsMap(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_EXISTS,env,last,&nonelast,&errlast);
  /* nonelast should be NIL */
  last = tackApply(Z_UNBOXED_FALSE,NIL,last);
  errlast = tackApply(Z_UNBOXED_TRUE,NIL,errlast);
  last = compileMergeBin(last,errlast);
  last = tackApply(Z_BOX_BOOL,last,last);
  return last;
}

PRIVATE ExprPtr compileSomeOf(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_SOME,env,last,&nonelast,&errlast);
  /* nonelast should be NIL */
  (void) compileRaiseImm(I_EMPTYEXC,env,last); /* raise Empty */
  /* if not found, raise Empty */
  return errlast;
}

PRIVATE ExprPtr compileSomeMap(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_SOME,env,last,&nonelast,&errlast);
  /* nonelast should be NIL */
  (void) compileRaiseImm(I_EMPTYEXC,env,last); /* raise Empty */
  /* if not found, raise Empty */
  return errlast;
}

PRIVATE ExprPtr compileWhile(cond,suchthat,body,ik,env,last,nonelastp,errlastp)
IN CONST ExprPtr cond,suchthat,body;
IN IterKind ik;
IN CONST compEnv *env;
INOUT ExprPtr last;
INOUT ExprPtr *nonelastp,*errlastp;
{
  struct compilesimpleof *data = (struct compilesimpleof *)alloctmp(sizeof(struct compilesimpleof));
  ExprPtr elselast,outlast,val,resr,start;

  fillData(data,ik);
  outlast = NIL;
  start = tackMerge(last); /* insert a (not yet useful) merge node */
  last = translateStmt(cond,env,start);
  last = tackApply(Z_UNBOX_BOOL,last,last);
  last = compileIf(last,last,&elselast);
  /* if cond failed, exit of the loop and return neutral */
  elselast = (*data->neutral)(env,elselast,nonelastp,errlastp);
  outlast = mergeContinuations(elselast,elselast,outlast);
  /* otherwise, while cond is true, look for first time when suchthat becomes true */
  if (suchthat!=trueValue)
    {
      last = translateStmt(suchthat,env,last);
      last = tackApply(Z_UNBOX_BOOL,last,last);
      last = compileIf(last,last,&elselast);
      /* if suchthat is false, loop back */
      (void) mergeContinuations(elselast,elselast,start); /* first elselast is dummy */
    }
  last = translateStmt(body,env,last);
  last = (*data->promote)(env,last);
  last = (*data->combine1)(env,last,errlastp);
  /* now, loop and combine */
  resr = tackApply(Z_BOX_LOCAL_REF,last,last);
  start = tackMerge(resr); /* insert a (not yet useful) merge node */
  last = translateStmt(cond,env,start);
  last = tackApply(Z_UNBOX_BOOL,last,last);
  last = compileIf(last,last,&elselast);
  /* if cond failed, exit of the loop and return !resr */
  elselast = tackApply(Z_UNBOX_LOCAL_REF,resr,elselast);
  outlast = mergeContinuations(elselast,elselast,outlast);
  /* otherwise, check suchthat */
  if (suchthat!=trueValue)
    {
      last = translateStmt(suchthat,env,last);
      last = tackApply(Z_UNBOX_BOOL,last,last);
      last = compileIf(last,last,&elselast);
      /* if suchthat is false, loop back */
      (void) mergeContinuations(elselast,elselast,start); /* first elselast is dummy */
    }
  last = translateStmt(body,env,last);
  last = (*data->promote)(env,last);
  val = tackApply(Z_UNBOX_LOCAL_REF,resr,last);
  last = (*data->combine)(last,env,val,errlastp);
  last = compileRefSet(Z_LOCAL_REF_SET,resr,last,last); /* store new result */
  (void) mergeContinuations(last,last,start); /* first last is dummy */
  /* and finish: */
  return outlast;
}

PRIVATE ExprPtr compileWhileNoFuss(e,ik,env,last)
IN CONST ExprPtr e;
IN IterKind ik;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),ik,
		      env,last,&nonelast,&errlast);
  return last;
}

PRIVATE ExprPtr compileInterWhile(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileOf(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_INTER,
		   env,last,&nonelast,&errlast);
  if (nonelast!=NIL)
    (void) compileRaiseImm(I_EMPTYEXC,env,nonelast); /* raise Empty */
  return last;
}

PRIVATE ExprPtr compileAllWhile(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileOf(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_ALL,
		   env,last,&nonelast,&errlast);
  /* nonelast should be NIL */
  last = tackApply(Z_UNBOXED_TRUE,NIL,last);
  errlast = tackApply(Z_UNBOXED_FALSE,NIL,errlast);
  last = compileMergeBin(last,errlast);
  last = tackApply(Z_BOX_BOOL,last,last);
  return last;
}

PRIVATE ExprPtr compileExistsWhile(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileOf(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_EXISTS,
		   env,last,&nonelast,&errlast);
  /* nonelast should be NIL */
  last = tackApply(Z_UNBOXED_FALSE,NIL,last);
  errlast = tackApply(Z_UNBOXED_TRUE,NIL,errlast);
  last = compileMergeBin(last,errlast);
  last = tackApply(Z_BOX_BOOL,last,last);
  return last;
}

PRIVATE ExprPtr compileSomeWhile(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr nonelast,errlast;

  errlast = NIL;
  last = compileOf(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_SOME,
		   env,last,&nonelast,&errlast);
  /* nonelast should be NIL */
  (void) compileRaiseImm(I_EMPTYEXC,env,last); /* raise Empty */
  /* if not found, raise Empty */
  return errlast;
}

PRIVATE ExprPtr translateHandle(body,handlers,env,last)
IN CONST ExprPtr body,handlers;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr elselast,outlast,ee;
  compEnv *cenv = (compEnv *)alloctmp(sizeof(compEnv));
  ExprPtr x,val,normlast,errlast,handler,match;
  ENTER;

  normlast = compileLoop(Z_START_HANDLE,NIL,last,&elselast);
  /* now, compile handlers */
  x = tackApply(Z_GETTHEEXPR,NIL,elselast); /* get theExpr */
  handler = mergeContinuations(x,x,NIL); /* insert label, so that we can tack
					    here goto's coming from Raise()s
					    in the same function */
  ee = compileKthEnv(0,env,&last);
  cenv->letconsenv = zApplyBin(Z_CONSENV,ee,KINT(COUNTEDN(handlers)));
  val = NIL; /* make dummy environment */
  cenv->env = MakePair(E_CONS,val,cenv->env); /* extend environment */
  (void) KEEP(cenv->env);
  cenv->level = env->level+1;
  MAKECONS(NewMap(),env->funs,cenv->funs);
  (void) KEEP(cenv->funs);
  cenv->letdelayref = MakeSing(E_REF,NewMap());
  (void) KEEP(cenv->letdelayref);
  cenv->handlers = env->handlers;
  cenv->constpatmemoref = env->constpatmemoref;
  outlast = errlast = NIL;
  match = COUNTEDVAL(handlers);
  last = compilePattern(CAR(match),x,cenv,val,&outlast,&errlast);
  /* compile decomposition of pattern */
  if (errlast!=NIL)
    (void) compileRaise(x,cenv,errlast); /* propagate exception raising */
  /* we assume that raise never returns (check with the debugger) */
  last = val = tackAssign(cenv->letconsenv,last);
  SETCAR(cenv->env,val);
  last = dumpLetStores(REFVAL(cenv->letdelayref),val,last);
  last = translateStmt(CDR(match),cenv,last);
  outlast = mergeContinuations(last,last,outlast);
  /* push handler onto static exception stack before compiling body,
     and pop it afterwards */
  MAKECONS(last,env->handlers,cenv->handlers);
  (void) KEEP(cenv->handlers);
  cenv->env = env->env;
  cenv->level = env->level;
  cenv->funs = env->funs;
  /* compile body */
  last = translateStmt(body,cenv,normlast);
  /* and pop handler */
  last = tackApply(Z_END_HANDLE,NIL,last);
  outlast = mergeContinuations(last,last,outlast);
  QUIT;
  return outlast;
}

PRIVATE ExprPtr translateMu1(rec,modlab,modval,env,last,share)
IN CONST ExprPtr rec,modlab,modval;
IN CONST compEnv *env;
INOUT ExprPtr last;
IN Boolean share;
{
  ExprPtr reclast,lbllast;

  reclast = translateStmt(rec,env,last);
  lbllast = compileLabel(modlab,reclast);
  last = translateStmt(modval,env,lbllast);
  last = tackApplyTer(share?Z_RMU1:Z_UNSHARED_RMU1,reclast,lbllast,last,last);
  return last;
}

PRIVATE ExprPtr translateMuGnrl(rec,mod,env,last,share)
IN CONST ExprPtr rec,mod;
IN CONST compEnv *env;
INOUT ExprPtr last;
IN Boolean share;
{
  ExprPtr reclast;

  reclast = translateStmt(rec,env,last);
  last = translateStmt(mod,env,reclast);
  last = tackApplyBin(share?Z_RMUX:Z_UNSHARED_RMUX,reclast,last,last);
  return last;
}

PRIVATE ExprPtr translatePack(e,type,env,last)
IN CONST ExprPtr e,type;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  last = translateStmt(e,env,last);
  type = compileConstant(type,last,true);
  last = tackApplyBin(Z_BOX_PACK,last,type,type);
  return last;
}

PRIVATE ExprPtr compileEqual(e1,e2,env,last)
IN CONST ExprPtr e1;
IN CONST ExprPtr e2;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  ExprPtr l1,l2;

  l1 = translateStmt(e1,env,last);
  l2 = translateStmt(e2,env,l1);
  last = tackApplyBin(Z_EQUAL,l1,l2,l2);
  return last;
}

DEF_PRIVATE_ACTION(translatestmt)
{
  struct translatestmt *data;

  data = (struct translatestmt *)env;
  return translateStmt(data->e,data->env,data->last);
}

PRIVATE ExprPtr translateStmt1(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
{
  struct translatestmt *data = (struct translatestmt *)alloctmp(sizeof(struct translatestmt));

  data->e = e;
  data->env = env;
  data->last = last;
  return CallCC(ACTION_NAME(translatestmt),(ExprPtr)data,true);
}

PRIVATE ExprPtr translateStmt(e,env,last)
IN CONST ExprPtr e;
IN CONST compEnv *env;
INOUT ExprPtr last;
	/* tacks program at the end of last,
	   returns the ExprPtr corresponding to
	   the output variable, where we find the result. As a label,
	   it is also the place where it is defined. */
{
  ENTER;

  if (STKFULL(last))
    return translateStmt1(e,env,last);
again:
  (void) alloctmp(0L);
  if (last!=NIL)
    switch (e->kind)
      {
      case E_NIL:
	last = compileNil(env,last);
	break;
      case E_INT:
	last = compileInt(e,last);
	break;
      case E_REAL: case E_TUPLE_0: case E_RECORD_0:
	last = compileConstant(e,last,true);
	break;
      case E_COMPLEX:
	last = translateBinary(Z_BOX_COMPLEX,REPART(e),IMPART(e),env,last);
	break;
      case E_STRING:
	last = compileString(e,last);
	break;
      case E_EMPTY:
	last = compileNewMap(env,last); /* build empty set */
	break;
      case E_REF:
	last = translateUnary(Z_BOX_REF,REFVAL(e),env,last);
	break;
      case E_TUPLE_1:
	last = translateUnary(Z_BOX_TUPLE_1,REFVAL(e),env,last);
	break;
      case E_UNSHARED_TUPLE_1:
	last = translateUnary(Z_BOX_UNSHARED_TUPLE_1,REFVAL(e),env,last);
	break;
      case E_RECORD_1:
	last = translateRecord1(Z_BOX_RECORD_1,e,env,last);
	break;
      case E_UNSHARED_RECORD_1:
	last = translateRecord1(Z_BOX_UNSHARED_RECORD_1,e,env,last);
	break;
      case E_DATACON:
	last = translateDataCon(COUNTEDN(e),COUNTEDVAL(e),env,last);
	break;
      case E_TUPLE_2:
	last = translateBinary(Z_BOX_TUPLE_2,CAR(e),CDR(e),env,last);
	break;
      case E_UNSHARED_TUPLE_2:
	last = translateBinary(Z_BOX_UNSHARED_TUPLE_2,CAR(e),CDR(e),env,last);
	break;
      case E_CONS:
	last = translateBinary(Z_BOX_CONS,CAR(e),CDR(e),env,last);
	break;
      case E_UNSHARED_CONS:
	last = translateBinary(Z_BOX_UNSHARED_CONS,CAR(e),CDR(e),env,last);
	break;
      case E_TUPLE_GNRL:
	last = translateTupleGnrl(Z_BOX_TUPLE_GNRL,e,env,last);
	break;
      case E_UNSHARED_TUPLE_GNRL:
	last = translateTupleGnrl(Z_BOX_UNSHARED_TUPLE_GNRL,e,env,last);
	break;
      case E_RECORD_GNRL:
	last = translateRecordGnrl(Z_BOX_RECORD_GNRL,e,env,last);
	break;
      case E_UNSHARED_RECORD_GNRL:
	last = translateRecordGnrl(Z_BOX_UNSHARED_RECORD_GNRL,e,env,last);
	break;
      case E_MAPLET:
	last = translateBinary(Z_BOX_SINGLETON,MAPLETX(e),MAPLETY(e),env,last);
	break;
      case E_OVERWRITE:
	last = translateBinary(Z_OVERWRITE,CAR(e),CDR(e),env,last);
	break;
      case E_UNDERWRITE:
	last = translateRevBinary(Z_OVERWRITE,CAR(e),CDR(e),env,last);
	break;
      case E_QUOTE:
	/* it is usually illegal to compile an E_QUOTE. However,
	   we allow it if it quotes a constant value, which would evaluate
	   to itself anyway. This is notably the case with fixity bindings,
	   which are E_COUNTED(symbol,prec). */
	/* it may happen if the compiled code contains a reference to 'it',
	   but as this is bad practice to compile code using 'it', I think it
	   is better to forbid it. */
	last = translateStmt(REFVAL(e),env,last);
	break;
      case E_COUNTED: /* fixity information */
	/* this is used mainly by the debugger to retrieve information
	   so as to print internal datastructures with the right infix or
	   prefix status. As we chose not to have a debugger for compiled
	   code, this is useless. */
	last = compileNil(env,last);
	break;
      case E_VARREF:
	last = compileLookup(VARRANK(e),VARLEVEL(e),env,last);
	break;
      case E_LET:
	last = compileLet(LETBIND(e),LETBODY(e),env,last);
	break;
      case E_POLYFUN:
	last = compilePolyFun(COUNTEDVAL(e),COUNTEDN(e),env,last);
	last = compileClosure(last,getEnv(env),last);
	break;
      case E_MEMO_FN:
	last = compileMemoFn(MEMOFUN(e),MEMOREM(e),env,last);
	break;
      case E_APPL: case E_GOTO:
	/* we actually lose the fact that E_GOTO is a tail call. */
	last = compileApply(APPLFUN(e),APPLARG(e),env,last);
	break;
      case E_GET:
	last = translateStmt(GETREF(e),env,last);
	last = tackApply(Z_UNBOX_REF,last,last);
	break;
      case E_SET:
	last = compileSet(SETREF(e),SETVAL(e),env,last);
	break;
      case E_MAPGET:
	last = compileMapGet(CAR(e),CDR(e),env,last);
	break;
      case E_TSELECT:
	last = compileTupleSelect(CAR(e),CDR(e),env,last);
	break;
      case E_TSEL1:
	last = translateStmt(CDR(e),env,last);
	last = tackApply(Z_UNBOX_REF,last,last);
	break;
      case E_TSEL2:
	last = translateStmt(CDR(e),env,last);
	last = tackApply((CAR(e)==ione)?Z_CAR:Z_CDR,last,last);
	break;
      case E_TSELGNRL:
	last = translateStmt(CDR(e),env,last);
	last = compileSelectGnrl1(NIL,CAR(e),last,env,last);
	break;
      case E_RSELECT:
	last = compileRecordSelect(CAR(e),CDR(e),env,last);
	break;
      case E_RSEL1:
	last = translateStmt(CDR(e),env,last);
	last = tackApply(Z_CDR,last,last);
	break;
      case E_RSELGNRL:
	last = translateStmt(CDR(e),env,last);
	last = compileSelectGnrl(CAR(e),last,env,last);
	break;
      case E_RMU1:
	last = translateMu1(CAR(e),CAR(CDR(e)),CDR(CDR(e)),env,last,true);
	break;
      case E_UNSHARED_RMU1:
	last = translateMu1(CAR(e),CAR(CDR(e)),CDR(CDR(e)),env,last,false);
	break;
      case E_RMUX:
	last = translateMuGnrl(CAR(e),CDR(e),env,last,true);
	break;
      case E_UNSHARED_RMUX:
	last = translateMuGnrl(CAR(e),CDR(e),env,last,false);
	break;
      case E_IF:
	last = translateIf(TESTIF(e),CAR(TESTFORK(e)),CDR(TESTFORK(e)),env,last);
	break;
      case E_PROG:
	last = translateStmt(PROGDO(e),env,last);
	e = PROGTHEN(e);
	goto again;
      case E_AND:
	last = compileAnd(PROGDO(e),PROGTHEN(e),env,last);
	break;
      case E_OR:
	last = compileOr(PROGDO(e),PROGTHEN(e),env,last);
	break;
      case E_EQUAL:
	last = compileEqual(CAR(e),CDR(e),env,last);
	last = tackApply(Z_BOX_BOOL,last,last);
	break;
      case E_NEQUAL:
	last = compileEqual(CAR(e),CDR(e),env,last);
	last = tackApply(Z_NOT,last,last);
	last = tackApply(Z_BOX_BOOL,last,last);
	break;
      case E_NOT:
	last = translateStmt(REFVAL(e),env,last);
	last = tackApply(Z_UNBOX_BOOL,last,last);
	last = tackApply(Z_NOT,last,last);
	last = tackApply(Z_BOX_BOOL,last,last);
	break;
      case E_PROG_OF:
	last = compileOfNoFuss(e,C_PROG,env,last);
	if (last!=NIL)
	  last = compileNil(env,last);
	break;
      case E_UNSHARED_LIST_OF: /*not done*/
      case E_LIST_OF:
	last = compileOfNoFuss(e,C_LIST,env,last);
endlistof:
	if (last!=NIL)
	  {
	    last = compileQueueToList(env,last);
	  }
	break;
      case E_SET_OF:
	last = compileOfNoFuss(e,C_SET,env,last);
	break;
      case E_UNSHARED_APPEND_OF: /*not done*/
      case E_APPEND_OF:
	last = compileOfNoFuss(e,C_APPEND,env,last);
	goto endlistof;
      case E_UNION_OF:
	last = compileOfNoFuss(e,C_UNION,env,last);
	break;
      case E_UNION_UNDER_OF:
	last = compileOfNoFuss(e,C_UNION_UNDER,env,last);
	break;
      case E_INTER_OF:
	last = compileInterOf(e,env,last);
	break;
      case E_MAP_OF:
	last = compileOfNoFuss(e,C_MAP,env,last);
	break;
      case E_MAP_UNDER_OF:
	last = compileOfNoFuss(e,C_MAP_UNDER,env,last);
	break;
      case E_ALL_OF:
	last = compileAllOf(e,env,last);
	break;
      case E_EXISTS_OF:
	last = compileExistsOf(e,env,last);
	break;
      case E_UNSHARED_SOME_OF: /*not done*/
      case E_SOME_OF:
	last = compileSomeOf(e,env,last);
	break;
      case E_PROG_MAP:
	last = compileMapNoFuss(e,C_PROG,env,last);
	if (last!=NIL)
	  last = compileNil(env,last);
	break;
      case E_UNSHARED_LIST_MAP: /*not done*/
      case E_LIST_MAP:
	last = compileMapNoFuss(e,C_LIST,env,last);
endlistmap:
	if (last!=NIL)
	  {
	    last = compileQueueToList(env,last);
	  }
	break;
      case E_SET_MAP:
	last = compileMapNoFuss(e,C_SET,env,last);
	break;
      case E_UNSHARED_APPEND_MAP: /*not done*/
      case E_APPEND_MAP:
	last = compileMapNoFuss(e,C_APPEND,env,last);
	goto endlistmap;
      case E_UNION_MAP:
	last = compileMapNoFuss(e,C_UNION,env,last);
	break;
      case E_UNION_UNDER_MAP:
	last = compileMapNoFuss(e,C_UNION_UNDER,env,last);
	break;
      case E_INTER_MAP:
	last = compileInterMap(e,env,last);
	break;
      case E_MAP_MAP:
	last = compileMapNoFuss(e,C_MAP,env,last);
	break;
      case E_MAP_UNDER_MAP:
	last = compileMapNoFuss(e,C_MAP_UNDER,env,last);
	break;
      case E_ALL_MAP:
	last = compileAllMap(e,env,last);
	break;
      case E_EXISTS_MAP:
	last = compileExistsMap(e,env,last);
	break;
      case E_UNSHARED_SOME_MAP: /*not done*/
      case E_SOME_MAP:
	last = compileSomeMap(e,env,last);
	break;
      case E_PROG_WHILE:
	last = compileWhileNoFuss(e,C_PROG,env,last);
	if (last!=NIL)
	  last = compileNil(env,last);
	break;
      case E_UNSHARED_LIST_WHILE: /*not done*/
      case E_LIST_WHILE:
	last = compileWhileNoFuss(e,C_LIST,env,last);
endlistwhile:
	if (last!=NIL)
	  {
	    last = compileQueueToList(env,last);
	  }
	break;
      case E_SET_WHILE:
	last = compileWhileNoFuss(e,C_SET,env,last);
	break;
      case E_UNSHARED_APPEND_WHILE: /*not done*/
      case E_APPEND_WHILE:
	last = compileWhileNoFuss(e,C_APPEND,env,last);
	goto endlistwhile;
      case E_UNION_WHILE:
	last = compileWhileNoFuss(e,C_UNION,env,last);
	break;
      case E_UNION_UNDER_WHILE:
	last = compileWhileNoFuss(e,C_UNION_UNDER,env,last);
	break;
      case E_INTER_WHILE:
	last = compileInterWhile(e,env,last);
	break;
      case E_MAP_WHILE:
	last = compileWhileNoFuss(e,C_MAP,env,last);
	break;
      case E_MAP_UNDER_WHILE:
	last = compileWhileNoFuss(e,C_MAP_UNDER,env,last);
	break;
      case E_ALL_WHILE:
	last = compileAllWhile(e,env,last);
	break;
      case E_EXISTS_WHILE:
	last = compileExistsWhile(e,env,last);
	break;
      case E_UNSHARED_SOME_WHILE: /*not done*/
      case E_SOME_WHILE:
	last = compileSomeWhile(e,env,last);
	break;
      case E_HANDLE:
	last = translateHandle(HBODY(e),HHANDLERS(e),env,last);
	break;
      case E_RAISE:
	last = translateStmt(RBODY(e),env,last);
	(void) compileRaise(last,env,last);
	last = NIL;
	break;
      case E_EXCEPTION:
	last = translateUnary((e->mark & MARK_SHARED)?Z_BOX_SHARED_EXCEPTION:
			      Z_BOX_UNSHARED_EXCEPTION,EXCVAL(e),env,last);
	break;
      case E_DELAY:
	last = translateUnary(Z_MAKE_PROMISE,DCODE(e),env,last);
	break;
      case E_PACK:
	last = translatePack(PACKVAL(e),PACKTYPE(e),env,last);
	break;
      case E_DEBUGZONE: case E_DEBUGFUNC:
      case E_DZ_BRK_ENTRY: case E_DF_BRK_ENTRY:
      case E_DZ_BRK_EXIT: case E_DF_BRK_EXIT:
      case E_DZ_BRK_BOTH: case E_DF_BRK_BOTH:
	e = DZCODE(e);
	goto again;
      case E_PROFILE:
      case E_CLR_PROFILE:
	e = PROFCODE(e);
	goto again;
	/* the following tags should really never be evaluated: */
      case E_EXTERNAL: case E_COMPILED_CLOSURE:
      case E_INTERPRETED_CLOSURE: case E_MEMO_CLOSURE:
      case E_STACK: case E_ENV:
	kinderror(message(430),e->kind);
    }
  QUIT; /* everything is kept in a translation structure,
	   so it is safe to QUIT. */
  return last; /* return variable */
}

#endif
