(* First-class stores, as used by threads (one thread, one store).
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

(* These stores are not undoable/redoable as Morrissett's ones.
   (In his paper about 'Refining first-class stores'.)
   This is liberally adapted from the appendix A of Cooper and
   Morrissett's paper on adding threads to SML. *)

type store = unit ref (* to provide unique IDs *)

abstype 'a var = VAR of (store -m> 'a ref) ref
with
   eqtype var

   exception Var (* instead of Undefined for Cooper and Morrissett *)
	(* raised when accessing an undefined local var *)

   fun new_store () = ref () (* creates a new store; named new_env(), originally *)

   val current_store = ref (new_store ())

   fun var x = VAR (ref {!current_store => ref x})
	(* originally, var was unit -> '_a ref, but I prefer this one
	   since it makes it look more like the ref constructor. *)

   fun getvar (VAR (ref m)) = (* gets value of variable in the current_store.
	it was called get originally. *)
       if !current_store inset m
	  then ?m (!current_store)
       else raise Var

   fun setvar (VAR (v as ref m),value) = (* sets value of variable in the current_store.
	it was called set originally, and was currified. *)
       let val cs = !current_store
       in
	  if cs inset m
	     then ?m cs := value
	  else v := m ++ {cs => ref value}
       end
end
