/* The HimML bytecode interpreter.
   Copyright (C) 1995-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "portable.h"
#if TRANSLATOR
#include "mem.h"
#include "eval.h"
#include "bc.h"
#include "expr.h"
#include "timing.h"
#include "debug.h"
#include "bciaux.h"
#include "rbci.h"
#include "evlproto.h"
#include "typproto.h"
#include "errproto.h"
#include "exeproto.h"
#include "cmpproto.h"
#include "stkproto.h"
#include "dbgproto.h"
#if OS==AMIGA_OS
#include <dos.h>
#endif
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "hakproto.h"
#ifdef HAS_PROTOTYPES
#include "sbci.h"
#endif

#define __SEG__ EVAL

#define CASE(opcode) case opcode
#define NEXT goto again
#define NEXTOP(opval) { op = opval; goto again_op; }
#define SWITCH

#undef DEBUG
#undef HDEBUG
#undef CALLDEBUG
#undef PROFDEBUG

#ifdef HIMMLRUN
#undef PROFILING
#define PROFILING 0
#endif

#define USE_GNU_TRICK

#ifdef USE_GNU_TRICK
/* If we use gcc v>=2, then the inner loop of the bytecode interpreter
   will use computed gotos instead of switch(). This is faster. */
#ifdef __GNUC__
#if __GNUC__ >= 2
#ifndef DEBUG
#undef SWITCH
#undef CASE
#define CASE(opcode) MERGE(lab_,opcode)
#undef NEXT
#define NEXT    goto *bclabels[*pc++]
#undef NEXTOP
#define NEXTOP(opval) goto *bclabels[opval]
#endif
#endif
#endif
#endif

#ifdef HDEBUG
PRIVATE nHandle = 0, nNextHandle = 0, nInlineRaise = 0, nLongJumpRaise = 0;
#endif

PUBLIC ExprPtr CGlobals[N_C_GLOBALS];
EXTERN ExprPtr toplevelCont;
EXTERN ExprPtr specialMarker;
EXTERN Boolean itBarrier;

EXTERN ExprPtr subscriptException,stringnthException;

#if DEBUGGING
PUBLIC Boolean debugEnabled = true;
EXTERN Boolean stopwanted;
PRIVATE struct bcframes *parentBcFrames;
PRIVATE struct bcstacks *parentStack;
PRIVATE Boolean raiseInPrimitive = false;
#endif

#if PROFILING
EXTERN long gc_utime,gc_stime,gc_rtime,gc_count;
EXTERN ExprPtr profileExprs;
#if PROFILING & PROFILING_PROPER_TIMES
PUBLIC ExprPtr currentProf = NOEXPR;
#endif

#if PROFILING & PROFILING_PROPER_TIMES
PRIVATE void profEnterProper(data,tms)
INOUT REGISTER Profile *data;
#if OS==AMIGA_OS
IN CONST unsigned int tms[];
#else
IN CONST struct tms *tms;
#endif
{
#if OS==AMIGA_OS
  data->last.proper.time.clock[0] = tms[0];
  data->last.proper.time.clock[1] = tms[1];
#else
  data->last.proper.time.tms_utime = tms->tms_utime;
  data->last.proper.time.tms_stime = tms->tms_stime;
#endif
  data->last.proper.gc_time.tms_utime = gc_utime;
  data->last.proper.gc_time.tms_stime = gc_stime;
  data->last.proper.gc_count = gc_count;
}
#else
#define profEnterProper(e,tms)
#endif

#if PROFILING & PROFILING_PROPER_TIMES
#if OS==AMIGA_OS
PRIVATE void subtime(c3,c1,c2)
IN CONST unsigned int c1[];
IN CONST unsigned int c2[];
OUT unsigned int c3[];
{
  unsigned int musec;

  musec = c1[1] - c2[1];
  if (musec>c1[1]) /* borrow */
    {
      c3[0] = c1[0]-c2[0]-1;
      c3[1] = 1000000+musec; /* this overflows, but the Amiga does not care */
    }
  else
    {
      c3[0] = c1[0]-c2[0];
      c3[1] = musec;
    }
}
#endif

PRIVATE void profExitProper(data,tms)
INOUT REGISTER Profile *data;
#if OS==AMIGA_OS
IN CONST unsigned int tms[];
#else
IN CONST struct tms *tms;
#endif
{
#if OS==AMIGA_OS
  subtime(data->stat.proper.time.clock,tms,data->last.proper.time.clock);
#else
  data->stat.proper.time.tms_utime += tms->tms_utime - data->last.proper.time.tms_utime;
  data->stat.proper.time.tms_stime += tms->tms_stime - data->last.proper.time.tms_stime;
#endif
  data->stat.proper.gc_time.tms_utime += gc_utime - data->last.proper.gc_time.tms_utime;
  data->stat.proper.gc_time.tms_stime += gc_stime - data->last.proper.gc_time.tms_stime;
  data->stat.proper.gc_count += gc_count - data->last.proper.gc_count;
}
#else
#define profExitProper(e,tms)
#endif

#ifdef PROFDEBUG
EXTERN extDesc profExtDesc;

Profile *PROFDATA(prof)
IN CONST ExprPtr prof;
{
  if (prof==NOEXPR)
    abort();
  if (prof->kind!=(CHAR)BUILD_HEADER_KIND)
    abort();
  prof = CAR(CDR(ARGS(prof)));
  if (prof->kind!=(CHAR)E_EXTERNAL)
    abort();
  if (EXTDESC(prof)!=&profExtDesc)
    abort();
  return (Profile *)EXTDATA(prof);
}
#else
Profile *PROFDATA(prof)
IN CONST ExprPtr prof;
{
  return (Profile *)EXTDATA(CAR(CDR(ARGS(prof))));
}
#endif

#ifdef OBSOLETE
#if PROFILING & PROFILING_PROPER_TIMES
PUBLIC void syncProfProper(save)
IN CONST ExprPtr save;
{
#if OS==AMIGA_OS
  unsigned int clock[2];

  if (save!=NULL)
    {
      (void) timer(clock);
      profEnterProper(PROFDATA(save),clock);
    }
#else
  struct tms tms;

  if (save!=NULL)
    {
      (void) times(&tms);
      profEnterProper(PROFDATA(save),&tms);
    }
#endif
  currentProf = save;
}
#endif
#endif

#if PROFILING & PROFILING_TOTAL_TIMES
PRIVATE void profEnter(e,tms)
INOUT ExprPtr e;
#if OS==AMIGA_OS
IN CONST unsigned int tms[];
#else
IN CONST struct tms *tms;
#endif
{
  REGISTER struct Profile *data;

  data = PROFDATA(e);
  if (data->nEntries++==0)
    {
#if OS==AMIGA_OS
      data->last.total.time.clock[0] = tms[0];
      data->last.total.time.clock[1] = tms[1];
#else
      data->last.total.time.tms_utime = tms->tms_utime;
      data->last.total.time.tms_stime = tms->tms_stime;
#endif
      data->last.total.gc_time.tms_utime = gc_utime;
      data->last.total.gc_time.tms_stime = gc_stime;
      data->last.total.gc_count = gc_count;
    }
}
#else
#define profEnter(e,tms)
#endif

#if PROFILING & PROFILING_TOTAL_TIMES
PRIVATE void profExit(e,tms)
INOUT ExprPtr e;
#if OS==AMIGA_OS
IN CONST unsigned int tms[];
#else
IN CONST struct tms *tms;
#endif
{
  REGISTER struct Profile *data;

  data = PROFDATA(e);
#ifdef PROFDEBUG
  if (data->nEntries==0)
    abort();
#endif
  if (--data->nEntries==0)
    {
#if OS==AMIGA_OS
      subtime(data->stat.total.time.clock,tms,data->last.total.time.clock);
#else
      data->stat.total.time.tms_utime += tms->tms_utime - data->last.total.time.tms_utime;
      data->stat.total.time.tms_stime += tms->tms_stime - data->last.total.time.tms_stime;
#endif
      data->stat.total.gc_time.tms_utime += gc_utime - data->last.total.gc_time.tms_utime;
      data->stat.total.gc_time.tms_stime += gc_stime - data->last.total.gc_time.tms_stime;
      data->stat.total.gc_count += gc_count - data->last.total.gc_count;
    }
}
#else
#define profExit(e,tms)
#endif
#endif

PUBLIC NORETURN void Toplevel(errcond)
IN CONST ExprPtr errcond;
{
  excHandler *handler;

#if PROFILING
  {
    ExprPtr ll,e;
#if OS==AMIGA_OS
    unsigned int clock[2];
#define TMS clock
#else
    struct tms tms;
#define TMS &tms
#endif

#if OS==AMIGA_OS
    (void) timer(clock);
#else
    times(&tms);
#endif

#if PROFILING & PROFILING_PROPER_TIMES
    if (currentProf!=NOEXPR && TYPE(currentProf)==EA_PROFILE)
      profExitProper(PROFDATA(currentProf),TMS);
    currentProf = NOEXPR;
#endif
    for (ll=profileExprs; ll!=NIL; ll=CDR(ll))
      {
	e = CAR(ll);
	if (TYPE(e)==EA_PROFILE)
	  {
	    struct Profile *data;

	    data = PROFDATA(e);
	    if (data->nEntries!=0)
	      {
		data->nEntries = 1;
		profExit(e,TMS);
	      }
	  }
      }
  }
#endif
  if (toplevelCont!=NOEXPR && toplevelCont!=NIL)
    Throw(toplevelCont,errcond);
  berror(message(397));
  /*NOTREACHED*/
}

PRIVATE NORETURN ExprPtr Raise0 P((IN CONST ExprPtr exn)) _NORETURN;

PRIVATE ExprPtr Raise0(exn)
IN CONST ExprPtr exn;
{
  REGISTER excHandler *exc;
  REGISTER ExprPtr oldCont;

  oldCont = NOEXPR;
  for (exc = excStack; exc!=NULL; exc = NEXTEXCHANDLER(exc))
    {
      switch (exc->tag)
	{
	case X_CONT:
	  oldCont = exc->data;
	  break;
	case X_PROF:
	  break;
	default: /* case X_HANDLER: */
	  excStack = exc;
	  THEEXPR = exn;
	  if (oldCont==NOEXPR)
	    {
	      LONGJUMP(exc->jb,1);
	      /*NOTREACHED*/
	    }
	  /* otherwise, reestablish old stack and raise exception
	     in this stack */
	  Throw1(oldCont,NIL,&exc->jb);
	  /*NOTREACHED*/
	  break;
	}
    }
#if DEBUGGING
  raiseInPrimitive = false;
#ifdef STUPID
  {
    ExprPtr exc = exn;

    if (exc->kind==(CHAR)E_CONS)
      exc = CAR(exc);
    if ((COUNTEDN(exc) & EXC_BENIGN)==0)
      {
	(void) DoBreak (NULL, NULL, NULL, NULL, NULL, exn);
	/* Enter the debugger, although we cannot do anything with it. */
      }
  }
#endif
#endif
#ifdef HAS_NORETURN
  (void) Error (exn);
#else
  return Error(exn);
#endif
}

PUBLIC ExprPtr Raise(exn)
IN CONST ExprPtr exn;
{
#if DEBUGGING
  if (debugEnabled)
    raiseInPrimitive = true;
#endif
#ifdef HAS_NORETURN
  (void) Raise0(exn);
#else
  return Raise0(exn);
#endif
}

PRIVATE void ncallSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
}

PRIVATE extDesc ncallDesc = {
  (CHAR *)"ncall",
  { 'n','c','a','l' }, NULL,
  NULL, /* not shared */
  NULL, /* not shared */
  NULL, /* print as default */
  ncallSweep,
  NULL, /* do not load */
};

DEF_ACTION(apply_bytecode_closure)
{
  va_list _args;
  ExprPtr uenv, res;

  uenv = ENV_UNDERLYING(env);
  M_VARCALL2(res,bcEval,(byteCodeChunk *)CAR(uenv),CDR(uenv),ENV_ARITY(env));
  /* call bcEval with varying number of arguments. */
  return res;
}

IMPORT ExprPtr bcEval0 P((IN CONST byteCodeChunk *bc,
			  IN CONST ExprPtr env, ...));

#ifdef HAS_STDARG
PUBLIC ExprPtr bcEval0(IN CONST byteCodeChunk *bc,
		      IN CONST ExprPtr env, ...)
#else
PUBLIC ExprPtr bcEval0(bc, env, va_alist)
IN CONST byteCodeChunk *bc;
IN CONST ExprPtr env;
va_dcl
#endif
{
  va_list _args;
  ExprPtr res;
  REGISTER OPCODE *pc;
  REGISTER ExprPtr *top;
  ExprPtr **maptop;
#define DO_ALLOCBC
#ifdef DO_ALLOCBC
  struct bcstacks *bcstacks_ptr;
#define BCSTACKS_PTR bcstacks_ptr
#define BCSTACKS (*bcstacks_ptr)
  struct bcframes *bcf_ptr;
#define BCF_PTR bcf_ptr
#define BCF (*bcf_ptr)
#define ALLOC_BC() { \
    REGISTER char *p = alloctmp_anchored (MODALIGN(sizeof (struct bcstacks)+ALIGNMENT-1) \
					  + sizeof (struct bcframes), (char *)&env); \
    bcstacks_ptr = (struct bcstacks *)p; \
    bcf_ptr = (struct bcframes *)(p+MODALIGN(sizeof (struct bcstacks)+ALIGNMENT-1)); \
    }
#define theFrames theHeapFrames
#else
  struct bcstacks bcstacks;
#define BCSTACKS_PTR (&bcstacks)
#define BCSTACKS bcstacks
  struct bcframes bcf;
#define BCF_PTR (&bcf)
#define BCF bcf
#define theFrames theStackFrames
#define ALLOC_BC()
#endif
#if PROFILING & PROFILING_PROPER_TIMES
  ExprPtr saveProf;
#endif
  VOLATILE struct bcframe *curbcframe;
  VOLATILE struct xhandlerinfo {
    struct xhandler *curxhandler;
    struct xhandler *endxhandler;
    struct xhandlers {
      struct xhandlers *next;
      struct xhandler exc[MAX_EXC_NEST];
    } *xhandlers;
  } xhi;
#ifdef SWITCH
  OPCODE op;
#else
  static void *bclabels[] = {
#include "labs.c"
  };
#endif

  ALLOC_BC();
#if DEBUGGING!=0
  &bcstacks_ptr; /* so that its value is saved across longjmps */
  &bcf_ptr; /* so that its value is saved across longjmps */
  BCF.parentBcFrames = parentBcFrames;
  BCSTACKS.parentStack = parentStack;
#endif
#if PROFILING
  {
    ExprPtr prof = bc->consts[0];
#if OS==AMIGA_OS
    unsigned int clock[2];
#define TMS clock
#else
    struct tms tms;
#define TMS &tms
#endif

#ifdef PROFDEBUG
    if (prof==NOEXPR || prof->kind!=(CHAR)BUILD_HEADER_KIND)
      abort();
#endif
    if (TYPE(prof)==EA_PROFILE)
      {
#if PROFILING & PROFILING_CALL_COUNTS
	PROFDATA(prof)->nCalls++;
#endif
#if PROFILING & PROFILING_PROPER_TIMES
#if OS==AMIGA_OS
	(void) timer(clock);
#else
	times(&tms);
#endif
#ifdef PROFDEBUG
	if (currentProf!=NOEXPR && currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
	  abort();
#endif
	if (currentProf!=NOEXPR && TYPE(currentProf)==EA_PROFILE)
	  profExitProper(PROFDATA(currentProf),TMS);
	profEnterProper(PROFDATA(prof),TMS);
#endif
	profEnter (prof, TMS);
      }
#if PROFILING & PROFILING_PROPER_TIMES
    saveProf = currentProf;
    currentProf = prof;
#endif
  }
#endif
  BCSTACKS._x.n = 0;
  BCSTACKS._x.next = theFrames;
  theFrames = (gcStackView *)&BCSTACKS._x;
  top = BCSTACKS._x.p;
  maptop = BCSTACKS.mapstack;
  curbcframe = BCF.bcframes;
  xhi.xhandlers = NULL;
  xhi.curxhandler = xhi.endxhandler = NULL;
  /* to keep the compiler happy (so that it does not complain about these
     being used but possibly not defined) */

#define PUSHTOP(x) top += x; BCSTACKS._x.n += x
#define POPTOP(x) top -= x; BCSTACKS._x.n -= x
#define SYNCTOP()

  /*restart:*/
  if (BCSTACKS._x.n+bc->stackSize > BC_STACK_SIZE)
    { /* stack overflow: call oneself recursively to reallocate a new stack */
      ExprPtr uenv;
      /*overflow:*/
#if DEBUGGING!=0
      parentBcFrames = BCF_PTR;
      BCF.curBcFrame = (struct bcframe *)curbcframe;
      parentStack = BCSTACKS_PTR;
#endif
      uenv = ENV_UNDERLYING(env);
      M_VARCALL2(res,bcEval,bc,uenv,ENV_ARITY(env));

      theFrames = BCSTACKS._x.next;
      if (xhi.xhandlers!=NULL) /* if we have installed an exception handler
				  in excStack, pop it. */
	excStack = excStack->next;
#if DEBUGGING!=0
      (void) freetmp (bcstacks_ptr);
#endif
#if PROFILING
      if (currentProf!=NOEXPR)
	{
#if OS==AMIGA_OS
	  unsigned int clock[2];
#define TMS clock
#else
	  struct tms tms;
#define TMS &tms
#endif

#if OS==AMIGA_OS
	  (void) timer(clock);
#else
	  times(&tms);
#endif
#ifdef PROFDEBUG
	  if (currentProf!=NOEXPR && currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
	    abort();
#endif
	  if (TYPE(currentProf)==EA_PROFILE)
	    profExit(currentProf,TMS);
#if PROFILING & PROFILING_PROPER_TIMES
	  if (TYPE(currentProf)==EA_PROFILE)
	    profExitProper(PROFDATA(currentProf),TMS);
#ifdef PROFDEBUG
	  if (saveProf!=NOEXPR && saveProf->kind!=(CHAR)BUILD_HEADER_KIND)
	    abort();
#endif
	  if (saveProf!=NOEXPR && TYPE(saveProf)==EA_PROFILE)
	    profEnterProper(PROFDATA(saveProf),TMS);
	  currentProf = saveProf;
#endif
	}
#endif
      return res;
    }

  pc = bc->instr;
#define RAWDATA bc->rawData

  top[0] = ENV_UNDERLYING(env);
  if (bc->nargs==1)
    {
      FUN_COLLECT(FC_IGN);
      top[1] = arg;
      PUSHTOP(2);
    }
  else if (bc->nargs==0)
    {
      PUSHTOP(1);
    }
  else
    {
      va_list _args;

      VA_START;
      if (ENV_ARITY(env)==bc->nargs)
	{
	  COUNTER i;

	  for (i=1; i<=bc->nargs; i++)
	    top[i] = va_arg(_args,ExprPtr);
	}
      else
	{
	  ExprPtr _arg = va_arg(_args,ExprPtr);

	  if (bc->nargs==2)
	    {
	      top[1] = CAR(_arg);
	      top[2] = CDR(_arg);
	    }
	  else
	    {
	      VecPtr vp = AVAL(_arg);
	      COUNTER i;

	      for (i=1; i<=bc->nargs; i++)
		top[i] = vp->expr[i-1];
	    }
	}
      PUSHTOP(bc->nargs+1);
      va_end(_args);
    }

#ifdef SWITCH
  again:
#endif
#ifdef SWITCH
  op = *pc++;
 again_op:
  switch (op)
    {
#else
      NEXT;
#endif
    CASE(BC_LOOP):
    CASE(BC_FUNCTION):
#if OS==AMIGA_OS
      {
#define BRK_CHK_FREQ 32
	static int i = 0;

	if (++i==BRK_CHK_FREQ)
	  {
	    i = 0;
	    chkabort();
	  }
      }
#endif
#if DEBUGGING
      if (stopwanted) /* can only be true if debugEnabled is true as well. */
	goto do_break;
#endif
      /*fallthrough*/
    CASE(BC_MERGE):
      NEXT;
    CASE(BC_BREAK):
#if DEBUGGING
      if (debugEnabled)
	{
	do_break:
	  BCF.curBcFrame = (struct bcframe *)curbcframe;
	  NEXTOP (DoBreak(bc,pc,top,BCF_PTR,BCSTACKS_PTR,NOEXPR));
	}
      else { NEXT; }
#else
    NEXT;
#endif
    CASE(BC_GOTO):
_goto:
      {
	REGISTER COUNTER n;

	n = (COUNTER)*pc++;
	POPTOP(n);
      }
      pc += *pc;
      NEXT;
    CASE(BC_GOTO_0):
      pc += pc[1];
      NEXT;
    CASE(BC_IF):
      POPTOP(1);
      if (top[0]==falseValue)
	goto _goto;
      pc += 2;
      NEXT;
    CASE(BC_CGLOBAL):
      top[0] = CGlobals[*pc++];
      PUSHTOP(1);
      NEXT;
    CASE(BC_CONST):
      top[0] = bc->consts[*pc++];
      PUSHTOP(1);
      NEXT;
    CASE(BC_GLOBVARREF):
      top[0] = EVARS(GLOBENV)->expr[*pc++];
      PUSHTOP(1);
      NEXT;
    CASE(BC_DROP):
      {
	REGISTER COUNTER n;

	n = (COUNTER)*pc++;
	POPTOP(n);
	NEXT;
      }
    CASE(BC_SWDROP):
      {
	REGISTER COUNTER n;

	n = (COUNTER)*pc++;
	top[n++] = top[-1];
	PUSHTOP(n);
	NEXT;
      }
    CASE(BC_NSWDROP):
      {
	REGISTER COUNTER n, k;

	n = (COUNTER)*pc++;
	k = (COUNTER)*pc++;
	while (k<0)
	  top[n++] = top[k++];
	PUSHTOP(n);
	NEXT;
      }
    CASE(BC_REF):
      SYNCTOP();
      top[-1] = MakeSing(E_REF,top[-1]);
      NEXT;
    CASE(BC_KTUPLE1):
      SYNCTOP();
      top[-1] = Ktuple1(top[-1]);
      NEXT;
    CASE(BC_MAKETUPLE1):
      SYNCTOP();
      top[-1] = MakeSing(E_UNSHARED_TUPLE_1,top[-1]);
      NEXT;
    CASE(BC_KONS):
      SYNCTOP();
      top[-2] = Kons(top[-2],top[-1]);
      POPTOP(1);
      NEXT;
    CASE(BC_CONS):
    CASE(BC_CONSMEMO):
      SYNCTOP();
      top[-2] = MakePair(E_CONS,top[-2],top[-1]);
      POPTOP(1);
      NEXT;
    CASE(BC_KDATA):
      SYNCTOP();
      top[-1] = Kdata((SHORT)pc[0],bc->consts[pc[1]],top[-1]);
      pc += 2;
      NEXT;
    CASE(BC_DATA):
      SYNCTOP();
      top[-1] = MakeData((SHORT)pc[0],bc->consts[pc[1]],top[-1]);
      pc += 2;
      NEXT;
    CASE(BC_FETCH):
    CASE(BC_STARTLISTSWEEP):
    CASE(BC_GETMAPSWEEP):
    CASE(BC_GETSUBMAPSWEEP):
      top[0] = top[*pc++];
      PUSHTOP(1);
      NEXT;
    CASE(BC_FETCH_m8): top[0] = top[-8]; PUSHTOP(1); NEXT;
    CASE(BC_FETCH_m7): top[0] = top[-7]; PUSHTOP(1); NEXT;
    CASE(BC_FETCH_m6): top[0] = top[-6]; PUSHTOP(1); NEXT;
    CASE(BC_FETCH_m5): top[0] = top[-5]; PUSHTOP(1); NEXT;
    CASE(BC_FETCH_m4): top[0] = top[-4]; PUSHTOP(1); NEXT;
    CASE(BC_FETCH_m3): top[0] = top[-3]; PUSHTOP(1); NEXT;
    CASE(BC_FETCH_m2): top[0] = top[-2]; PUSHTOP(1); NEXT;
    CASE(BC_FETCH_m1): top[0] = top[-1]; PUSHTOP(1); NEXT;
    CASE(BC_CAR):
      top[-1] = CAR(top[-1]);
      NEXT;
    CASE(BC_FETCHCAR):
      top[0] = CAR(top[*pc++]);
      PUSHTOP(1);
      NEXT;
    CASE(BC_FETCHCAR_m8): top[0] = CAR(top[-8]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCAR_m7): top[0] = CAR(top[-7]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCAR_m6): top[0] = CAR(top[-6]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCAR_m5): top[0] = CAR(top[-5]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCAR_m4): top[0] = CAR(top[-4]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCAR_m3): top[0] = CAR(top[-3]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCAR_m2): top[0] = CAR(top[-2]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCAR_m1): top[0] = CAR(top[-1]); PUSHTOP(1); NEXT;
    CASE(BC_CDR):
      top[-1] = CDR(top[-1]);
      NEXT;
    CASE(BC_FETCHCDR):
      top[0] = CDR(top[*pc++]);
      PUSHTOP(1);
      NEXT;
    CASE(BC_FETCHCDR_m8): top[0] = CDR(top[-8]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCDR_m7): top[0] = CDR(top[-7]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCDR_m6): top[0] = CDR(top[-6]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCDR_m5): top[0] = CDR(top[-5]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCDR_m4): top[0] = CDR(top[-4]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCDR_m3): top[0] = CDR(top[-3]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCDR_m2): top[0] = CDR(top[-2]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHCDR_m1): top[0] = CDR(top[-1]); PUSHTOP(1); NEXT;
    CASE(BC_CARCDR):
      {
	REGISTER ExprPtr a;

	a = top[-1];
	top[0] = CDR(a);
	top[-1] = CAR(a);
	PUSHTOP(1);
	NEXT;
      }
    CASE(BC_CDRCAR):
      {
	REGISTER ExprPtr a;

	a = top[-1];
	top[0] = CAR(a);
	top[-1] = CDR(a);
	PUSHTOP(1);
	NEXT;
      }
    CASE(BC_REFVAL):
      top[-1] = REFVAL(top[-1]);
      NEXT;
    CASE(BC_FETCHREFVAL):
      top[0] = REFVAL(top[*pc++]);
      PUSHTOP(1);
      NEXT;
    CASE(BC_FETCHREFVAL_m8): top[0] = REFVAL(top[-8]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHREFVAL_m7): top[0] = REFVAL(top[-7]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHREFVAL_m6): top[0] = REFVAL(top[-6]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHREFVAL_m5): top[0] = REFVAL(top[-5]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHREFVAL_m4): top[0] = REFVAL(top[-4]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHREFVAL_m3): top[0] = REFVAL(top[-3]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHREFVAL_m2): top[0] = REFVAL(top[-2]); PUSHTOP(1); NEXT;
    CASE(BC_FETCHREFVAL_m1): top[0] = REFVAL(top[-1]); PUSHTOP(1); NEXT;
    CASE(BC_KRECORD1):
      SYNCTOP();
      top[-1] = Krecord1(bc->consts[*pc++],top[-1]);
      NEXT;
    CASE(BC_MAKERECORD1):
      SYNCTOP();
      top[-1] = MakePair(E_RECORD_1,bc->consts[*pc++],top[-1]);
      NEXT;
    CASE(BC_KTUPLE2):
      SYNCTOP();
      top[-2] = Ktuple2(top[-2],top[-1]);
      POPTOP(1);
      NEXT;
    CASE(BC_MAKETUPLE2):
      SYNCTOP();
      top[-2] = MakePair(E_TUPLE_2,top[-2],top[-1]);
      POPTOP(1);
      NEXT;
    CASE(BC_KTUPLEGNRL):
      SYNCTOP();
      {
	REGISTER INT n;
	REGISTER ExprPtr *ep;

	n = (INT)*pc++;
	ep = top-n;
	ep[0] = KtupleGnrl(n,ep);
	n--;
	POPTOP(n);
	NEXT;
      }
    CASE(BC_MAKETUPLEGNRL):
      SYNCTOP();
      {
	REGISTER INT n;
	REGISTER ExprPtr *ep;

	n = (INT)*pc++;
	ep = top-n;
	ep[0] = MakeRecord(E_TUPLE_GNRL,NIL,n,ep);
	n--;
	POPTOP(n);
	NEXT;
      }
    CASE(BC_KRECORDGNRL):
      SYNCTOP();
      {
	REGISTER INT n;
	REGISTER ExprPtr *ep;

	n = (INT)*pc++;
	ep = top-n;
	ep[0] = KrecordGnrl(bc->consts[*pc++],n,ep);
	n--;
	POPTOP(n);
	NEXT;
      }
    CASE(BC_MAKERECORDGNRL):
      SYNCTOP();
      {
	REGISTER INT n;
	REGISTER ExprPtr *ep;

	n = (INT)*pc++;
	ep = top-n;
	ep[0] = MakeRecord(E_RECORD_GNRL,bc->consts[*pc++],n,ep);
	n--;
	POPTOP(n);
	NEXT;
      }
    CASE(BC_MAPLET):
      SYNCTOP();
      top[-2] = MapletKons(top[-2],top[-1]);
      POPTOP(1);
      NEXT;
    CASE(BC_OVERWRITE):
      SYNCTOP();
      top[-2] = MapOverwrite(top[-2],top[-1]);
      POPTOP(1);
      NEXT;
    CASE(BC_UNDERWRITE):
      SYNCTOP();
      top[-2] = MapOverwrite(top[-1],top[-2]);
      POPTOP(1);
      NEXT;
    CASE(BC_VARREF0):
      top[0] = EVARS(top[pc[0]])->expr[pc[1]];
      pc += 2;
      PUSHTOP(1);
      NEXT;
    CASE(BC_VARREF1):
      top[0] = EVARS(ENEXT(top[pc[0]]))->expr[pc[1]];
      pc += 2;
      PUSHTOP(1);
      NEXT;
    CASE(BC_VARREF2):
      top[0] = EVARS(ENEXT(ENEXT(top[pc[0]])))->expr[pc[1]];
      pc += 2;
      PUSHTOP(1);
      NEXT;
    CASE(BC_VARREF3):
      top[0] = EVARS(ENEXT(ENEXT(ENEXT(top[pc[0]]))))->expr[pc[1]];
      pc += 2;
      PUSHTOP(1);
      NEXT;
    CASE(BC_VARREF):
      {
	REGISTER ExprPtr env;
	REGISTER COUNTER n;

	for (env = top[*pc++], n = (COUNTER)*pc++; n--!=0; env = ENEXT(env));
	top[0] = EVARS(env)->expr[*pc++];
	PUSHTOP(1);
	NEXT;
      }
    CASE(BC_CLOSURE):
      SYNCTOP();
      top[-1] = MakePair(E_TUPLE_2,CAR(bc->consts[*pc++]),top[-1]);
      top[-1] = MakeCompiled(E_INTERPRETED_CLOSURE,ACTION_NAME(apply_bytecode_closure),top[-1]);
      NEXT;
    CASE(BC_CLOSURE0):
      SYNCTOP();
      top[0] = MakeCompiled(E_INTERPRETED_CLOSURE,ACTION_NAME(apply_bytecode_closure),
			    bc->consts[*pc++]);
      PUSHTOP(1);
      NEXT;
    CASE(BC_NOEXPR):
      {
	REGISTER COUNTER i;

	i = (COUNTER)*pc++;
	BCSTACKS._x.n += i;
	while (i--)
	  *top++ = NOEXPR;
	NEXT;
      }
    CASE(BC_TUCK):
      POPTOP(1);
      top[*pc++] = top[0];
      NEXT;
    CASE(BC_FAKEENV):
      SYNCTOP();
      {
	REGISTER ExprPtr env;

	MAKEENV1(EVARS(GLOBENV),top[-1],env);
	top[-1] = env;
	NEXT;
      }
    CASE(BC_CONSENV):
      SYNCTOP();
      top[-1] = envCons ((COUNTER)*pc++,top[-1]);
      NEXT;
    CASE(BC_LET):
      {
	REGISTER ExprPtr *ep;
	REGISTER COUNTER k;
	REGISTER ExprPtr env;

	env = top[*pc++];
	ep = EVARS(env)->expr;
	while ((k = *pc++))
	  {
	    ep[*pc++] = top[k];
	  }
	HITBARRIER(env);
	NEXT;
      }
    CASE(BC_LET1):
      {
	REGISTER ExprPtr env;
	REGISTER COUNTER k;

	env = top[*pc++];
	k = *pc++;
	EVARS(env)->expr[*pc++] = top[k];
	HITBARRIER(env);
	NEXT;
      }
    CASE(BC_QENV):
      {
	REGISTER ExprPtr e;

	e = top[-1];
	top[-1] = (e->kind==(CHAR)E_INTERPRETED_CLOSURE)?(CDR(QENV(e))):QENV(e);
      }
      NEXT;
    CASE(BC_IS):
      if (top[-1]!=bc->consts[pc[2]])
	goto _goto;
      pc += 3;
      NEXT;
    CASE(BC_ISDATA):
      if (top[-1]->magic!=(SHORT)pc[2])
	goto _goto;
      pc += 3;
      NEXT;
    CASE(BC_SWITCH):
      pc += top[-1]->magic + *pc;
      pc += *pc;
      NEXT;
    CASE(BC_ISEQUAL):
      POPTOP(2);
      if (top[0]!=top[1])
	goto _goto;
      pc += 2;
      NEXT;
    CASE(BC_NTH):
      top[-1] = AVAL(top[-1])->expr[*pc++];
      NEXT;
    CASE(BC_NTH_0): top[-1] = AVAL(top[-1])->expr[0]; NEXT;
    CASE(BC_NTH_1): top[-1] = AVAL(top[-1])->expr[1]; NEXT;
    CASE(BC_NTH_2): top[-1] = AVAL(top[-1])->expr[2]; NEXT;
    CASE(BC_NTH_3): top[-1] = AVAL(top[-1])->expr[3]; NEXT;
    CASE(BC_NTH_4): top[-1] = AVAL(top[-1])->expr[4]; NEXT;
    CASE(BC_NTH_5): top[-1] = AVAL(top[-1])->expr[5]; NEXT;
    CASE(BC_NTH_6): top[-1] = AVAL(top[-1])->expr[6]; NEXT;
    CASE(BC_NTH_7): top[-1] = AVAL(top[-1])->expr[7]; NEXT;
    CASE(BC_3ARGS):
      {
	REGISTER ExprPtr a;

	a = top[-1];
	top[1] = AVAL(a)->expr[2];
	top[0] = AVAL(a)->expr[1];
	top[-1] = AVAL(a)->expr[0];
	PUSHTOP(2);
	NEXT;
      }
    CASE(BC_NARGS):
      {
	REGISTER VecPtr vp;
	COUNTER i, n = (COUNTER)*pc++;

	vp = AVAL(top[-1]);
	for (i=0; i<n; i++)
	  top[i-1] = vp->expr[i];
	PUSHTOP(n-1);
	NEXT;
      }
    CASE(BC_TYPEMATCH):
      if (TypeMatch(top[-1],bc->consts[pc[2]]))
	{
	  POPTOP(1);
	  goto _goto;
	}
      POPTOP(1);
      pc += 3;
      NEXT;
    CASE(BC_CARDEQ):
      if (MapCardDec(top[-1],(INT)pc[2])!=0)
	goto _goto;
      pc += 3;
      NEXT;
    CASE(BC_CARDGEQ):
      if (MapCardDec(top[-1],(INT)pc[2])>0)
	goto _goto;
      pc += 3;
      NEXT;
    CASE(BC_SELECT):
      top[-1] = AVAL(top[-1])->expr[IVAL(MapGetSure(bc->consts[*pc++],ASHAPE(top[-1])))-1];
      NEXT;
    CASE(BC_ISNEMAP):
      if (top[-1]==NewMap())
	goto _goto;
      pc += 2;
      NEXT;
    CASE(BC_MAPGET):
      {
	Boolean found;
	REGISTER ExprPtr val;

	POPTOP(1);
	val = MapGet(top[0],top[-1],&found);
	if (!found)
	  goto _goto;
	top[-1] = val;
	pc += 2;
	NEXT;
      }
    CASE(BC_NCARGET):
      {
	Boolean found;
	REGISTER ExprPtr val;

	val = NGet(top[-1], CAR(top[-2]), &found);
	if (!found)
	  goto _goto;
	POPTOP(1);
	top[-1] = val;
	pc += 2;
	NEXT;
      }
    CASE(BC_NCARPUT):
      setNPut (top[-2], top[-1], &CAR(top[-3]), top[-3]);
      POPTOP(2);
      top[-1] = top[1];
      NEXT;
    CASE(BC_MAPREMOVE):
      SYNCTOP();
      top[-2] = MapKill(top[-2],top[-1]);
      POPTOP(1);
      NEXT;
    CASE(BC_RETURN):
      res = top[-1];
      if (curbcframe!=BCF.bcframes)
	{
	  POPTOP(pc[0]);
	  --curbcframe;
	  bc = curbcframe->bc;
	  pc = curbcframe->pc;
	  top[-1] = res;
#if PROFILING
	  if (currentProf!=NOEXPR)
	    {
	      ExprPtr prof;
#if OS==AMIGA_OS
	      unsigned int clock[2];
#define TMS clock
#else
	      struct tms tms;
#define TMS &tms
#endif

#if OS==AMIGA_OS
	      (void) timer(clock);
#else
	      times(&tms);
#endif
#ifdef PROFDEBUG
	      if (currentProf!=NOEXPR && currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
		abort();
#endif
	      if (TYPE(currentProf)==EA_PROFILE)
		profExit(currentProf,TMS);
#if PROFILING & PROFILING_PROPER_TIMES
	      if (TYPE(currentProf)==EA_PROFILE)
		profExitProper(PROFDATA(currentProf),TMS);
	      prof = bc->consts[0];
#ifdef PROFDEBUG
	      if (prof==NOEXPR || currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
		abort();
#endif
	      if (TYPE(prof)==EA_PROFILE)
		profEnterProper(PROFDATA(prof),TMS);
	      currentProf = prof;
#endif
	    }
#endif
	  NEXT;
	}
      theFrames = BCSTACKS._x.next;
      if (xhi.xhandlers!=NULL) /* if we have installed an exception handler
				  in excStack, pop it. */
	excStack = excStack->next;
#if DEBUGGING!=0
      (void) freetmp (bcstacks_ptr);
#endif
#if PROFILING
      if (currentProf!=NOEXPR)
	{
#if OS==AMIGA_OS
	  unsigned int clock[2];
#define TMS clock
#else
	  struct tms tms;
#define TMS &tms
#endif

#if OS==AMIGA_OS
	  (void) timer(clock);
#else
	  times(&tms);
#endif
#ifdef PROFDEBUG
	  if (currentProf!=NOEXPR && currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
	    abort();
#endif
	  if (TYPE(currentProf)==EA_PROFILE)
	    profExit(currentProf,TMS);
#if PROFILING & PROFILING_PROPER_TIMES
	  if (TYPE(currentProf)==EA_PROFILE)
	    profExitProper(PROFDATA(currentProf),TMS);
#ifdef PROFDEBUG
	  if (saveProf!=NOEXPR && saveProf->kind!=(CHAR)BUILD_HEADER_KIND)
	    abort();
#endif
	  if (saveProf!=NOEXPR && TYPE(saveProf)==EA_PROFILE)
	    profEnterProper(PROFDATA(saveProf),TMS);
	  currentProf = saveProf;
#endif
	}
#endif
      return res;
    CASE(BC_MAPADD):
      top[-3] = MapAdd(top[-3],top[-2],top[-1]);
      POPTOP(2);
      NEXT;
    CASE(BC_MAPADD2):
      top[-3] = MapAdd(top[-2],top[-1],top[-3]);
      POPTOP(2);
      NEXT;
    CASE(BC_SETCAR):
      SETCAR(top[-1],top[-2]);
      POPTOP(2);
      NEXT;
    CASE(BC_N_CALL):
      {
	COUNTER n = (COUNTER)*pc++;
	ExprPtr *args = &top[-n];
	REGISTER ExprPtr f;

	f = args[-1];
	if (f->kind==(CHAR)E_INTERPRETED_CLOSURE)
	  {
	    curbcframe->bc = bc;
	    curbcframe->pc = pc;
	    curbcframe++;
	    f = QENV(f);
	    bc = (byteCodeChunk *)CAR(f);
#if PROFILING
	    {
	      ExprPtr prof = bc->consts[0];
#if OS==AMIGA_OS
	      unsigned int clock[2];
#define TMS clock
#else
	      struct tms tms;
#define TMS &tms
#endif

#ifdef PROFDEBUG
	      if (prof==NOEXPR || prof->kind!=(CHAR)BUILD_HEADER_KIND)
		abort();
#endif
	      if (TYPE(prof)==EA_PROFILE)
		{
#if PROFILING & PROFILING_CALL_COUNTS
		  PROFDATA(prof)->nCalls++;
#endif
#if PROFILING & PROFILING_PROPER_TIMES
#if OS==AMIGA_OS
		  (void) timer(clock);
#else
		  times(&tms);
#endif
#ifdef PROFDEBUG
		  if (currentProf!=NOEXPR && currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
		    abort();
#endif
		  if (currentProf!=NOEXPR && TYPE(currentProf)==EA_PROFILE)
		    profExitProper(PROFDATA(currentProf),TMS);
		  profEnterProper(PROFDATA(prof),TMS);
#endif
		  profEnter (prof, TMS);
		}
#if PROFILING & PROFILING_PROPER_TIMES
	      currentProf = prof;
#endif
	    }
#endif
	    SYNCTOP();
	    if (BCSTACKS._x.n+bc->stackSize > BC_STACK_SIZE+1+n)
	      {
#if DEBUGGING!=0
		parentBcFrames = BCF_PTR;
		BCF.curBcFrame = (struct bcframe *)curbcframe;
		parentStack = BCSTACKS_PTR;
#endif
		M_VARCALL2_CC(res,bcEval,bc,ENV_CONS(CDR(f),n),n,(args-1));
		POPTOP(n);
		{
		overflow_on_call:
		  --curbcframe;
		  bc = curbcframe->bc;
		  pc = curbcframe->pc;
		  top[-1] = res;
#if PROFILING
		  if (currentProf!=NOEXPR)
		    {
		      ExprPtr prof;
#if OS==AMIGA_OS
		      unsigned int clock[2];
#define TMS clock
#else
		      struct tms tms;
#define TMS &tms
#endif
		      
#if OS==AMIGA_OS
		      (void) timer(clock);
#else
		      times(&tms);
#endif
#ifdef PROFDEBUG
		      if (currentProf!=NOEXPR && currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
			abort();
#endif
		      if (TYPE(currentProf)==EA_PROFILE)
			profExit(currentProf,TMS);
#if PROFILING & PROFILING_PROPER_TIMES
		      if (TYPE(currentProf)==EA_PROFILE)
			profExitProper(PROFDATA(currentProf),TMS);
		      prof = bc->consts[0];
#ifdef PROFDEBUG
		      if (prof==NOEXPR || currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
			abort();
#endif
		      if (TYPE(prof)==EA_PROFILE)
			profEnterProper(PROFDATA(prof),TMS);
		      currentProf = prof;
#endif
		    }
#endif
		  NEXT;
		}
	      }
	    args[-1] = CDR(f);
	    if (bc->nargs==1)
	      switch (n)
		{
		case 0: top[0] = tuple0; PUSHTOP(1); break;
		case 1: break;
		case 2: top[-2] = Ktuple2(top[-2],top[-1]); POPTOP(1); break;
		default: args[0] = KtupleGnrl(n,args); POPTOP(n-1); break;
		}
	    pc = bc->instr;
	    NEXT;
	  }
	SYNCTOP();
#if DEBUGGING!=0
	curbcframe->bc = bc;
	curbcframe->pc = pc;
#endif
#ifdef CALLDEBUG
	if (f->kind!=(CHAR)E_COMPILED_CLOSURE)
	  {
	    berror ("BC_N_CALL on bad kind : %d", f->kind);
	  }
#endif
	M_VARCALL_CC(res,(*ACTION(f)),ENV_CONS(QENV(f),n),n,(args-1));
	args[-1] = res;
	POPTOP(n);
	NEXT;
      }
    CASE(BC_CALL):
      {
	REGISTER ExprPtr f;

	f = top[-2];
	if (f->kind==(CHAR)E_INTERPRETED_CLOSURE/* && curbcframe<BCF.bcframes+MAX_RECURSE*/)
	  {
	    curbcframe->bc = bc;
	    curbcframe->pc = pc;
	    curbcframe++;
	    f = QENV(f);
	    bc = (byteCodeChunk *)CAR(f);
#if PROFILING
	    {
	      ExprPtr prof = bc->consts[0];
#if OS==AMIGA_OS
	      unsigned int clock[2];
#define TMS clock
#else
	      struct tms tms;
#define TMS &tms
#endif

#ifdef PROFDEBUG
	      if (prof==NOEXPR || prof->kind!=(CHAR)BUILD_HEADER_KIND)
		abort();
#endif
	      if (TYPE(prof)==EA_PROFILE)
		{
#if PROFILING & PROFILING_CALL_COUNTS
		  PROFDATA(prof)->nCalls++;
#endif
#if PROFILING & PROFILING_PROPER_TIMES
#if OS==AMIGA_OS
		  (void) timer(clock);
#else
		  times(&tms);
#endif
#ifdef PROFDEBUG
		  if (currentProf!=NOEXPR && currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
		    abort();
#endif
		  if (currentProf!=NOEXPR && TYPE(currentProf)==EA_PROFILE)
		    profExitProper(PROFDATA(currentProf),TMS);
		  profEnterProper(PROFDATA(prof),TMS);
#endif
		  profEnter (prof, TMS);
		}
#if PROFILING & PROFILING_PROPER_TIMES
	      currentProf = prof;
#endif
	    }
#endif
	    SYNCTOP();
	    if (BCSTACKS._x.n+bc->stackSize > BC_STACK_SIZE+2)
	      {
#if DEBUGGING!=0
		parentBcFrames = BCF_PTR;
		BCF.curBcFrame = (struct bcframe *)curbcframe;
		parentStack = BCSTACKS_PTR;
#endif
		env = ENV_CONS(CDR(f),1);
		res = bcEval(bc,env,top[-1]);
		POPTOP(1);
		goto overflow_on_call;
	      }
	    top[-2] = CDR(f);
	    switch (bc->nargs)
	      {
	      case 0: POPTOP(1); break;
	      case 1: break;
	      case 2: top[0] = CDR(top[-1]); top[-1] = CAR(top[-1]); PUSHTOP(1); break;
	      default:
		{
		  COUNTER i;
		  ExprPtr *ep = AVAL(top[-1])->expr, *epp;

		  for (i=bc->nargs, epp = &top[-1]; --i>=0; )
		    *epp++ = *ep++;
		  PUSHTOP(bc->nargs-1);
		  break;
		}
	      }
	    /*top[1] = top[-1];
	    PUSHTOP(2);*/
	    pc = bc->instr;
	    NEXT;
	  }
	SYNCTOP();
#if DEBUGGING!=0
	curbcframe->bc = bc;
	curbcframe->pc = pc;
#endif
#ifdef CALLDEBUG
	if (f->kind!=(CHAR)E_COMPILED_CLOSURE)
	  {
	    berror ("BC_CALL on bad kind : %d", f->kind);
	  }
#endif
	top[-2] = (*ACTION(f))(ENV_CONS(QENV(f),1), top[-1]);
	POPTOP(1);
	NEXT;
      }
    CASE(BC_SETREF):
      SETREFVAL(top[-2],top[-1]);
      POPTOP(2);
      NEXT;
    CASE(BC_KMU1):
      SYNCTOP();
      top[-2] = Kmu1(top[-2],bc->consts[*pc++],top[-1]);
      POPTOP(1);
      NEXT;
    CASE(BC_MU1):
      SYNCTOP();
      top[-2] = Mu1(top[-2],bc->consts[*pc++],top[-1]);
      POPTOP(1);
      NEXT;
    CASE(BC_KMUGNRL):
      SYNCTOP();
      top[-2] = KmuGnrl(top[-2],top[-1]);
      POPTOP(1);
      NEXT;
    CASE(BC_MUGNRL):
      SYNCTOP();
      top[-2] = MuGnrl(top[-2],top[-1]);
      POPTOP(1);
      NEXT;
    CASE(BC_SWAP): /* currently unused by bc.c */
      {
	REGISTER ExprPtr aux;

	aux = top[-1]; top[-1] = top[-2]; top[-2] = aux;
      }
      NEXT;
    CASE(BC_NEWQUEUE): /* build an empty circular list */
      top[0] = NIL;
      PUSHTOP(1);
      NEXT;
    CASE(BC_CONCATQUEUE): /* build top[*pc] @ top[-1], destructively */
      /* The idea is that the value of the circular list representing a1, ..., an, n>=1,
	 is the pointer to its *last* element an (which then points to a1, then a2,
	 and so on). */
      {
	REGISTER ExprPtr q;

	q = top[*pc++];
	if (q==NIL)
	  ;
	else if (top[-1]==NIL)
	  top[-1] = q;
	else
	  {
	    REGISTER ExprPtr e1,e2;

	    e1 = CDR(q);
	    e2 = CDR(top[-1]);
	    SETCDR(q,e2);
	    SETCDR(top[-1],e1);
	  }
	NEXT;
      }
    CASE(BC_NEW1QUEUE):
      SYNCTOP();
      {
	REGISTER ExprPtr e;

	EALLOC_HOMO(e);
	e->kind = (CHAR)E_CONS;
	CAR(e) = top[-1];
	CDR(e) = e;
	top[-1] = e;
      }
      NEXT;
    CASE(BC_QUEUELIST):
      if (top[-1]!=NIL)
	{
	  SYNCTOP();
	  top[-1] = QueueToList(top[-1]);
	}
      NEXT;
    CASE(BC_LISTQUEUE):
      if (top[-1]!=NIL)
	{
	  SYNCTOP();
	  top[-1] = ListToQueue(top[-1]);
	}
      NEXT;
    CASE(BC_MAPINTER):
    CASE(BC_MAPDOMRESTRTO):
      SYNCTOP();
      top[-2] = MapDRT(top[-1],top[-2]);
      POPTOP(1);
      NEXT;
    CASE(BC_MAPDOMRESTRBY):
      SYNCTOP();
      top[-2] = MapDRB(top[-1],top[-2]);
      POPTOP(1);
      NEXT;
    CASE(BC_DIFF):
      SYNCTOP();
      top[-2] = MapDRB(top[-2],top[-1]);
      POPTOP(1);
      NEXT;
    CASE(BC_DELTA):
      SYNCTOP();
      top[-2] = MapD(top[-1],top[-2]);
      POPTOP(1);
      NEXT;
    CASE(BC_STARTMAPSWEEP):
      {
	REGISTER ExprPtr e,*starttop;

	e = top[*pc++];
	starttop = top++;
	maptop[0] = top;
	while (e->kind!=(CHAR)E_MAPLET)
	  {
	    REGISTER ExprPtr aux;

	    aux = MAPLEFT(e);
	    if (aux==empty_map)
	      e = MAPRIGHT(e);
	    else
	      {
		*top++ = MAPRIGHT(e);
		e = aux;
	      }
	  }
	*starttop = e;
	maptop[1] = top;
	maptop += 2;

	for (starttop += (1+MAP_CHAIN_LENGTH); top<starttop;) *top++ = NOEXPR;
      }
      BCSTACKS._x.n += 1+MAP_CHAIN_LENGTH;
      NEXT;
    CASE(BC_STARTSUBMAPSWEEP):
      {
	ExprPtr map;

	map =  top[*pc++];
	top[0] = NewMap();
	top[1] = NIL;
	PUSHTOP(2);
	SYNCTOP();
	(void) MapSweep(map,do_initialsubmap,(char *)&top[-1]);
      }
      NEXT;
    CASE(BC_GETLISTSWEEP):
      top[0] = CAR(top[*pc++]);
      PUSHTOP(1);
      NEXT;
    CASE(BC_INCLISTSWEEP):
      {
	REGISTER ExprPtr *ep = &top[pc[2]];

	if ((*ep = CDR(*ep))==NIL)
	  goto _goto;
	pc += 3;
	NEXT;
      }
    CASE(BC_INCMAPSWEEP):
      {
	REGISTER ExprPtr e;
	REGISTER ExprPtr *tpp;
	REGISTER ExprPtr **mtop;

	mtop = maptop - pc[3];
	tpp = mtop[1];
	do {
	  if (tpp==mtop[0]) /* end of map */
	    goto _goto;
	  e = *--tpp; /* pop from local map iteration stack */
	  *tpp = NOEXPR; /* allow this entry to be reclaimed by GC */
	} while (e==empty_map);
	while (e->kind!=(CHAR)E_MAPLET)
	  {
	    REGISTER ExprPtr aux;

	    aux = MAPLEFT(e);
	    if (aux==empty_map)
	      e = MAPRIGHT(e);
	    else
	      {
		*tpp++ = MAPRIGHT(e); /* push on local map iteration stack */
		e = aux;
	      }
	  }
	mtop[1] = tpp;
	top[pc[2]] = e;
	pc += 4;
	NEXT;
      }
    CASE(BC_INCSUBMAPSWEEP):
      SYNCTOP();
      {
	REGISTER ExprPtr *ep = &top[pc[2]];
	REGISTER ExprPtr l = ep[1];

	while (true)
	  {
	    switch (l->kind)
	      {
	      case E_NIL:
		goto _goto;
	      case E_OR:
		l->kind = (CHAR)E_AND;
		ep[0] = MapPutMapletTrie(CAR(l),ep[0],(CODE)(1L<<NLOWBITS));
		pc += 3;
		NEXT;
	      case E_AND:
		l->kind = (CHAR)E_OR;
		ep[0] = MapRemoveTrie(CAR(CAR(l)),ep[0],(CODE)(1L<<NLOWBITS));
		break;
	      }
	    l = CDR(l);
	  }
	/* ep[0] is a list of elements, where the tags on cons-cells are E_OR (true)
	   or E_AND (false); the code above builds the subset of all elements that are
	   marked true.  Moreover, if we see the sequence of trues and falses as
	   bits in the binary expansion of a number, it increments this number.
	   Each call to the BC_INCSUBMAPSWEEP opcode does k+1 map operations,
	   where k is the number of trailing true bits.  On average, and when there
	   are n bits (the original set has cardinality n), k is 2-(n+2)/2^n,
	   so we do roughly 3 map operations per iteration. */
	/*NOTREACHED*/
	NEXT;
      }
    CASE(BC_MAPDROP):
      maptop -= (COUNTER)*pc++;
      NEXT;
    CASE(BC_QUEUEUNSHAREDLIST):
      if (top[-1]!=NIL)
	{
	  REGISTER ExprPtr aux;

	  aux = top[-1];
	  top[-1] = CDR(aux);
	  SETCDR(aux,NIL);
	}
      NEXT;
    CASE(BC_HANDLE):
#ifdef HDEBUG
      nHandle++;
#endif
      if (xhi.xhandlers==NULL)
	{ /* first handler: allocate excHandler and array of xhandlers */
	  {
	    REGISTER excHandler *exc;

	    exc = (excHandler *)alloctmp_anchored(sizeof(excHandler),(char *)&env);
	    exc->next = excStack;
	    exc->tag = X_HANDLER;
	    exc->data = NIL;
	    exc->code = NIL;
	    excStack = exc;
	    if (SETJUMP(exc->jb)!=0)
	      {
		(void) alloctmp (0);
#ifdef HDEBUG
		nLongJumpRaise++;
#endif
#if DEBUGGING
		/* It may happen that an exception was raised inside a primitive;
		   then we cannot possibly have caught it through a BC_BREAK_ON_RAISE
		   instruction.  We try to catch it here:
		   - pc[-1] will be BC_CALL or BC_TAILCALL;
		   - top[-2] will be a E_COMPILED_CLOSURE;
		   But we detect it by the fact that raiseInPrimitive would be true.
		   This also happens with BC_AREF, BC_ASET.
		   */
		if (raiseInPrimitive) /* can only be true if debugEnabled==true. */
		  {
		    ExprPtr exn,exc;

		    raiseInPrimitive = false;
		    pc = curbcframe->pc;
		    bc = curbcframe->bc;
		    top = (ExprPtr *)(BCSTACKS._x.p + BCSTACKS._x.n);
		  do_raise:
		    exc = exn = THEEXPR;
		    if (exc->kind==(CHAR)E_CONS)
		      exc = CAR(exc);
		    if ((COUNTEDN(exc) & EXC_BENIGN)==0)
		      {
			BCF.curBcFrame = (struct bcframe *)curbcframe;
			(void) DoBreak (bc,pc,top,BCF_PTR,BCSTACKS_PTR,exn);
			THEEXPR = exn;
		      }
		    /* consider break happened at pc-1, namely the point
		       were CALL or TAILCALL was to be found. */
		  }
#endif
		if (xhi.xhandlers!=NULL && xhi.curxhandler!=xhi.xhandlers->exc)
		  goto raise;
#if PROFILING
		if (currentProf!=NOEXPR)
		  {
		    ExprPtr prof;
#if OS==AMIGA_OS
		    unsigned int clock[2];
#define TMS clock
#else
		    struct tms tms;
#define TMS &tms
#endif

#if OS==AMIGA_OS
		    (void) timer(clock);
#else
		    times(&tms);
#endif
		    if (TYPE(currentProf)==EA_PROFILE)
		      profExit(currentProf,TMS);
		    while (curbcframe!=BCF.bcframes)
		      {
			--curbcframe;
			prof = curbcframe->bc->consts[0];
			if (TYPE(prof)==EA_PROFILE)
			  profExit (prof, TMS);
		      }
#if PROFILING & PROFILING_PROPER_TIMES
#ifdef PROFDEBUG
		    if (currentProf==NOEXPR || currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
		      abort();
#endif
		    if (TYPE(currentProf)==EA_PROFILE)
		      profExitProper(PROFDATA(currentProf),TMS);
#ifdef PROFDEBUG
		    if (saveProf!=NOEXPR && saveProf->kind!=(CHAR)BUILD_HEADER_KIND)
		      abort();
#endif
		    if (saveProf!=NOEXPR && TYPE(saveProf)==EA_PROFILE)
		      profEnterProper(PROFDATA(saveProf),TMS);
		    currentProf = saveProf;
#endif
		  }
#endif
		if (xhi.xhandlers!=NULL)
		  {
		    excStack = excStack->next; /* the next Raise() will exit bcEval0() */
		    freetmp((char *)xhi.xhandlers);
		    xhi.xhandlers = NULL;
		  }
		top[-1] = Raise0(THEEXPR);
		NEXT;
	      }
	  }
	  goto newxhandler;
	}
      if (xhi.curxhandler==xhi.endxhandler) /* exception stack overflow */
	{
	  REGISTER struct xhandlers *new;

newxhandler:
	  new = (struct xhandlers *)alloctmp_anchored(sizeof(struct xhandlers),
						      (char *)&env);
	  new->next = (struct xhandlers *)xhi.xhandlers;
	  xhi.xhandlers = new;
	  xhi.curxhandler = &new->exc[0];
	  xhi.endxhandler = &new->exc[MAX_EXC_NEST];
	}
      { /* normal case: push handler and resume execution */
	REGISTER struct xhandler *xh;

	xh = (struct xhandler *)xhi.curxhandler++;
	xh->bc = bc;
	xh->pc = pc;
	xh->n = BCSTACKS._x.n;
	xh->maptop = maptop;
	xh->bcframe = (struct bcframe *)curbcframe;
      }
      pc += 2;
      NEXT;
    CASE(BC_NEXTHANDLE):
#ifdef HDEBUG
      nNextHandle++;
#endif
#if DEBUGGING
    if (debugMode>=DBM_STEPPING)
      {
	goto do_break; /* catches stepping through exception raising. */
      }
#endif
    /*FALLTHROUGH*/
    CASE(BC_NEXTHANDLE_NODEBUG): /* This is not a legal opcode, rather
				    NEXTHANDLE_NODEBUG is returned by DoBreak()
				    when debugger was entered on NEXTHANDLE (just above),
				    so as not to enter DoBreak() again. */
      if (--xhi.curxhandler==xhi.xhandlers->exc)
	{
	  struct xhandlers *next;

	  next = xhi.xhandlers->next;
	  if (next==NULL)
	    {
	      /* do nothing: don't deallocate handler, in case we need to reallocate it
		 later on. */
	    }
	  else
	    {
	      freetmp((char *)xhi.xhandlers);
	      xhi.xhandlers = next;
	      xhi.endxhandler = &xhi.xhandlers->exc[MAX_EXC_NEST];
	      xhi.curxhandler = xhi.endxhandler;
	    }
	}
      NEXT;
    CASE(BC_RAISE):
      SYNCTOP();
#ifdef HDEBUG
      nInlineRaise++;
#endif
#if DEBUGGING
      if (debugEnabled)
	{
	  REGISTER ExprPtr exc;

	  exc = top[-1];
	  if (exc->kind==(CHAR)E_CONS)
	    exc = CAR(exc);
	  if ((COUNTEDN(exc) & EXC_BENIGN)==0)
	    {
	      BCF.curBcFrame = (struct bcframe *)curbcframe;
	      (void) DoBreak(bc,pc,top,BCF_PTR,BCSTACKS_PTR,top[-1]);
	    }
	}
#endif
      if (xhi.xhandlers!=NULL && xhi.curxhandler!=xhi.xhandlers->exc)
	{
	  THEEXPR = top[-1];
raise:
	  /* reload registers: */
	  {
	    REGISTER struct xhandler *xh;

#ifdef HDEBUG
	    if (xhi.xhandlers==NULL || xhi.curxhandler==xhi.xhandlers->exc)
	      xh = NULL;
#endif
	    xh = (struct xhandler *)xhi.curxhandler;
	    bc = xh[-1].bc;
	    pc = xh[-1].pc;
	    top = (ExprPtr *)(BCSTACKS._x.p+(BCSTACKS._x.n = xh[-1].n+1));
	    top[-1] = THEEXPR; /* push exception on top of stack */
	    maptop = xh[-1].maptop;
#if PROFILING
	    if (curbcframe==xh[-1].bcframe) /* raise from inside the same function
					       (may happen if raise was called from
					       a primitive). */
	      ;
	    else if (currentProf!=NOEXPR)
	      {
		ExprPtr prof;
#if OS==AMIGA_OS
		unsigned int clock[2];
#define TMS clock
#else
		struct tms tms;
#define TMS &tms
#endif

#if OS==AMIGA_OS
		(void) timer(clock);
#else
		times(&tms);
#endif
#ifdef PROFDEBUG
		if (currentProf==NOEXPR || currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
		  abort();
#endif
		if (TYPE(currentProf)==EA_PROFILE)
		  profExit(currentProf,TMS);
		while (--curbcframe!=xh[-1].bcframe)
		  {
		    prof = curbcframe->bc->consts[0];
		    if (TYPE(prof)==EA_PROFILE)
		      profExit (prof, TMS);
		  }
#if PROFILING & PROFILING_PROPER_TIMES
		if (TYPE(currentProf)==EA_PROFILE)
		  profExitProper(PROFDATA(currentProf),TMS);
		prof = bc->consts[0];
#ifdef PROFDEBUG
		if (prof==NOEXPR || prof->kind!=(CHAR)BUILD_HEADER_KIND)
		  abort();
#endif
		if (TYPE(prof)==EA_PROFILE)
		  profEnterProper(PROFDATA(prof),TMS);
		currentProf = prof;
#endif
	      }
	    else curbcframe = xh[-1].bcframe;
#else
	    curbcframe = xh[-1].bcframe;
#endif
	  }
	  goto _goto;
	}
#if PROFILING
      if (currentProf!=NOEXPR)
	{
	  ExprPtr prof;
#if OS==AMIGA_OS
	  unsigned int clock[2];
#define TMS clock
#else
	  struct tms tms;
#define TMS &tms
#endif

#if OS==AMIGA_OS
	  (void) timer(clock);
#else
	  times(&tms);
#endif
	  if (TYPE(currentProf)==EA_PROFILE)
	    profExit(currentProf,TMS);
	  while (curbcframe!=BCF.bcframes)
	    {
	      --curbcframe;
	      prof = curbcframe->bc->consts[0];
	      if (TYPE(prof)==EA_PROFILE)
		profExit (prof, TMS);
	    }
#if PROFILING & PROFILING_PROPER_TIMES
#ifdef PROFDEBUG
	  if (currentProf==NOEXPR || currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
	    abort();
#endif
	  if (TYPE(currentProf)==EA_PROFILE)
	    profExitProper(PROFDATA(currentProf),TMS);
#ifdef PROFDEBUG
	  if (saveProf!=NOEXPR && saveProf->kind!=(CHAR)BUILD_HEADER_KIND)
	    abort();
#endif
	  if (saveProf!=NOEXPR && TYPE(saveProf)==EA_PROFILE)
	    profEnterProper(PROFDATA(saveProf),TMS);
	  currentProf = saveProf;
#endif
	}
#endif
      if (xhi.xhandlers!=NULL)
	{
	  excStack = excStack->next; /* the next Raise() will exit bcEval0() */
	  freetmp((char *)xhi.xhandlers);
	  xhi.xhandlers = NULL;
	}
      top[-1] = Raise0(top[-1]);
      NEXT;
    CASE(BC_EXCEPTION):
      SYNCTOP();
      top[-1] = MakeException(top[-1],EXC_CHECKED);
      NEXT;
    CASE(BC_EXCEPTION_BENIGN):
      SYNCTOP();
      top[-1] = MakeException(top[-1],EXC_BENIGN);
      NEXT;
    CASE(BC_DELAY):
      SYNCTOP();
      top[-1] = MakePromise(top[-1]);
      NEXT;
    CASE(BC_PACK):
      SYNCTOP();
      top[-2] = MakePack(E_PACK,top[-2],top[-1]);
      POPTOP(1);
      NEXT;
    CASE(BC_IADD):
      top[-2] = Kint(IVAL(top[-2])+IVAL(top[-1]));
      POPTOP(1);
      NEXT;
    CASE(BC_ISUB):
      top[-2] = Kint(IVAL(top[-2])-IVAL(top[-1]));
      POPTOP(1);
      NEXT;
    CASE(BC_IMUL):
      top[-2] = Kint(IVAL(top[-2])*IVAL(top[-1]));
      POPTOP(1);
      NEXT;
    CASE(BC_IDIV):
      top[-2] = Kint(idiv(IVAL(top[-2]),IVAL(top[-1])));
      POPTOP(1);
      NEXT;
    CASE(BC_IMOD):
      top[-2] = Kint(imod(IVAL(top[-2]),IVAL(top[-1])));
      POPTOP(1);
      NEXT;
    CASE(BC_ISNZERO):
      if (top[-1]==izero)
	goto _goto;
      pc += 2;
      NEXT;
    CASE(BC_IOR):
      top[-2] = Kint(IVAL(top[-2]) | IVAL(top[-1]));
      POPTOP(1);
      NEXT;
    CASE(BC_IXOR):
      top[-2] = Kint(IVAL(top[-2]) ^ IVAL(top[-1]));
      POPTOP(1);
      NEXT;
    CASE(BC_IAND):
      top[-2] = Kint(IVAL(top[-2]) & IVAL(top[-1]));
      POPTOP(1);
      NEXT;
    CASE(BC_ILSL):
      top[-2] = Kint(IVAL(top[-2]) << IVAL(top[-1]));
      POPTOP(1);
      NEXT;
    CASE(BC_ILSR):
      top[-2] = Kint(((unsigned long)IVAL(top[-2])) >> IVAL(top[-1]));
      POPTOP(1);
      NEXT;
    CASE(BC_IASR):
      top[-2] = Kint(IVAL(top[-2]) >> IVAL(top[-1]));
      POPTOP(1);
      NEXT;
    CASE(BC_IABS):
      {
	REGISTER INT n = IVAL(top[-1]);

	if (n<0)
	  top[-1] = Kint(-n);
      }
      NEXT;
    CASE(BC_INEG):
      top[-1] = Kint(-IVAL(top[-1]));
      NEXT;
    CASE(BC_ISQR):
      {
	REGISTER INT n = IVAL(top[-1]);

	top[-1] = Kint(n*n);
      }
      NEXT;
    CASE(BC_INOT):
      top[-1] = Kint(~IVAL(top[-1]));
      NEXT;
    CASE(BC_ILT):
      POPTOP(2);
      if (IVAL(top[0])>=IVAL(top[1]))
	goto _goto;
      pc += 2;
      NEXT;
    CASE(BC_ILE):
      POPTOP(2);
      if (IVAL(top[0])>IVAL(top[1]))
	goto _goto;
      pc += 2;
      NEXT;
    CASE(BC_IGT):
      POPTOP(2);
      if (IVAL(top[0])<=IVAL(top[1]))
	goto _goto;
      pc += 2;
      NEXT;
    CASE(BC_IGE):
      POPTOP(2);
      if (IVAL(top[0])<IVAL(top[1]))
	goto _goto;
      pc += 2;
      NEXT;
    CASE(BC_IMIN):
      if (IVAL(top[-2])>IVAL(top[-1]))
	top[-2] = top[-1];
      POPTOP(1);
      NEXT;
    CASE(BC_IMAX):
      if (IVAL(top[-2])<IVAL(top[-1]))
	top[-2] = top[-1];
      POPTOP(1);
      NEXT;
    CASE(BC_RLT):
      POPTOP(2);
      if (top[0]->kind!=(CHAR)E_REAL || top[1]->kind!=(CHAR)E_REAL)
	{
	raise_complex:
	  THEEXPR = INCOMPEXC;
	  goto do_raise;
	}
      if (XVAL(top[0])>=XVAL(top[1]))
	goto _goto;
      pc += 2;
      NEXT;
    CASE(BC_RLE):
      POPTOP(2);
      if (top[0]->kind!=(CHAR)E_REAL || top[1]->kind!=(CHAR)E_REAL)
	goto raise_complex;
      if (XVAL(top[0])>XVAL(top[1]))
	goto _goto;
      pc += 2;
      NEXT;
    CASE(BC_RGT):
      POPTOP(2);
      if (top[0]->kind!=(CHAR)E_REAL || top[1]->kind!=(CHAR)E_REAL)
	goto raise_complex;
      if (XVAL(top[0])<=XVAL(top[1]))
	goto _goto;
      pc += 2;
      NEXT;
    CASE(BC_RGE):
      POPTOP(2);
      if (top[0]->kind!=(CHAR)E_REAL || top[1]->kind!=(CHAR)E_REAL)
	goto raise_complex;
      if (XVAL(top[0])<XVAL(top[1]))
	goto _goto;
      pc += 2;
      NEXT;
    CASE(BC_RMIN): 
      if (top[-2]->kind!=(CHAR)E_REAL || top[-1]->kind!=(CHAR)E_REAL)
	goto raise_complex;
     if (XVAL(top[-2])>XVAL(top[-1]))
	top[-2] = top[-1];
      POPTOP(1);
      NEXT;
    CASE(BC_RMAX):
      if (top[-2]->kind!=(CHAR)E_REAL || top[-1]->kind!=(CHAR)E_REAL)
	goto raise_complex;
      if (XVAL(top[-2])<XVAL(top[-1]))
	top[-2] = top[-1];
      POPTOP(1);
      NEXT;
    CASE(BC_FLOOR):
      if (top[-1]->kind!=(CHAR)E_REAL)
	goto raise_complex;
      {
	REAL x;

	x = floor(XVAL(top[-1]));
	top[-1] = Kreal(&x);
	NEXT;
      }
    CASE(BC_CEIL):
      if (top[-1]->kind!=(CHAR)E_REAL)
	goto raise_complex;
      {
	REAL x;

	x = ceil(XVAL(top[-1]));
	top[-1] = Kreal(&x);
	NEXT;
      }
    CASE(BC_AREF):
      {
	REGISTER COUNTER m;
	REGISTER ExprPtr a;
	REGISTER VecPtr vp;

	m = IVAL(top[-1]);
	if (m<0)
	  {
	  raise_subscript:
	    THEEXPR = subscriptException;
	    goto do_raise;
	    /* top[-2] = Raise0(subscriptException); */
	  }
	else
	  {
	    a = top[-2];
#ifdef OBSOLETE
	    switch (a->kind)
	      {
	      case E_ARRAY_0:
		goto raise_subscript;
		/* top[-2] = Raise0(subscriptException);
		   break; */
	      case E_ARRAY_1:
		if (m!=0)
		  {
		    goto raise_subscript;
		    /* top[-2] = Raise0(subscriptException); */
		  }
		else top[-2] = REFVAL(a);
		break;
	      case E_ARRAY_2:
		if (m>1)
		  {
		    goto raise_subscript;
		    /* top[-2] = Raise0(subscriptException); */
		  }
		else top[-2] = (m==0)?CAR(a):CDR(a);
		break;
	      default:
#endif
		vp = AVAL(a);
		if (m>=vp->n)
		  {
		    goto raise_subscript;
		    /* top[-2] = Raise0(subscriptException); */
		  }
		else top[-2] = vp->expr[m];
#ifdef OBSOLETE
		break;
	      }
#endif
	  }
	POPTOP(1);
	NEXT;
      }
    CASE(BC_ASET):
      {
	REGISTER ExprPtr a, y;
	REGISTER COUNTER m;

	m = IVAL(top[-2]);
	if (m<0)
	  {
	    goto raise_subscript;
	    /* (void) Raise0(subscriptException); */
	  }
	else
	  {
	    a = top[-3];
	    y = top[-1];
#ifdef OBSOLETE
	    switch (a->kind)
	      {
	      case E_ARRAY_0:
		goto raise_subscript;
		/* (void) Raise0(subscriptException);
		   break; */
	      case E_ARRAY_1:
		if (m!=0)
		  {
		    goto raise_subscript;
		    /* (void) Raise0(subscriptException); */
		  }
		else
		  {
		    SETREFVAL(a,y);
		  }
		break;
	      case E_ARRAY_2:
		if (m>1)
		  {
		    goto raise_subscript;
		    /* (void) Raise0(subscriptException); */
		  }
		else if (m==0)
		  {
		    SETCAR(a,y);
		  }
		else
		  {
		    SETCDR(a,y);
		  }
		break;
	      default:
#endif
		{
		  REGISTER VecPtr vp;

		  vp = AVAL(a);
		  if (m>=vp->n)
		    {
		      goto raise_subscript;
		      /* (void) Raise0(subscriptException); */
		    }
		  else
		    {
		      vp->expr[m] = y;
		      HITBARRIER(a);
		    }
		}
#ifdef OBSOLETE
		break;
	      }
#endif
	  }
	POPTOP(3);
	NEXT;
      }
    CASE(BC_IAREF):
      {
	REGISTER INT m;
	REGISTER INT *ip;

	m = IVAL(top[-1]);
	if (m<0)
	  goto raise_subscript;
	ip = (INT *)EXTDATA (top[-2]);
	if (m>=ip[0])
	  goto raise_subscript;
	ip++;
	top[-2] = Kint (ip [m]);
	POPTOP(1);
	NEXT;
      }
    CASE(BC_IASET):
      {
	REGISTER ExprPtr a;
	REGISTER INT m, ival;
	REGISTER INT *ip;

	a = top[-3];
	m = IVAL(top[-2]);
	ival = IVAL(top[-1]);
	if (m<0)
	  goto raise_subscript;
	ip = (INT *)EXTDATA(a);
	if (m>=ip[0])
	  goto raise_subscript;
	ip++;
	ip[m] = ival;
	POPTOP(3);
	NEXT;
      }
    CASE(BC_NUMINT):
      top[-1] = Kint((INT)XVAL(top[-1]));
      NEXT;
    CASE(BC_INTNUM):
      {
	REAL x;

	x = (REAL)IVAL(top[-1]);
	top[-1] = Kreal(&x);
      }
      NEXT;
    CASE(BC_SIZE):
      top[-1] = Kint(SLEN(top[-1]));
      NEXT;
    CASE(BC_ORD):
      if (SLEN(top[-1])==0)
	{
	raise_stringnth:
	  THEEXPR = stringnthException;
	  goto do_raise;
	}
      top[-1] = Kint((INT)STEXT(top[-1])[0]);
      NEXT;
    CASE(BC_ORDOF):
      {
	REGISTER INT n = IVAL(top[-1]);

	if (n<0 || n>=SLEN(top[-2]))
	  goto raise_stringnth;
	top[-2] = Kint ((INT)STEXT(top[-2])[n]);
	POPTOP(1);
	NEXT;
      }
    CASE(BC_SUBSTR):
      {
	REGISTER INT m, n;

	m = IVAL(top[-2]);
	if (m<0)
	  goto raise_stringnth;
	n = IVAL(top[-1]);
	if (n>SLEN(top[-3]))
	  goto raise_stringnth;
	if (n<m)
	  n = m;
#if STOP_AND_COPY_VECTORS
	{
	  REGISTER INT size = n-m;
	  CHAR *s = (CHAR *)alloctmp_anchored (size, (char *)&env);
	  safememcpy(s,(char *)&STEXT(top[-3])[m], size);
	  top[-3] = Kstring (size, s);
	}
#else
	top[-3] = Kstring (n-m, &STEXT(top[-3])[m]);
#endif
	POPTOP(2);
	NEXT;
      }
    CASE(BC_CALLBIND): /* mostly do as though we were calling an interpreted closure
			(see CASE(BC_CALL)) */
      {
	REGISTER ExprPtr f;

	f = bc->consts[*pc++];
	if (f->kind!=(CHAR)E_TUPLE_2) /* compiled initialization function. */
	  {
	    top[-2] = (*ACTION(f)) (ENV_CONS(top[-2],1), top[-1]);
	    POPTOP(1);
	    NEXT;
	  }
	curbcframe->bc = bc;
	curbcframe->pc = pc;
	curbcframe++;
	bc = (byteCodeChunk *)CAR(f);
#if PROFILING
	{
	  ExprPtr prof = bc->consts[0];
#if OS==AMIGA_OS
	  unsigned int clock[2];
#define TMS clock
#else
	  struct tms tms;
#define TMS &tms
#endif

#ifdef PROFDEBUG
	  if (prof==NOEXPR || currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
	    abort();
#endif
	  if (TYPE(prof)==EA_PROFILE)
	    {
#if PROFILING & PROFILING_CALL_COUNTS
	      PROFDATA(prof)->nCalls++;
#endif
#if PROFILING & PROFILING_PROPER_TIMES
#if OS==AMIGA_OS
	      (void) timer(clock);
#else
	      times(&tms);
#endif
#ifdef PROFDEBUG
	      if (currentProf!=NOEXPR && currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
		abort();
#endif
	      if (currentProf!=NOEXPR && TYPE(currentProf)==EA_PROFILE)
		profExitProper(PROFDATA(currentProf),TMS);
	      profEnterProper(PROFDATA(prof),TMS);
#endif
	      profEnter (prof, TMS);
	    }
#if PROFILING & PROFILING_PROPER_TIMES
	  currentProf = prof;
#endif
	}
#endif
	SYNCTOP();
	if (BCSTACKS._x.n+bc->stackSize > BC_STACK_SIZE+2)
	  {
#if DEBUGGING!=0
	    parentBcFrames = BCF_PTR;
	    BCF.curBcFrame = (struct bcframe *)curbcframe;
	    parentStack = BCSTACKS_PTR;
#endif
	    res = bcEval(bc,ENV_CONS(top[-2],1),top[-1]);
	    POPTOP(1);
	    goto overflow_on_call;
	  }
	switch (bc->nargs)
	  {
	  case 0: POPTOP(1); break;
	  case 1: break;
	  case 2: top[0] = CDR(top[-1]); top[-1] = CAR(top[-1]); PUSHTOP(1); break;
	  default:
	    {
	      COUNTER i;
	      ExprPtr *ep = AVAL(top[-1])->expr, *epp;

	      for (i=bc->nargs, epp = &top[-1]; --i>=0; )
		*epp++ = *ep++;
	      PUSHTOP(bc->nargs-1);
	      break;
	    }
	  }
	pc = bc->instr;
	NEXT;
      }

#include "hastag.c"
    CASE(BC_ILLEGAL):
#ifdef SWITCH
    default:
#endif
      kinderror(message(606),pc[-1]);
      /*NOTREACHED*/
#ifdef SWITCH
      break;
    }
#endif
#if DEBUGGING!=0
  (void) freetmp (bcstacks_ptr);
#endif
#if PROFILING
  if (currentProf!=NOEXPR)
    {
#if OS==AMIGA_OS
      unsigned int clock[2];
#define TMS clock
#else
      struct tms tms;
#define TMS &tms
#endif

#if OS==AMIGA_OS
      (void) timer(clock);
#else
      times(&tms);
#endif
#ifdef PROFDEBUG
      if (currentProf==NOEXPR || currentProf->kind!=(CHAR)BUILD_HEADER_KIND)
	abort();
#endif
      if (TYPE(currentProf)==EA_PROFILE)
	profExit(currentProf,TMS);
#if PROFILING & PROFILING_PROPER_TIMES
      if (TYPE(currentProf)==EA_PROFILE)
	profExitProper(PROFDATA(currentProf),TMS);
#ifdef PROFDEBUG
      if (saveProf!=NOEXPR && saveProf->kind!=(CHAR)BUILD_HEADER_KIND)
	abort();
#endif
      if (saveProf!=NOEXPR && TYPE(saveProf)==EA_PROFILE)
	profEnterProper(PROFDATA(saveProf),TMS);
      currentProf = saveProf;
#endif
    }
#endif
  return NIL;
#undef EXCSTACK
#undef RAWDATA
}

PRIVATE COUNTER getLevel(env)
IN CONST REGISTER ExprPtr env;
{
  REGISTER COUNTER n;

  for (n=0; env!=NIL; n++, env = ENEXT(env));
  return n;
}

PUBLIC ExprPtr Eval(e,env)
IN CONST ExprPtr e;
IN CONST ExprPtr env;
{
  ExprPtr code;
  ENTER;

  code = Compile(e,getLevel(env),false);
#if DEBUGGING!=0
  parentBcFrames = NULL;
  parentStack = NULL;
#endif
  e = bcEval((byteCodeChunk *)CAR(code),env,NIL);
  KEEPONLY(e);
  return e;
}

#define INITENV(vp,from,to) \
{ \
  REGISTER ExprPtr *_ep,*_epend; \
 \
  for (_ep = &vp->expr[from], _epend = &vp->expr[to]; _ep<_epend;) \
    *_ep++ = NOEXPR; \
}

PUBLIC Boolean EvalBindingsRC(bindings,env,from,to)
IN CONST ExprPtr bindings;
OUT ExprPtr env;
IN COUNTER from,to;
/* evaluates bindings that are to be put into vp,
   from index 'from' to 'to'. There should be no
   values stored there. */
{
  Boolean nomatch;
  VecPtr vp;
  ExprPtr code;
  ENTER;

  vp = EVARS(env);
  INITENV(vp,from,to);
  vp->n = to;
  code = Compile(bindings,getLevel(env)-1,true);
#if DEBUGGING!=0
  parentBcFrames = NULL;
  parentStack = NULL;
#endif
  nomatch = (bcEval((byteCodeChunk *)CAR(code),env,NIL)==trueValue);
  QUIT;
  return nomatch;
}

PUBLIC Boolean EvalBindingsRCIt(bindings,env,from,to)
IN CONST ExprPtr bindings;
OUT ExprPtr env;
IN COUNTER from,to;
/* same as EvalBindingsRC(), but updates
   refcount of the special variable 'it'	*/
{
  ExprPtr saveit,it;
  VecPtr vp;
  Boolean nomatch;

  if (env==getTopEnv())
    {
      vp = EVARS(env);
      saveit = vp->expr[itIndex()];
      nomatch = EvalBindingsRC(bindings,env,from,to);
#if STOP_AND_COPY_VECTORS
      vp = EVARS(env);
#endif
      it = vp->expr[itIndex()];
      if (it!=saveit)
	{
	  itBarrier = true;
	}
      return nomatch;
    }
  return EvalBindingsRC(bindings,env,from,to);
}

#endif
