/* Bytecode interpreter callcc trampoline.
   Copyright (C) 2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#if TRANSLATOR
#include "expr.h"
#include "refcount.h"
#include "bc.h"
#include "rbc.h"

#define __SEG__ BYTECODE

#define DEBUG

/* The following should not be placed in the same file as bcEval0()
   (i.e, bci.c),
   otherwise the compiler might have the really bad idea of inlining it.
   If it inlined it, then it would only detect that the stack is full
   after allocating all the local arrays of bcEval0() on the stack, and
   the system would then core dump after trashing its own stack. */

IMPORT ExprPtr bcEval0 P((IN CONST byteCodeChunk *bc,
			  IN CONST ExprPtr env, ...));

DEF_PRIVATE_ACTION_1ARG(bceval)
{
  ExprPtr res;
  struct bceval *data = (struct bceval *)ENV_UNDERLYING(env);

  M_VARCALL2_CC(res,bcEval0,data->bc,data->arg[0],ENV_ARITY(_env),data->arg);
  return res;
}
ACTION_END

#ifdef HAS_STDARG
PUBLIC ExprPtr bcEval(IN CONST byteCodeChunk *bc,
		      IN CONST ExprPtr env, ...)
#else
PUBLIC ExprPtr bcEval(bc, env, va_alist)
IN CONST byteCodeChunk *bc;
IN CONST ExprPtr env;
va_dcl
#endif
{
  va_list _args;

  if (STKFULL(bc))
    {
      COUNTER i, n = ENV_ARITY(env);
      struct bceval *data = (struct bceval *) alloctmp(OFFSETOF(struct bceval,arg[n+1]));
      data->bc = bc;
      VA_START;
      data->arg[0] = env;
      for (i=1; i<=n; i++)
	data->arg[i] = va_arg(_args,ExprPtr);
      va_end(_args);
      return CallCC(ACTION_NAME(bceval),(ExprPtr)data,true);
    }
  {
    ExprPtr res;
    ExprPtr uenv = ENV_UNDERLYING(env);

    M_VARCALL2(res,bcEval0,bc,uenv,ENV_ARITY(env));
    return res;
  }
}

#endif
