/* Basic structure allocation functions (unshared), plus initialization.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#include "expr.h"
#include "mem.h"
#include "special.h"
#include "expproto.h"
#include "memproto.h"
#if PLDI93_HACK
#include "mem.h"
#endif
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "hakproto.h"
#ifdef HAS_PROTOTYPES
#include "sexpr.h"
#endif

#undef DEBUG1

#define __SEG__ EXPR

INT stackAlloced = 0;

PUBLIC BucketPtr theBucketList;
PUBLIC BucketPtr theHomoBucketList,*endHomoBucketList;
PUBLIC BucketPtr theHeteroBucketList,*endHeteroBucketList;
PUBLIC ExprPtr *nextHomoExpr,*nextHeteroExpr;
PUBLIC ExprPtr *lastHomoExpr,*lastHeteroExpr;
PUBLIC BucketPtr currentHomoBucket,currentHeteroBucket;

PUBLIC BucketPtr theSlotBucketList;
PUBLIC ExprPtr theAvailSlotList;
PUBLIC ExprPtr reserveSlot;
	/* like reserveCell for cells */

PUBLIC ExprPtr nihil;

PUBLIC INT nCells = 0L;
EXTERN SHORT generation_count;

EXTERN EXTENDED fone;

EXTERN INT liveCells[2][MAX_GENERATIONS];

#define DEFAULT_GROW_FACTOR 2
PUBLIC EXTENDED growFactor = DEFAULT_GROW_FACTOR;
PUBLIC INT growBuckets = DEFAULT_GROW_FACTOR*BUCKET_SIZE;
#if STOP_AND_COPY_VECTORS
#define COPY_SLOWDOWN 3
 /* rule of thumb: the number of times slower it is
    to move an average block than to mark it with GC_MARK() */
PUBLIC EXTENDED copySlowdown = COPY_SLOWDOWN;
PUBLIC EXTENDED scGrowFactor = COPY_SLOWDOWN*DEFAULT_GROW_FACTOR;
PUBLIC EXTENDED minScGrowFactor = COPY_SLOWDOWN;
PUBLIC EXTENDED maxScGrowFactor = 4*COPY_SLOWDOWN*DEFAULT_GROW_FACTOR;
EXTERN INT minAllocedPages,allocedPages;
#endif

EXTERN Boolean itBarrier;
EXTERN INT maxCells;
EXTERN ExprPtr subscriptException;

#ifdef DEBUG
PUBLIC void ELcheck()
{
  BucketPtr bp;
  ExprPtr *epp,*endp;

  for (epp=nextHomoExpr, endp=lastHomoExpr; epp<endp; epp++)
    if ((*epp)->kind!=(CHAR)E_FREE)
      CRASH();
  for (bp=currentHomoBucket; (bp = bp->next)!=(BucketPtr)0L; )
    for (epp=bp->newcells+bp->firstNewCell+bp->occupancy[generation_count],
	 endp=bp->newcells+BUCKET_SIZE; epp<endp; epp++)
      if ((*epp)->kind!=(CHAR)E_FREE)
	CRASH();
  for (epp=nextHeteroExpr, endp=lastHeteroExpr; epp<endp; epp++)
    if ((*epp)->kind!=(CHAR)E_FREE)
      CRASH();
  for (bp=currentHeteroBucket; (bp = bp->next)!=(BucketPtr)0L; )
    for (epp=bp->newcells+bp->firstNewCell+bp->occupancy[generation_count],
	 endp=bp->newcells+BUCKET_SIZE; epp<endp; epp++)
      if ((*epp)->kind!=(CHAR)E_FREE)
	CRASH();
}
#endif

PUBLIC BucketPtr NewBucket(new,force)
IN allocFunction new;
IN Boolean force;
{
  BucketPtr bp;
  Expression *ep,*oldep,**epp;
  COUNTER i;

  if (nCells>maxCells && !force) /* don't exceed limit, unless we need to */
    return (BucketPtr)0L;
  bp = (BucketPtr)(*new)((long)sizeof(Bucket));
  if (bp!=(BucketPtr)0L)
    {
      bp->next = theBucketList;
      theBucketList = bp;
      bp->firstNewCell = 0;
      for (i=0; i<MAX_GENERATIONS; i++)
	bp->occupancy[i] = 0;
      i=0; ep=bp->cell;
      epp = bp->newcells;
      /* it is important that we add new cells *to the end* of theAvailExprList.
         This improves locality of data when reallocating just after GCs. */
      oldep = NOEXPR;
      for (; i<BUCKET_SIZE; i++, ep++)
	{
	  *epp++ = ep;
	  ep->kind = (CHAR)E_FREE;
	  ep->mark = 0;
	  NEXTFREE(ep) = oldep;
	  ELINK(ep) = NEWGEN;
	  oldep = ep;
	}
      nCells += BUCKET_SIZE;
    }
  return bp;
}

PUBLIC void bumpHomoExpr()
{
  do {
    currentHomoBucket = currentHomoBucket->next;
    if (currentHomoBucket==(BucketPtr)0L)
      {
	if (theBucketList!=(BucketPtr)0L)
	  {
	    currentHomoBucket = *endHomoBucketList = theBucketList;
	    endHomoBucketList = &currentHomoBucket->next;
	    theBucketList = *endHomoBucketList;
	    *endHomoBucketList = (BucketPtr)0L;
	  }
	else
	  {
	    GCgrow(0, NULL);
	    return;
	  }
      }
    nextHomoExpr = currentHomoBucket->newcells+currentHomoBucket->firstNewCell+
      currentHomoBucket->occupancy[generation_count];
    lastHomoExpr = currentHomoBucket->newcells+BUCKET_SIZE;
  } while (nextHomoExpr==lastHomoExpr);
}

PUBLIC void bumpHeteroExpr()
{
  do {
    currentHeteroBucket = currentHeteroBucket->next;
    if (currentHeteroBucket==(BucketPtr)0L)
      {
	if (theBucketList!=(BucketPtr)0L)
	  {
	    currentHeteroBucket = *endHeteroBucketList = theBucketList;
	    endHeteroBucketList = &currentHeteroBucket->next;
	    theBucketList = *endHeteroBucketList;
	    *endHeteroBucketList = (BucketPtr)0L;
	  }
	else
	  {
	    GCgrow(0, NULL);
	    return;
	  }
      }
    nextHeteroExpr = currentHeteroBucket->newcells+currentHeteroBucket->firstNewCell+
      currentHeteroBucket->occupancy[generation_count];
    lastHeteroExpr = currentHeteroBucket->newcells+BUCKET_SIZE;
  } while (nextHeteroExpr==lastHeteroExpr);
}

PUBLIC BucketPtr NewSlotBucket(new)
IN allocFunction new;
{
  BucketPtr bp;
  Expression *ep,*oldep;
  COUNTER i;

  bp = (BucketPtr)(*new)((long)sizeof(Bucket));
  if (bp!=(BucketPtr)0L)
    {
      bp->next = theSlotBucketList;
      theSlotBucketList = bp;
      oldep = NOEXPR;
      for (i=0, ep=bp->cell; i<BUCKET_SIZE; i++, ep++)
	{
	  ep->kind = (CHAR)E_SLOT_BRANCH;
	  ep->mark = 0;
	  NEXTFREE(ep) = oldep;
	  oldep = ep;
	}
      if (reserveSlot==NOEXPR)
	{
	  theAvailSlotList = oldep;
	}
      else
	{ /* append at the end */
	  NEXTFREE(reserveSlot) = oldep;
	}
      reserveSlot = bp->cell; /* the last allocated slot */
    }
  return bp;
}

PUBLIC COUNTER VecBucketN[] = {
  0,
  256, 128, 85, 64,
  51, 42, 36, 32,
  28, 25, 23, 21
  };

#define VBSIZE(n) VARSIZE(VecBucket,contents,n)

PRIVATE COUNTER VecBucketSize[] = { /* entry i uses i*VecBucketN[i] elements */
  0,
  VBSIZE(256), VBSIZE(256), VBSIZE(255), VBSIZE(256),
  VBSIZE(255), VBSIZE(252), VBSIZE(252), VBSIZE(256),
  VBSIZE(252), VBSIZE(250), VBSIZE(253), VBSIZE(252)
  };

PUBLIC VecBucketPtr theVecBucketList[] = {
  NULL,
  NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL
  };

PUBLIC VecBucketPtr theAvailVecBuckets[] = {
  NULL,
  NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL
  };

PUBLIC VecBucketPtr NewVecBucket(new,n)
IN allocFunction new;
IN INT n;
{
  VecBucketPtr bp,aux;
  VecBucketPtr *ep;
  INT i,ni;

  bp = (VecBucketPtr) (*new)((long)VecBucketSize[n]);
  if (bp!=(VecBucketPtr)0L)
    {
      bp->next = theVecBucketList[n];
      theVecBucketList[n] = bp;
      aux = theAvailVecBuckets[n];
      for (i = 0, ep = bp->contents, ni = VecBucketN[n]; i<ni; i++, ep += n)
	{
	  *ep = aux;
	  aux = (VecBucketPtr)ep;
	}
      theAvailVecBuckets[n] = aux;
    }
  return bp;
}

PUBLIC char dummy;

PUBLIC VOIDPTR VecAlloc(size)
IN INT size;
/* returns a possibly unaligned datum. It will always be
   aligned like a char *, but possibly not as much as
   what malloc() returns. */
{
  char *p;

  VECALLOC(p,size);
  return (VOIDPTR)p;
}

PUBLIC void VecFree(p,size)
IN VOIDPTR p;
IN INT size;
{
  VECFREE(p,size);
}

PUBLIC ExprPtr MakePair(kind,car,cdr)
IN ExprKind kind;
IN CONST ExprPtr car,cdr;
{
  REGISTER ExprPtr e;

  EALLOC_HOMO(e);
  e->kind = (CHAR)kind;
  CAR(e) = car;
  CDR(e) = cdr;
  return e;
}

#ifdef HAS_PROTOTYPES
PUBLIC ExprPtr MakeData(IN SHORT magic,
			IN CONST ExprPtr car,
			IN CONST ExprPtr cdr)
#else
PUBLIC ExprPtr MakeData(magic,car,cdr)
IN SHORT magic;
IN CONST ExprPtr car,cdr;
#endif
{
  REGISTER ExprPtr e;

  EALLOC_HOMO(e);
  e->kind = (CHAR)E_DATA;
  e->magic = magic;
  CAR(e) = car;
  CDR(e) = cdr;
  return e;
}

PUBLIC ExprPtr MakeCompiled(kind,action,env)
IN ExprKind kind;
IN compiledAction action;
IN CONST ExprPtr env;
{
  REGISTER ExprPtr e;

  EALLOC_HOMO(e);
  e->kind = (CHAR)kind;
  ACTION(e) = action;
  QENV(e) = env;
  return e;
}

DEF_PRIVATE_ACTION_0ARG(get_forced_promise)
{
  return ENV_UNDERLYING(env);
}
ACTION_END

DEF_PRIVATE_ACTION_0ARG(force_promise)
{ /* env is a couple (fun,prom) where fun is the function to compute the value and
     prom is the promise itself, which is a compiled closure. */
  ExprPtr f,val;
  ExprPtr uenv;

  uenv = ENV_UNDERLYING(env);
  f = CAR(uenv);
  val = (*ACTION(f))(ENV_CONS(QENV(f),0));
  f = CDR(uenv); /* promise */
  ACTION(f) = ACTION_NAME(get_forced_promise);
  SETQ(QENV(f),val);
  HITBARRIER(f);
  return val;
}
ACTION_END

PUBLIC ExprPtr MakePromise(fun)
IN CONST ExprPtr fun;
{
  ExprPtr e,c;

  BEGINQUICKKEEP();
  MAKECONS(fun,NIL,c); /* NIL is dummy */
  QUICKKEEP(c);
  e = MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(force_promise),c);
  SETCDR(c,e); /* make CAR of QENV of the promise point to the promise itself */
  ENDQUICKKEEP();
  return e;
}

PUBLIC ExprPtr MakeStack(jb,chunk)
IN CONST stackDesc *jb;
IN CONST StackChunk *HEAP chunk;
{
  REGISTER ExprPtr e;
  stackDesc *new;

  CHECKHETEROFREE();
#if STOP_AND_COPY_VECTORS
  scAlloc(*(char **)&new,sizeof(stackDesc));
#else
  new = (stackDesc *)MALLOC(sizeof(stackDesc));
  /* cannot use scAlloc, since we need new to be aligned. */
#endif
  *new = *jb;
  EALLOC_HETERO_NOGC(e);
  e->kind = (CHAR)E_STACK;
  SJMPBUF(e) = new;
  SCHUNK(e) = chunk;
  stackAlloced++;
  return e;
}

PUBLIC ExprPtr MakeRecord(kind,shape,n,exprs)
IN ExprKind kind;
IN CONST ExprPtr shape;
IN INT n;
IN ExprPtr *exprs;
{
  VecPtr vpout;
  ExprPtr e;

  CHECKHETEROFREE();
  scAlloc(*(char **)&vpout,VECSIZE(n));
  vpout->n = n;
  safememcpy(vpout->expr,exprs,n*sizeof(ExprPtr));
  MAKEKRECORD_NOGC(kind,shape,vpout,e);
  return e;
}

PUBLIC ExprPtr ListToArray0(l)
IN CONST ExprPtr l;
{
#ifdef OBSOLETE
  ExprPtr e1,e2;
#endif
  ExprPtr rest,e,*ep;
  VecPtr vp;
  COUNTER i,n;
  INT nn;

#ifdef OBSOLETE
  if (l==NIL)
    return array0;
  e1 = CAR(l);
  rest = CDR(l);
  if (rest==NIL)
    {
      MAKESING(E_ARRAY_1,e1,e);
      return e;
    }
  e2 = CAR(rest);
  rest = CDR(rest);
  if (rest==NIL)
    {
      EALLOC_HOMO(e);
      e->kind = (CHAR)E_ARRAY_2;
      CAR(e) = e1;
      CDR(e) = e2;
      return e;
    }
  nn = ListLength(rest)+2;
  n = nn;
  if ((INT)n!=nn)
    return Raise(subscriptException);
  CHECKHETEROFREE();
  scAlloc(*(char **)&vp,VECSIZE(n));
  vp->n = n;
  vp->expr[0] = e1;
  vp->expr[1] = e2;
  ep = &vp->expr[2];
  for (i=2; i<n; i++, rest=CDR(rest))
    {
      e = CAR(rest);
      *ep++ = e;
    }
#else
  nn = ListLength(l);
  n = nn;
  if ((INT)n!=nn)
    return Raise(subscriptException);
  CHECKHETEROFREE();
  scAlloc(*(char **)&vp,VECSIZE(n));
  vp->n = n;
  ep = &vp->expr[0];
  for (i=0, rest=l; i<n; i++, rest=CDR(rest))
    {
      e = CAR(rest);
      *ep++ = e;
    }
#endif
  EALLOC_HETERO_NOGC(e);
  e->kind = (CHAR)E_ARRAY_GNRL;
  /*e->mark = (CHAR)0;*/
  AVAL(e) = vp;
  ASHAPE(e) = NIL;
  return e;
}

PUBLIC ExprPtr MakeArray(n,e)
IN COUNTER n;
IN CONST ExprPtr e;
{
  ExprPtr ep,*elt;
  VecPtr vp;

#ifdef OBSOLETE
  switch (n)
    {
    case 0:
      return array0;
    case 1:
      MAKESING(E_ARRAY_1,e,ep);
      return ep;
    case 2:
      EALLOC_HOMO(ep);
      ep->kind = (CHAR)E_ARRAY_2;
      CAR(ep) = e;
      CDR(ep) = e;
      return ep;
    default:
#endif
      CHECKHETEROFREE();
      scAlloc(*(char **)&vp,VECSIZE(n));
      vp->n = n;
      EALLOC_HETERO_NOGC(ep);
      ep->kind = (CHAR)E_ARRAY_GNRL;
      /*ep->mark = (CHAR)0;*/
      AVAL(ep) = vp;
      ASHAPE(ep) = NIL;
      elt = vp->expr;
      while (n--!=0)
	{
	  *elt++ = e;
	}
      return ep;
#ifdef OBSOLETE
    }
#endif
}

PUBLIC ExprPtr MakeExternal(desc,data)
IN CONST extDescPtr desc;
IN CONST char *data;
{
  REGISTER ExprPtr e;

  EALLOC_HETERO(e);
  e->kind = (CHAR)E_EXTERNAL;
  EXTDESC(e) = desc;
  EXTDATA(e) = data;
  return e;
}

PUBLIC ExprPtr MakeCounted(kind,expr,n)
IN ExprKind kind;
IN CONST ExprPtr expr;
IN INT n;
{
  REGISTER ExprPtr e;

  EALLOC_HOMO(e);
  e->kind = (CHAR)kind;
  COUNTEDVAL(e) = expr;
  COUNTEDN(e) = n;
  return e;
}

#ifdef HAS_PROTOTYPES
PUBLIC ExprPtr MakePolyfun (IN SHORT magic,
			    IN CONST ExprPtr expr,
			    IN INT n)
#else
PUBLIC ExprPtr MakePolyfun(magic,expr,n)
     IN SHORT magic;
     IN CONST ExprPtr expr;
     IN INT n;
#endif
{
  ExprPtr e;

  e = MakeCounted(E_POLYFUN,expr,n);
  e->magic = magic;
  return e;
}

PUBLIC int CompareStrings(s1,s2)
IN CONST ExprPtr s1,s2;
{
  REGISTER INT n;
  INT nn;

  n = SLEN(s1);
  nn = SLEN(s2);
  if (n<nn)
    return -1;
  if (n>nn)
    return 1;
  {
    REGISTER CHAR *s,*t,c,cc;

    s = STEXT(s1);
    t = STEXT(s2);
    while (n--!=0)
      {
	c = *s++;
	cc = *t++;
	if (c<cc)
	  return -1;
	if (c>cc)
	  return 1;
      }
  }
  return 0;
}

PRIVATE int cmpstr(m1,m2)
IN CONST ANSI_CONST VOIDPTR m1;
IN CONST ANSI_CONST VOIDPTR m2;
{/* note that we need to have an ordering of the fields that does not
    depend on the system ordering, for example, to make compilation to C
    of positions valid. We choose the order that compares lengths first,
    then the characters in the lexicographic ordering (this has the advantage
    of letting 1<2<...<9<10<11<..., but this has no influence here). */
  return CompareStrings(((field *)m1)->label,((field *)m2)->label);
}

#ifdef HAS_PROTOTYPES
PUBLIC void sortFieldArray(IN COUNTER n,INOUT field *array)
#else
PUBLIC void sortFieldArray(n,array)
IN COUNTER n;
INOUT field *array;
#endif
{
  qsort(array,(long)n,(long)sizeof(field),cmpstr);
}

PUBLIC ExprPtr fieldsToShape1(n,n0,array)
IN COUNTER n,n0;
IN CONST field *array;
{
  if (n==0)
    return NewMap();
  if (n==1)
    return MapSingleton(array->label,KINT(n0));
  {
    COUNTER m;

    m = n>>1;
    return MapMerge(fieldsToShape1(m,n0,array),
		    fieldsToShape1(n-m,n0+m,array+m));
  }
}

PUBLIC void fieldsToExprs(exprs,array,n)
OUT ExprPtr *exprs;
IN CONST field *array;
IN COUNTER n;
{
  REGISTER COUNTER i;

  for (i=0; i<n; i++)
    *exprs++ = (array++)->val;
}

PUBLIC ExprPtr MakeRecordGnrl(n,array,shape)
IN COUNTER n; /* n>=2 */
IN CONST field *array;
IN CONST ExprPtr shape;
{
  VecPtr vp;
  REGISTER ExprPtr *ep,e;
  REGISTER field *fields;
  REGISTER COUNTER i;

  CHECKHETEROFREE();
  scAlloc(*(char **)&vp,VECSIZE(n));
  vp->n = n;
  for (i=0, ep=vp->expr, fields=array; i<n; i++)
    {
      e = (fields++)->val;
      *ep++ = e;
    }
  MAKERECORD_NOGC(shape,vp,e);
  return KEEP(e);
}

#if PLDI93_HACK
PUBLIC ExprPtr MakeReal(x)
IN CONST REAL *x;
{
  REGISTER ExprPtr e;

  EALLOC_HOMO(e);
  e->kind = (CHAR)E_REAL;
  XVAL(e) = *x;
  return e;
}

EXTERN ExprPtr voidString,*charStrings;

PUBLIC ExprPtr MakeString1(len,name,copy)
IN REGISTER INT len;
IN REGISTER CONST CHAR *name;
IN Boolean copy;
{
  REGISTER ExprPtr cons;

  switch (len)
    {
    case 0:
      if (!copy)
	MARKFREE(name);
      return voidString;
    case 1:
      if (!copy)
	MARKFREE(name);
      return charStrings[*name];
    default:
      EALLOC_HETERO(cons);
      cons->kind = (CHAR)E_STRING;
      SLEN(cons) = 0L; /* to prevent stop-and-copy gc from moving name,
      			which is normally allocated on the heap */
      STEXT(cons) = name;
      if (copy)
	{
	  BEGINQUICKKEEP();
	  QUICKKEEP(cons);
	  scAlloc(*(char **)&STEXT(cons),len);
	  safememcpy(STEXT(cons),name,len);
	  ENDQUICKKEEP();
	}
      SLEN(cons) = len;
      return cons;
    }
}

PUBLIC ExprPtr MakeStr(name)
IN CONST CHAR *name;
{
  return MakeString1(strlen((char *)name),name,true);
}

PUBLIC ExprPtr keepCons(a,b)
IN CONST ExprPtr a,b;
{
  ExprPtr c;

  c = MakePair(E_CONS,a,b);
  return KEEP(c);
}
#endif

#define EXTDESC_HASHSIZE 17
PRIVATE extDesc *extDescHash[EXTDESC_HASHSIZE] = { NULL, };

PUBLIC extDesc *getExtDesc(id)
IN CONST CHAR id[];
{
  int code;
  extDesc *ed;

  code = ((unsigned long)((((((id[0] << 8) + id[1]) << 8) + id[2]) << 8) + id[3])) % EXTDESC_HASHSIZE;
  for (ed = extDescHash[code]; ed!=NULL; ed=ed->next)
    if (ed->id[0]==id[0] && ed->id[1]==id[1] && ed->id[2]==id[2] && ed->id[3]==id[3])
      return ed;
  return NULL;
}

#undef __SEG__
#define __SEG__ INIT

PRIVATE void initBucketLists()
{
  currentHomoBucket = *endHomoBucketList = theBucketList;
  endHomoBucketList = &currentHomoBucket->next;
  theBucketList = *endHomoBucketList;
  *endHomoBucketList = (BucketPtr)0L;
  nextHomoExpr = currentHomoBucket->newcells;
  lastHomoExpr = currentHomoBucket->newcells+BUCKET_SIZE;

  currentHeteroBucket = *endHeteroBucketList = theBucketList;
  endHeteroBucketList = &currentHeteroBucket->next;
  theBucketList = *endHeteroBucketList;
  *endHeteroBucketList = (BucketPtr)0L;
  nextHeteroExpr = currentHeteroBucket->newcells;
  lastHeteroExpr = currentHeteroBucket->newcells+BUCKET_SIZE;
}

PUBLIC void InitExprs()
{
  Boolean cont;
  COUNTER i;

  for (i=0; i<MAX_GENERATIONS; i++)
    liveCells[0][i] = liveCells[1][i] = 0L;
  theBucketList = (BucketPtr)0L;
  theHomoBucketList = (BucketPtr)0L;
  endHomoBucketList = &theHomoBucketList;
  theHeteroBucketList = (BucketPtr)0L;
  endHeteroBucketList = &theHeteroBucketList;
  nextHomoExpr = lastHomoExpr = nextHeteroExpr = lastHeteroExpr = (ExprPtr *)0L;
  theSlotBucketList = (BucketPtr)0L;
  theAvailSlotList = NOEXPR;
  reserveSlot = NOEXPR;
#if STOP_AND_COPY_VECTORS
  twinPageList = (PagePtr)0L;
  curPage = (PagePtr)0L;
  spaceIndex = 0; /* important since during initialization, pages
		     are actually single, not twin */
  nPages = 0;
#endif
  for (i=0; i<VECBN_N; i++)
    {
      theVecBucketList[i] = (VecBucketPtr)0L;
      theAvailVecBuckets[i] = (VecBucketPtr)0L;
    }
#if SPECIAL_MALLOC
  AllocRainyDayFund(MemSize()/4);
#endif
  cont = true;
  for (i=0; i<10 && cont; i++)
    {
#if STOP_AND_COPY_VECTORS
      if (NewPage(malloc_fail)==(PagePtr)0L)
	cont = false;
#endif
      if (NewBucket(malloc_fail,false)==(BucketPtr)0L) /* preallocate some cells */
	cont = false;
    }
#if STOP_AND_COPY_VECTORS
  curPage = twinPageList;
  if (twinPageList==(PagePtr)0L)
    nomem();
  minAllocedPages = allocedPages;
#endif
  initBucketLists();
  EALLOC_HOMO(nihil);
  ELINK(nihil) = NOEXPR; /* used in gc, to terminate the write barrier */
  nihil->kind = (CHAR)E_NIL;
}

PUBLIC void installGrowFactor(grow)
IN CONST double *grow;
{
  growFactor = *grow;
  growBuckets = (INT)(growFactor*BUCKET_SIZE);
#if STOP_AND_COPY_VECTORS
  scGrowFactor = minScGrowFactor = copySlowdown*(*grow);
#endif
}

PUBLIC void installMaxGrowFactor(grow)
IN CONST double *grow;
{
#if STOP_AND_COPY_VECTORS
  maxScGrowFactor = copySlowdown*(*grow);
#endif
}

PUBLIC COUNTER globLimit=0;
PUBLIC BucketPtr oldHomoBucketList = NULL,oldHeteroBucketList = NULL;

PRIVATE void Olden(bucklist)
INOUT BucketPtr bucklist;
{
  REGISTER ExprPtr ep;

  {
    struct Bucket *buck;
    REGISTER COUNTER i;

    for (buck=bucklist; buck!=(BucketPtr)0L; buck=buck->next)
      for (i=0, ep=buck->cell; i<BUCKET_SIZE; i++, ep++)
	{
	  GC_MARK(ep);
	  if (ELINK(ep)==NEWGEN)
	    ELINK(ep) = OLDGEN; /* shared objects are untouched by this */
	}
  }
}

PUBLIC void registerExtDesc(desc)
IN CONST extDesc *desc;
{
  int code;
  extDesc **edp,*ed;

  code = ((unsigned long)((((((desc->id[0] << 8) + desc->id[1]) << 8) + desc->id[2]) << 8) + desc->id[3])) % EXTDESC_HASHSIZE;
  edp = &extDescHash[code];
  ed = *edp;
  desc->next = ed;
  *edp = desc;
}

PUBLIC void cutoffInitials()
{
  Boolean cont;
  COUNTER i;

  Olden(theHomoBucketList);
  oldHomoBucketList = theHomoBucketList;
  Olden(theHeteroBucketList);
  oldHeteroBucketList = theHeteroBucketList;
  theHomoBucketList = theHeteroBucketList = (BucketPtr)0L;
  endHomoBucketList = &theHomoBucketList;
  endHeteroBucketList = &theHeteroBucketList;
  nextHomoExpr = lastHomoExpr = nextHeteroExpr = lastHeteroExpr = (ExprPtr *)0L;
  nCells = 0L;
  for (i=0; i<MAX_GENERATIONS; i++)
    liveCells[0][i] = liveCells[1][i] = 0L;
  /*
    theSlotBucketList = (BucketPtr)0L;
    theAvailSlotList = reserveSlot;
    :are never reclaimed, so don't touch them.
    */
#if STOP_AND_COPY_VECTORS
  twinPageList = (PagePtr)0L;
  curPage = (PagePtr)0L;
  spaceIndex = 0;
  nPages = allocedPages = 0L;
#endif
  /*
    theSmallBucketList = (SmallBucketPtr)0L;
    theAvailSmallBuckets = (SmallBucketPtr)0L;
    theMedBucketList = (MedBucketPtr)0L;
    theAvailMedCells = (struct MedCell *)0L;
    :are never swept, so don't reinitialize them.
    */
  cont = true;
  for (i=0; i<10 && cont; i++)
    {
#if STOP_AND_COPY_VECTORS
      if (NewPage(malloc_fail)==(PagePtr)0L)
	cont = false;
#endif
      if (NewBucket(malloc_fail,false)==(BucketPtr)0L) /* preallocate some cells */
	cont = false;
    }
#if STOP_AND_COPY_VECTORS
  curPage = twinPageList;
  if (twinPageList==(PagePtr)0L)
    nomem();
  minAllocedPages = allocedPages;
#endif
  initBucketLists();
  /* it would be too complicated and not interesting
     enough to implement write barriers for type assums, states
     and compile assums, so we don't do it */
  cutoffInitialHeaders();
  cutoffInitialTypeAssums();
  cutoffInitialTypeStates();
  cutoffInitialCompileAssums();
  cutoffInitialBCs();
  globLimit = EVARS(GLOBENV)->n;
  itBarrier = false;
}
