%{
/* HimML grammar (without Standard ML modules).
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#include "abssyn.h"
#include "expr.h"
#include "gram.h"
#include "type.h"
#include "compile.h"
#include "input.h"
#include "timing.h"
#include "debug.h"
#include "prtproto.h"
#include "grmproto.h"
#include "memproto.h"
#include "evlproto.h"
#include "errproto.h"
#include "cmpproto.h"
#include "typproto.h"
#include "stkproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#else
#include <errno.h>
#endif
#ifdef HAS_PROTOTYPES
#include "sgram.h"
#endif

#define __SEG__ PARSE

#ifndef bcopy
#define bcopy(src,dst,size) safememcpy(dst,src,(long)size)
#endif

#ifndef YYERROR_VERBOSE
#define YYERROR_VERBOSE
#endif

EXTERN SymPtr equalSym;
EXTERN Boolean wantCloseFile,syntax_analysis;
EXTERN ExprPtr theSyntax;
EXTERN SymPtr ItSymbol;

EXTERN SymPtr trueSym;

#define yyerror(msg) yy_error(yylloc.first_line,yylloc.first_column,yylloc.last_line,yylloc.last_column,msg)

PRIVATE ExprPtr cons(a,b)
IN CONST ExprPtr a,b;
{
  ExprPtr aux;

  MAKECONS(a,b,aux);
  return KEEP(aux);
}

PRIVATE ExprPtr list1(a)
IN CONST ExprPtr a;
{
  ExprPtr aux;

  MAKECONS(a,NIL,aux);
  return KEEP(aux);
}

PRIVATE ExprPtr getInt(text, loc)
IN CONST CHAR *text;
IN CONST struct yyltype *loc;
{
  Boolean neg;
  long i;

  errno = 0;
  if ((neg = (text[0]=='~')))
    text++;
  if (text[0]=='0')
    switch (text[1])
      {
      case 'x': case 'X':
	i = (long)strtoul ((char *)text+2, NULL, 16);
	break;
      case 'o': case 'O':
	i = (long)strtoul ((char *)text+2, NULL, 8);
	break;
      case 'b': case 'B':
	i = (long)strtoul ((char *)text+2, NULL, 2);
	break;
      default:
	i = strtol ((char *)text+1, NULL, 10);
	break;
      }
  else i = strtol ((char *)text, NULL, 10);
  if (errno)
    {
      TZone zone;
      makeZone(&zone,loc,loc);
      type_warn(&zone,(CHAR *)message(75));
    }
  if (neg)
    i = -i;
  return KINT(i);
}

PRIVATE void checkLabel(loc)
IN CONST struct yyltype *loc;
{
  if (lexPath!=NIL)
    {
      TZone zone;
      makeZone(&zone,loc,loc);
      type_error(&zone,(CHAR *)message(386));
    }
}

PRIVATE COUNTER quantifier[4][2] = {
  { A_PROG_OF, A_PROG_MAP },
  { A_ALL_OF, A_ALL_MAP },
  { A_EXISTS_OF, A_EXISTS_MAP },
  { A_SOME_OF, A_SOME_MAP }
};

PRIVATE COUNTER qwhile[4] = {
  A_WHILE,
  A_WHILE_ALL,
  A_WHILE_EXISTS,
  A_WHILE_SOME
};

#if FLOAT_FORMAT==IEEE754
#define makeSpecNum(zone,x) makeNum(zone,x)
#else
#define makeSpecNum(zone,x) NIL
#endif

#ifdef OBSOLETE
PRIVATE void checkDebug(loc1,loc2)
IN CONST struct yyltype *loc1,*loc2;
{
  if (debugMode==DBM_IN_DEBUGGER)
    {
      TZone zone;
      makeZone(&zone,loc1,loc1);
      type_error(&zone,(CHAR *)message(387));
    }
}
#endif

#define checkDebug(loc1,loc2)

PRIVATE ExprPtr do_load (loc1, loc2, s, recompile)
IN CONST struct yyltype *loc1,*loc2;
IN CONST char *s;
IN Boolean recompile;
{
  ExprPtr res;

  (void) getString(loc2,loc2,(CHAR *)s,false);
  {
    int len = strlen(s);
    TZone *zone = (TZone *)alloctmp_anchored(sizeof(TZone),(char *)&len);
    char *text = (char *)alloctmp_anchored(len+1,(char *)&len);

    strcpy(text,s);
    makeZone(zone,loc1,loc2);
    res = moduleLoad(zone,(CHAR *)text,recompile);
    if (res==NIL)
      {
	char *buf = alloctmp_anchored(len+64,(char *)&len);

	sprintf(buf,message(recompile?502:921),text);
	type_error(zone,(CHAR *)buf);
      }
    checkConsEnv(zone,res);
    augmentLexicalEnv(res);
    res = list1(build1(loc1,loc2,A_BASIS,res));
  }
  return res;
}

#if DEBUGGING
PUBLIC Boolean evaluatingp=false,stopwanted=false;
#endif

%}

%union {
  int i;
  ExprPtr e;
}

%token kw_abstype kw_all kw_and kw_andalso kw_as kw_case kw_do kw_datatype
%token kw_else kw_end kw_exception kw_exists kw_fn kw_fun kw_handle kw_if
%token kw_in kw_in_list kw_in_map kw_in_set kw_infix kw_infixr
%token kw_let kw_local kw_nonfix kw_of kw_orelse
%token kw_dimension kw_raise kw_rec kw_some kw_such_that kw_then kw_type
%token kw_scale kw_val kw_with kw_withtype kw_while kw_open_paren kw_close_paren
%token kw_open_bracket kw_close_bracket kw_open_array_ref kw_array_assign
%token kw_open_brace_under kw_open_brace kw_close_brace
%token kw_open_record kw_close_record kw_open_array kw_close_array
%token kw_comma kw_colon
%token kw_semicolon kw_ellipsis kw_underscore
%token kw_vbar kw_vbar_vbar kw_equal kw_big_arrow kw_arrow kw_star
%token precedence_constant string_constant
%token integer_constant real_constant complex_constant
%token hex_integer_constant octal_integer_constant binary_integer_constant
%token kw_identifier kw_cons_id
%token kw_type_variable
%token kw_left_0 kw_right_0
%token kw_left_1 kw_right_1
%token kw_left_2 kw_right_2
%token kw_left_3 kw_right_3
%token kw_left_4 kw_right_4
%token kw_left_5 kw_right_5
%token kw_left_6 kw_right_6
%token kw_left_7 kw_right_7
%token kw_left_8 kw_right_8
%token kw_left_9 kw_right_9
%token kw_left_0_constructor kw_right_0_constructor
%token kw_left_1_constructor kw_right_1_constructor
%token kw_left_2_constructor kw_right_2_constructor
%token kw_left_3_constructor kw_right_3_constructor
%token kw_left_4_constructor kw_right_4_constructor
%token kw_left_5_constructor kw_right_5_constructor
%token kw_left_6_constructor kw_right_6_constructor
%token kw_left_7_constructor kw_right_7_constructor
%token kw_left_8_constructor kw_right_8_constructor
%token kw_left_9_constructor kw_right_9_constructor
%token kw_scale_star kw_scale_power kw_union kw_map_arrow
%token kw_iterate kw_eqtype kw_sharp kw_open_mu kw_use
%token kw_pinf kw_minf
%token kw_nan0 kw_nan1 kw_nan2 kw_nan3 kw_nan4 kw_nan5 kw_nan6 kw_nan7
%token kw_mnan1 kw_mnan2 kw_mnan3 kw_mnan4 kw_mnan5 kw_mnan6 kw_mnan7
%token kw_sub_map kw_delay kw_pack /* kw_extend suppressed */
%token kw_memo_fn kw_memo_fun kw_extern kw_pragma

%token kw_sig kw_signature kw_structure kw_sharing kw_open kw_openx kw_include kw_functor
%token kw_abstraction kw_struct kw_compile

%right kw_big_arrow kw_arrow kw_map_arrow
%left kw_fn
%nonassoc kw_case kw_of
%nonassoc kw_while kw_do
%nonassoc kw_if kw_then kw_else
%left kw_raise
%left kw_handle
%right kw_orelse
%right kw_andalso
%left kw_as

%type <e> real integer number number_or_int
%type <e> scale_id simple_scale scale optional_scale
%type <e> numerical_constant special_constant
%type <e> identifier long_constructor
%type <e> long_id_or_constructor /*long_symbol*/
%type <e> /*constructor id_or_constructor*/ symbol
%type <e> label labeled_expression expression_row optional_expression_row
%type <e> expression_list optional_expression_list
%type <e> expression_sequence implicit_progn
%type <e> domain_sweep domain_sweep_list
%type <e> such_that optional_such_that
%type <e> domain_description
%type <e> pure_maplet maplet maplet_list
%type <i> open_brace optional_precedence
%type <e> atomic_expression application_expression delay_expression
%type <e> array_expression
%type <e> left_0_expression right_0_expression
%type <e> left_1_expression right_1_expression
%type <e> left_2_expression right_2_expression
%type <e> left_3_expression right_3_expression
%type <e> left_4_expression right_4_expression
%type <e> left_5_expression right_5_expression
%type <e> left_6_expression right_6_expression
%type <e> left_7_expression right_7_expression
%type <e> left_8_expression right_8_expression
%type <e> left_9_expression right_9_expression
%type <e> long_left_0 long_right_0
%type <e> long_left_1 long_right_1
%type <e> long_left_2 long_right_2
%type <e> long_left_3 long_right_3
%type <e> long_left_4 long_right_4
%type <e> long_left_5 long_right_5
%type <e> long_left_6 long_right_6
%type <e> long_left_7 long_right_7
%type <e> long_left_8 long_right_8
%type <e> long_left_9 long_right_9
%type <e> infix_expression
%type <e> colon_expression
%type <e> andalso_expression orelse_expression
%type <e> match_expression raise_expression
%type <e> closed_match match match_rule
%type <e> expression semicolon
%type <e> optional_declaration_list declaration_list closed_declaration_list
%type <e> declaration load_declaration simple_declaration fixity_declaration
%type <e> nosemi_declaration extern_declaration
%type <e> value_binding nonrec_value_binding simple_value_binding
%type <e> function_value_binding function_value_binding_l
%type <e> closed_function_value_binding optional_with_fun
%type <e> simple_function_value_binding one_function_value_binding
%type <e> atomic_pattern_list infix_function_header
%type <e> dimension_binding dimension_binding_l simple_dimension_binding
%type <e> scale_binding scale_binding_l simple_scale_binding
%type <e> type_binding type_binding_l simple_type_binding
%type <e> optional_withtype withtype
%type <e> type_variable_sequence type_variable_sequence_l type_variable
%type <e> datatype_binding simple_datatype_binding
%type <e> constructor_binding simple_constructor_binding symbol_list
%type <e> optional_of_type of_type
%type <e> exception_binding exception_binding_l simple_exception_binding
%type <e> optional_type_row type_row labeled_type
%type <e> atomic_type constructor_type
%type <e> simple_dimension_type dimension_type
%type <e> tuple_type tuple_type_l function_type
%type <e> type type_list
%type <e> optional_colon_type colon_type
%type <e> optional_pattern_row pattern_row labeled_pattern
%type <e> optional_as_pattern as_pattern
%type <e> /*optional_pattern_list*/ pattern_list
%type <e> optional_pattern_l_list pattern_l_list
%type <e> optional_union_pattern union_pattern
%type <e> maplet_pattern maplet_pattern_list
%type <e> atomic_pattern application_pattern pack_pattern
%type <e> left_0_pattern right_0_pattern
%type <e> left_1_pattern right_1_pattern
%type <e> left_2_pattern right_2_pattern
%type <e> left_3_pattern right_3_pattern
%type <e> left_4_pattern right_4_pattern
%type <e> left_5_pattern right_5_pattern
%type <e> left_6_pattern right_6_pattern
%type <e> left_7_pattern right_7_pattern
%type <e> left_8_pattern right_8_pattern
%type <e> left_9_pattern right_9_pattern
%type <e> long_left_0_constructor long_right_0_constructor
%type <e> long_left_1_constructor long_right_1_constructor
%type <e> long_left_2_constructor long_right_2_constructor
%type <e> long_left_3_constructor long_right_3_constructor
%type <e> long_left_4_constructor long_right_4_constructor
%type <e> long_left_5_constructor long_right_5_constructor
%type <e> long_left_6_constructor long_right_6_constructor
%type <e> long_left_7_constructor long_right_7_constructor
%type <e> long_left_8_constructor long_right_8_constructor
%type <e> long_left_9_constructor long_right_9_constructor
%type <e> infix_pattern colon_pattern pattern
%type <e> optional_ellipsis_type ellipsis_type
%type <e> infix_identifier optional_eqtype eqtype_desc simple_eqtype_desc
%type <i> quantifier bar fn
%type <e> main toplevel_declaration

%type <e> value_description value_description_l
%type <e> simple_value_description type_description type_description_l
%type <e> simple_type_description
/*%type <e> exception_description exception_description_l simple_exception_description*/

%start main

%%

main : toplevel_declaration kw_semicolon
	{
	  theSyntax = buildrev(&@1,&@2,A_DECL_LIST,$1);
	  YYACCEPT;
	}
	;

	/* a toplevel_declaration is a reversed list of bindings or NIL */
toplevel_declaration : expression /* val it = expr; only form of syntax allowed
				     in debug mode. */
		{ $$ = exprToDecls(&@1,&@1,$1,ItSymbol);
		}
	| toplevel_declaration nosemi_declaration
		{ checkDebug(&@1,&@2);
		    $$ = Nappend($2,$1); }
/*	| toplevel_declaration kw_signature signature_binding
		{ checkDebug(&@1,&@3); $$ = cons($3,$1); }
*/
	|
	   { $$ = NIL; }
	;

load_declaration : kw_open string_constant
		{
		  $$ = do_load (&@1, &@2, yytext, true);
		}
		| kw_openx string_constant
		{
		  $$ = do_load (&@1, &@2, yytext, false);
		}
		| kw_compile string_constant
		{
		  (void) getString (&@2, &@2, (CHAR *)yytext, false);
		  {
		    int len = strlen (yytext);
		    char *text = (char *)alloctmp_anchored(len+1,(char *)&len);
		    ExprPtr mod;

		    strcpy(text,yytext);
		    mod = moduleCompile((CHAR *)text);
		    if (mod==NIL)
		      {
			TZone *zone = (TZone *)alloctmp_anchored(sizeof(TZone),(char *)&len);
			char *buf = alloctmp_anchored(len+64,(char *)&len);

			makeZone(zone,&@1,&@2);
			sprintf(buf,message(502),text);
			type_error(zone,(CHAR *)buf);
		      }
		    $$ = NIL;
		  }
		}
		| kw_use string_constant
		{
			ExprPtr name;

			(void) getString(&@2,&@2,(CHAR *)yytext,false);
			if (!OpenFile((CHAR *)yytext))
			  {
			    char *buf = alloctmp_anchored(strlen(yytext)+64,(char *)&name);

			    sprintf(buf,message(389),yytext);
			    {
			      TZone *zone = (TZone *)alloctmp_anchored(sizeof(TZone),(char *)&name);

			      makeZone(zone,&@1,&@2);
			      type_error(zone,(CHAR *)buf);
			    }
			  }
			name = Kstr(yyfilename);
			(void) KEEP(name);
			$$ = list1(build1(&@1,&@2,A_ENTER_FILE,name));
		}
		;

sharp : kw_sharp { pushMarker(); }
	;

open_record : kw_open_record { pushMarker(); }
	;

open_paren : kw_open_paren { pushMarker(); }
	;

open_bracket : kw_open_bracket { pushMarker(); }
	;

open_brace : kw_open_brace { $$ = 0; pushMarker(); }
	| kw_open_brace_under { $$ = 1; pushMarker(); }
	;

quantifier : kw_iterate { pushMarker(); $$ = 0; }
	| kw_all { pushMarker(); $$ = 1; }
	| kw_exists { pushMarker(); $$ = 2; }
	| kw_some { pushMarker(); $$ = 3; }
	;

bar : kw_vbar { $$ = 0; }
	| kw_vbar_vbar { $$ = 1; }
	;

atomic_expression : special_constant
	| long_id_or_constructor
	| open_record optional_expression_row kw_close_record
	{ $$ = popMarker(BUILD(&@1,&@3,A_RECORD,$2)); }
	| sharp label
	{ $$ = popMarker(BUILD(&@1,&@2,A_SELECT_FN,ARGS($2))); }
	| sharp integer_constant
	{ $$ = popMarker(build1(&@1,&@2,A_SELECT_FN,getInt((CHAR *)yytext, &@2))); }
	| sharp precedence_constant
	{ $$ = popMarker(build1(&@1,&@2,A_SELECT_FN,getInt((CHAR *)yytext, &@2))); }
	| open_paren optional_expression_list kw_close_paren
		{ if ($2!=NIL && CDR($2)==NIL)
			{ $$ = CAR($2); }
			else
				$$ = BUILD(&@1,&@3,A_TUPLE,$2);
			$$ = popMarker($$); }
	| open_paren expression kw_semicolon
		expression_sequence kw_close_paren
		{ $$ = popMarker(build1rev(&@1,&@5,A_PROGN,$2,$4)); }
	| open_bracket kw_close_bracket
		{ $$ = popMarker(BUILD(&@1,&@2,A_LIST,NIL)); }
	| open_bracket expression kw_close_bracket
		{ $$ = popMarker(build1(&@1,&@3,A_LIST,$2)); }
	| open_bracket expression kw_comma
		expression_list kw_close_bracket
		{ $$ = popMarker(build1rev(&@1,&@5,A_LIST,$2,$4)); }
	| open_bracket expression bar
		domain_description kw_close_bracket
		{ $$ = popMarker(build1list(&@1,&@5,
			A_LIST_OF+$3,$2,$4)); }
	| open_bracket expression bar /* should be kw_vbar really */
		kw_while expression optional_such_that kw_close_bracket
		{ $$ = popMarker(build3(&@1,&@7,A_WHILE_LIST,$5,$2,$6)); }
	| kw_open_array { pushMarker(); } optional_expression_list kw_close_array
		{ $$ = Kstr((CHAR *)message(357)); (void) KEEP($$);
		  $$ = popMarker(build2(&@1,&@4,A_APPLY,
					build1(&@1,&@4,A_VARIABLE,$$),
					BUILD(&@1,&@4,A_LIST,$3)));
		}
	| open_brace kw_close_brace
		{ $$ = popMarker(BUILD(&@1,&@2,A_MAP,NIL)); }
	| open_brace maplet kw_close_brace
		{ $$ = popMarker(build1(&@1,&@3,A_MAP,$2)); }
	| open_brace maplet kw_comma
		maplet_list kw_close_brace
		{ $$ = popMarker(build1rev(&@1,&@5,$1?A_MAP_UNDER:A_MAP,$2,$4)); }
	| open_brace maplet bar
		domain_description kw_close_brace
		{ $$ = popMarker(build1list(&@1,&@5,
			($1?A_MAP_UNDER_OF:A_MAP_OF)+$3,
			$2,$4)); }
	| open_brace maplet bar /* should be kw_vbar really */
		kw_while expression optional_such_that kw_close_brace
		{ $$ = popMarker(build3(&@1,&@7,
			($1?A_WHILE_MAP_UNDER:A_WHILE_MAP),$5,$2,$6)); }
	| quantifier expression bar domain_description kw_end
		{ $$ = popMarker(build1list(&@1,&@5,quantifier[$1][$3],$2,$4)); }
	| quantifier expression bar /* should be kw_vbar really */
		kw_while expression optional_such_that kw_end
		{ $$ = popMarker(build3(&@1,&@7,qwhile[$1],$5,$2,$6)); }
	| let_begin
		closed_declaration_list kw_in
		implicit_progn let_end
		{ $$ = popMarker(build2(&@1,&@5,A_LET,$2,$4)); }
	;

	/* maplet_lists are reversed lists of maplets */
maplet_list : maplet
		{ $$ = list1($1); }
	| maplet_list kw_comma maplet
		{ $$ = cons($3,$1); }
	;

maplet : expression
	{ $$ = build2(&@1,&@1,A_MAPLET,$1,zeroTuple(&@1)); }
	| pure_maplet
	;

pure_maplet : expression kw_big_arrow expression
	{ $$ = build2(&@1,&@3,A_MAPLET,$1,$3); }
	;

	/* domain_descriptions are couples (such_that,list) */
domain_description : { pushMarker(); }
	domain_sweep_list optional_declaration_list optional_such_that
	{ $$ = popMarker(cons($4,Nreverse1($2,ARGS($3)))); }
	;

	/* domain_sweep_lists are reversed lists */
domain_sweep_list : domain_sweep
	{ $$ = list1($1); }
	| domain_sweep_list kw_and domain_sweep
	{ $$ = cons($3,$1); }
	;

domain_sweep : pattern kw_in_set expression
	{ $$ = build2(&@1,&@3,A_IN_SET,
			build2(&@1,&@1,A_PAT_MAPLET,$1,
				BUILD(&@1,&@1,A_PAT_ANY,NIL)),
			$3); }
	| maplet_pattern kw_in_map expression
	{ $$ = build2(&@1,&@3,A_IN_SET,$1,$3); }
	| pattern kw_in_list expression
	{ $$ = build2(&@1,&@3,A_IN_LIST,$1,$3); }
	| pattern kw_sub_map expression
	{ $$ = build2(&@1,&@3,A_SUB_MAP,$1,$3); }
	;

optional_such_that : such_that
	| empty
	{ $$ = build1(&@1,&@1,A_VARIABLE,trueSym); }
	;

empty :
	;

such_that : kw_such_that expression { $$ = $2; }
	;

let_begin : kw_let { pushEnv(); pushMarker(); }
	;

let_end : kw_end { popEnv(); }
	;

implicit_progn : expression_sequence
	{ if (CDR($1)==NIL)
	$$ = CAR($1);
	else
	$$ = buildrev(&@1,&@1,A_PROGN,$1); }
	;

	/* expression sequences are reversed lists of expressions */
expression_sequence : expression
	{ $$ = list1($1); }
	| expression_sequence kw_semicolon expression
	{ $$ = cons($3,$1); }
	;

	/* expression lists are lists of expressions */
optional_expression_list : expression_list
	{ $$ = Nreverse($1); }
	| { $$ = NIL; }
	;

	/* expression lists are reversed lists of expressions */
expression_list : expression
	{ $$ = list1($1); }
	| expression_list kw_comma expression
	{ $$ = cons($3,$1); }
	;

	/* optional_expression_rows are lists of record maplets */
optional_expression_row : expression_row
	{ $$ = Nreverse($1); }
	| { $$ = NIL; }
	;

	/* expression_rows are reversed lists of record maplets */
expression_row : labeled_expression
	{ $$ = list1($1); }
	| expression_row kw_comma labeled_expression
	{ checkUniqueLabel($3,$1);
	$$ = cons($3,$1); }
	;

labeled_expression : label kw_equal expression
	{ $$ = build2(&@1,&@3,A_RECORD_MAPLET,$1,$3); }
	;

label : label_id
	{ checkLabel(&@1);
		$$ = build1(&@1,&@1,A_LABEL,lexId); }
	;

symbol : label_id
	{ $$ = lexVar(&@1,false); }
	;

label_id : kw_identifier
	| kw_cons_id
	| infix_identifier_1
	| kw_left_0_constructor
	| kw_right_0_constructor
	| kw_left_1_constructor
	| kw_right_1_constructor
	| kw_left_2_constructor
	| kw_right_2_constructor
	| kw_left_3_constructor
	| kw_right_3_constructor
	| kw_left_4_constructor
	| kw_right_4_constructor
	| kw_left_5_constructor
	| kw_right_5_constructor
	| kw_left_6_constructor
	| kw_right_6_constructor
	| kw_left_7_constructor
	| kw_right_7_constructor
	| kw_left_8_constructor
	| kw_right_8_constructor
	| kw_left_9_constructor
	| kw_right_9_constructor
	;

application_expression : atomic_expression
	| application_expression atomic_expression
	{ $$ = build2(&@1,&@2,A_APPLY,$1,$2); }
	| application_expression kw_open_mu { pushMarker(); }
		optional_expression_row kw_close_record
		{ $$ = popMarker(build2(&@1,&@5,A_MU,
				 $1,BUILD(&@2,&@5,A_RECORD,$4))); }
	| array_expression
	;

array_expression : application_expression kw_open_array_ref { pushMarker(); }
			expression kw_close_paren
		{ $$ = Kstr((CHAR *)message(354)); (void) KEEP($$);
		  $$ = popMarker(applyBin(&@1,&@5,
					build1(&@2,&@2,A_VARIABLE,$$),
					$1,$4));
		}
	;

delay_expression : application_expression
	| kw_delay { pushMarker(); }
	delay_expression
	{ $$ = popMarker(build1(&@1,&@3,A_DELAY,$3)); }
	| kw_pack { pushMarker(); }
	delay_expression
	{ $$ = popMarker(build1(&@1,&@3,A_PACK,$3)); }
	;

left_9_expression : delay_expression
	| left_9_expression long_left_9 delay_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_9 : kw_left_9
		{ $$ = lexVar(&@1,true); }
		| long_left_9_constructor
		;

right_9_expression : left_9_expression
	| left_9_expression long_right_9 right_9_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_9 : kw_right_9
		{ $$ = lexVar(&@1,true); }
		| long_right_9_constructor
		;

left_8_expression : right_9_expression
	| left_8_expression long_left_8 right_9_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_8 : kw_left_8
		{ $$ = lexVar(&@1,true); }
		| long_left_8_constructor
		;

right_8_expression : left_8_expression
	| left_8_expression long_right_8 right_8_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_8 : kw_right_8
		{ $$ = lexVar(&@1,true); }
		| long_right_8_constructor
		;

left_7_expression : right_8_expression
	| left_7_expression long_left_7 right_8_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_7 : kw_left_7
		{ $$ = lexVar(&@1,true); }
		| long_left_7_constructor
		;

right_7_expression : left_7_expression
	| left_7_expression long_right_7 right_7_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_7 : kw_right_7
		{ $$ = lexVar(&@1,true); }
		| long_right_7_constructor
		;

left_6_expression : right_7_expression
	| left_6_expression long_left_6 right_7_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_6 : kw_left_6
		{ $$ = lexVar(&@1,true); }
		| long_left_6_constructor
		;

right_6_expression : left_6_expression
	| left_6_expression long_right_6 right_6_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_6 : kw_right_6
		{ $$ = lexVar(&@1,true); }
		| long_right_6_constructor
		;

left_5_expression : right_6_expression
	| left_5_expression long_left_5 right_6_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_5 : kw_left_5
		{ $$ = lexVar(&@1,true); }
		| long_left_5_constructor
		;

right_5_expression : left_5_expression
	| left_5_expression long_right_5 right_5_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_5 : kw_right_5
		{ $$ = lexVar(&@1,true); }
		| long_right_5_constructor
		;

left_4_expression : right_5_expression
	| left_4_expression long_left_4 right_5_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_4 : kw_left_4
		{ $$ = lexVar(&@1,true); }
	| kw_equal
		{ SETQ(lexId,equalSym);
		  SETQ(lexPath,NIL);
		  $$ = lexVar(&@1,true); }
	| long_left_4_constructor
	;

right_4_expression : left_4_expression
	| left_4_expression long_right_4 right_4_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_4 : kw_right_4
		{ $$ = lexVar(&@1,true); }
		| long_right_4_constructor
		;

left_3_expression : right_4_expression
	| left_3_expression long_left_3 right_4_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	| array_expression kw_array_assign right_4_expression
	{ $$ = Kstr((CHAR *)message(355)); (void) KEEP($$);
	  $$ = build2(&@1,&@3,A_APPLY,
		      build1(&@1,&@3,A_VARIABLE,$$),
		      build3(&@1,&@3,A_TUPLE,
				CAR(ARGS(CAR(CDR(ARGS($1))))),
				CAR(CDR(ARGS(CAR(CDR(ARGS($1)))))),
				$3));
	}
	;

long_left_3 : kw_left_3
		{ $$ = lexVar(&@1,true); }
		| long_left_3_constructor
		;

right_3_expression : left_3_expression
	| left_3_expression long_right_3 right_3_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_3 : kw_right_3
		{ $$ = lexVar(&@1,true); }
		| long_right_3_constructor
		;

left_2_expression : right_3_expression
	| left_2_expression long_left_2 right_3_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_2 : kw_left_2
		{ $$ = lexVar(&@1,true); }
		| long_left_2_constructor
		;

right_2_expression : left_2_expression
	| left_2_expression long_right_2 right_2_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_2 : kw_right_2
		{ $$ = lexVar(&@1,true); }
		| long_right_2_constructor
		;

left_1_expression : right_2_expression
	| left_1_expression long_left_1 right_2_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_1 : kw_left_1
		{ $$ = lexVar(&@1,true); }
		| long_left_1_constructor
		;

right_1_expression : left_1_expression
	| left_1_expression long_right_1 right_1_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_1 : kw_right_1
		{ $$ = lexVar(&@1,true); }
		| long_right_1_constructor
		;

left_0_expression : right_1_expression
	| left_0_expression long_left_0 right_1_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_0 : kw_left_0
		{ $$ = lexVar(&@1,true); }
		| long_left_0_constructor
		;

right_0_expression : left_0_expression
	| left_0_expression long_right_0 right_0_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_0 : kw_right_0
		{ $$ = lexVar(&@1,true); }
		| long_right_0_constructor
		;

infix_expression : right_0_expression
	;

colon_expression : infix_expression
	| colon_expression colon_type
	{ $$ = build2(&@1,&@2,A_CAST,$1,$2); }
	;

andalso_expression : colon_expression
	| colon_expression kw_andalso andalso_expression
	{ $$ = build2(&@1,&@3,A_AND,$1,$3); }
	;

orelse_expression : andalso_expression
	| andalso_expression kw_orelse orelse_expression
	{ $$ = build2(&@1,&@3,A_OR,$1,$3); }
	;

match_expression : orelse_expression
	| orelse_expression
	kw_handle { pushMarker(); }
	closed_match /* shift-reduce on vbar */
	{ TYPE($4) = A_MATCH;
	$$ = popMarker(build2(&@1,&@4,A_HANDLE,$1,$4)); }
	;

raise_expression : match_expression
	| kw_raise raise_expression
	{ $$ = build1(&@1,&@2,A_RAISE,$2); }
	;

while : kw_while { pushMarker(); }
	;

expression : raise_expression
	| kw_if {pushMarker();} expression
	kw_then expression
	kw_else expression
	{ $$ = popMarker(build3(&@1,&@7,A_IF,$3,$5,$7)); }
	| while expression
	kw_do expression
	{ $$ = popMarker(build3(&@1,&@4,A_WHILE,$2,$4,
				build1(&@1,&@4,A_VARIABLE,trueSym))); }
	| kw_case {pushMarker();} expression
	kw_of closed_match /* shift-reduce on vbar */
	{ $$ = popMarker(build2(&@1,&@5,A_APPLY,$5,$3)); }
	| fn {pushMarker();} match /* shift-reduce on vbar */
	{ $$ = popMarker(buildrev(&@1,&@3,$1,$3)); }
	;

fn : kw_fn { $$ = A_FN; }
	| kw_memo_fn { $$ = A_MEMO_FN; }
	;

closed_match : match
	{ $$ = buildrev(&@1,&@1,A_FN,$1); }
	;

	/* matches are reversed lists of match_rules */
match : match_rule
	{ $$ = list1($1); }
	| match kw_vbar match_rule /* shift-reduce on vbar */
	{ $$ = cons($3,$1); }
	;

match_rule : { pushMarker(); }
	pattern kw_big_arrow expression
	{ $$ = popMarker(build2(&@1,&@4,A_LAMBDA,$2,$4)); }
	;

special_constant : numerical_constant
	| string_constant
	{ $$ = build1(&@1,&@1,A_STRING,
		      getString(&@1,&@1,(CHAR *)yytext,true)); }
	;

numerical_constant : number {unit_recognized=true;}
			optional_scale
		{ unit_recognized = false;
		  $$ = $1;
		  if ($3!=NIL)
		    $$ = build2(&@1,&@3,A_SET_UNIT,$$,$3);
		}
	| integer {unit_recognized=true;}
			optional_scale
		{ unit_recognized = false;
		  $$ = $1;
		  if ($3!=NIL)
		    {
		      TYPE($$) = A_NUMBER;
		      $$ = build2(&@1,&@3,A_SET_UNIT,$$,$3);
		    }
		}
	;

optional_scale : kw_scale_star scale { $$ = $2; }
	| { $$ = NIL; }
	;

scale : simple_scale
	| scale kw_scale_star simple_scale
	{ $$ = build2(&@1,&@3,A_UNIT_STAR,$1,$3); }
	;

simple_scale : scale_id
	| scale_id kw_scale_power number_or_int
	{ $$ = build2(&@1,&@3,A_UNIT_POWER,$1,$3); }
	;

scale_id : label_id
	{ $$ = lexVar(&@1,true);
	  TYPE($$) = A_UNIT_NAME; }
	;

number : real
	| complex_constant
	{ $$ = makeNum(&@1,getComplex(&@1,&@1,(CHAR *)yytext)); }
	;

number_or_int : integer { TYPE($$) = A_NUMBER; }
     	      | number
	      ;

real : real_constant
	{ $$ = makeNum(&@1,getReal(&@1,&@1,(CHAR *)yytext)); }
	| kw_pinf
	{ $$ = makeSpecNum(&@1,pinf); }
	| kw_minf
	{ $$ = makeSpecNum(&@1,minf); }
	| kw_nan0
	{ $$ = makeSpecNum(&@1,NaNs[0+MAX_NAN]); }
	| kw_nan1
	{ $$ = makeSpecNum(&@1,NaNs[1+MAX_NAN]); }
	| kw_nan2
	{ $$ = makeSpecNum(&@1,NaNs[2+MAX_NAN]); }
	| kw_nan3
	{ $$ = makeSpecNum(&@1,NaNs[3+MAX_NAN]); }
	| kw_nan4
	{ $$ = makeSpecNum(&@1,NaNs[4+MAX_NAN]); }
	| kw_nan5
	{ $$ = makeSpecNum(&@1,NaNs[5+MAX_NAN]); }
	| kw_nan6
	{ $$ = makeSpecNum(&@1,NaNs[6+MAX_NAN]); }
	| kw_nan7
	{ $$ = makeSpecNum(&@1,NaNs[7+MAX_NAN]); }
	| kw_mnan1
	{ $$ = makeSpecNum(&@1,NaNs[-1+MAX_NAN]); }
	| kw_mnan2
	{ $$ = makeSpecNum(&@1,NaNs[-2+MAX_NAN]); }
	| kw_mnan3
	{ $$ = makeSpecNum(&@1,NaNs[-3+MAX_NAN]); }
	| kw_mnan4
	{ $$ = makeSpecNum(&@1,NaNs[-4+MAX_NAN]); }
	| kw_mnan5
	{ $$ = makeSpecNum(&@1,NaNs[-5+MAX_NAN]); }
	| kw_mnan6
	{ $$ = makeSpecNum(&@1,NaNs[-6+MAX_NAN]); }
	| kw_mnan7
	{ $$ = makeSpecNum(&@1,NaNs[-7+MAX_NAN]); }
	;

integer : integer_constant
	{ $$ = makeInt(&@1,getReal(&@1,&@1,(CHAR *)yytext)); }
	| hex_integer_constant
	{ $$ = makeInt(&@1, getInt((CHAR *)yytext, &@1)); }
	| octal_integer_constant
	{ $$ = makeInt(&@1, getInt((CHAR *)yytext, &@1)); }
	| binary_integer_constant
	{ $$ = makeInt(&@1, getInt((CHAR *)yytext, &@1)); }
	| precedence_constant
	{ $$ = makeInt(&@1,getReal(&@1,&@1,(CHAR *)yytext)); }
	;

identifier : kw_identifier
		       	{ $$ = lexVar(&@1,false); }
			;

long_constructor : kw_cons_id
						{ $$ = lexVar(&@1,true); }
					;

long_id_or_constructor : kw_identifier
							{ $$ = lexVar(&@1,true); }
						| kw_cons_id
							{ $$ = lexVar(&@1,true); }
						;

/*
long_symbol : label_id
				{ $$ = lexVar(&@1,true); }
			;

constructor : kw_cons_id
				{ $$ = lexVar(&@1,false); }
	;
*/

optional_declaration_list : closed_declaration_list
	| empty
	{ $$ = BUILD(&@1,&@1,A_DECL_LIST,NIL); }
	;

closed_declaration_list : declaration_list
	{ $$ = buildrev(&@1,&@1,A_DECL_LIST,$1); }
	;

	/* a declaration_list is a reversed list of bindings */
declaration_list : declaration
	| declaration_list declaration
	{ $$ = Nappend($2,$1); }
	;

semicolon : kw_semicolon {
		if (wantCloseFile)
		{
		  wantCloseFile = false;
		  CloseFile();
		  if (yyfilename==NULL)
		    $$ = Build(&@1,&@1,A_EXIT_FILE,NIL);
		  else
		    {
		      ExprPtr name;
		      
		      name = Kstr(yyfilename);
		      (void) KEEP(name);
		      $$ = build1(&@1,&@1,A_ENTER_FILE,name);
		    }
		  $$ = list1($$);
		}
		else $$ = NIL;
		}
;

	/* a declaration is a reversed list of bindings */
declaration : nosemi_declaration
	| semicolon
	;

	/* a nosemi_declaration is a reversed list of bindings */
nosemi_declaration : { pushMarker(); }
	simple_declaration
	{ $$ = popMarker($2); }
	| local_begin { pushMarker(); }
	optional_declaration_list kw_in
	{ pushEnv(); }
	optional_declaration_list
	{ setLocalEnv(); popEnv(); }
	local_end
	{ $$ = popMarker(list1(build2(&@1,&@8,A_LOCAL,$3,$6)));
	}
	| kw_extern extern_declaration { checkDebug(&@1,&@2); $$ = $2; }
	;

local_begin : kw_local { pushEnv(); }
	;

local_end : kw_end { popEnv(); }
	;

	/* a simple_declaration is a reversed list of bindings */
simple_declaration : kw_val value_binding
	{ $$ = list1($2); }
	| kw_fun function_value_binding
	{ $$ = list1($2); }
	| kw_memo_fun function_value_binding
		optional_with_fun
	{ $$ = list1($2); patchMemoFunList(ARGS($2),CAR($3),CDR($3)); }
	| kw_dimension dimension_binding
	{ $$ = list1($2); }
	| kw_scale scale_binding
	{ $$ = list1($2); }
	| kw_type type_binding
	{ $$ = list1($2); }
	| kw_datatype datatype_binding optional_withtype
	{ ExprPtr dtb,prime;
	ENTER;
	dtb = Nreverse($2);
	prime = build1list(&@2,&@3,A_PRIME_DATABIND,$3,dtb);
	checkWithType(dtb,$3);
	$$ = list1(build1(&@1,&@3,A_DATATYPE,prime));
	/* $3 is shared ! */
	if (ARGS($3)!=NIL)
	$$ = cons($3,$$);
	KEEPONLY($$); }
	| kw_abstype {pushEnv(); }
	datatype_binding {pushEnv();}
	optional_withtype
	kw_with
	optional_eqtype
	declaration_list
	kw_end
	{ ExprPtr dtb,prime,declist;
	ENTER;
	setLocalEnv(); popEnv(); popEnv();
	checkEqTypes($7,$3);
	dtb = Nreverse($3);
	prime = build1list(&@3,&@5,A_PRIME_DATABIND,$5,dtb);
	checkWithType(dtb,$5);
	declist = Nreverse($8);
	declist = (ARGS($5)!=NIL)?cons($5,declist):declist;
	/* $5 is shared ! */
	declist = BUILD(&@5,&@8,A_DECL_LIST,declist);
	$$ = list1(build2list(&@1,&@9,A_ABSTYPE,prime,declist,$7));
	KEEPONLY($$); }
	| kw_exception exception_binding
	{ $$ = list1($2); }
	| fixity_declaration
/*	| kw_open long_symbol_list
		{ $$ = buildrev(&@1,&@2,A_OPEN_DECL,$2);
			longLexOpenList(ARGS($$),false);
			$$ = list1($$); }
*/
	| load_declaration
	| kw_pragma
	{
	$$ = Kstr((CHAR *)yytext);
	(void) KEEP($$);
	$$ = list1(build1(&@1,&@1,A_PRAGMA,$$));
	}
	;

extern_declaration : kw_datatype type_description
			{ TYPE($2) = A_EXT_DATATYPE; $$ = list1($2); }
		| kw_eqtype type_description
			{ TYPE($2) = A_EXT_EQTYPE; $$ = list1($2); }
		| kw_dimension dimension_binding
			{ TYPE($2) = A_EXT_QUANTITY; $$ = list1($2); }
		| kw_val value_description
			{ TYPE($2) = A_EXT_VAL; $$ = list1($2); }
		/*| kw_exception exception_description
			{ TYPE($2) = A_EXT_EXC; $$ = list1($2); }*/
		;

	/* optional_with_fun returns (tyvarseq,functionbindings) */
optional_with_fun : kw_with function_value_binding { $$ = ARGS($2); }
		| empty { $$ = list1(BUILD(&@1,&@1,A_TYPE_VAR_SEQ,NIL)); }
;

fixity_declaration : kw_infix optional_precedence symbol_list
						{ $$ = setFix(&@1,&@3,$3,(COUNTER)($2+1)); }
					| kw_infixr optional_precedence symbol_list
						{ $$ = setFix(&@1,&@3,$3,(COUNTER)(-$2-1)); }
					| kw_nonfix symbol_list
						{ $$ = setFix(&@1,&@2,$2,(COUNTER)0); }
					;

optional_withtype : withtype
	| empty { $$ = BUILD(&@1,&@1,A_TYPE,NIL); }
	;

withtype : kw_withtype type_binding
	{ $$ = $2; }
	;

value_binding : nonrec_value_binding
	{ ExprPtr tyvars;
	tyvars = BUILD(&@1,&@1,A_TYPE_VAR_SEQ,
	listTyVars($1,NIL));
	/* list of unguarded tyvars, not of scoped tyvars */
	$$ = build1rev(&@1,&@1,A_VAL,tyvars,$1); }
	| kw_rec nonrec_value_binding
	{ ExprPtr tyvars;
	tyvars = BUILD(&@2,&@2,A_TYPE_VAR_SEQ,
	listTyVars($2,NIL));
	/* list of unguarded tyvars, not of scoped tyvars */
	$$ = build1rev(&@1,&@2,A_VALREC,tyvars,$2); }
	;

	/* nonrec_value_bindings are reversed lists of bindings */
nonrec_value_binding : simple_value_binding
	{ $$ = list1($1); }
	| nonrec_value_binding kw_and simple_value_binding
	{ /*checkUniqueVal($3,$1);*/
		$$ = cons($3,$1); }
	;

simple_value_binding : pattern kw_equal expression
	{ $$ = build2(&@1,&@3,A_BIND,$1,$3); }
	;

function_value_binding : function_value_binding_l
	{ ExprPtr tyvars;
	tyvars = BUILD(&@1,&@1,A_TYPE_VAR_SEQ,
	listTyVars($1,NIL));
	/* list of unguarded tyvars, not of scoped tyvars */
	$$ = build1rev(&@1,&@1,A_VALREC,tyvars,$1); }
	;

	/* function_value_binding_ls are reversed lists of bindings */
function_value_binding_l : closed_function_value_binding
	{ $$ = list1($1); }
	| function_value_binding_l kw_and
	closed_function_value_binding
	{ /*checkUniqueVal($3,$1);*/
	$$ = cons($3,$1); }
	;

closed_function_value_binding : simple_function_value_binding
	{ /*checkUniqueValInFunBind($1);*/
	$$ = buildFunBind(&@1,&@1,$1); }
	;

	/* simple_function_value_bindings are reversed lists of
	zonex::f::revpatlist::expr */
simple_function_value_binding : one_function_value_binding
	{ $$ = list1(cons(BUILD(&@1,&@1,-1,NIL),
	$1)); }
	| simple_function_value_binding kw_vbar
	/* shift-reduce on vbar */
	one_function_value_binding
	{ ENTER;
	if (CAR(ARGS(CAR($3)))!=
	CAR(ARGS(CAR(CDR(CAR($1))))))
	type_error(ZONE(CAR($3)),(CHAR *)message(390));
	if (ListLength(CAR(CDR($3)))!=ListLength(CAR(CDR(CDR(CAR($1))))))
	type_error(ZONE(CAR($3)),(CHAR *)message(391));
	$$ = cons(cons(BUILD(&@3,&@3,-1,NIL),
	$3),
	$1);
	KEEPONLY($$); }
	;

	/* one_function_value_bindings are f::revpatlist::expr */
one_function_value_binding : identifier atomic_pattern_list
	optional_colon_type
	kw_equal expression
	{ ExprPtr expr;
	ENTER;
	TYPE($1) = A_REAL_VAR;
	if ($3==NIL)
		{ expr = $5; }
	else
		expr = build2(&@3,&@5,A_CAST,$5,$3);
	$$ = cons($1,cons($2,expr));
	KEEPONLY($$); }
	| infix_function_header
	optional_colon_type
	kw_equal expression
	{ ExprPtr expr;
	ENTER;
	if ($2==NIL)
	{ expr = $4; }
	else
	expr = build2(&@2,&@4,A_CAST,$4,$2);
	$$ = cons(CAR($1),
	cons(list1(CDR($1)),
	expr));
	KEEPONLY($$); }
	| open_paren
	infix_function_header
	kw_close_paren
	optional_colon_type
	kw_equal expression
	{ ExprPtr expr;
	if ($4==NIL)
	{ expr = $6; }
	else
	expr = build2(&@4,&@6,A_CAST,$6,$4);
	$$ = popMarker(cons(CAR($2),
	cons(list1(CDR($2)),
	expr))); }
	| open_paren
	infix_function_header
	kw_close_paren
	atomic_pattern_list
	optional_colon_type
	kw_equal expression
	{ ExprPtr expr;
	if ($5==NIL)
	{ expr = $7; }
	else
	expr = build2(&@5,&@7,A_CAST,$7,$5);
	$$ = popMarker(cons(CAR($2),
	cons(Nappend($4,
	list1(CDR($2))),
	expr))); }
	;

	/* infix_function_headers are f::pattern */
infix_function_header : atomic_pattern infix_identifier atomic_pattern
	{ ENTER;
	TYPE($2) = A_REAL_VAR;
	$$ = cons($2,build2(&@1,&@3,A_PAT_TUPLE,
	$1,$3));
	KEEPONLY($$); }
	;

	/* atomic_pattern_lists are reversed lists of patterns */
atomic_pattern_list : atomic_pattern
	{ $$ = list1($1); }
	| atomic_pattern_list atomic_pattern
	{ $$ = cons($2,$1); }
	;

dimension_binding : dimension_binding_l
	{ $$ = buildrev(&@1,&@1,A_QUANTITY,$1); }
	;

	/* dimension_binding_ls are reversed lists of dimension declarations */
dimension_binding_l : simple_dimension_binding
	{ $$ = list1($1); }
	| dimension_binding_l kw_and simple_dimension_binding
	{ checkUniqueQuantityAndUnit($3,$1);
	$$ = cons($3,$1); }
	;

simple_dimension_binding : symbol
	{ $$ = build2(&@1,&@1,A_NEWQUAN,$1,$1); }
	/* beware : $1 is shared ! */
	| symbol kw_open_paren symbol kw_close_paren
	{ $$ = build2(&@1,&@4,A_NEWQUAN,$1,$3); }
	;

scale_binding : scale_binding_l
	{ $$ = buildrev(&@1,&@1,A_UNIT,$1); }
	;

	/* scale_binding_ls are reversed lists of scale declarations */
scale_binding_l : simple_scale_binding
	{ $$ = list1($1); }
	| scale_binding_l kw_and simple_scale_binding
	{ checkUniqueUnit($3,$1);
	$$ = cons($3,$1); }
	;

simple_scale_binding : {unit_recognized=true;}
	number_or_int kw_scale_star symbol kw_equal numerical_constant
	{ unit_recognized=false;
	  $$ = build3(&@1,&@6,A_NEWUNIT,$4,$2,$6); }
	;

type_binding : type_binding_l
	{ $$ = buildrev(&@1,&@1,A_TYPE,$1); }
	;

	/* type_binding_ls are reversed lists of type bindings */
type_binding_l : simple_type_binding
	{ $$ = list1($1); }
	| type_binding_l kw_and simple_type_binding
	{ checkUniqueIdInType($3,$1);
		$$ = cons($3,$1); }
	;

simple_type_binding : type_variable_sequence symbol
	kw_equal type
	{ checkQuantifiedType($1,$4);
		$$ = build3(&@1,&@4,A_TYPE_BIND,$1,$2,$4); }
	;

type_variable_sequence : empty { $$ = BUILD(&@1,&@1,A_TYPE_VAR_SEQ,NIL); }
			| type_variable { $$ = build1(&@1,&@1,A_TYPE_VAR_SEQ,$1); }
			| open_paren type_variable
				type_variable_sequence_l kw_close_paren
				{ checkUniqueTypeSeq($2,$3);
					$$ = build1rev(&@1,&@4,A_TYPE_VAR_SEQ,$2,$3);
					$$ = popMarker($$); }
						;

	/* type_variable_sequence_ls are reversed lists of variables */
type_variable_sequence_l : { $$ = NIL; }
							| type_variable_sequence_l kw_comma type_variable
							{ checkUniqueTypeSeq($3,$1);
								$$ = cons($3,$1); }
							;

	/* datatype_bindings are reversed lists of datatype bindings */
datatype_binding : simple_datatype_binding
	{ $$ = list1($1); }
	| datatype_binding kw_and simple_datatype_binding
	{ checkUniqueIdInType($3,$1);
	checkUniqueConstructors($3,$1);
	$$ = cons($3,$1); }
	;

simple_datatype_binding : type_variable_sequence symbol
	kw_equal constructor_binding
	{ checkQuantifiedDatatype($1,$4);
	$$ = build2rev(&@1,&@4,A_DATATYPE_BIND,
	$1,$2,$4); }
	;

	/* constructor_bindings are reversed lists of constructor bindings */
constructor_binding : simple_constructor_binding
	{ $$ = list1($1); }
	| constructor_binding kw_vbar simple_constructor_binding
	{ checkUniqueCon($3,$1);
	$$ = cons($3,$1); }
	;

/* don't forget to record the new status of the constructor */
simple_constructor_binding : symbol optional_of_type
	{ setConstructor(CAR(ARGS($1)));
	$$ = ($2==NIL)?
	build1(&@1,&@2,A_CONSTRUCTOR,$1):
	build2(&@1,&@2,A_CONSTRUCTOR,$1,$2); }
	;

optional_of_type : of_type
	| { $$ = NIL; }
	;

of_type : kw_of type
	{ $$ = $2; }
	;

	/* optional_eqtypes are lists of type names */
optional_eqtype : { $$ = NIL; }
	| kw_eqtype eqtype_desc { $$ = Nreverse($2); }
	;

	/* type_descs are reversed lists of type names */
eqtype_desc : simple_eqtype_desc
	{ $$ = list1($1); }
	| eqtype_desc kw_and simple_eqtype_desc
	{ $$ = cons($3,$1); }
	;

simple_eqtype_desc : symbol
				{ $$ = build2(&@1,&@1,A_EQTYPE_INFO,$1,
						BUILD(&@1,&@1,A_RECORD,NIL)); }
/*#ifdef QUOTIENT_TYPES
			| symbol kw_open_brace
				expression_row kw_close_brace
				{ $$ = build2(&@1,&@4,A_EQTYPE_INFO,$1,
						buildrev(&@2,&@4,A_RECORD,$3)); }
*/
			;

exception_binding : exception_binding_l
	{ $$ = buildrev(&@1,&@1,A_EXCEPTION,$1); }
	;

	/* exception_binding_ls are reversed lists of exception bindings */
exception_binding_l : simple_exception_binding
	{ $$ = list1($1); }
	| exception_binding_l kw_and simple_exception_binding
	{ checkUniqueException($3,$1);
		$$ = cons($3,$1); }
	;

simple_exception_binding : symbol optional_of_type
	{ setConstructor(CAR(ARGS($1)));
	$$ = ($2==NIL)?
	build1(&@1,&@2,A_NEWEXC,$1):
	build2(&@1,&@2,A_NEWEXC,$1,$2); }
	| symbol kw_equal long_constructor
	{ setConstructor(CAR(ARGS($1)));
	$$ = build2(&@1,&@3,A_EXC_BIND,$1,$3); }
	;

atomic_pattern : kw_underscore
	{ $$ = BUILD(&@1,&@1,A_PAT_ANY,NIL); }
	| special_constant
	{ $$ = build1(&@1,&@1,A_PAT_CONST,$1); }
	| identifier
	{ TYPE($1) = A_REAL_VAR /*A_PAT_VAR*/; }
	| long_constructor
	{ TYPE($1) = A_PAT_VAR; }
	| open_record optional_pattern_row kw_close_record
	{ $$ = popMarker(BUILD(&@1,&@3,A_PAT_RECORD,$2)); }
	| open_paren pattern kw_close_paren
	{ $$ = popMarker($2); }
	| open_paren kw_close_paren
	{ $$ = popMarker(BUILD(&@1,&@2,A_PAT_TUPLE,NIL)); }
	| open_paren pattern kw_comma pattern_list kw_close_paren
	{ $$ = popMarker(build1list(&@1,&@3,A_PAT_TUPLE,$2,$4)); }
	| open_bracket optional_pattern_l_list kw_close_bracket
	{ $$ = popMarker($2); }
	| open_brace kw_close_brace
	{ $$ = popMarker(BUILD(&@1,&@2,A_PAT_EMPTY,NIL)); }
	| open_brace maplet_pattern_list
	{ union_recognized++; } kw_close_brace
	optional_union_pattern
	{ union_recognized--;
	installMapPattern(&@5,$2,$5);
	$$ = popMarker($2); }
	;

optional_union_pattern : union_pattern
	| { $$ = NIL; }
	;

maplet_pattern_list : maplet_pattern
	{ $$ = build2(&@1,&@1,A_PAT_MAPADD,$1,NIL);
	/* to be replaced later, by installMapPattern() */
	}
	| kw_ellipsis
	{ $$ = BUILD(&@1,&@1,A_PAT_ANYMAP,NIL); }
	| maplet_pattern kw_comma maplet_pattern_list
	{ $$ = build2(&@1,&@3,A_PAT_MAPADD,$1,$3); }
	;

maplet_pattern : pattern
	{ $$ = build2(&@1,&@1,A_PAT_MAPLET,$1,zeroPatTuple(&@1)); }
	| pattern kw_big_arrow pattern
	{ $$ = build2(&@1,&@3,A_PAT_MAPLET,$1,$3); }
	;

union_pattern : kw_union atomic_pattern
	{ $$ = $2; }
	;

optional_pattern_row : pattern_row
	| { $$ = NIL; }
	;

optional_pattern_l_list : pattern_l_list
	| empty
	{ $$ = BUILD(&@1,&@1,A_PAT_NIL,NIL); }
	;

	/* pattern_rows are lists of labeled_patterns or ellipsis */
pattern_row : kw_ellipsis
	{ $$ = list1(BUILD(&@1,&@1,
	A_PAT_RECORD_ELLIPSIS,NIL)); }
	| labeled_pattern
	{ $$ = list1($1); }
	| labeled_pattern kw_comma pattern_row
	{ checkUniqueLabel($1,$3);
	$$ = cons($1,$3); }
	;

	/* pattern_lists are lists of patterns or ellipsis */
pattern_list : kw_ellipsis
	{ $$ = list1(BUILD(&@1,&@1,
	A_PAT_TUPLE_ELLIPSIS,NIL)); }
	| pattern
	{ $$ = list1($1); }
	| pattern kw_comma pattern_list
	{ $$ = cons($1,$3); }
	;

pattern_l_list : kw_ellipsis
	{ $$ = BUILD(&@1,&@1,A_PAT_ANYLIST,NIL); }
	| pattern
	{ $$ = build2(&@1,&@1,A_PAT_APPLY,
	build1(&@1,&@1,A_PAT_VAR,consSym),
	build2(&@1,&@1,A_PAT_TUPLE,
	$1,patNil(&@1))); }
	| pattern kw_comma pattern_l_list
	{ $$ = build2(&@1,&@3,A_PAT_APPLY,
	build1(&@2,&@2,A_PAT_VAR,consSym),
	build2(&@1,&@3,A_PAT_TUPLE,$1,$3)); }
	;

labeled_pattern : label kw_equal pattern
	{ $$ = build2(&@1,&@3,A_PAT_RECORD_MAPLET,$1,$3); }
	| identifier optional_colon_type optional_as_pattern
	{ ExprPtr pat,label;
	ENTER;
	pat = $1;
	TYPE(pat) = A_REAL_VAR /*A_PAT_VAR*/;
	label = build1(&@1,&@1,A_LABEL,CAR(ARGS(pat)));
	if ($3!=NIL)
	pat = build2(&@1,&@3,A_PAT_AS,pat,$3);
	if ($2!=NIL)
	pat = build2(&@1,&@2,A_PAT_CAST,pat,$2);
	$$ = build2(&@1,&@3,A_PAT_RECORD_MAPLET,
	label,pat);
	KEEPONLY($$); }
	;

optional_as_pattern : as_pattern
	| { $$ = NIL; }
	;

as_pattern : kw_as pattern
	{ $$ = $2; }
	;

application_pattern : atomic_pattern
	| long_constructor atomic_pattern
	{ TYPE($1) = A_PAT_VAR;
	    $$ = build2(&@1,&@2,A_PAT_APPLY,$1,$2); }
	;

pack_pattern : application_pattern
	| kw_pack { pushMarker(); }
		pack_pattern
	{ $$ = popMarker(build1(&@1,&@3,A_PAT_PACK,$3)); }
	;

left_9_pattern : pack_pattern
	| left_9_pattern long_left_9_constructor pack_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_9_constructor : kw_left_9_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_9_pattern : left_9_pattern
	| left_9_pattern long_right_9_constructor right_9_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_9_constructor : kw_right_9_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_8_pattern : right_9_pattern
	| left_8_pattern long_left_8_constructor right_9_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_8_constructor : kw_left_8_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_8_pattern : left_8_pattern
	| left_8_pattern long_right_8_constructor right_8_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_8_constructor : kw_right_8_constructor
							{ $$ = lexVar(&@1,true); }
							;

left_7_pattern : right_8_pattern
	| left_7_pattern long_left_7_constructor right_8_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_7_constructor : kw_left_7_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_7_pattern : left_7_pattern
	| left_7_pattern long_right_7_constructor right_7_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_7_constructor : kw_right_7_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_6_pattern : right_7_pattern
	| left_6_pattern long_left_6_constructor right_7_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_6_constructor : kw_left_6_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_6_pattern : left_6_pattern
	| left_6_pattern long_right_6_constructor right_6_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_6_constructor : kw_right_6_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_5_pattern : right_6_pattern
	| left_5_pattern long_left_5_constructor right_6_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_5_constructor : kw_left_5_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_5_pattern : left_5_pattern
	| left_5_pattern long_right_5_constructor right_5_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_5_constructor : kw_right_5_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_4_pattern : right_5_pattern
	| left_4_pattern long_left_4_constructor right_5_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_4_constructor : kw_left_4_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_4_pattern : left_4_pattern
	| left_4_pattern long_right_4_constructor right_4_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_4_constructor : kw_right_4_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_3_pattern : right_4_pattern
	| left_3_pattern long_left_3_constructor right_4_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_3_constructor : kw_left_3_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_3_pattern : left_3_pattern
	| left_3_pattern long_right_3_constructor right_3_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_3_constructor : kw_right_3_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_2_pattern : right_3_pattern
	| left_2_pattern long_left_2_constructor right_3_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_2_constructor : kw_left_2_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_2_pattern : left_2_pattern
	| left_2_pattern long_right_2_constructor right_2_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_2_constructor : kw_right_2_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_1_pattern : right_2_pattern
	| left_1_pattern long_left_1_constructor right_2_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_1_constructor : kw_left_1_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_1_pattern : left_1_pattern
	| left_1_pattern long_right_1_constructor right_1_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_1_constructor : kw_right_1_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_0_pattern : right_1_pattern
	| left_0_pattern long_left_0_constructor right_1_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_0_constructor : kw_left_0_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_0_pattern : left_0_pattern
	| left_0_pattern long_right_0_constructor right_0_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_0_constructor : kw_right_0_constructor
							{ $$ = lexVar(&@1,true); }
						;

infix_pattern : right_0_pattern
	;

colon_pattern : infix_pattern
	| infix_pattern colon_type
	{ $$ = build2(&@1,&@2,A_PAT_CAST,$1,$2); }
	;

pattern : colon_pattern
	| colon_pattern as_pattern
	{ if (TYPE($1)!=A_REAL_VAR && (TYPE($1)!=A_PAT_CAST ||
				       TYPE(CAR(ARGS($1)))!=A_REAL_VAR))
	    type_error(ZONE($1),(CHAR *)message(612));
	  $$ = build2(&@1,&@2,A_PAT_AS,$1,$2); }
	;

optional_colon_type : colon_type
	| { $$ = NIL; }
	;

colon_type : kw_colon type
	{ $$ = $2; }
	;

atomic_type : type_variable
	| open_record optional_type_row kw_close_record
	{ $$ = popMarker(BUILD(&@1,&@3,A_RECORD_TYPE,$2)); }
	| open_paren type kw_close_paren
	{ $$ = popMarker($2); }
	;

type_variable : kw_type_variable
	{ $$ = build1(&@1,&@1,A_TYPE_VAR,lexId); }
	;

constructor_type : atomic_type
	| constructor_type symbol
	{ $$ = build2(&@1,&@2,A_TYPE_APPLY,$2,$1); }
	| symbol
	{ $$ = build1(&@1,&@1,A_TYPE_APPLY,$1); }
	| open_paren type kw_comma type_list kw_close_paren
		symbol
	{ $$ = popMarker(build2rev(&@1,&@6,A_TYPE_APPLY,$6,$2,$4)); }
	;

simple_dimension_type : constructor_type
	| constructor_type kw_scale_power number_or_int
	{ $$ = build2(&@1,&@3,A_TYPE_POWER,$1,$3); }
	;

dimension_type : simple_dimension_type
	| dimension_type kw_scale_star simple_dimension_type
	{ $$ = build2(&@1,&@3,A_TYPE_STAR,$1,$3); }
	;

tuple_type : tuple_type_l
	{ if (CDR($1)==NIL)
	{
	$$ = CAR($1);
	if (TYPE($$)==A_TYPE_ELLIPSIS)
	type_error(ZONE($$),(CHAR *)message(392));
	}
	else
	$$ = BUILD(&@1,&@1,A_TUPLE_TYPE,$1); }
	;

	/* tuple_type_ls are lists of types */
tuple_type_l : dimension_type
	{ $$ = list1($1); }
	| ellipsis_type
	{ $$ = list1($1); }
	| dimension_type kw_star tuple_type_l
	{ $$ = cons($1,$3); }
	;

ellipsis_type : kw_ellipsis kw_colon type_variable
	{ $$ = build1(&@1,&@3,A_TYPE_ELLIPSIS,$3); }
	;

function_type : tuple_type
	| tuple_type kw_arrow function_type
	{ $$ = build2(&@1,&@3,A_FUNCTION_TYPE,$1,$3); }
	| tuple_type kw_map_arrow function_type
	{ $$ = build2(&@1,&@3,A_MAP_TYPE,$1,$3); }
	;

type : { star_recognized++;
	pushMarker();
	}
	function_type
	{ star_recognized--;
	$$ = popMarker($2);
	}
	;

	/* type lists are reversed lists of types */
type_list : type
	{ $$ = list1($1); }
	| type_list kw_comma type
	{ $$ = cons($3,$1); }
	;

optional_type_row : type_row optional_ellipsis_type
		{ $$ = Nreverse1($1,$2); }
	| optional_ellipsis_type
	;

optional_ellipsis_type : { $$ = NIL; }
	| kw_comma ellipsis_type
		{ $$ = list1($2); }
	;

	/* type_rows are reversed lists of labeled types */
type_row : labeled_type
	{ $$ = list1($1); }
	| type_row kw_comma labeled_type
	{ checkUniqueLabel($3,$1);
	$$ = cons($3,$1); }
	;

labeled_type : label colon_type
	{ $$ = build2(&@1,&@2,A_MAPLET_TYPE,$1,$2); }
	;

infix_identifier : infix_identifier_1
						{ $$ = lexVar(&@1,false); }
					;

infix_identifier_1 : kw_left_0
	| kw_right_0
	| kw_left_1
	| kw_right_1
	| kw_left_2
	| kw_right_2
	| kw_left_3
	| kw_right_3
	| kw_left_4
	| kw_right_4
	| kw_left_5
	| kw_right_5
	| kw_left_6
	| kw_right_6
	| kw_left_7
	| kw_right_7
	| kw_left_8
	| kw_right_8
	| kw_left_9
	| kw_right_9
	;

optional_precedence : precedence_constant { $$ = yytext[0]-'0'; }
	| { $$ = 0; }
	;

	/* symbol_lists are in reverse */
symbol_list : label_id
		{ checkLabel(&@1); $$ = list1(lexId); }
	| kw_equal
		{ $$ = list1(equalSym); }
	| symbol_list label_id
		{ checkLabel(&@2); $$ = cons(lexId,$1); }
	| symbol_list kw_equal
		{ $$ = cons(equalSym,$1); }
	;

value_description : value_description_l
					{ $$ = buildrev(&@1,&@1,A_VAL_SPEC,$1); }
					;

	/* value_description_ls are reversed lists */
value_description_l : simple_value_description
						{ $$ = list1($1); }
					| value_description_l kw_and simple_value_description
						{ checkUniqueValDesc($3,$1);
							$$ = cons($3,$1); }
					;

simple_value_description : identifier kw_colon type
								{ $$ = build2(&@1,&@3,A_VAL_BIND_SPEC,$1,$3); }
							;

type_description : type_description_l
						{ $$ = buildrev(&@1,&@1,A_TYPE_SPEC,$1); }
					;

	/* type_description_ls are reversed lists */
type_description_l : simple_type_description
						{ $$ = list1($1); }
					| type_description_l kw_and simple_type_description
						{ checkUniqueIdInType($3,$1);
							$$ = cons($3,$1); }
					;

simple_type_description : type_variable_sequence symbol
						{ $$ = build2(&@1,&@2,A_TYPE_BIND_SPEC,$1,$2); }
						;

/*
exception_description : exception_description_l
							{ $$ = buildrev(&@1,&@1,A_EXCEPTION_DESC,$1); }
						;
*/

	/* exception_desciption_ls are reversed lists */
/*
exception_description_l : simple_exception_description
							{ $$ = list1($1); }
						| exception_description_l kw_and simple_exception_description
							{ checkUniqueException($3,$1);
								$$ = cons($3,$1); }
						;

simple_exception_description : symbol optional_of_type
									{ setConstructor(CAR(ARGS($1)));
										$$ = ($2==NIL)?
												build1(&@1,&@1,A_NEWEXC_DESC,$1):
												build2(&@1,&@2,A_NEWEXC_DESC,$1,$2); }
								;
*/

%%
