/* Parsing support functions, himmlrun version.
   Copyright (C) 2003-2005,2011 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <stdio.h>
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#else
#include <errno.h>
#endif
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "abssyn.h"
#include "expr.h"
#include "gram.h"
#include "himml.tab.h"
#include "type.h"
#include "input.h"
#include "compile.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#include "grmproto.h"
#include "errproto.h"
#include "typproto.h"
#include "memproto.h"
#include "prtproto.h"
#include "cmpproto.h"
#include "stkproto.h"
#include "evlproto.h"
#include "dirproto.h"
#include "exgproto.h"
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/types.h>
#include <sys/stat.h>
#endif
#if OS==AMIGA_OS
#include <stat.h>
#endif
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rgramaux.h"
#ifdef HAS_PROTOTYPES
#include "sgramaux.h"
#endif

#define DEBUG

#define __SEG__ PARSE
#define HML_EXTENSION

EXTERN ExprPtr keepLexEnv,keepConsEnv;

PUBLIC ExprPtr parseMarker;
PUBLIC ExprPtr parseMarkerStack;
PUBLIC ExprPtr toplevelCont=NOEXPR;
PUBLIC ExprPtr theSyntax;
PUBLIC SymPtr ItSymbol;

PUBLIC ExprPtr theTypes,theBindings;

PUBLIC Boolean syntax_analysis=false;

EXTERN int yynerrs;

EXTERN INT currentAge;

PUBLIC struct cslist *Checksums=NULL;

PRIVATE ExprPtr openModRef = NULL;

PRIVATE DECL_POW10(my_pow10)
#define POW10 (my_pow10+10)

#if DEBUGGING
EXTERN Boolean evaluatingp,stopwanted, debugEnabled;
#endif

PRIVATE ExprPtr cons(a,b)
IN CONST ExprPtr a,b;
{
  REGISTER ExprPtr res;

  MAKECONS(a,b,res);
  QUICKKEEP(res);
  return res;
}

PRIVATE ExprPtr Cons(a,b)
IN CONST ExprPtr a,b;
{
  ExprPtr res;

  MAKECONS(a,b,res);
  return KEEP(res);
}

PUBLIC ExprPtr usePaths; /* string list ref */

PUBLIC hFILE *xfopen(name,mode,realname)
char *name,*mode;
char **realname;
{
  char *s;
  hFILE *file;
  COUNTER len;
  ExprPtr paths,path;

  len = strlen(name)+1;
  *realname = (char *)MALLOC_FAIL(len);
  if (*realname==NULL)
    return NULL;
  strcpy(*realname,name);
  file = hfopen(name,mode);
  if (file==NULL)
    {
      MARKFREE(*realname);
      for (paths = REFVAL(usePaths); paths!=NIL; paths = CDR(paths))
	{
	  path = CAR(paths);
	  *realname = (char *)MALLOC_FAIL(SLEN(path)+len+1);
	  if (*realname==NULL)
	    return NULL;
	  safememcpy(*realname,STEXT(path),SLEN(path));
	  s = *realname+SLEN(path);
	  if (SLEN(path)!=0 && s[-1]!=DIR_CHAR && s[-1]!=VOL_CHAR)
	    *s++ = DIR_CHAR;
	  strcpy(s,name);
	  file = hfopen(*realname,mode);
	  if (file!=NULL)
	    break;
	  MARKFREE(*realname);
	  *realname = NULL;
	}
    }
  return file;
}

PUBLIC void InitMarkerStack()
{
  SETQ(parseMarkerStack,NIL);
}

PUBLIC ExprPtr getString(start,end,text,build)
IN CONST struct yyltype *start,*end;
IN REGISTER CHAR *text;
IN Boolean build;
{
  ExprPtr str;
  CHAR *begin;
  REGISTER INT len;
  REGISTER CHAR *dest,c;

  begin = text;
  len = 0;
  for (dest = text, text++; c = *text++, c!='\"'; )
    {
      switch (c)
	{
	case '\\':
	  c = *text++;
	  switch (c)
	    {
	    case '\"':   case '\\':
	      *dest++ = c;
	      len++;
	      break;
	    case 'n':
	      *dest++ = '\n';
	      len++;
	      break;
	    case 't':
	      *dest++ = '\t';
	      len++;
	      break;
	    case 'v':
	      *dest++ = '\v';
	      len++;
	      break;
	    case 'f':
	      *dest++ = '\f';
	      len++;
	      break;
	    case '^':
	      *dest++ = (*text++)-'@';
	      len++;
	      break;
	    case '0':	case '1':	case '2':	case '3':
	    case '4':	case '5':	case '6':	case '7':
	    case '8':	case '9':
	      *dest = (c-'0')*100;
	      *dest += ((*text++)-'0')*10;
	      *dest++ += ((*text++)-'0');
	      len++;
	      break;
	    default:
	      while (c = *text++, c!='\\');
	      break;
	    }
	  break;
	case '\n':
	  if (start!=(struct yyltype *)0L && end!=(struct yyltype *)0L)
	    {
	      TZone zone;

	      *dest = 0;
	      makeZone(&zone,start,end);
	      type_error(&zone,(CHAR *)message(646));
	    }
	  break;
	default:
	  *dest++ = c;
	  len++;
	  break;
	}
    }
  *dest = 0;
  return build?(str = Kstring(len,begin), KEEP(str)):NIL;
}

PUBLIC void pushEnv()
{
  ExprPtr aux;
  ENTER;

  MAKECONS1(NewMap(),lexEnv,aux);
  lexEnv = aux;
  MAKECONS1(NewMap(),consEnv,aux);
  consEnv = aux;
#if MODULE_SYSTEM & STD_ML_MODULE
  MAKECONS1(NewMap(),strEnv,aux);
  strEnv = aux;
  MAKECONS1(NewMap(),sigEnv,aux);
  sigEnv = aux;
#endif
  QUIT;
}

PUBLIC void makeZone(zone,yylsp1,yylsp2)
OUT TZone *zone;
IN CONST struct yyltype *yylsp1,*yylsp2;
{
  zone->line1 = (unsigned short)yylsp1->first_line;
  zone->pos1 = (unsigned short)yylsp1->first_column;
  zone->line2 = (unsigned short)yylsp2->last_line;
  zone->pos2 = (unsigned short)yylsp2->last_column;
}

#define BH_BUCKET_SIZE 32

typedef struct bhBucket {
  struct bhBucket *next;
  SHORT firstNewCell;
  SHORT occupancy[MAX_GENERATIONS];
#if BH_BUCKET_GAP!=0
  char filler[BH_BUCKET_GAP];
#endif
  struct build_header bh[BH_BUCKET_SIZE];
  struct build_header *bhcells[BH_BUCKET_SIZE];
} bhBucket;

PRIVATE bhBucket *allocBuckets=(bhBucket *)0L;
PRIVATE struct build_header *freeBuildHeaders=(struct build_header *)0L;
EXTERN SHORT generation_count;

PRIVATE struct build_header *reserve_bh=(struct build_header *)0L;

PRIVATE void newBucket()
{
  bhBucket *bhb;
  REGISTER struct build_header *bh,*oldbh,**bhp;
  REGISTER COUNTER i;

  bhb = (bhBucket *)MALLOC(sizeof(bhBucket));
  bhb->next = allocBuckets;
  bhb->firstNewCell = 0;
  for (i=0; i<MAX_GENERATIONS; i++)
    bhb->occupancy[i] = 0;
  allocBuckets = bhb;
  oldbh = (struct build_header *)0L; /* we add buckets at the end of the list */
  for (i = 0, bh = bhb->bh, bhp = bhb->bhcells; i<BH_BUCKET_SIZE; i++, bh++)
    {
      *bhp++ = bh;
      bh->kind = E_FREE;
      bh->mark = 0;
      /*bh->gcmark = 0;*/
      bh->args = (ExprPtr)oldbh;
      oldbh = bh;
    }
  if (reserve_bh==(struct build_header *)0L)
    {
      freeBuildHeaders = oldbh;
    }
  else
    { /* append at the end */
      reserve_bh->args = (ExprPtr)oldbh;
    }
  reserve_bh = bhb->bh; /* the last allocated cell */
}

PUBLIC void FreeBuildHeaders()
{
  REGISTER struct build_header *bh;
  REGISTER COUNTER i,n,occupancy;
  struct build_header **bhp,**firstbhp;
  bhBucket *buck;

  for (buck=allocBuckets; occupancy = 0, buck!=(bhBucket *)0L; buck = buck->next)
    {
      n = BH_BUCKET_SIZE-buck->firstNewCell;
      for (i=0, firstbhp = bhp = buck->bhcells+buck->firstNewCell; i<n; i++, bhp++)
	{
	  bh = *bhp;
	  if (GC_MARKED(bh))
	    {
	      GC_UNMARK(bh);
	      occupancy++;
	      if (firstbhp!=bhp) /* firstbhp is always equal to bhp, or points to the first
				    free cell after the firstNewCell position */
		{ /* exchange position of bh with that of the free cell *firstbhp */
		  *bhp = *firstbhp;
		  *firstbhp = bh;
		}
	      firstbhp++;
	      /* this all sorts the newcells array, from the firstNewCell position up,
		 so that all addresses of still allocated cells are in an interval
		 [firstNewCell..firstNewCell+buck->occupancy[generation_count] [ */
	    }
	  else if (bh->kind!=E_FREE)
	    {
	      bh->kind = E_FREE;
	      bh->args = (ExprPtr)freeBuildHeaders;
	      freeBuildHeaders = bh;
	    }
	}
      buck->occupancy[generation_count] = occupancy;
    }
}

PUBLIC void FreeBuildHeader(bh)
IN struct build_header *bh;
{
  bh->kind = E_FREE; /* space won't be got back until a FreeBuildHeaders() */
  bh->args = (ExprPtr)freeBuildHeaders;
  freeBuildHeaders = bh;
}

PUBLIC void cutoffInitialHeaders()
{
  freeBuildHeaders = reserve_bh = (struct build_header *)0L;
  allocBuckets = (bhBucket *)0L;
}

#ifdef GC_SLIDING
PUBLIC void slideOldHeaders()
{
  REGISTER bhBucket *bhbp;
  REGISTER COUNTER i;
  REGISTER SHORT *occ;

  for (bhbp = allocBuckets; bhbp!=(bhBucket *)0L; bhbp=bhbp->next)
    {
      bhbp->occupancy[0] += bhbp->occupancy[1];
      for (i=1, occ=&bhbp->occupancy[1]; i<generation_count; i++, occ++)
	occ[0] = occ[1];
      occ[0] = 0;
    }
}
#endif

PUBLIC void cutoffOldHeaders()
{
  bhBucket *bhbp;
  REGISTER struct build_header **bhp;
  REGISTER COUNTER i;
  REGISTER COUNTER n;

  for (bhbp = allocBuckets; bhbp!=(bhBucket *)0L; bhbp=bhbp->next)
    {
      n = bhbp->occupancy[generation_count];
      if (n==0)
	continue;
      /* mark in advance all "newly old" cells */
      for (i=0, bhp=bhbp->bhcells + bhbp->firstNewCell; i<n; i++, bhp++)
	{
	  GC_MARK((*bhp)); /* to speed up future mark phases */
	}
      bhbp->firstNewCell += n;
    }
}

PUBLIC void refreshOldHeaders(oldgen)
IN COUNTER oldgen;
{
  REGISTER COUNTER n;
  REGISTER bhBucket *bhbp;

  for (bhbp=allocBuckets; bhbp!=(bhBucket *)0L; bhbp=bhbp->next)
    {
      n = bhbp->firstNewCell;
      {
	REGISTER COUNTER i;
	REGISTER SHORT *occ;

	for (i = generation_count, occ = bhbp->occupancy+i; i<oldgen; i++)
	  n -= *occ++;
      }
      {
	REGISTER struct build_header **bhp;
	REGISTER COUNTER i;

	for (i = bhbp->firstNewCell, bhp = bhbp->bhcells+i; --i>=n; )
	  GC_UNMARK((*--bhp));
      }
      bhbp->firstNewCell = n;
    }
}

PUBLIC ExprPtr ZBuild(zone,f,info)
IN CONST TZone *zone;
IN COUNTER f;
IN CONST ExprPtr info;
{
  struct build_header *bh;

  if (freeBuildHeaders==reserve_bh)
    newBucket();
  bh = freeBuildHeaders;
  freeBuildHeaders = (struct build_header *)bh->args;
  bh->kind = BUILD_HEADER_KIND;
  bh->mark = 0;
  /*bh->gcmark = 0;*/
  bh->type = f;
  bh->z = *zone;
  bh->args = info;
  return (ExprPtr)bh;
}

PUBLIC ExprPtr Build1(start,end,f,info)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr info;
{
  TZone zone;

  makeZone(&zone,start,end);
  return ZBuild(&zone,f,info);
}

PUBLIC ExprPtr Build(start,end,f,info)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr info;
{
  ExprPtr aux;
  TZone zone;

  makeZone(&zone,start,end);
  aux = ZBuild(&zone,f,info);
  return KEEP(aux);
}

PUBLIC ExprPtr build1(start,end,f,a)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a;
{
  ExprPtr aux;

  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,NIL));
  ENDQUICKKEEP();
  return KEEP(aux);
}

PUBLIC ExprPtr build2(start,end,f,a,b)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a,b;
{
  ExprPtr aux;

  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,cons(b,NIL)));
  ENDQUICKKEEP();
  return KEEP(aux);
}

PUBLIC ExprPtr build3(start,end,f,a,b,c)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a,b,c;
{
  ExprPtr aux;

  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,cons(b,cons(c,NIL))));
  ENDQUICKKEEP();
  return KEEP(aux);
}

PUBLIC ExprPtr buildrev(start,end,f,list)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr list;
{
  ENTER;
  list = Build1(start,end,f,Nreverse(list));
  KEEPONLY(list);
  return list;
}

PUBLIC ExprPtr build1rev(start,end,f,a,list)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a,list;
{
  ExprPtr aux;
  ENTER;

  list = Nreverse(list);
  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,list));
  ENDQUICKKEEP();
  KEEPONLY(aux);
  return aux;
}

PUBLIC ExprPtr build2rev(start,end,f,a,b,list)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a,b,list;
{
  ExprPtr aux;
  ENTER;

  list = Nreverse(list);
  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,cons(b,list)));
  ENDQUICKKEEP();
  KEEPONLY(aux);
  return aux;
}

PUBLIC ExprPtr build1list(start,end,f,a,list)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a,list;
{
  ExprPtr aux;

  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,list));
  ENDQUICKKEEP();
  return KEEP(aux);
}

PUBLIC ExprPtr build2list(start,end,f,a,b,list)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a,b,list;
{
  ExprPtr aux;

  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,cons(b,list)));
  ENDQUICKKEEP();
  return KEEP(aux);
}

PUBLIC void do_printfile(text,len,stream)
REGISTER IN CONST CHAR *text;
REGISTER IN COUNTER len;
REGISTER INOUT char *stream;
{
  REGISTER CHAR c;

  while (len--!=0)
    {
      c = *text++;
      hputc(c,(hFILE *)stream);
    }
}

PRIVATE void basisSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  Basis *b;

  b = (Basis *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      {
	CHAR c;

	c = b->syntax_analysis;
	outChars(&c,1);
	if (errno) return;
#if DEBUGGING
	c = b->evaluatingp;
#else
	c = 0;
#endif
	outChars(&c,1);
	if (errno) return;
#if DEBUGGING
	c = b->stopwanted;
#else
	c = 0;
#endif
	outChars(&c,1);
	if (errno) return;
	c = b->unit_recognized;
	outChars(&c,1);
	if (errno) return;
	outInt(0L /*b->condState*/);
	if (errno) return;
	outInt(0L /*b->condLevel*/);
	if (errno) return;
	outInt(b->star_recognized);
	if (errno) return;
	outInt(b->union_recognized);
	if (errno) return;
      }
      goto markdata;
    case X_MARK:
      { /* We never save the excHandler part of bases,
	   so this code is only for the X_MARK case: */
	REGISTER excHandler *exc;

	for (exc = b->excStack; exc!=NULL; exc = exc->next)
	  switch (exc->tag)
	    {
	    case X_HANDLER:
	      (*doext->action)(exc->code);
	      /*fallthrough*/
	    case X_PROF:
	    case X_CONT:
	      (*doext->action)(exc->data);
	      break;
	    }
      }
      /*fallthrough*/
    case X_MARKDATA:
markdata:
      (*doext->action)(b->parseMarker);
      (*doext->action)(b->parseMarkerStack);
      (*doext->action)(b->lexEnv);
      (*doext->action)(b->consEnv);
      (*doext->action)(b->strEnv);
      (*doext->action)(b->sigEnv);
      (*doext->action)(b->toplevelCont);
      (*doext->action)(b->basisState);
      (*doext->action)(b->basisAssum);
      (*doext->action)(b->basisTable);
      (*doext->action)(b->basisEnv);
      (*doext->action)(b->bindings);
#ifdef OBSOLETE
      (*doext->action)(b->debugApplStk);
#endif
      break;
    case X_FREE:
      MARKFREE(extdata); /* free the C structure */
      break;
    }
}

PRIVATE ExprPtr basisLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  Basis *b;
  ExprPtr e;
  CHAR c;
  ENTER;

  b = (Basis *)MALLOC(sizeof(Basis));
  inChars(&c,1);
  if (errno)
    {
err:
      MARKFREE(b);
      return NIL;
    }
  b->syntax_analysis = c;
  inChars(&c,1);
  if (errno) goto err;
#if DEBUGGING
  b->evaluatingp = c;
#endif
  inChars(&c,1);
  if (errno) goto err;
#if DEBUGGING
  b->stopwanted = c;
#endif
  inChars(&c,1);
  if (errno) goto err;
  b->unit_recognized = c;
  /*b->condState =*/ (void) inInt();
  if (errno) goto err;
  /*b->condLevel =*/ (void) inInt();
  if (errno) goto err;
  b->star_recognized = inInt();
  if (errno) goto err;
  b->union_recognized = inInt();
  if (errno) goto err;
  b->parseMarker = (*load)(sharedref);
  if (errno) goto err;
  b->parseMarkerStack = (*load)(sharedref);
  if (errno) goto err;
  b->lexEnv = (*load)(sharedref);
  if (errno) goto err;
  b->consEnv = (*load)(sharedref);
  if (errno) goto err;
  b->strEnv = (*load)(sharedref);
  if (errno) goto err;
  b->sigEnv = (*load)(sharedref);
  if (errno) goto err;
  b->toplevelCont = (*load)(sharedref);
  if (errno) goto err;
  b->excStack = NULL; /* cannot load an exception stack */
  b->basisState = (*load)(sharedref);
  if (errno) goto err;
  b->basisAssum = (*load)(sharedref);
  if (errno) goto err;
  b->basisTable = (*load)(sharedref);
  if (errno) goto err;
  b->basisEnv = (*load)(sharedref);
  if (errno) goto err;
  b->bindings = (*load)(sharedref);
  if (errno) goto err;
#ifdef OBSOLETE
  b->debugApplStk = (*load)(sharedref);
  if (errno) goto err;
#endif
  e = (shared?Kexternal:MakeExternal)(desc,(char *)b);
  KEEPONLY(e);
  return e;
}

PRIVATE extDesc basisDesc = {
  (CHAR *)"basis",
  { 'b','a','s','e' }, NULL,
  NULL, /* not shared */
  NULL, /* not shared */
  NULL, /* print as default */
  basisSweep,
  basisLoad
  };

PUBLIC ExprPtr basisNew(copy)
IN Boolean copy;
{
  Basis *b;
  ExprPtr e,aux;
  ENTER;

  b = (Basis *)MALLOC(sizeof(Basis));
  b->parseMarker = parseMarker;
  b->parseMarkerStack = parseMarkerStack;
  b->syntax_analysis = syntax_analysis;
#if DEBUGGING
  b->evaluatingp = evaluatingp;
  b->stopwanted = stopwanted;
#ifdef OBSOLETE
  b->debugApplStk = getApplStk();
#endif
#endif
  b->unit_recognized = unit_recognized;
  b->star_recognized = star_recognized;
  b->union_recognized = union_recognized;
  if (copy)
    {
      b->lexEnv = copyLexEnv();
      b->consEnv = copyConsEnv();
#if MODULE_SYSTEM & STD_ML_MODULE
      b->strEnv = copyStrEnv();
      b->sigEnv = copySigEnv();
#else
      b->strEnv = b->sigEnv = NIL;
#endif
    }
  else
    {
      b->lexEnv = lexEnv;
      b->consEnv = consEnv;
#if MODULE_SYSTEM & STD_ML_MODULE
      b->strEnv = strEnv;
      b->sigEnv = sigEnv;
#else
      b->strEnv = b->sigEnv = NIL;
#endif
    }
  /*b->condState = condState;*/
  /*b->condLevel = condLevel;*/
  b->toplevelCont = toplevelCont;
  b->excStack = excStack;
  aux = basisState;
  if (copy)
    aux = DupTypeState(aux);
  b->basisState = aux;
  aux = basisAssum;
  if (copy)
    aux = DupTypeAssum(aux);
  b->basisAssum = aux;
  aux = basisTable;
  if (copy)
    aux = DupCompAssum(aux);
  b->basisTable = aux;
  b->basisEnv = basisEnv;
  b->bindings = NIL;
  e = MakeExternal(&basisDesc,(char *)b);
  KEEPONLY(e);
  return e;
}

PUBLIC void basisInstall(e,errp,copy)
IN CONST ExprPtr e;
IN Boolean errp,copy;
{
  Basis *b;
  ExprPtr aux;

  b = (Basis *)EXTDATA(e);
  SETQ(parseMarker,b->parseMarker);
  SETQ(parseMarkerStack,b->parseMarkerStack);
  unit_recognized = b->unit_recognized;
  star_recognized = b->star_recognized;
  union_recognized = b->union_recognized;
  if (errp)
    {
      SETQ(lexEnv,b->lexEnv);
      if (copy)
	{
	  aux = copyLexEnv();
	  SETQ(lexEnv,aux);
	}
      SETQ(consEnv,b->consEnv);
      if (copy)
	{
	  aux = copyConsEnv();
	  SETQ(consEnv,aux);
	}
#if MODULE_SYSTEM & STD_ML_MODULE
      SETQ(strEnv,b->strEnv);
      if (copy)
	{
	  aux = copyStrEnv();
	  SETQ(strEnv,aux);
	}
      SETQ(sigEnv,b->sigEnv);
      if (copy)
	{
	  aux = copySigEnv();
	  SETQ(sigEnv,aux);
	}
#endif
      aux = b->basisState;
      if (copy)
	aux = DupTypeState(aux);
      SETQ(basisState,aux);
      aux = b->basisAssum;
      if (copy)
	aux = DupTypeAssum(aux);
      SETQ(basisAssum,aux);
      aux = b->basisTable;
      if (copy)
	aux = DupCompAssum(aux);
      SETQ(basisTable,aux);
      SETQ(basisEnv,b->basisEnv);
    }
  /*condState = b->condState;*/
  /*condLevel = b->condLevel;*/
  SETQ(toplevelCont,b->toplevelCont);
  setHandlerStack(b->excStack);
#if DEBUGGING
  evaluatingp = b->evaluatingp;
  stopwanted = debugEnabled && b->stopwanted;
#ifdef OBSOLETE
  setApplStk(b->debugApplStk);
#endif
#endif
  syntax_analysis = b->syntax_analysis;
}

DEF_ACTION_1ARG(precompile)
{ /* arg is now the toplevel continuation. */
  ExprPtr syntax;

  SETQ(toplevelCont,arg);
  parseMarker = getProtectWrapper();
  if (yyparse()!=0)
    {
      theTypes = NIL;
      theBindings = NIL;
      return ione; /* error */
    }
  InitMarkerStack();
  (void) alloctmp(0L);
#if CORE_TRACE
  if (coreTraceFile!=NULL)
    hfflush(coreTraceFile);
#endif
  syntax = theSyntax;
  setProtectWrapper(parseMarker,syntax);
  syntax_analysis = false;
  theTypes = tcDeclaration(syntax,basisAssum,basisState);
  (void) alloctmp(0L);
  if (yynerrs!=0)
    {
      theTypes = NIL;
      theBindings = NIL;
      type_error(NULL,NULL); /* abort elaboration */
    }
  if (yy_input_device==YY_INPUT_STDIN)
    {
      COUNTER save_maxlines,maxlines;

      save_maxlines = crtab_maxlines;
      crtab_maxlines = maxlines = IVAL(REFVAL(maxLinesRef));
#if IMPERATIVE_TYPES==EFFECT_TYVAR
      fxReset();
#endif
      (void) (typeEnvPrint(theTypes,do_printfile,hstdout,basisState)
#if IMPERATIVE_TYPES==EFFECT_TYVAR
	      || FxPretty(theFx,do_printfile,hstdout,basisState)
	      || fxFlush(do_printfile,hstdout,basisState)
#endif
	      );
#if IMPERATIVE_TYPES==EFFECT_TYVAR
      resetFx();
#endif
      if (crtab_maxlines<=0)
	hfputs("\n...\n",hstdout);
      (void) hfflush(hstdout);
      crtab_maxlines = save_maxlines;
    }
  (void) alloctmp(0L);
  SETQ(C_TYSTATE(basisTable),basisState);
  theBindings = cp(syntax,basisTable);
  SETQ(S_PACKERS(basisState),NIL);
  SETQ(S_ANNOTATIONS(basisState),NIL);
  SETQ(C_TYSTATE(basisTable),NIL);
  if (yynerrs!=0)
    {
      theTypes = NIL;
      theBindings = NIL;
      type_error(NULL,NULL); /* abort elaboration */
    }
  return izero; /* toplevel phase correctly completed */
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(himmlparse)
{ /* arg is now the toplevel continuation. */
  ExprPtr res;
  ExprPtr types,bindings,state,assum,table,envv;
  VecPtr vp;
  COUNTER n;

  res = ACTION_NAME(precompile)(env,arg); /* no need to call ENV_CONS(ENV_UNDERLYING(env),1),
					     since this is already equal to env. */
  types = theTypes;
  bindings = theBindings;
  if (res!=izero)
    {
      theTypes = NIL;
      theBindings = NIL;
      return res;
    }
  resetHandlerStack();
#if DEBUGGING
  evaluatingp = true;
  stopwanted = false;
#ifdef OBSOLETE
  initApplStk();
#endif
#endif
  /* save the following; we shall need them after evaluating;
     but we may return from Eval*() functions by non-local
     Throw()s, so we have to reinstall a coherent version of the
     world. */
  state = DupTypeState(basisState);
  assum = DupTypeAssum(basisAssum);
  table = DupCompAssum(basisTable);
  envv = KEEP(basisEnv);
  vp = EVARS(basisEnv);
  n = nLocals(basisTable);
#ifdef OBSOLETE
  if (debugMode==DBM_IN_DEBUGGER)
    {
      (void) EvalBindingsRC(bindings,env,0,2);
      setNLocals(table,0);
    }
  else
#endif
    {
      if (EvalBindingsRCIt(bindings,
			   basisEnv
			   /* there is something which could fail to work here:
			      normally, vp->n should be the max size of vp,
			      otherwise copying might shrink vp to the current
			      vp->n; this is not so, because of the special
			      treatment of GLOBENV in GCmark().
			      */
			   ,vp->n,n))
	{
#if DEBUGGING
	  evaluatingp = false;
#endif
	  (void) Raise(BINDEXC); /* because evaluatingp is false,
				    Raise won't return */
	}
    }
#if DEBUGGING
#ifdef OBSOLETE
  currentDebugZone = NIL;
#endif
  evaluatingp = false;
#endif
  /* reload coherent version of the world. */
  SETQ(basisState,state);
  SETQ(basisAssum,assum);
  SETQ(basisTable,table);
  SETQ(basisEnv,envv);
#if STOP_AND_COPY_VECTORS
  vp = EVARS(basisEnv);
#endif
  /*vp->n = n;*/ /* done by EvalBindingsRC() */
  AddTypeAssum(basisAssum,types);
  (void) alloctmp(0L);
  if (yy_input_device==YY_INPUT_STDIN)
    {
      valsPrint(types,do_printfile,hstdout);
      hfflush(hstdout);
    }
  setProtectWrapper(parseMarker,NIL);
  (void) alloctmp(0L);
  return izero; /* toplevel phase correctly completed */
}
ACTION_END

PUBLIC void ResetCurrentAge(age)
IN INT age;
{
  if (age<currentAge)
    {
      struct cslist *csl,**cslp;

      currentAge = age;
      for (cslp = &Checksums; csl = *cslp, csl!=NULL;)
	{
	  if (csl->age > age)
	  {
	    *cslp = csl->next;
	    MARKFREE(csl);
	  }
	else
	  cslp = &csl->next;
      }
    }
}

PUBLIC int HimMLParse1(action)
IN compiledAction action;
{
  ExprPtr basis,res;
  char *mark;
  INT age;
  ENTER;

  basis = basisNew(true);
  InitMarkerStack();
  unit_recognized = false;
  star_recognized = 0;
  union_recognized = 0;
#if CORE_TRACE
  resetTracingYyLex();
#endif
  setTopEnv();
  syntax_analysis = true;
  /*condState = 0;*/
  /*condLevel = 0;*/
  mark = getMarkFiles();
  age = currentAge;
  res = CallCC(action,NIL,true);
  (void) alloctmp(0L);
  if (res==itwo) /* no memory left */
    GCgrow(10, NULL); /* get back memory urgently */
  ResetCurrentAge(age);
  if (res!=izero)
    {
#if CORE_TRACE
      if (coreTraceFile!=NULL)
	hfflush(coreTraceFile);
#endif
      basisInstall(basis,true,false);
      ResetFiles(mark,true);
      yyinit();
    }
  else
    {
      basisInstall(basis,false,false);
    }
  (void) alloctmp(0L);
  QUIT;
  return (int)IVAL(res);
}

PUBLIC int HimMLParse()
{
  int res;

  res = HimMLParse1(ACTION_NAME(himmlparse));
  currentAge++; /* because we evaluated some declarations, our age increases;
		   we may throw an ancient continuation (return to the past),
		   but not one that was stored in some future (returns to the
		   future are not allowed). */
  return res;
}

PRIVATE void recordOpenMod(name,dirp)
IN CONST char *name;
IN Boolean dirp;
{
  if (openModRef!=NULL)
    {
      INT len;
      char *s;
      ExprPtr str,e;
      ENTER;

      len = strlen(name);
      s = name+len;
      if (len>=3 && (s[-1]=='l' && s[-2]=='m' && s[-3]=='.'))
	{ len -= 3; /* chop off '.ml' */ }
      else if (len>=4 && (s[-1]=='x' && s[-2]=='l' && s[-3]=='m' && s[-4]=='.'))
	{ len -= 4; /* chop off '.mlx' */ }
#ifdef HML_EXTENSION
      else if (len>=4 && (s[-1]=='l' && s[-2]=='m' && s[-3]=='h' && s[-4]=='.'))
	{ len -= 4; /* chpo off '.hml' */ }
#endif
      if (dirp)
	{
	  CHAR *pwd,*xname;

	  pwd = (CHAR *)alloctmp(PWD_BUFSIZE+len+1);
	  pwd = getDir(pwd,PWD_BUFSIZE);
	  xname = pwd+strlen((char *)pwd);
	  if (xname>pwd && xname[-1]!=VOL_CHAR && xname[-1]!=DIR_CHAR)
	    *xname++ = DIR_CHAR;
	  safememcpy(xname,name,len);
	  str = Kstring((xname-pwd)+len,pwd);
	}
      else
	str = Kstring(len,(CHAR *)name);
      (void) KEEP(str);
      e = Cons(str,REFVAL(openModRef));
      SETREFVAL(openModRef,e);
      QUIT;
    }
}

READFUNC(readfile)
{
  REGISTER hFILE *hf;
  REGISTER COUNTER i;
  REGISTER int c;

  hf = READDATA(hFILE *);
  for (i=0; i<len; i++)
    {
      c = hgetc(hf);
      if (c==-1)
	{
	  errno = EIO; /* so that inTransfer1() detects a problem. */
	  return;
	}
      *text++ = c;
    }
}

PRIVATE INT inI(mod)
IN hFILE *mod;
{
  INT n;
  COUNTER i;

  {
    BEGIN_TO_INT(n);
    for (i=0; i<sizeof(INT); i++)
      TO_INT(n)[i] = hfgetc(mod);
    END_TO_INT(n);
  }
  return n;
}

PRIVATE Boolean checkOpenMods(mlxt,mod)
IN CONST time_t mlxt;
IN hFILE *mod;
{
  INT i,n,len,j;
  char *buf,*s;
  struct stat st;

  n = inI(mod);
  if (errno) return false;
  if (hfgetc(mod)!='\n' || errno) return false;
  for (i=0; i<n; i++)
    {
      len = inI(mod);
      if (errno) return false;
      buf = alloctmp(len+5);
      for (j=0, s=buf; j<len; j++)
	*s++ = hfgetc(mod);
      *s++ = '.';
      *s++ = 'm';
      *s++ = 'l';
      *s = 0;
      if (stat(buf,&st)==0 && st.st_mtime>mlxt)
	return false;
      *s++ = 'x';
      *s = 0;
      if (stat(buf,&st)==0 && st.st_mtime>mlxt)
	return false;
      if (hfgetc(mod)!='\n' || errno) return false;
      recordOpenMod(buf,false);
    }
  return true;
}

PRIVATE Boolean skipOpenMods(mod)
IN hFILE *mod;
{
  INT i,n,len,j;

  n = inI(mod);
  if (errno) return false;
  if (hfgetc(mod)!='\n' || errno) return false;
  for (i=0; i<n; i++)
    {
      len = inI(mod);
      if (errno) return false;
      for (j=0; j<len; j++)
	(void) hfgetc(mod);
      if (hfgetc(mod)!='\n' || errno) return false;
    }
  return true;
}

PRIVATE hFILE *openModule(name,realname,basisp,recompile)
IN CONST CHAR *name;
OUT char **realname;
OUT ExprPtr *basisp;
IN Boolean recompile;
{
  hFILE *module;
  COUNTER len;
  CHAR *xname,*s;

  len = strlen((char *)name);
  /* otherwise, try adding .mlx at the end of the name */
  xname = (CHAR *)alloctmp_anchored(len+5, (char *)&realname);
  strcpy((char *)xname,(char *)name);
  s = xname+len;
  *basisp = NIL;
  module = xfopen((char *)xname,"r",realname);
  if (module!=NULL)
    { /* if module exists under this name, then: */
      {
	char *buf;

	{
	  int c;

	  while (c=hfgetc(module), c!=-1 && c!='\n');
	  /* skip first line, which should be of the form #!/usr/local/bin/himmlrun
	     for Unix systems to be able to launch complete bytecode files. */
	}
	len = strlen(message(807)); /* then, check that it has the magic header */
	buf = alloctmp_anchored(len+1, (char *)&realname);
	readfile((CHAR *)buf,len,(char *)module);
	buf[len] = 0;
	if (errno || strcmp(buf,message(807))!=0)
	  goto wrongmlx;
      }
      { /* then, that it has a valid checksum */
	Checksum cs;

	readfile(cs,sizeof(Checksum),(char *)module);
	if (errno || GetChecksum(cs)==(struct cslist *)0L)
	  goto wrongmlx;
      }
      {
	if (!skipOpenMods(module))
	  goto wrongmlx;
      }
      return module;
wrongmlx:
      MARKFREE(*realname);
      hfclose(module);
    }
    {
      *basisp = NIL;
    }
  return NULL; /* don't open the file, we've got the basis already */
}

PUBLIC ExprPtr moduleLoad(zone,name,recompile)
IN CONST TZone *zone;
IN CONST CHAR *name;
IN Boolean recompile;
{
  hFILE *new;
  CHAR *realname;
  ExprPtr basis;

  new = openModule(name,(char **)&realname,&basis,recompile);
  if (new==NULL)
    return basis;
  basis = inTransfer(readfile,(char *)new);
  if (errno)
    {
      CHAR *tmp = (CHAR *)alloctmp(128+strlen((char *)realname));

      (void) hfclose(new);
      sprintf((char *)tmp,message(811),realname,errno,ioerrormsg(errno));
      type_error(zone,tmp);
    }
  (void) hfclose(new);
  return basis;
}

PUBLIC struct cslist *CurrentChecksum()
{
  struct cslist *aux;

  while (Checksums->age>currentAge)
    {
      aux = Checksums->next;
      MARKFREE(Checksums);
      Checksums = aux;
    }
  return Checksums;
}

PUBLIC struct cslist *GetChecksum(cs)
IN CONST Checksum cs;
{
  struct cslist *csl;
  REGISTER COUNTER i;
  REGISTER CHAR *s,*t;

  for (csl = Checksums; csl!=NULL; csl = csl->next)
    {
      for (i=0, s=csl->cs, t=cs; i<CHECKSUM_WIDTH; i++)
	if (*s++ != *t++)
	  break;
      if (i==CHECKSUM_WIDTH)
	break;
    }
  return csl;
}

PUBLIC void StoreInitialBasis()
{
  struct cslist *new;
  ExprPtr assum,state,table,e;
  ENTER;

  new = (struct cslist *)MALLOC(sizeof(struct cslist));
  new->next = Checksums;
  new->age = currentAge;
  new->basis = basisNew(true);
  assum = Cons(MapRemove(itSym,A_IDS(basisAssum)),
	       Cons(A_TYFCNS(basisAssum),
#if MODULE_SYSTEM & STD_ML_MODULE
		    Cons(A_STRUCTS(basisAssum),
			 Cons(A_SIGS(basisAssum),
#endif
			      Cons(A_UNITS(basisAssum),
				   NIL)
#if MODULE_SYSTEM & STD_ML_MODULE
			      ))
#endif
		    ));
  state = Cons(S_NONEQS(basisState),
	       Cons(S_TYCONS(basisState),
		    NIL));
  table = Cons(MapRemove(itSym,C_IDS(basisTable)),
#if MODULE_SYSTEM & STD_ML_MODULE
	       Cons(C_STRUCTS(basisTable),
		    Cons(C_SIGS(basisTable),
#endif
			 Cons(C_CONS(basisTable),
			      NIL)
#if MODULE_SYSTEM & STD_ML_MODULE
			 ))
#endif
	       );
  /* we remove 'it' from the computation of the checksum, as its
     type can vary, although this should have no influence on the
     checksum, since modules are not allowed to depend on the value
     of it. */
  e = Cons(assum,Cons(state,table));
  CheckSum(e,new->cs);
  if (GetChecksum(new->cs)==NULL) /* never record twice the same initial basis */
    Checksums = new;
  else
    {
      MARKFREE(new);
    }
  QUIT;
}

DEF_ACTION_0ARG(stampworld)
{
  StoreInitialBasis();
  return tuple0;
}
ACTION_END

/***** The following are functions that provide a reasonably secure
  hash function for structures in memory (i.e, a reliable checksum).
  This is needed to ensure that things we saved to disk in some environment
  can be reloaded later on, by checking that the environments are the
  same with probability close to 1. *****/

/* The first trick is to encode couples (a,b):
   Let M be a prime, and c, d positive numbers.
   Assume we have checksums for a and b as integers, then we compute
   <a,b> = c*a+b+d mod M. As M is prime, the chance that <a,b>=<a',b'>
   is 1/M. Indeed, this is the chance that cx+y=0 mod M for any two
   x and y; for each x, this is the chance that y = -cx mod M, i.e. 1/M.
   We then choose several different values for c and d (CHECKSUM_WIDTH values),
   which we choose relatively prime, so that the chance of collision
   decreases again.
*/

PRIVATE unsigned short multipliers[] = { 71, 13, 97, 7, 11, 59, 37, 61 };
PRIVATE unsigned short adders[] = { 17, 31, 89, 3, 53, 47, 5, 23 };

/*
   PUBLIC void checksumCouple1(cs1,cs2,cs)
   IN CONST Checksum *cs1,*cs2;
   OUT Checksum *cs;
   {
   cs->line1 = (9821L*(unsigned long)cs1->line1+(unsigned long)cs2->line1+13849L) & 0xFFFFL;
   cs->line2 = (17341L*(unsigned long)cs1->line2+(unsigned long)cs2->line2+38653L) & 0xFFFFL;
   cs->pos1 = (8827L*(unsigned long)cs1->pos1+(unsigned long)cs2->pos1+14279L) & 0xFFFFL;
   cs->pos2 = (18563L*(unsigned long)cs1->pos2+(unsigned long)cs2->pos2+13443L) & 0xFFFFL;
   }
*/

PRINTFUNC(checksumPrint)
{
  struct checksum *data = PRINTDATA(struct checksum *);
  REGISTER COUNTER i,j;
  REGISTER CHAR *cs;
  REGISTER unsigned short *mul,*add;
  REGISTER unsigned short c;

  for (i=0; i<len; i++)
    {
      c = (unsigned short)*text++;
      for (j = 0, cs = data->cs, mul = multipliers, add = adders; j<CHECKSUM_WIDTH; j++)
	{
	  *cs = (*mul++)*((unsigned short)(*cs))+c+(*add++);
	  cs++;
	}
    }
}

PUBLIC void CheckSum(e,cs)
IN CONST ExprPtr e;
OUT Checksum cs;
{
  struct checksum *data = (struct checksum *)alloctmp(sizeof(struct checksum));
  COUNTER i;

  for (i=0; i<CHECKSUM_WIDTH; i++)
    data->cs[i] = 0;
  outTransfer(e,checksumPrint,(char *)data,true);
  memcpy(cs,data->cs,sizeof(Checksum));
}

#undef __SEG__
#define __SEG__ INIT

PRIVATE ExprPtr S(i)
IN INT i;
{
  return Kstr((CHAR *)message(i));
}

PUBLIC void InitLex()
{
  ExprPtr aux;
  ENTER;

  lexEnv=consEnv=NIL;
#if MODULE_SYSTEM & STD_ML_MODULE
  strEnv=sigEnv=NIL;
#endif
  keepLexEnv=keepConsEnv=NIL;
#if MODULE_SYSTEM & STD_ML_MODULE
  keepStrEnv=keepSigEnv=NIL;
#endif
  lexId=lexPath=NIL;
  pushEnv();
  setConstructor(nilSym);
  setConstructor(consSym);
  setConstructor(refsym);
  setConstructor(trueSym);
  setConstructor(falseSym);
  setConstructor(bindSym);
  setConstructor(matchSym);
  setConstructor(emptySym);
  setConstructor(mapgetSym);
  setConstructor(parsweepSym);
  setConstructor(nomemSym);
  setConstructor(stackSym);
  setConstructor(incompSym);
  setConstructor(nthSym);
  setConstructor(rangeSym);
  setConstructor(asciiSym);
  setConstructor(stringnthSym);
  setConstructor(noninvSym);
  setConstructor(nonsenseSym);
  setConstructor(futureSym);
  setConstructor(catchSym);
  setConstructor(subscriptSym);
  setConstructor(arithSym);
  setConstructor(ioSym);
  setConstructor(reSym);
  setConstructor(ufindSym);
/*
infix   0 before
infix   3 o := <<< >>>
infix   4 = <> < > <= >= inset inmap submap subset strless
infixr  5 @ (* infix in ML *) O
infixr  5 ::
infix   6 + - ^ ++ U <| <-| |> |->
infix   7 * div mod divmod / & \ delta intersects
infixr  8 **
infix   9 nth to sub
*/
#define I(i) Kint(i)
  aux = MapMerge(MapMerge(MapPair(S(359),I(1), /* infix 0 before */
				  S(301),I(4)), /* infix 3 o */
			  MapPair(S(343),I(4), /* infix 3 := */
				  S(276),I(5))), /* infix 4 = */
		 MapMerge(MapPair(S(277),I(5), /* infix 4 <> */
				  S(579),I(5)), /* infix 4 < */
			  MapPair(S(580),I(5), /* infix 4 > */
				  S(581),I(5)))); /* infix 4 <= */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(582),I(5), /* infix 4 >= */
				  S(328),I(5)), /* infix 4 inset */
			  MapPair(S(329),I(5), /* infix 4 inmap */
				  S(327),I(5)))); /* infix 4 submap */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(326),I(5), /* infix 4 subset */
				  S(358),I(5)), /* infix 4 strless */
			  MapPair(S(306),I(-6), /* infixr 5 @ */
				  S(302),I(-6)))); /* infixr 5 O */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(193),I(-6), /* infixr 5 :: */
				  S(583),I(7)), /* infix 6 + */
			  MapPair(S(584),I(7), /* infix 6 - */
				  S(307),I(7)))); /* infix 6 ^ */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(217),I(7), /* infix 6 ++ */
				  S(223),I(7)), /* infix 6 U */
			  MapPair(S(318),I(7), /* infix 6 <| */
				  S(319),I(7)))); /* infix 6 <-| */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(320),I(7), /* infix 6 |> */
				  S(321),I(7)), /* infix 6 |-> */
			  MapPair(S(222),I(8), /* infix 7 * */
				  S(585),I(8)))); /* infix 7 div */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(586),I(8), /* infix 7 mod */
				  S(587),I(8)), /* infix 7 divmod */
			  MapPair(S(266),I(8), /* infix 7 #/ */
				  S(323),I(8)))); /* infix 7 & */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(324),I(8), /* infix 7 \ */
				  S(325),I(8)), /* infix 7 delta */
			  MapPair(S(622),I(8), /* infix 7 intersects */
				  S(225),I(-9)))); /* infixr 8 #^ */
  aux = MapMerge(aux,
		 MapPair(S(335),I(10), /* infix 9 nth */
			 S(337),I(10))); /* infix 9 to */
			  /* MapSingleton(S(354),I(10)))); // infix 9 sub */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(263),I(7), /* infix 6 #+ */
				  S(264),I(7)), /* infix 6 #- */
			  MapPair(S(265),I(8), /* infix 7 #* */
				  S(315),I(8)))); /* infix 7 #div */
  aux = MapMerge(aux,
		 MapPair(S(316),I(8), /* infix 7 #mod */
			 S(317),I(8))); /* infix 7 #divmod */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(278),I(5), /* infix 4 #< */
				  S(279),I(5)), /* infix 4 #> */
			  MapPair(S(280),I(5), /* infix 4 #<= */
				  S(281),I(5)))); /* infix 4 #>= */
  setcarMapMerge(lexEnv,aux);
  /*theFileStack = NULL;*/
  yyfilename = NULL;
  QUIT;
}

PUBLIC void InitGram()
{
  parseMarker=NIL;
  parseMarkerStack=NIL;
  toplevelCont = NIL;
  ItSymbol = itSym;
  registerExtDesc(&basisDesc);
  theTypes = NIL;
  theBindings = NIL;
}
