/* Hlex scanner engine.
   Copyright (C) 2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#else
#include <errno.h>
#endif
#include "portable.h"
#include "eval.h"
#include "mem.h"
#include "expr.h"
#include "transfer.h"
#include "hlexproto.h"
#include "rhlex.h"
#ifdef HAS_PROTOTYPES
#include "shlex.h"
#endif

#define __SEG__ HLEX

PUBLIC ExprPtr flexErrExc;

EXTERN SymPtr readsym, readlinesym, stdinval;
EXTERN ExprPtr none,optttm;
EXTERN extDesc iarrayExtDesc;
EXTERN ExprPtr voidString;

/* Returned upon end-of-file. */
#define YY_NULL 0

/* Promotes a possibly negative, possibly signed char to an unsigned
 * integer for use as an array index.  If the signed char is negative,
 * we want to instead treat it as an 8-bit unsigned char, hence the
 * double cast.
 */
#define YY_SC_TO_UI(c) ((unsigned int) (unsigned char) c)

/* Enter a start condition.  This macro really ought to take a parameter,
 * but we do it the disgusting crufty way forced on us by the ()-less
 * definition of BEGIN.
 */
/*
#define BEGIN yy_start = 1 + 2 *
*/

PRIVATE void yyBegin (yyd, state)
INOUT struct yydata *yyd;
IN COUNTER state;
{
  yyd->yy_start = 1 + 2 * state;
}

DEF_ACTION_2ARGS(hbegin) /* hbegin : yydata * int -> unit */
{
  yyBegin ((struct yydata *)EXTDATA (arg1), IVAL (arg2));
  return tuple0;
}
ACTION_END

/* Translate the current start state into a value that can be later handed
 * to BEGIN to return to the state.
 */
/*
#define YY_START ((yy_start - 1) / 2)
*/

PRIVATE COUNTER yyStart (yyd)
IN CONST struct yydata *yyd;
{
  return (yyd->yy_start-1) >> 1;
}

DEF_ACTION_1ARG(hstart) /* hstart : yydata -> int */
{
  INT state = (INT)yyStart ((struct yydata *)EXTDATA (arg));
  return Kint (state);
}
ACTION_END

/* Action number for EOF rule of a given start state. */
/*
#define YY_STATE_EOF(state) (YY_END_OF_BUFFER + state + 1)
*/

PRIVATE COUNTER yyStateEof (yytbl, state)
IN CONST struct yytables *yytbl;
IN COUNTER state;
{
  return yytbl->yy_end_of_buffer+state+1;
}

/* Special action meaning "start processing a new file". */
#define YY_NEW_FILE(yydx) yyrestart (yydx, yyd->yyread, yyd->yyreadln)

#define YY_END_OF_BUFFER_CHAR 0

/* Size of default input buffer. */
#define YY_BUF_SIZE 16384

#define EOB_ACT_CONTINUE_SCAN 0
#define EOB_ACT_END_OF_FILE 1
#define EOB_ACT_LAST_MATCH 2

#define unput(yydx, c) yyunput (yydx, c, yyd->yytext)

/* We provide macros for accessing buffer states in case in the
 * future we want to put the buffer states in a more general
 * "scanner state".
 */
#define YY_CURRENT_BUFFER ((YY_BUFFER_STATE)EXTDATA(yyd->yy_cbuf))


/* Flag which is used to allow yywrap()'s to do buffer switches
 * instead of setting up a fresh yyd->yyin.  A bit of a hack ...
 */

#define YY_FLUSH_BUFFER yy_flush_buffer (yyd, YY_CURRENT_BUFFER)

#define yy_set_interactive(yydx,yyd,is_interactive) \
{ \
  if (yyd->yy_cbuf==NIL) \
    { \
      yyd->yy_cbuf = build_buffer (yyd, yyd->yyread, yyd->yyreadln, YY_BUF_SIZE ); \
      HITBARRIER (yydx); \
    } \
  if (is_interactive) \
    YY_CURRENT_BUFFER->yy_flags |= YY_IS_INTERACTIVE_MASK; \
  else YY_CURRENT_BUFFER->yy_flags &= ~YY_IS_INTERACTIVE_MASK; \
}

DEF_ACTION_2ARGS(hsetinteractive) /* hsetinteractive : yydata * bool -> unit */
{
  ExprPtr yydx = arg1;
  struct yydata *yyd = (struct yydata *)EXTDATA (yydx);

  yy_set_interactive (yydx, yyd, arg2==trueValue);
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(hisinteractive) /* hisinteractive : yydata -> bool */
{
  struct yydata *yyd = (struct yydata *)EXTDATA (arg);

  if (yyd->yy_cbuf==NIL)
    {
      yyd->yy_cbuf = build_buffer (yyd, yyd->yyread, yyd->yyreadln, YY_BUF_SIZE );
      HITBARRIER (arg);
    }
  return (YY_CURRENT_BUFFER->yy_flags & YY_IS_INTERACTIVE_MASK)?trueValue:falseValue;
}
ACTION_END

#define yy_set_bol(yydx,yyd,at_bol) \
{ \
  if (yyd->yy_cbuf==NIL) \
    { \
      yyd->yy_cbuf = build_buffer (yyd, yyd->yyread, yyd->yyreadln, YY_BUF_SIZE ); \
      HITBARRIER (yydx); \
    } \
  if (at_bol) \
    YY_CURRENT_BUFFER->yy_flags |= YY_AT_BOL_MASK; \
  else YY_CURRENT_BUFFER->yy_flags &= ~YY_AT_BOL_MASK; \
}

DEF_ACTION_2ARGS(hsetbol) /* hsetbol : yydata * bool -> unit */
{
  ExprPtr yydx = arg1;
  struct yydata *yyd = (struct yydata *)EXTDATA (yydx);

  yy_set_bol (yydx, yyd, arg2==trueValue);
  return tuple0;
}
ACTION_END

#define YY_AT_BOL(yyd) ((((YY_BUFFER_STATE)EXTDATA((yyd)->yy_cbuf))->yy_flags & YY_AT_BOL_MASK)!=0)

DEF_ACTION_1ARG(hatbol) /* hatbol : yydata -> bool */
{
  struct yydata *yyd = (struct yydata *)EXTDATA (arg);

  return YY_AT_BOL(yyd)?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_1ARG(htext) /* : yydata -> string */
{
  struct yydata *yyd = (struct yydata *)EXTDATA(arg);

  return Kstring (yyd->yyleng, (CHAR *)yyd->yytext);
}
ACTION_END

/* Done after the current pattern has been matched and before the
 * corresponding action - sets up yytext.
 */
#define YY_DO_BEFORE_ACTION \
	yyd->yytext = yy_bp; \
	yyd->yyleng = (INT) (yy_cp - yy_bp); \
	yyd->yy_hold_char = *yy_cp; \
	*yy_cp = '\0'; \
	yyd->yy_c_buf_p = yy_cp;

#define YY_MORE_ADJ 0
#define YY_RESTORE_YY_MORE_OFFSET

#define YY_ALWAYS_INTERACTIVE 1

/* Amount of stuff to slurp up with each read. */
#ifndef YY_READ_BUF_SIZE
#define YY_READ_BUF_SIZE 8192
#endif

/* Copy whatever the last rule matched to the standard output. */

#ifdef OBSOLETE
/* Gets input and stuffs it into "buf".  number of characters read, or YY_NULL,
 * is returned in "result".
 */
#ifndef YY_INPUT
#define YY_INPUT(buf,result,max_size) \
{ \
  ExprPtr str; \
  ENTER; \
  str = (YY_CURRENT_BUFFER->yy_flags & YY_IS_INTERACTIVE_MASK)? \
    (*ACTION (yyd->yyreadln)) (ENV_CONS(QENV (yyd->yyreadln),0)):  \
    (*ACTION (yyd->yyread)) (ENV_CONS(QENV (yyd->yyread),1), KINT (max_size));	   \
  safememcpy (buf, STEXT(str), result = SLEN(str)); \
  QUIT; \
}
#endif
#endif

/*
#ifndef yyterminate
#define yyterminate() return YY_NULL
#endif
*/

/*
PRIVATE ExprPtr yyterminate ()
{
  ExprPtr arg;

  arg = Kdata (0, optttm, izero); // SOME 0
  return arg;
}
*/

/* Number of entries by which start-condition stack grows. */
#ifndef YY_START_STACK_INCR
#define YY_START_STACK_INCR 25
#endif

/* Report a fatal error. */
/*
#ifndef YY_FATAL_ERROR
#define YY_FATAL_ERROR(msg) yy_fatal_error (msg)
#endif
*/

/* Default declaration of generated scanner - a define so the user can
 * easily add parameters.
 */
#ifndef YY_DECL
#define YY_DECL 
#endif

/* Code executed at the beginning of each rule, after yyd->yytext and yyd->yyleng
 * have been set up.
 */
#define YY_USER_ACTION YY_UPDATE_YYLLOC(ip[0] = ip[2]; ip[1] = ip[3];)
#define YY_UPDATE_YYLLOC(init) { /* update yylloc */ \
  REGISTER INT *ip = 1+(INT *)EXTDATA (yyd->yylloc); \
  REGISTER char *p = yyd->yytext; \
  REGISTER char *pend = p + yyd->yyleng; \
  init \
  for (; p<pend; ) \
    switch (*p++) \
      { \
      case '\n': ip[2]++; /* last_line++; */ ip[3] = 0; /* last_column = 0; */ break; \
      default: ip[3]++; /* last_column++; */ break; \
      } \
}

/* Code executed at the end of each rule. */
#ifndef YY_BREAK
#define YY_BREAK break;
#endif

#define YY_RULE_SETUP \
if (yyd->yyleng > 0) \
{ \
  if (yyd->yytext[yyd->yyleng - 1] == '\n') \
    YY_CURRENT_BUFFER->yy_flags |= YY_AT_BOL_MASK; \
  else YY_CURRENT_BUFFER->yy_flags &= ~YY_AT_BOL_MASK; \
  YY_USER_ACTION \
}

PRIVATE void fillWithStream (stream, readp, readlnp)
IN CONST ExprPtr stream;
OUT ExprPtr *readp, *readlnp;
{
  ExprPtr idx, shape;
  VecPtr tbl;

  shape = ASHAPE (stream);
  tbl = AVAL (stream);
  idx = MapGetSure (readsym, shape);
  *readp = tbl->expr[IVAL(idx)-1];
  idx = MapGetSure (readlinesym, shape);
  *readlnp = tbl->expr[IVAL(idx)-1];
}

PRIVATE ExprPtr yylex (yytbl, yydx)
IN CONST struct yytables *yytbl;
INOUT ExprPtr yydx;
{
  register yy_state_type yy_current_state;
  register char *yy_cp, *yy_bp;
  register INT yy_act;
  register struct yydata *yyd = (struct yydata *)EXTDATA(yydx);
  register INT yy_act_n = AVAL(yytbl->yy_action_table)->n;

  if (yyd->yy_init & 1)
    {
      yyd->yy_init &= ~1;
#ifdef YY_USER_INIT
      YY_USER_INIT;
#endif
      if (!yyd->yy_start)
	yyd->yy_start = 1;	/* first start state */
      if (yyd->yyread==NOEXPR)
	fillWithStream (stdinval, &yyd->yyread, &yyd->yyreadln);
      if (yyd->yy_cbuf==NIL)
	{
	  yyd->yy_cbuf = build_buffer (yyd, yyd->yyread, yyd->yyreadln, YY_BUF_SIZE);
	  HITBARRIER (yydx);
	}
      yy_load_buffer_state (yyd);
    }

  if (yyd->yy_init & 2) /* we have left yylex() by abnormal means
			   (an exception was raised, typically). */
    {
      yyrestart (yydx, yyd->yyread, yyd->yyreadln);
    }
  yyd->yy_init |= 2; /* try to catch abnormal returns. */

  while (true)		/* loops until end-of-file is reached */
    {
      yy_cp = yyd->yy_c_buf_p;
      /* Support of yyd->yytext. */
      *yy_cp = yyd->yy_hold_char;
      /* yy_bp points to the position in yy_ch_buf of the start of
       * the current run.
       */
      yy_bp = yy_cp;
      yy_current_state = yyd->yy_start;
      yy_current_state += YY_AT_BOL(yyd);
    yy_match:
      do
	{
	  register YY_CHAR yy_c = yytbl->yy_ec[YY_SC_TO_UI(*yy_cp)];
	  if (yytbl->yy_accept[yy_current_state])
	    {
	      yyd->yy_last_accepting_state = yy_current_state;
	      yyd->yy_last_accepting_cpos = yy_cp;
	    }
	  while (yytbl->yy_chk[yytbl->yy_base[yy_current_state] + yy_c] != yy_current_state)
	    {
	      yy_current_state = (INT) yytbl->yy_def[yy_current_state];

	      if (yy_current_state >= yytbl->yy_accept_n)
		yy_c = yytbl->yy_meta[(unsigned int) yy_c];
	    }
	  yy_current_state = yytbl->yy_nxt[yytbl->yy_base[yy_current_state] + (unsigned int) yy_c];
	  ++yy_cp;
	}
      while (yytbl->yy_base[yy_current_state] != yytbl->yy_jambase);
    yy_find_action:
      yy_act = yytbl->yy_accept[yy_current_state];
      if (yy_act == 0)
	{ /* have to back up */
	  yy_cp = yyd->yy_last_accepting_cpos;
	  yy_current_state = yyd->yy_last_accepting_state;
	  yy_act = yytbl->yy_accept[yy_current_state];
	}

      YY_DO_BEFORE_ACTION;

    do_action:	/* This label is used only to access EOF actions. */
      if (yy_act==0)
	{ /* must back up */
	  /* undo the effects of YY_DO_BEFORE_ACTION */
	  *yy_cp = yyd->yy_hold_char;
	  yy_cp = yyd->yy_last_accepting_cpos;
	  yy_current_state = yyd->yy_last_accepting_state;
	  goto yy_find_action;
	}
      if (yy_act==yytbl->yy_end_of_buffer)
	{
	  /* Amount of text matched not including the EOB char. */
	  INT yy_amount_of_matched_text = (INT) (yy_cp - yyd->yytext) - 1;

	  /* Undo the effects of YY_DO_BEFORE_ACTION. */
	  *yy_cp = yyd->yy_hold_char;
	  YY_RESTORE_YY_MORE_OFFSET

	    if ((YY_CURRENT_BUFFER->yy_flags & YY_BUFFER_STATUS_MASK) == YY_BUFFER_NEW)
	      {
		/* We're scanning a new file or input source.  It's
		 * possible that this happened because the user
		 * just pointed yyd->yyin at a new source and called
		 * yylex().  If so, then we have to assure
		 * consistency between YY_CURRENT_BUFFER and our
		 * globals.  Here is the right place to do so, because
		 * this is the first action (other than possibly a
		 * back-up) that will match for the new input source.
		 */
		YY_BUFFER_STATE b = YY_CURRENT_BUFFER;

		yyd->yy_n_chars = b->yy_n_chars;
		b->yyread = yyd->yyread;
		b->yyreadln = yyd->yyreadln;
		b->yy_flags = (b->yy_flags & ~YY_BUFFER_STATUS_MASK) | YY_BUFFER_NORMAL;
	      }

	  /* Note that here we test for yyd->yy_c_buf_p "<=" to the position
	   * of the first EOB in the buffer, since yyd->yy_c_buf_p will
	   * already have been incremented past the NUL character
	   * (since all states make transitions on EOB to the
	   * end-of-buffer state).  Contrast this with the test
	   * in input().
	   */
	  if ( yyd->yy_c_buf_p <= &YY_CURRENT_BUFFER->yy_ch_buf[yyd->yy_n_chars] )
	    { /* This was really a NUL. */
	      yy_state_type yy_next_state;

	      yyd->yy_c_buf_p = yyd->yytext + yy_amount_of_matched_text;

	      yy_current_state = yy_get_previous_state (yytbl, yyd);

	      /* Okay, we're now positioned to make the NUL
	       * transition.  We couldn't have
	       * yy_get_previous_state() go ahead and do it
	       * for us because it doesn't know how to deal
	       * with the possibility of jamming (and we don't
	       * want to build jamming into it because then it
	       * will run more slowly).
	       */

	      yy_next_state = yy_try_NUL_trans (yyd, yytbl, yy_current_state);

	      yy_bp = yyd->yytext + YY_MORE_ADJ;

	      if (yy_next_state)
		{
		  /* Consume the NUL. */
		  yy_cp = ++yyd->yy_c_buf_p;
		  yy_current_state = yy_next_state;
		  goto yy_match;
		}
	      else
		{
		  yy_cp = yyd->yy_c_buf_p;
		  goto yy_find_action;
		}
	    }
	  else switch (yy_get_next_buffer (yydx))
	    {
	    case EOB_ACT_END_OF_FILE:
	      {
		yyd->yy_did_buffer_switch_on_eof = 0;

		if ((*ACTION (yyd->yywrap)) (ENV_CONS(QENV (yyd->yywrap),0))==trueValue)
		  {
		    /* Note: because we've taken care in
		     * yy_get_next_buffer() to have set up
		     * yyd->yytext, we can now set up
		     * yyd->yy_c_buf_p so that if some total
		     * hoser (like flex itself) wants to
		     * call the scanner after we return the
		     * YY_NULL, it'll still work - another
		     * YY_NULL will get returned.
		     */
		    yyd->yy_c_buf_p = yyd->yytext + YY_MORE_ADJ;
		    yy_act = yyStateEof (yytbl, yyStart (yyd));
		    goto do_action;
		  }
		else
		  {
		    if (!yyd->yy_did_buffer_switch_on_eof)
		      YY_NEW_FILE (yydx);
		  }
		break;
	      }
	    case EOB_ACT_CONTINUE_SCAN:
	      yyd->yy_c_buf_p = yyd->yytext + yy_amount_of_matched_text;
	      yy_current_state = yy_get_previous_state (yytbl, yyd);
	      yy_cp = yyd->yy_c_buf_p;
	      yy_bp = yyd->yytext + YY_MORE_ADJ;
	      goto yy_match;
	    case EOB_ACT_LAST_MATCH:
	      yyd->yy_c_buf_p = &YY_CURRENT_BUFFER->yy_ch_buf[yyd->yy_n_chars];
	      yy_current_state = yy_get_previous_state (yytbl, yyd);
	      yy_cp = yyd->yy_c_buf_p;
	      yy_bp = yyd->yytext + YY_MORE_ADJ;
	      goto yy_find_action;
	    }
	}
      else if (yy_act<yy_act_n)
	{
	  YY_RULE_SETUP
	    {
	      ExprPtr act = AVAL(yytbl->yy_action_table)->expr[yy_act];
	      ExprPtr res = (*ACTION (act)) (ENV_CONS(QENV (act),1), yydx);
	      if (res->magic==0) /* SOME token */
		{
		  yyd->yy_init &= ~2; /* this is a normal return. */
		  return CDR (res); /* return token */
		}
	    }
	  /*YY_BREAK*/
	}
      else
	{
	  ExprPtr e = MakePair (E_CONS, flexErrExc, izero);
	  (void) KEEP (e);
	  (void) Raise (e);
	  /*YY_FATAL_ERROR("fatal flex scanner internal error--no action found" );*/
	} /* end of action switch */
    } /* end of scanning one token */
} /* end of yylex */

PRIVATE INT align_int (n)
INT n;
{
  return (INT)(((n+sizeof(INT)-1)/sizeof(INT))*sizeof(INT));
}

PRIVATE INT yy_table_size (naccept, nec, nmeta, nnul, nbase, ndef, nnxt, nchk, intstartp)
INT naccept, nec, nmeta, nnul, nbase, ndef, nnxt, nchk;
OUT INT *intstartp;
{
  INT size;

  size = sizeof (struct yytables);
  size += (naccept+nbase+ndef+nnxt+nchk) * sizeof(SHORT);
  size = align_int (size);
  *intstartp = size;
  size += (nec+nmeta+nnul) * sizeof(INT);
  return size;
}

PRIVATE SHORT *list_to_shortp (l, p)
IN CONST ExprPtr l;
IN CONST SHORT *p;
{
  while (ISCONS (l))
    {
      *p++ = (SHORT)(IVAL (CAR (l)));
      l = CDR (l);
    }
  return p;
}

PRIVATE INT *list_to_intp (l, p)
IN CONST ExprPtr l;
IN CONST INT *p;
{
  while (ISCONS (l))
    {
      *p++ = (INT)(IVAL (CAR (l)));
      l = CDR (l);
    }
  return p;
}

PRIVATE void yytblSweep (extdata, doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  struct yytables *yytbl;

  yytbl = (struct yytables *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      {
	outInt (yytbl->yy_accept_n);
	if (errno) return;
	outInt (yytbl->yy_ec_n);
	if (errno) return;
	outInt (yytbl->yy_meta_n);
	if (errno) return;
	outInt (yytbl->yy_nultrans_n);
	if (errno) return;
	outInt (yytbl->yy_base_n);
	if (errno) return;
	outInt (yytbl->yy_def_n);
	if (errno) return;
	outInt (yytbl->yy_nxt_n);
	if (errno) return;
	outInt (yytbl->yy_chk_n);
	if (errno) return;
	outInt ((INT)yytbl->yy_num_rules);
	if (errno) return;
	outInt ((INT)yytbl->yy_end_of_buffer);
	if (errno) return;
	outInt (yytbl->yy_jambase);
	if (errno) return;
	outInt (yytbl->yy_jamstate);
	if (errno) return;
	outInt (yytbl->yy_NUL_ec);
	if (errno) return;
	/* all SHORT *'s are assumed to be contiguous (see yy_build_tables (),
	   yytblLoad()); same thing for INT *'s. */
	outShorts (yytbl->yy_accept, yytbl->yy_accept_n + yytbl->yy_base_n
		   + yytbl->yy_def_n + yytbl->yy_nxt_n + yytbl->yy_chk_n);
	if (errno) return;
	outInts (yytbl->yy_ec, yytbl->yy_ec_n + yytbl->yy_meta_n + yytbl->yy_nultrans_n);
	if (errno) return;
      }
    /*FALLTHROUGH*/
    case X_MARK:
    case X_MARKDATA:
      (*doext->action) (yytbl->yy_action_table);
      break;
    case X_FREE:
      MARKFREE (extdata);
      break;
    }
}

PRIVATE ExprPtr yytblLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  struct yytables *yytbl;
  INT naccept, nec, nmeta, nnul, nbase, ndef, nnxt, nchk, size, intstart;
  ExprPtr e;

  naccept = inInt ();
  if (errno) return NIL;
  nec = inInt ();
  if (errno) return NIL;
  nmeta = inInt ();
  if (errno) return NIL;
  nnul = inInt ();
  if (errno) return NIL;
  nbase = inInt ();
  if (errno) return NIL;
  ndef = inInt ();
  if (errno) return NIL;
  nnxt = inInt ();
  if (errno) return NIL;
  nchk = inInt ();
  if (errno) return NIL;
  size = yy_table_size (naccept, nec, nmeta, nnul, nbase, ndef, nnxt, nchk, &intstart);
  yytbl = (struct yytables *)MALLOC (size);
  yytbl->yy_accept_n = naccept;
  yytbl->yy_ec_n = nec;
  yytbl->yy_meta_n = nmeta;
  yytbl->yy_nultrans_n = nnul;
  yytbl->yy_base_n = nbase;
  yytbl->yy_def_n = ndef;
  yytbl->yy_nxt_n = nnxt;
  yytbl->yy_chk_n = nchk;
  yytbl->yy_num_rules = (COUNTER) inInt ();
  if (errno)
    {
    err:
      DISPOSPTR (yytbl, size);
      return NIL;
    }
  yytbl->yy_end_of_buffer = (COUNTER) inInt ();
  if (errno) goto err;
  yytbl->yy_jambase = inInt ();
  if (errno) goto err;
  yytbl->yy_jamstate = inInt ();
  if (errno) goto err;
  yytbl->yy_NUL_ec = inInt ();
  if (errno) goto err;
  yytbl->yy_accept = (SHORT *)(((char *)yytbl)+sizeof(struct yytables));
  inShorts (yytbl->yy_accept, naccept + nbase + ndef + nnxt + nchk);
  if (errno) goto err;
  yytbl->yy_base = yytbl->yy_accept + naccept;
  yytbl->yy_def = yytbl->yy_base + nbase;
  yytbl->yy_nxt = yytbl->yy_def + ndef;
  yytbl->yy_chk = yytbl->yy_nxt + nnxt;
  yytbl->yy_ec = (INT *)(((char *)yytbl)+intstart);
  inInts (yytbl->yy_ec, nec + nmeta + nnul);
  if (errno) goto err;
  yytbl->yy_meta = yytbl->yy_ec + nec;
  yytbl->yy_NUL_trans = yytbl->yy_meta + nmeta;
  yytbl->yy_action_table = (*load) (sharedref);
  if (errno) goto err;
  e = MakeExternal (desc, (char *)yytbl);
  (void) KEEP(e);
  return e;
}

PUBLIC extDesc yytblExtDesc = {
  (CHAR *)"flex-tables",
  { 'f','l','x','t' }, NULL,
  NULL, /* no hash function, yytables are not shared */
  NULL, /* no comparison function, yytables are not shared */
  NULL, /* no print function */
  yytblSweep,
  yytblLoad
};

PRIVATE ExprPtr yy_build_tables (arg)
IN CONST ExprPtr arg;
/* arg : (yydata -> 'token option) list * int * int * int list * int list * int list
         (0)actions                    (1)#rules (2)eob (3)accept (4)ec      (5)meta

	 * int list * int list * int list * int list * int list * int * int * int
	 (6)base      (7)def     (8)nxt     (9)chk    (10)NUL_trans (11)jambase (12)jamstate (13)NUL_ec
    -> 'token yytables
 */
{
  VecPtr vp;
  ExprPtr accept, ec, meta, base, def, nxt, chk, nultrans;
  INT naccept, nec, nmeta, nbase, ndef, nnxt, nchk, nnul;
  INT size, intstart;
  void *p;
  struct yytables *yytbl;
  ExprPtr e, actions;
  ENTER;

  vp = AVAL(arg);
  actions = ListToArray0 (vp->expr[0]);
  (void) KEEP (actions);
  vp = AVAL(arg);
  accept = vp->expr[3];
  ec = vp->expr[4];
  meta = vp->expr[5];
  base = vp->expr[6];
  def = vp->expr[7];
  nxt = vp->expr[8];
  chk = vp->expr[9];
  nultrans = vp->expr[10];
  naccept = ListLength (accept);
  nec = ListLength (ec);
  nmeta = ListLength (meta);
  nnul = ListLength (nultrans);
  nbase = ListLength (base);
  ndef = ListLength (def);
  nnxt = ListLength (nxt);
  nchk = ListLength (chk);
  size = yy_table_size (naccept, nec, nmeta, nnul, nbase, ndef, nnxt, nchk, &intstart);
  yytbl = (struct yytables *)MALLOC (size);
  vp = AVAL(arg);
  yytbl->yy_action_table = actions;
  yytbl->yy_num_rules = IVAL (vp->expr[1]);
  yytbl->yy_end_of_buffer = IVAL (vp->expr[2]);
  p = (void *)(((char *)yytbl)+sizeof(struct yytables));
  yytbl->yy_accept_n = naccept;
  p = list_to_shortp (accept, yytbl->yy_accept = (SHORT *)p);
  yytbl->yy_base_n = nbase;
  p = list_to_shortp (base, yytbl->yy_base = (SHORT *)p);
  yytbl->yy_def_n = ndef;
  p = list_to_shortp (def, yytbl->yy_def = (SHORT *)p);
  yytbl->yy_nxt_n = nnxt;
  p = list_to_shortp (nxt, yytbl->yy_nxt = (SHORT *)p);
  yytbl->yy_chk_n = nchk;
  (void) list_to_shortp (chk, yytbl->yy_chk = (SHORT *)p);
  p = (void *)(((char *)yytbl)+intstart);
  yytbl->yy_ec_n = nec;
  p = list_to_intp (ec, yytbl->yy_ec = (INT *)p);
  yytbl->yy_meta_n = nmeta;
  p = list_to_intp (meta, yytbl->yy_meta = (INT *)p);
  yytbl->yy_nultrans_n = nnul;
  (void) list_to_intp (nultrans, yytbl->yy_NUL_trans = (INT *)p);
  yytbl->yy_jambase = IVAL (vp->expr[11]);
  yytbl->yy_jamstate = IVAL (vp->expr[12]);
  yytbl->yy_NUL_ec = IVAL (vp->expr[13]);
  e = MakeExternal (&yytblExtDesc, (char *)yytbl);
  QUIT;
  return e;
}

DEF_ACTION(hbuildyytbl)
#define N_ARGS 14 /* 14 arguments */
{
  va_list _args;
  ExprPtr arg;

  VA_START;
  if (ENV_ARITY(env)==N_ARGS) {
    ExprPtr args[N_ARGS];
    COUNTER i;
    ENTER;

    for (i=0; i<N_ARGS; i++)
      args[i] = va_arg(_args,ExprPtr);
    va_end(_args);
    arg = MakeRecord(E_TUPLE_GNRL, NIL, N_ARGS, args);
    (void) KEEP (arg);
    arg = yy_build_tables (arg);
    QUIT;
  }
  else
    {
      arg = va_arg(_args,ExprPtr);
      va_end(_args);
      arg = yy_build_tables (arg);
    }
  return arg;
}
#undef N_ARGS

PRIVATE void yybufSweep (extdata, doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  YY_BUFFER_STATE buf;

  buf = (YY_BUFFER_STATE)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      outInt (buf->yy_buf_size);
      if (errno) return;
      outInt ((INT)(buf->yy_buf_pos - buf->yy_ch_buf));
      if (errno) return;
      outChars (&buf->yy_flags, 1);
      if (errno) return;
      outInt ((INT)buf->yy_n_chars);
      if (errno) return;
      outChars ((CHAR *)buf->yy_ch_buf, buf->yy_n_chars+2);
      if (errno) return;
      (*doext->action) (buf->yyread);
      if (errno) return;
      (*doext->action) (buf->yyreadln);
      break;
    case X_MARK:
    case X_MARKDATA:
      (*doext->action) (buf->yyread);
      (*doext->action) (buf->yyreadln);
      break;
    case X_FREE:
      MARKFREE (buf->yy_ch_buf);
      MARKFREE (extdata);
      break;
    }
}

PRIVATE ExprPtr yybufLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  YY_BUFFER_STATE buf;
  INT size, pos, nchars;
  CHAR flags;
  char *chbuf;
  ExprPtr read, readln, e;

  size = inInt ();
  if (errno) return NIL;
  pos = inInt ();
  if (errno) return NIL;
  inChars (&flags, 1);
  if (errno) return NIL;
  nchars = inInt ();
  if (errno) return NIL;
  chbuf = (char *)MALLOC (size+2);
  inChars ((CHAR *)chbuf, nchars+2);
  if (errno)
    {
    err:
      DISPOSPTR (chbuf, size+2);
      return NIL;
    }
  read = (*load) (sharedref);
  if (errno) goto err;
  readln = (*load) (sharedref);
  if (errno) goto err;
  buf = (YY_BUFFER_STATE) MALLOC (sizeof (struct yy_buffer_state));
  buf->yyread = read;
  buf->yyreadln = readln;
  buf->yy_buf_size = size;
  buf->yy_ch_buf = chbuf;
  buf->yy_buf_pos = chbuf+pos;
  buf->yy_n_chars = nchars;
  buf->yy_flags = flags;
  e = MakeExternal (desc, (char *)buf);
  (void) KEEP (e);
  return e;
}

PUBLIC extDesc yybufExtDesc = {
  (CHAR *)"flex-buffer",
  { 'f','l','x','b' }, NULL,
  NULL, /* no hash function, yy_buffer_states are not shared */
  NULL, /* no comparison function, yy_buffer_states are not shared */
  NULL, /* no print function */
  yybufSweep,
  yybufLoad
};

PRIVATE void outPtr (p, base)
IN CONST char *p, *base;
{
  if (p==NULL)
    outInt ((INT)-1);
  else outInt ((INT) (p-base));
}

PRIVATE char *inPtr (base)
IN CONST char *base;
{
  INT i;

  i = inInt ();
  if (errno || i<0) return NULL;
  return base+i;
}

PRIVATE void yydataSweep (extdata, doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  struct yydata *yyd;
  char c;

  yyd = (struct yydata *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      (*doext->action) (yyd->yy_cbuf);
      if (errno) return;
      if (yyd->yy_cbuf!=NIL)
	{
	  char *base = YY_CURRENT_BUFFER->yy_ch_buf;

	  outPtr (yyd->yytext, base);
	  if (errno) return;
	  outPtr (yyd->yy_c_buf_p, base);
	  if (errno) return;
	  outPtr (yyd->yy_last_accepting_cpos, base);
	  if (errno) return;
	}
      outInt ((INT)yyd->yy_n_chars);
      if (errno) return;
      outInt ((INT)yyd->yyleng);
      if (errno) return;
      outInt ((INT)yyd->yy_start);
      if (errno) return;
      outInt (yyd->yy_last_accepting_state);
      if (errno) return;
      outChars ((CHAR *)&yyd->yy_hold_char, 1);
      if (errno) return;
      c = (yyd->yy_init << 1) | (yyd->yy_did_buffer_switch_on_eof?1:0);
      outChars ((CHAR *)&c, 1);
      if (errno) return;
      outInt (yyd->yy_start_stack_ptr);
      if (errno) return;
      outInt (yyd->yy_start_stack_depth);
      if (errno) return;
      if (yyd->yy_start_stack_depth!=0)
	{
	  outInts (yyd->yy_start_stack, yyd->yy_start_stack_depth);
	  if (errno) return;
	}
      (*doext->action) (yyd->yyread);
      if (errno) return;
      (*doext->action) (yyd->yyreadln);
      if (errno) return;
      (*doext->action) (yyd->yywrap);
      if (errno) return;
      (*doext->action) (yyd->yylloc);
      break;
    case X_MARK:
    case X_MARKDATA:
      (*doext->action) (yyd->yy_cbuf);
      (*doext->action) (yyd->yyread);
      (*doext->action) (yyd->yyreadln);
      (*doext->action) (yyd->yywrap);
      (*doext->action) (yyd->yylloc);
      break;
    case X_FREE:
      if (yyd->yy_start_stack!=NULL)
	{
	  MARKFREE (yyd->yy_start_stack);
	}
      MARKFREE (extdata);
      break;
    }
}

PRIVATE ExprPtr yydataLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  ExprPtr cbuf,read,readln,wrap,yylloc;
  char *yytext, *cbufp, *lastcpos;
  COUNTER nchars,len,start;
  INT last,ptr,depth;
  char hold, c;
  struct yydata *yyd;
  INT *stack;
  ExprPtr e;
  INT size;

  cbuf = (*load) (sharedref);
  if (errno) return NIL;
  yytext = cbufp = lastcpos = NULL;
  if (cbuf!=NIL)
    {
      char *base = ((YY_BUFFER_STATE)EXTDATA(cbuf))->yy_ch_buf;

      yytext = inPtr (base);
      if (errno) return NIL;
      cbufp = inPtr (base);
      if (errno) return NIL;
      lastcpos = inPtr (base);
      if (errno) return NIL;
    }
  nchars = (COUNTER)inInt ();
  if (errno) return NIL;
  len = (COUNTER)inInt ();
  if (errno) return NIL;
  start = (COUNTER)inInt ();
  if (errno) return NIL;
  last = inInt ();
  if (errno) return NIL;
  inChars ((CHAR *)&hold, 1);
  if (errno) return NIL;
  inChars ((CHAR *)&c, 1);
  if (errno) return NIL;
  ptr = inInt ();
  if (errno) return NIL;
  depth = inInt ();
  if (errno) return NIL;
  if (depth==0)
    {
      stack = NULL;
      size = 0;
    }
  else
    {
      size = depth * sizeof (INT);
      stack = (INT *)MALLOC (size);
      inInts (stack, depth);
      if (errno)
	{
	err:
	  if (stack!=NULL)
	    {
	      DISPOSPTR (stack, size);
	    }
	  return NIL;
	}
    }
  read = (*load) (sharedref);
  if (errno) goto err;
  readln = (*load) (sharedref);
  if (errno) goto err;
  wrap = (*load) (sharedref);
  if (errno) goto err;
  yylloc = (*load) (sharedref);
  if (errno) goto err;
  yyd = (struct yydata *)MALLOC(sizeof (struct yydata));
  yyd->yy_cbuf = cbuf;
  yyd->yytext = yytext;
  yyd->yy_c_buf_p = cbufp;
  yyd->yy_last_accepting_cpos = lastcpos;
  yyd->yy_n_chars = nchars;
  yyd->yyleng = len;
  yyd->yy_start = start;
  yyd->yy_last_accepting_state = last;
  yyd->yy_hold_char = hold;
  yyd->yy_init = (c >> 1);
  yyd->yy_did_buffer_switch_on_eof = ((c&1)!=0);
  yyd->yy_start_stack_ptr = ptr;
  yyd->yy_start_stack_depth = depth;
  yyd->yy_start_stack = stack;
  yyd->yyread = read;
  yyd->yyreadln = readln;
  yyd->yywrap = wrap;
  yyd->yylloc = yylloc;
  e = MakeExternal (desc, (char *)yyd);
  (void) KEEP (e);
  return e;
}

PUBLIC extDesc yydataExtDesc = {
  (CHAR *)"flex-data",
  { 'f','l','x','d' }, NULL,
  NULL, /* no hash function, yydata are not shared */
  NULL, /* no comparison function, yydata are not shared */
  NULL, /* no print function */
  yydataSweep,
  yydataLoad
};

PRIVATE ExprPtr yy_new_data (stream, yywrap)
IN CONST ExprPtr stream;
IN CONST ExprPtr yywrap;
{
  struct yydata *yyd;
  ExprPtr e;
  ENTER;

  yyd = (struct yydata *)MALLOC (sizeof (struct yydata));
  fillWithStream (stream, &yyd->yyread, &yyd->yyreadln);
  yyd->yy_cbuf = NIL;
  {
    INT *pos;

    pos = (INT *)MALLOC (5 * sizeof (INT));
    pos[0] = 4;
    pos[1] = pos[2] = pos[3] = pos[4] = 0;
    yyd->yylloc = e = MakeExternal (&iarrayExtDesc, (char *)pos);
    (void) KEEP (e);
  }
  yyd->yytext = NULL; /* initialization not needed, will get inited by yylex() */
  yyd->yy_c_buf_p = NULL;
  yyd->yy_start = 0;
  yyd->yy_last_accepting_cpos = NULL; /* initialization not needed, will get inited by yylex() */
  yyd->yywrap = yywrap;
  yyd->yy_start_stack_ptr = 0;
  yyd->yy_start_stack_depth = 0;
  yyd->yy_start_stack = NULL;
  yyd->yy_init = 1;
  yyd->yy_did_buffer_switch_on_eof = false;
  e = MakeExternal (&yydataExtDesc, (char *)yyd);
  QUIT;
  return e;
}

DEF_ACTION_2ARGS(hnewdata) /* : 'a instream * (unit -> bool) -> yydata */
{
  return yy_new_data (arg1, arg2);
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_hlex)
{
  return yylex ((struct yytables *)EXTDATA(ENV_UNDERLYING(env)), arg);
}
ACTION_END

DEF_ACTION_1ARG(hlex) /* : 'token yytables -> yydata -> 'token */
{
  return MakeCompiled (E_COMPILED_CLOSURE, ACTION_NAME (q_hlex), arg);
}
ACTION_END

/* yy_get_next_buffer - try to read in a new buffer
 *
 * Returns a code representing an action:
 *	EOB_ACT_LAST_MATCH -
 *	EOB_ACT_CONTINUE_SCAN - continue scanning from current position
 *	EOB_ACT_END_OF_FILE - end of file
 */

PRIVATE INT yy_get_next_buffer(yydx)
INOUT ExprPtr yydx;
{
  REGISTER struct yydata *yyd = (struct yydata *)EXTDATA (yydx);
  REGISTER char *dest = YY_CURRENT_BUFFER->yy_ch_buf;
  REGISTER char *source = yyd->yytext;
  REGISTER INT number_to_move, i;
  INT ret_val;

  if (yyd->yy_c_buf_p > &YY_CURRENT_BUFFER->yy_ch_buf[yyd->yy_n_chars + 1])
    {
      ExprPtr e = MakePair (E_CONS, flexErrExc, ione);
      (void) KEEP (e);
      (void) Raise (e);
      return EOB_ACT_END_OF_FILE;
      /*YY_FATAL_ERROR("fatal flex scanner internal error--end of buffer missed" );*/
    }
  if ((YY_CURRENT_BUFFER->yy_flags & YY_FILL_BUFFER_MASK) == 0)
    { /* Don't try to fill the buffer, so this is an EOF. */
      if ( yyd->yy_c_buf_p - yyd->yytext - YY_MORE_ADJ == 1 )
	{
	  /* We matched a single character, the EOB, so
	   * treat this as a final EOF.
	   */
	  return EOB_ACT_END_OF_FILE;
	}
      else
	{
	  /* We matched some text prior to the EOB, first
	   * process it.
	   */
	  return EOB_ACT_LAST_MATCH;
	}
    }
  /* Try to read more data. */
  /* First move last chars to start of buffer. */
  number_to_move = (INT) (yyd->yy_c_buf_p - yyd->yytext) - 1;
  for ( i = 0; i < number_to_move; ++i )
    *(dest++) = *(source++);
  if ((YY_CURRENT_BUFFER->yy_flags & YY_BUFFER_STATUS_MASK) == YY_BUFFER_EOF_PENDING )
    /* don't do the read, it's not guaranteed to return an EOF,
     * just force an EOF
     */
    YY_CURRENT_BUFFER->yy_n_chars = yyd->yy_n_chars = 0;
  else
    {
      INT num_to_read;
      ExprPtr str;
      ENTER;

      /* if interactive, we first read in more data; otherwise, we'll
	 just read whatever is needed later. */
      str = NOEXPR;
      if (YY_CURRENT_BUFFER->yy_flags & YY_IS_INTERACTIVE_MASK)
	  {
	    /* just a shorter name for the current buffer */
	    YY_BUFFER_STATE b = YY_CURRENT_BUFFER;
	    INT new_size;
	    INT old_size = b->yy_buf_size;

	    str = (*ACTION (yyd->yyreadln)) (ENV_CONS(QENV (yyd->yyreadln),0));
	    new_size = SLEN(str) + number_to_move;
	    if (new_size > old_size)
	      {
		b->yy_buf_size = new_size;
		b->yy_ch_buf = (char *)
		  /* Include room in for 2 EOB chars. */
		  yy_flex_realloc ((VOIDPTR)b->yy_ch_buf, old_size + 2, b->yy_buf_size + 2);
	      }
	  }
      else
	{
	  num_to_read =	YY_CURRENT_BUFFER->yy_buf_size - number_to_move - 1;
	  while ( num_to_read <= 0 )
	    { /* Not enough room in the buffer - grow it. */
	      /* just a shorter name for the current buffer */
	      YY_BUFFER_STATE b = YY_CURRENT_BUFFER;
	      INT yy_c_buf_p_offset = (INT) (yyd->yy_c_buf_p - b->yy_ch_buf);

	      if (b->yy_flags & YY_IS_OUR_BUFFER_MASK)
		{
		  INT old_size = b->yy_buf_size + 2;
		  INT new_size = b->yy_buf_size * 2;

		  if ( new_size <= 0 )
		    b->yy_buf_size += b->yy_buf_size / 8;
		  else
		    b->yy_buf_size *= 2;
		  b->yy_ch_buf = (char *)
		    /* Include room in for 2 EOB chars. */
		    yy_flex_realloc ((VOIDPTR)b->yy_ch_buf, old_size, b->yy_buf_size + 2);
		}
	      else
		/* Can't grow it, we don't own it. */
		b->yy_ch_buf = 0;
	      if ( ! b->yy_ch_buf )
		{
		  ExprPtr e = MakePair (E_CONS, flexErrExc, itwo);
		  (void) KEEP (e);
		  (void) Raise (e);
		  return EOB_ACT_END_OF_FILE;
		  /*YY_FATAL_ERROR("fatal error - scanner input buffer overflow" );*/
		}
	      yyd->yy_c_buf_p = &b->yy_ch_buf[yy_c_buf_p_offset];
	      num_to_read = YY_CURRENT_BUFFER->yy_buf_size - number_to_move - 1;
	    }
	  /* Read in more data. */
	  {
	    /* but not more than YY_READ_BUF_SIZE each time. */
	    if ( num_to_read > YY_READ_BUF_SIZE )
	      num_to_read = YY_READ_BUF_SIZE;
	    str = (*ACTION (yyd->yyread)) (ENV_CONS(QENV (yyd->yyread),1), KINT (num_to_read));
	  }
	}
      safememcpy (&YY_CURRENT_BUFFER->yy_ch_buf[number_to_move],
		  STEXT(str),
		  yyd->yy_n_chars = SLEN(str));
      YY_CURRENT_BUFFER->yy_n_chars = yyd->yy_n_chars;
      QUIT;
    }
  if (yyd->yy_n_chars == 0)
    {
      if (number_to_move == YY_MORE_ADJ)
	{
	  ret_val = EOB_ACT_END_OF_FILE;
	  YY_NEW_FILE (yydx);
	}
      else
	{
	  ret_val = EOB_ACT_LAST_MATCH;
	  YY_CURRENT_BUFFER->yy_flags =
	    (YY_CURRENT_BUFFER->yy_flags & ~YY_BUFFER_STATUS_MASK)
	    | YY_BUFFER_EOF_PENDING;
	}
    }
  else
    ret_val = EOB_ACT_CONTINUE_SCAN;
  yyd->yy_n_chars += number_to_move;
  YY_CURRENT_BUFFER->yy_ch_buf[yyd->yy_n_chars] = YY_END_OF_BUFFER_CHAR;
  YY_CURRENT_BUFFER->yy_ch_buf[yyd->yy_n_chars + 1] = YY_END_OF_BUFFER_CHAR;
  yyd->yytext = &YY_CURRENT_BUFFER->yy_ch_buf[0];
  return ret_val;
}

/* yy_get_previous_state - get the state just before the EOB char was reached */

PRIVATE yy_state_type yy_get_previous_state (yytbl, yyd)
IN CONST struct yytables *yytbl;
INOUT struct yydata *yyd;
{
  register yy_state_type yy_current_state;
  register char *yy_cp;

  yy_current_state = yyd->yy_start;
  yy_current_state += YY_AT_BOL(yyd);
  if (yytbl->yy_nultrans_n!=0)
    {
      for (yy_cp = yyd->yytext + YY_MORE_ADJ; yy_cp < yyd->yy_c_buf_p; ++yy_cp)
	{
	  if ( yytbl->yy_accept[yy_current_state] )
	    {
	      yyd->yy_last_accepting_state = yy_current_state;
	      yyd->yy_last_accepting_cpos = yy_cp;
	    }
	  if ( *yy_cp )
	    {
	      register YY_CHAR yy_c = yytbl->yy_ec[YY_SC_TO_UI(*yy_cp)];
	      if ( yytbl->yy_accept[yy_current_state] )
		{
		  yyd->yy_last_accepting_state = yy_current_state;
		  yyd->yy_last_accepting_cpos = yy_cp;
		}
	      while ( yytbl->yy_chk[yytbl->yy_base[yy_current_state] + yy_c] != yy_current_state )
		{
		  yy_current_state = (int) yytbl->yy_def[yy_current_state];
		  if ( yy_current_state >= yytbl->yy_accept_n )
		    yy_c = yytbl->yy_meta[(unsigned int) yy_c];
		}
	      yy_current_state = yytbl->yy_nxt[yytbl->yy_base[yy_current_state] + (unsigned int) yy_c];
	    }
	  else
	    yy_current_state = yytbl->yy_NUL_trans[yy_current_state];
	}
    }
  else
    {
      for (yy_cp = yyd->yytext + YY_MORE_ADJ; yy_cp < yyd->yy_c_buf_p; ++yy_cp)
	{
	  register YY_CHAR yy_c = (*yy_cp ? yytbl->yy_ec[YY_SC_TO_UI(*yy_cp)] : yytbl->yy_NUL_ec);

	  if (yytbl->yy_accept[yy_current_state])
	    {
	      yyd->yy_last_accepting_state = yy_current_state;
	      yyd->yy_last_accepting_cpos = yy_cp;
	    }
	  while (yytbl->yy_chk[yytbl->yy_base[yy_current_state] + yy_c] != yy_current_state)
	    {
	      yy_current_state = (INT) yytbl->yy_def[yy_current_state];
	      if (yy_current_state >= yytbl->yy_accept_n)
		yy_c = yytbl->yy_meta[(unsigned int) yy_c];
	    }
	  yy_current_state = yytbl->yy_nxt[yytbl->yy_base[yy_current_state] + (unsigned int) yy_c];
	}
    }
  return yy_current_state;
}

/* yy_try_NUL_trans - try to make a transition on the NUL character
 *
 * synopsis
 *	next_state = yy_try_NUL_trans (yyd, yytbl, current_state);
 */

PRIVATE yy_state_type yy_try_NUL_trans (yyd, yytbl, yy_current_state)
INOUT struct yydata *yyd;
IN CONST struct yytables *yytbl;
IN yy_state_type yy_current_state;
{
  register Boolean yy_is_jam;

  if (yytbl->yy_nultrans_n!=0)
    {
      yy_current_state = yytbl->yy_NUL_trans[yy_current_state];
      yy_is_jam = (yy_current_state == 0);
    }
  else
    {
      register char *yy_cp = yyd->yy_c_buf_p;
      register YY_CHAR yy_c = yytbl->yy_NUL_ec;

      if (yytbl->yy_accept[yy_current_state])
	{
	  yyd->yy_last_accepting_state = yy_current_state;
	  yyd->yy_last_accepting_cpos = yy_cp;
	}
      while (yytbl->yy_chk[yytbl->yy_base[yy_current_state] + yy_c] != yy_current_state)
	{
	  yy_current_state = (INT) yytbl->yy_def[yy_current_state];
	  if (yy_current_state >= yytbl->yy_accept_n)
	    yy_c = yytbl->yy_meta[(unsigned int) yy_c];
	}
      yy_current_state = yytbl->yy_nxt[yytbl->yy_base[yy_current_state] + (unsigned int) yy_c];
      yy_is_jam = (yy_current_state == yytbl->yy_jamstate);
    }
  return yy_is_jam ? 0 : yy_current_state;
}

PRIVATE void yyunput (yydx, c, yy_bp)
INOUT ExprPtr yydx;
IN int c;
register char *yy_bp;
{
  REGISTER struct yydata *yyd = (struct yydata *)EXTDATA(yydx);
  REGISTER char *yy_cp = yyd->yy_c_buf_p;
  YY_BUFFER_STATE b;

  if (yyd->yy_cbuf==NIL)
    {
      yyd->yy_cbuf = build_buffer (yyd, yyd->yyread, yyd->yyreadln, YY_BUF_SIZE );
      HITBARRIER (yydx);
    }

  b = YY_CURRENT_BUFFER;

  /* undo effects of setting up yyd->yytext */
  *yy_cp = yyd->yy_hold_char;
  if (yy_cp < b->yy_ch_buf + 2)
    { /* need to shift things up to make room */
      /* +2 for EOB chars. */
      register INT number_to_move = yyd->yy_n_chars + 2;
      register char *dest = &b->yy_ch_buf[b->yy_buf_size + 2];
      register char *source = &b->yy_ch_buf[number_to_move];

      while ( source > b->yy_ch_buf )
	*--dest = *--source;
      yy_cp += (INT) (dest - source);
      yy_bp += (INT) (dest - source);
      b->yy_n_chars = yyd->yy_n_chars = b->yy_buf_size;
      if (yy_cp < b->yy_ch_buf + 2)
	{
	  ExprPtr aux = KINT (3);
	  ExprPtr e = MakePair (E_CONS, flexErrExc, aux);
	  (void) KEEP (e);
	  (void) Raise (e);
	  return;
	  /*YY_FATAL_ERROR( "flex scanner push-back overflow" );*/
	}
    }
  *--yy_cp = (char) c;
  yyd->yytext = yy_bp;
  yyd->yy_hold_char = *yy_cp;
  yyd->yy_c_buf_p = yy_cp;
}

DEF_ACTION_2ARGS(hunput) /* yydata * int -> unit */
{
  ExprPtr yydx = arg1;
  struct yydata *yyd = ((struct yydata *)EXTDATA(yydx));
  int c = (int)IVAL(arg2);

  unput (yydx, c);
  return tuple0;
}
ACTION_END

PRIVATE int input (yydx)
INOUT ExprPtr yydx;
{
  REGISTER struct yydata *yyd = (struct yydata *)EXTDATA(yydx);
  int c;

  if (yyd->yy_cbuf==NIL)
    {
      yyd->yy_cbuf = build_buffer (yyd, yyd->yyread, yyd->yyreadln, YY_BUF_SIZE );
      HITBARRIER (yydx);
    }

  *yyd->yy_c_buf_p = yyd->yy_hold_char;
  if (*yyd->yy_c_buf_p == YY_END_OF_BUFFER_CHAR)
    {
      /* yyd->yy_c_buf_p now points to the character we want to return.
       * If this occurs *before* the EOB characters, then it's a
       * valid NUL; if not, then we've hit the end of the buffer.
       */
      if (yyd->yy_c_buf_p < &YY_CURRENT_BUFFER->yy_ch_buf[yyd->yy_n_chars])
	/* This was really a NUL. */
	*yyd->yy_c_buf_p = '\0';
      else
	{ /* need more input */
	  INT offset = yyd->yy_c_buf_p - yyd->yytext;
	  ++yyd->yy_c_buf_p;
	  switch (yy_get_next_buffer (yydx))
	    {
	    case EOB_ACT_LAST_MATCH:
	      /* This happens because yy_g_n_b()
	       * sees that we've accumulated a
	       * token and flags that we need to
	       * try matching the token before
	       * proceeding.  But for input(),
	       * there's no matching to consider.
	       * So convert the EOB_ACT_LAST_MATCH
	       * to EOB_ACT_END_OF_FILE.
	       */
	      /* Reset buffer status. */
	      YY_NEW_FILE (yydx);
	      /* fall through */
	    case EOB_ACT_END_OF_FILE:
	      {
		if ((*ACTION (yyd->yywrap)) (ENV_CONS(QENV (yyd->yywrap),0))==trueValue)
		  return EOF;
		if (!yyd->yy_did_buffer_switch_on_eof)
		  YY_NEW_FILE (yydx);
		return input (yydx);
	      }
	    case EOB_ACT_CONTINUE_SCAN:
	      yyd->yy_c_buf_p = yyd->yytext + offset;
	      break;
	    }
	}
    }
  c = *(unsigned char *) yyd->yy_c_buf_p;	/* cast for 8-bit char's */
  *yyd->yy_c_buf_p = '\0';	/* preserve yyd->yytext */
  yyd->yy_hold_char = *++yyd->yy_c_buf_p;
  if (c=='\n')
    YY_CURRENT_BUFFER->yy_flags |= YY_AT_BOL_MASK;
  else YY_CURRENT_BUFFER->yy_flags &= ~YY_AT_BOL_MASK;
  return c;
}

DEF_ACTION_1ARG(hinput) /* : yydata -> int */
{
  INT c;

  c = input (arg);
  return Kint (c);
}
ACTION_END

PRIVATE void yyrestart (yydx, read, readln)
INOUT ExprPtr yydx;
IN CONST ExprPtr read;
IN CONST ExprPtr readln;
{
  REGISTER struct yydata *yyd = (struct yydata *)EXTDATA (yydx);

  if (yyd->yy_cbuf==NIL)
    {
      yyd->yy_cbuf = build_buffer (yyd, yyd->yyread, yyd->yyreadln, YY_BUF_SIZE);
      HITBARRIER (yydx);
    }
  yy_init_buffer (yyd, YY_CURRENT_BUFFER, read, readln);
  yy_load_buffer_state (yyd);
}

PRIVATE void yy_switch_to_buffer (yydx, newbuf)
INOUT ExprPtr yydx;
IN CONST ExprPtr newbuf;
{
  REGISTER struct yydata *yyd = (struct yydata *)EXTDATA(yydx);

  if (yyd->yy_cbuf==newbuf)
    return;
  if (yyd->yy_cbuf!=NIL)
    {
      /* Flush out information for old buffer. */
      YY_BUFFER_STATE b = YY_CURRENT_BUFFER;

      *yyd->yy_c_buf_p = yyd->yy_hold_char;
      b->yy_buf_pos = yyd->yy_c_buf_p;
      b->yy_n_chars = yyd->yy_n_chars;
    }
  yyd->yy_cbuf = newbuf;
  HITBARRIER (yydx);
  yy_load_buffer_state (yyd);
  /* We don't actually know whether we did this switch during
   * EOF ((*yyd->yywrap)()) processing, but the only time this flag
   * is looked at is after (*yyd->yywrap)() is called, so it's safe
   * to go ahead and always set it.
   */
  yyd->yy_did_buffer_switch_on_eof = true;
}

DEF_ACTION_2ARGS(hswitchbuf) /* : yydata * yybuf -> unit */
{
  yy_switch_to_buffer (arg1, arg2);
  return tuple0;
}
ACTION_END

PRIVATE void yy_load_buffer_state (yyd)
INOUT struct yydata *yyd;
{
  YY_BUFFER_STATE b = YY_CURRENT_BUFFER;

  yyd->yy_n_chars = b->yy_n_chars;
  yyd->yytext = yyd->yy_c_buf_p = b->yy_buf_pos;
  yyd->yyread = b->yyread;
  yyd->yyreadln = b->yyreadln;
  yyd->yy_hold_char = *yyd->yy_c_buf_p;
}

PRIVATE YY_BUFFER_STATE yy_create_buffer (yyd, read, readln, size)
INOUT struct yydata *yyd;
IN CONST ExprPtr read;
IN CONST ExprPtr readln;
IN INT size;
{
  YY_BUFFER_STATE b;

  b = (YY_BUFFER_STATE) MALLOC (sizeof (struct yy_buffer_state));
  b->yy_buf_size = size;
  /* yy_ch_buf has to be 2 characters longer than the size given because
   * we need to put in 2 end-of-buffer characters.
   */
  b->yy_ch_buf = (char *) MALLOC (b->yy_buf_size + 2);
  b->yy_flags |= YY_IS_OUR_BUFFER_MASK | YY_IS_INTERACTIVE_MASK;
  yy_init_buffer (yyd, b, read, readln);
  return b;
}

PRIVATE ExprPtr build_buffer (yyd, read, readln, size)
INOUT struct yydata *yyd;
IN CONST ExprPtr read;
IN CONST ExprPtr readln;
IN INT size;
{
  ExprPtr e;
  YY_BUFFER_STATE b;

  b = yy_create_buffer (yyd, read, readln, size);
  e = MakeExternal (&yybufExtDesc, (char *)b);
  return e;
}

DEF_ACTION_2ARGS(hnewbuf) /* : yydata * 'a instream -> yybuf */
{
  struct yydata *yyd = (struct yydata *)EXTDATA(arg1);
  ExprPtr stream = arg2;
  ExprPtr read, readln;

  fillWithStream (stream, &read, &readln);
  return build_buffer (yyd, read, readln, YY_BUF_SIZE);
}
ACTION_END

/*
PRIVATE void yy_delete_buffer (yyd, b)
INOUT struct yydata *yyd;
YY_BUFFER_STATE b;
{
  if (!b)
    return;
  if (b == YY_CURRENT_BUFFER)
    YY_CURRENT_BUFFER = (YY_BUFFER_STATE) 0;
  if (b->yy_flags & YY_IS_OUR_BUFFER_MASK)
    MARKFREE ((VOIDPTR) b->yy_ch_buf);
  MARKFREE ((VOIDPTR) b);
}
*/

PRIVATE void yy_init_buffer (yyd, b, read, readln)
INOUT struct yydata *yyd;
INOUT YY_BUFFER_STATE b;
IN CONST ExprPtr read;
IN CONST ExprPtr readln;
{
  yy_flush_buffer (yyd, b);
  b->yyread = read;
  b->yyreadln = readln;
  b->yy_flags |= YY_FILL_BUFFER_MASK;
  b->yy_flags |= YY_IS_INTERACTIVE_MASK; /* by default, buffers are interactive */
}

PRIVATE void yy_flush_buffer (yyd, b)
INOUT struct yydata *yyd;
INOUT YY_BUFFER_STATE b;
{
  if (!b)
    return;
  b->yy_n_chars = 0;
  /* We always need two end-of-buffer characters.  The first causes
   * a transition to the end-of-buffer state.  The second causes
   * a jam in that state.
   */
  b->yy_ch_buf[0] = YY_END_OF_BUFFER_CHAR;
  b->yy_ch_buf[1] = YY_END_OF_BUFFER_CHAR;
  b->yy_buf_pos = &b->yy_ch_buf[0];
  b->yy_flags |= YY_AT_BOL_MASK;
  b->yy_flags = (b->yy_flags & ~YY_BUFFER_STATUS_MASK) | YY_BUFFER_NEW;
  if (b == YY_CURRENT_BUFFER)
    yy_load_buffer_state (yyd);
}

DEF_ACTION_2ARGS(hflushbuf) /* : yydata * yybuf -> unit */
{
  yy_flush_buffer ((struct yydata *)EXTDATA(arg1), (YY_BUFFER_STATE)EXTDATA(arg2));
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(hbuf) /* : yydata -> yybuf */
{
  REGISTER struct yydata *yyd = (struct yydata *)EXTDATA (arg);

  if (yyd->yy_cbuf==NIL)
    {
      yyd->yy_cbuf = build_buffer (yyd, yyd->yyread, yyd->yyreadln, YY_BUF_SIZE);
      HITBARRIER (arg);
    }
  return yyd->yy_cbuf;
}
ACTION_END

PRIVATE void yy_push_state (yyd, new_state)
INOUT struct yydata *yyd;
IN INT new_state;
{
  if (yyd->yy_start_stack_ptr >= yyd->yy_start_stack_depth)
    {
      INT old_size, new_size;

      old_size = yyd->yy_start_stack_depth * sizeof(INT);
      yyd->yy_start_stack_depth += YY_START_STACK_INCR;
      new_size = yyd->yy_start_stack_depth * sizeof(INT);
      if (!yyd->yy_start_stack)
	yyd->yy_start_stack = (INT *) MALLOC (new_size);
      else
	yyd->yy_start_stack = (INT *) yy_flex_realloc((VOIDPTR)yyd->yy_start_stack,
						      old_size, new_size);
      if (!yyd->yy_start_stack)
	{
	  ExprPtr aux = KINT (4);
	  ExprPtr e = MakePair (E_CONS, flexErrExc, aux);
	  (void) KEEP (e);
	  (void) Raise (e);
	  return;
	  /*YY_FATAL_ERROR("out of memory expanding start-condition stack" );*/
	}
    }
  yyd->yy_start_stack[yyd->yy_start_stack_ptr++] = yyStart (yyd);
  yyBegin (yyd, new_state);
}

DEF_ACTION_2ARGS(hpushstate) /* yydata * int -> unit */
{
  yy_push_state ((struct yydata *)EXTDATA(arg1), IVAL(arg2));
  return tuple0;
}
ACTION_END

PRIVATE void yy_pop_state(yyd)
INOUT struct yydata *yyd;
{
  if (yyd->yy_start_stack_ptr <= 0)
    {
      ExprPtr aux = KINT (5);
      ExprPtr e = MakePair (E_CONS, flexErrExc, aux);
      (void) KEEP (e);
      (void) Raise (e);
      return;
      /*YY_FATAL_ERROR( "start-condition stack underflow" );*/
    }
  --yyd->yy_start_stack_ptr;
  yyBegin (yyd, yyd->yy_start_stack[yyd->yy_start_stack_ptr]);
}

DEF_ACTION_1ARG(hpopstate) /* : yydata -> unit */
{
  yy_pop_state ((struct yydata *)EXTDATA(arg));
  return tuple0;
}
ACTION_END

PRIVATE INT yy_top_state(yyd)
IN CONST struct yydata *yyd;
{
  if (yyd->yy_start_stack_ptr<=0)
    {
      ExprPtr aux = KINT (5);
      ExprPtr e = MakePair (E_CONS, flexErrExc, aux);
      (void) KEEP (e);
      return IVAL (Raise (e));
    }
  return yyd->yy_start_stack[yyd->yy_start_stack_ptr - 1];
}

DEF_ACTION_1ARG(htopstate) /* : yydata -> int */
{
  INT i;

  i = yy_top_state ((struct yydata *)EXTDATA(arg));
  return Kint (i);
}
ACTION_END

/* Redefine yyless() so it works in section 3 code. */

#undef yyless
#define yyless(n) \
	do \
		{ \
		/* Undo effects of setting up yytext. */ \
		yyd->yytext[yyd->yyleng] = yyd->yy_hold_char; \
		yyd->yy_c_buf_p = yyd->yytext + n; \
		yyd->yy_hold_char = *yyd->yy_c_buf_p; \
		*yyd->yy_c_buf_p = '\0'; \
		yyd->yyleng = n; \
                YY_UPDATE_YYLLOC(;) \
		} \
	while ( 0 )

DEF_ACTION_2ARGS(hless) /* yydata * int -> unit */
{
  struct yydata *yyd = (struct yydata *)EXTDATA(arg1);
  COUNTER n = IVAL(arg2);

  if (n<0 || n>yyd->yyleng)
    {
      ExprPtr aux = KINT ((n<0)?6:7);
      ExprPtr e = MakePair (E_CONS, flexErrExc, aux);
      (void) KEEP (e);
      (void) Raise (e);
    }
  if (n!=yyd->yyleng)
    {
      yyless (n);
    }
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(hlen) /* yydata -> int */
{
  struct yydata *yyd = (struct yydata *)EXTDATA(arg);
  COUNTER n = yyd->yyleng;

  return Kint (n);
}
ACTION_END

DEF_ACTION_1ARG(hyylloc) /* yydata -> intarray */
{
  struct yydata *yyd = (struct yydata *)EXTDATA(arg);
  return yyd->yylloc;
}
ACTION_END

DEF_ACTION_1ARG(herrormsg) /* int -> string */
{
  INT i = IVAL (arg);
  CHAR *s;

  if (i<0 || i>8)
    return voidString;
  s = (CHAR *)message (692+i);
  return Kstr (s);
}
ACTION_END

/* Internal utility routines. */

PRIVATE VOIDPTR yy_flex_realloc (ptr, oldsize, size)
INOUT VOIDPTR ptr;
IN INT oldsize;
IN INT size;
{
  VOIDPTR p;

  if (size<=oldsize)
    return ptr;
  p = (VOIDPTR) MALLOC(size);
  safememcpy (p, ptr, oldsize);
  MARKFREE (ptr);
  return p;
}
