/* Simple-minded resource compiler for Mac version of HimML.
   Intended for compiling constant strings into 'byt#' resources.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <stdio.h>
#ifndef THINK_C
#include <types.h>
#include <memory.h>
#include <resources.h>
#include <files.h>
#endif
#else
#include <console.h>
#endif

void usage()
{
  fprintf(stderr,"Usage: hrez <message-file> <executable-file>.\n");
  exit(1);
}

void badformat()
{
  fprintf(stderr,"Message file corrupted.\n");
  exit(1);
}

typedef void (*parseAction)(unsigned char c,void *data);

void parse(file,fun,data)
FILE *file;
parseAction fun;
void *data;
{
  int c;

  while ((c = fgetc(file)), c!=-1)
    {
      switch (c)
	{
	case ' ': case '\n': case '\t': case ',':
	  break;
	case '\'':
	  c = fgetc(file);
	  if (c=='\\')
	    c = fgetc(file);
	  if (c==-1)
	    badformat();
	  (*fun)((unsigned char)c,data);
	  c = fgetc(file);
	  if (c!='\'')
	    badformat();
	  break;
	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	  {
	    int a;

	    a = c-'0';
	    while ((c = fgetc(file)), c>='0' && c<='9')
	      a = 10*a+(c-'0');
	    (*fun)((unsigned char)c,data);
	    if (c!='\'')
	      badformat();
	  }
	  break;
	default:
	  badformat();
	  break;
	}
    }
}

void countChars(c,data)
unsigned char c;
void *data;
{
  ++((long *)data);
}

void storeChars(c,data)
unsigned char c;
void *data;
{
  *(*((unsigned char **)data))++ = c;
}

main(argc,argv)
int argc;
char *argv[];
{
  FILE *mess;
  long size;
  Handle h;
  int rfile;
  int id;

#ifdef THINK_C
  argc = ccommand(&argv);
#endif
  if (argc!=3)
    usage();
  mess = fopen(argv[1],"r");
  if (mess==NULL)
    usage();
  rfile = UseResourceFile(argv[2]); /* !!! see docs */
  size = 0;
  parse(mess,countChars,&size);
  fseek(mess,0L,0);
  h = GetResourceByName('byt#',argv[1]);
  if (h==(Handle)0)
    {
      h = NewHandle(size);
      do {
	id = UniqueId();
      } while (id<256);
      AttachResource(h,id,argv[1]); /* !!! see docs */
    }
  else
    {
      SetHandleSize(h,size);
    }
  parse(mess,storeChars,h);
  ChangedResource(h);
  WriteResource(h); /* !!! and check for error conditions */
  HFClose(rfile); /* !!! see docs */
  fclose(argv[1]);
  exit(0);
}
