/* Hyacc parser engine.
   Copyright (C) 2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <setjmp.h>
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#else
#include <errno.h>
#endif
#include "portable.h"
#include "eval.h"
#include "mem.h"
#include "expr.h"
#include "transfer.h"
#include "prtproto.h"
#include "hyaccproto.h"
#include "rhyacc.h"
#include "ex1proto.h"
#ifdef HAS_PROTOTYPES
#include "shyacc.h"
#endif

#define __SEG__ HYACC

PUBLIC ExprPtr hyaccErrExc; /* internal parser errors (exception Hyacc of int). */
PUBLIC ExprPtr parseErrExc; /* parse errors: exception Parse of string list;
			       Parse ["id", "string"] means we were expecting
			       an id or a string, but we didn't. */
EXTERN ExprPtr none,optttm;
EXTERN extDesc iarrayExtDesc;
EXTERN ExprPtr voidString;

/* Notes:
   !!! currently YYLEX is assumed to return an int, and this is in conflict with
   the way files are generated by hlex, which expects returned tokens to be of
   some eqtype.  (Maybe sacrifice generality and have yylex return ints only.)
   !!! semantic actions cannot do an abort or an accept... but they may raise
   an exception instead.
   !!! do a HITBARRIER(hydx) whenever we push or change data on the semantic value stack!

   YYLSP_NEEDED should be defined (means that some construct @n was used.)
   YYSTYPE should be ExprPtr, seen by HimML as an instance of some datatype, e.g. for:
%union {
  int i;
  struct expr *e;
  char *s;
}
we have:
datatype YYSTYPE = yys_0 of int | yys_1 of expr | yys_2 of string;

  YYPURE should be defined (our parsers are always pure).
  YYDEBUG is false, but yytname is accessible.
  yyoverflow is undefined.
  YYERROR_VERBOSE always defined.
  suppress YYFAIL, YYBACKUP
 */

#define YYSTYPE ExprPtr

#define YYTRANSLATE(x) ((unsigned)(x) <= hytbl->max_user_token ? hytbl->yytranslate[x] : hytbl->nsyms)

#define YYPUSHLOC()   { \
    /* *++yylsp = yylloc; */ \
    COUNTER i,n; \
    INT *ip; \
    YYLTYPE *yyls0; \
\
    ip = (INT *)EXTDATA(hyd->pos); \
    n = *ip++; \
    if (n>4) \
      n = 4; /* only use first 4 entries (should actually be of length 4 exactly). */ \
    yyls0 = ++hyd->yylsp; \
    for (i=0; i<n; i++) \
      yyls0->pos[i] = ip[i]; \
    for (; i<4; i++) \
      yyls0->pos[i] = 0; /* fill in with default value. */ \
  }

#define YYLEX (((*ACTION(hyd->lex)) (ENV_CONS(QENV(hyd->lex),1), hyd->lexarg)))

/* Skeleton output parser for bison,
   Copyright (C) 1984, 1989, 1990 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* As a special exception, when this file is copied by Bison into a
   Bison output file, you may use that output file without restriction.
   This special exception was added by the Free Software Foundation
   in version 1.24 of Bison.  */

/* This was copied from a file generated by Bison (same for hlex.c),
   replacing explicit constants by fields of hytbl, and so on.
*/

DEF_ACTION_1ARG(hmaxdepth) /* : ('token, 'val) hydata -> int, where 'val=YYSTYPE */
{
  struct hydata *hyd = (struct hydata *)EXTDATA(arg);
  return Kint (hyd->maxdepth);
}
ACTION_END

DEF_ACTION_2ARGS(hsetmaxdepth) /* : ('token, 'val) hydata * int -> unit, where 'val=YYSTYPE */
{
  struct hydata *hyd = (struct hydata *)EXTDATA(arg1);
  hyd->maxdepth = IVAL(arg2);
  return tuple0;
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_hyval) /* : ('token, 'val) hydata -> int -> 'val ,where val=YYSTYPE */
{
  struct hydata *hyd = (struct hydata *)EXTDATA(ENV_UNDERLYING(env));
  INT n = IVAL(arg);
  ExprPtr *yyvsp1;

  if (hyd->yyvs==NULL)
    goto err;
  if (n>0)
    {
      ExprPtr e;

    err:
      e = MakePair (E_CONS, hyaccErrExc, ione);
      (void) KEEP (e);
      /*yyerror("illegal $ value");*/
      return Raise (e);
    }
  yyvsp1 = &hyd->yyvsp[n];
  if (yyvsp1<hyd->yyvs)
    goto err;
  return *yyvsp1;
}
ACTION_END

DEF_ACTION_1ARG(hyval)
{
  return MakeCompiled (E_COMPILED_CLOSURE, ACTION_NAME (q_hyval), arg);
}
ACTION_END

DEF_ACTION_1ARG(hydefval) /* ('token, 'val) hydata -> 'val */
{
  struct hydata *hyd = (struct hydata *)EXTDATA (arg);
  return hyd->default_value;
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_hyloc) /* : ('token, 'val) hydata -> int -> intarray (of length 4),
				    where val=YYSTYPE;
				    creates a fresh intarray, which can also be modified,
				    although it is not exactly the spirit. */
{
  struct hydata *hyd = (struct hydata *)EXTDATA(ENV_UNDERLYING(env));
  INT n = IVAL(arg);
  YYLTYPE *yylsp1;
  INT *pos;
  ExprPtr e;

  if (hyd->yyls==NULL)
    goto err;
  if (n>0)
    {
      ExprPtr e;
    err:
      e = MakePair (E_CONS, hyaccErrExc, itwo);
      (void) KEEP (e);
      /*yyerror("illegal @ value");*/
      return Raise (e);
    }
  yylsp1 = &hyd->yylsp[n];
  if (yylsp1<hyd->yyls)
    goto err;
  pos = (INT *)MALLOC (5 * sizeof (INT));
  pos[0] = 4;
  pos[1] = yylsp1->pos[0];
  pos[2] = yylsp1->pos[1];
  pos[3] = yylsp1->pos[2];
  pos[4] = yylsp1->pos[3];
  e = MakeExternal (&iarrayExtDesc, (char *)pos);
  return e;
}
ACTION_END

DEF_ACTION_1ARG(hyloc)
{
  return MakeCompiled (E_COMPILED_CLOSURE, ACTION_NAME (q_hyloc), arg);
}
ACTION_END

#define yyerrok		(hyd->yyerrstatus = 0)

DEF_ACTION_1ARG(hyerrok) /* : ('token, 'val) hydata -> unit */
{
  struct hydata *hyd = (struct hydata *)EXTDATA(arg);
  yyerrok;
  return tuple0;
}
ACTION_END

#define yyclearin	(hyd->yychar = YYEMPTY)
#define YYEMPTY		-2

DEF_ACTION_1ARG(hyclearin) /* : ('token, 'val) hydata -> unit */
{
  struct hydata *hyd = (struct hydata *)EXTDATA(arg);
  yyclearin;
  return tuple0;
}
ACTION_END

#define YYEOF		0
#define YYERROR		goto yyerrlab1
#define YYRECOVERING()  (!!hyd->yyerrstatus)

DEF_ACTION_1ARG(hyrecovering) /* : ('token, 'val) hydata -> bool */
{
  struct hydata *hyd = (struct hydata *)EXTDATA(arg);
  return hyd->yyerrstatus?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(hysetdebug) /* : ('token, 'val) hydata * bool -> unit */
{
  struct hydata *hyd = (struct hydata *)EXTDATA(arg1);
  hyd->yydebug = (arg2==trueValue);
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(hychar) /* : ('token, 'val) hydata -> int */
{
  struct hydata *hyd = (struct hydata *)EXTDATA(arg);
  return Kint (hyd->yychar);
}
ACTION_END

#ifdef OBSOLETE
#define YYBACKUP(token, value) \
do								\
  if (hyd->yychar == YYEMPTY && yylen == 1)				\
    { hyd->yychar = (token), yylval = (value);			\
      yychar1 = YYTRANSLATE (hyd->yychar);				\
      YYPOPSTACK;						\
      goto yybackup;						\
    }								\
  else								\
    { (void) Raise (hyaccErrExc); /*yyerror ("syntax error: cannot back up");*/ YYERROR; }	\
while (0)
#endif

#define YYTERROR	1
#define YYERRCODE	256

/*  YYINITDEPTH indicates the initial size of the parser's stacks	*/

#ifndef	YYINITDEPTH
#define YYINITDEPTH 200
#endif

/*  YYMAXDEPTH is the maximum size the stacks can grow to
    (effective only if the built-in stack extension method is used).  */

#if YYMAXDEPTH == 0
#undef YYMAXDEPTH
#endif

#ifndef YYMAXDEPTH
#define YYMAXDEPTH 10000
#endif

#if __GNUC__ > 1		/* GNU C and GNU C++ define this.  */
#define __yy_memcpy(TO,FROM,COUNT)	__builtin_memcpy(TO,FROM,COUNT)
#else				/* not GNU C or C++ */
#ifndef __cplusplus

/* This is the most reliable way to avoid incompatibilities
   in available built-in functions on various systems.  */
static void
__yy_memcpy (to, from, count)
     char *to;
     char *from;
     int count;
{
  REGISTER char *f = from;
  REGISTER char *t = to;
  REGISTER int i = count;

  while (i-- > 0)
    *t++ = *f++;
}

#else /* __cplusplus */

/* This is the most reliable way to avoid incompatibilities
   in available built-in functions on various systems.  */
static void
__yy_memcpy (char *to, char *from, int count)
{
  REGISTER char *f = from;
  REGISTER char *t = to;
  REGISTER int i = count;

  while (i-- > 0)
    *t++ = *f++;
}

#endif
#endif

PRIVATE void emptyParseStack (hyd)
INOUT struct hydata *hyd;
{
  if (hyd->yyss!=NULL)
    {
      MARKFREE (hyd->yyss);
      hyd->yyss = hyd->yyssp = NULL;
    }
  if (hyd->yyvs!=NULL)
    {
      MARKFREE (hyd->yyvs);
      hyd->yyvs = hyd->yyvsp = NULL;
    }
  if (hyd->yyls!=NULL)
    {
      MARKFREE (hyd->yyls);
      hyd->yyls = hyd->yylsp = NULL;
    }
}

PRIVATE void prtTname (hytbl, i)
IN CONST struct hytables *hytbl;
IN COUNTER i;
{
  ExprPtr tname = AVAL(hytbl->yytname)->expr[i];
  (void) prtstderrAction (tname, NIL);
}

PRIVATE ExprPtr hyparse(hytbl, hydx) /* returns NONE if aborts, SOME v if accepts and v : YYSTYPE is the yyval of
					the main production. */
IN CONST struct hytables *hytbl;
INOUT ExprPtr hydx;
{
  REGISTER struct hydata *hyd = (struct hydata *)EXTDATA (hydx);
  COUNTER yychar1 = 0;		/*  lookahead token as an internal (translated) token number */
  YYSTYPE yylval = NIL;
  YYSTYPE yyval;		/*  the variable used to return		*/
				/*  semantic values from the action	*/
				/*  routines				*/
  int yylen;
  COUNTER yydebug = hyd->yydebug && hytbl->yyprhs_n!=0 && hytbl->yyrhs_n!=0 && hytbl->yyrline_n!=0;
  ENTER;

  if (yydebug)
    {
      hfprintf (hstderr, "Starting parse\n");
      hfflush (hstderr);
    }

  emptyParseStack (hyd);
  hyd->yyss = (SHORT *)MALLOC (hyd->initdepth * sizeof (SHORT)); /* the state stack */
  hyd->yyvsp = hyd->yyvs = (ExprPtr *)MALLOC (hyd->initdepth * sizeof (ExprPtr)); /* the semantic value stack */
  *hyd->yyvsp = NIL; /* dummy value, never used; we still have to set it to something, otherwise
			GC will barf. */
  hyd->yylsp = hyd->yyls = (YYLTYPE *)MALLOC (hyd->initdepth * sizeof (YYLTYPE)); /* the location stack */
#define YYPOPSTACK   (hyd->yyvsp--, hyd->yyssp--, hyd->yylsp--)
  /* no need to HITBARRIER() in YYPOPSTACK(), this will just sometimes make the GC believe
     it has to keep extraneous entries when it could actually free them. */
  hyd->yystacksize = hyd->initdepth;

  hyd->yystate = 0;
  hyd->yyerrstatus = 0;
  hyd->yynerrs = 0;
  hyd->yychar = YYEMPTY;		/* Cause a token to be read.  */

  /* Initialize stack pointers.
     Waste one element of value and location stack
     so that they stay on the same level as the state stack.
     The wasted elements are never initialized.  */

  hyd->yyssp = hyd->yyss - 1;
  /*hyd->yyvsp = hyd->yyvs; // already done */
  /*yylsp = hyd->yyls; // already done */

/* Push a new state, which is found in  yystate  .  */
/* In all cases, when you get here, the value and location stacks
   have just been pushed. so pushing a state here evens the stacks.  */
yynewstate:

  *++hyd->yyssp = hyd->yystate;

  if (hyd->yyssp >= hyd->yyss + hyd->yystacksize - 1)
    {
      /* Give user a chance to reallocate the stack */
      /* Use copies of these so that the &'s don't force the real ones into memory. */
      ExprPtr *yyvs1 = hyd->yyvs;
      SHORT *yyss1 = hyd->yyss;
      YYLTYPE *yyls1 = hyd->yyls;

      /* Get the current used size of the three stacks, in elements.  */
      int n = hyd->yyssp - hyd->yyss;
      int size = n + 1;

      /* Extend the stack our own way.  */
      if (hyd->yystacksize >= hyd->maxdepth)
	{
	  ExprPtr e = MakePair (E_CONS, hyaccErrExc, izero);
	  (void) KEEP (e);
	  emptyParseStack (hyd);
	  (void) Raise (e);
	  /*yyerror("parser stack overflow");*/
	  QUIT;
	  return none;
	}
      hyd->yystacksize *= 2;
      if (hyd->yystacksize > hyd->maxdepth)
	hyd->yystacksize = hyd->maxdepth;
      hyd->yyss = (SHORT *) MALLOC (hyd->yystacksize * sizeof (SHORT));
      __yy_memcpy ((char *)hyd->yyss, (char *)yyss1, size * sizeof (SHORT));
      hyd->yyssp = hyd->yyss + n;
      hyd->yyvs = (ExprPtr *) MALLOC (hyd->yystacksize * sizeof (ExprPtr));
      __yy_memcpy ((char *)hyd->yyvs, (char *)yyvs1, size * sizeof (ExprPtr));
      hyd->yyvsp = hyd->yyvs + n;
      hyd->yyls = (YYLTYPE *) MALLOC (hyd->yystacksize * sizeof (YYLTYPE));
      __yy_memcpy ((char *)hyd->yyls, (char *)yyls1, size * sizeof (YYLTYPE));
      hyd->yylsp = hyd->yyls + n;

      if (yydebug)
	{
	  hfprintf(hstderr, "Stack size increased to %d\n", hyd->yystacksize);
	  hfflush (hstderr);
	}

      if (hyd->yyssp >= hyd->yyss + hyd->yystacksize - 1)
	{
	  emptyParseStack (hyd);
	  QUIT;
	  return none;
	  /*YYABORT;*/
	}
    }

  if (yydebug)
    {
      hfprintf(hstderr, "Entering state %d\n", hyd->yystate);
      hfflush(hstderr);
    }

  goto yybackup;
 yybackup:

/* Do appropriate processing given the current state.  */
/* Read a lookahead token if we need one and don't already have one.  */
/* yyresume: */

  /* First try to decide what to do without reference to lookahead token.  */

  hyd->yyn = hytbl->yypact[hyd->yystate];
  if (hyd->yyn == hytbl->yyflag)
    goto yydefault;

  /* Not known => get a lookahead token if don't already have one.  */

  /* hyd->yychar is either YYEMPTY or YYEOF
     or a valid token in external form.  */

  if (hyd->yychar == YYEMPTY)
    {
      if (yydebug)
	hfprintf(hstderr, "Reading a token: ");

      SETREFVAL (hyd->yylval, hyd->default_value);
      hyd->yychar = IVAL (YYLEX);
      yylval = REFVAL (hyd->yylval);
      /*KEEPONLY (yylval); // useless, yylval is already protected through hyd->yylval */
    }

  /* Convert token to internal form (in yychar1) for indexing tables with */

  if (hyd->yychar <= 0)		/* This means end of input. */
    {
      yychar1 = 0;
      hyd->yychar = YYEOF;		/* Don't call YYLEX any more */
      if (yydebug)
	{
	  hfprintf(hstderr, "Now at end of input.\n");
	  hfflush(hstderr);
	}
    }
  else
    {
      yychar1 = YYTRANSLATE(hyd->yychar);
      if (yydebug)
	{
	  hfprintf (hstderr, "Next token is %d (", hyd->yychar);
	  prtTname (hytbl, yychar1);
	  hfprintf (hstderr, ", value=");
	  /* Give the individual parser a way to print the precise meaning
	     of a token, for further debugging info.  */
	  valPrintNoType (yylval, do_printfile, (char *)hstderr, 0);
	  hfprintf (hstderr, ")\n");
	  hfflush (hstderr);
	}
    }

  hyd->yyn += yychar1;
  if (hyd->yyn < 0 || hyd->yyn > hytbl->yylast || hytbl->yycheck[hyd->yyn] != yychar1)
    goto yydefault;

  hyd->yyn = hytbl->yytable[hyd->yyn];

  /* yyn is what to do for this token type in this state.
     Negative => reduce, -yyn is rule number.
     Positive => shift, yyn is new state.
       New state is final state => don't bother to shift,
       just return success.
     0, or most negative number => error.  */

  if (hyd->yyn < 0)
    {
      if (hyd->yyn == hytbl->yyflag)
	goto yyerrlab;
      hyd->yyn = -hyd->yyn;
      goto yyreduce;
    }
  else if (hyd->yyn == 0)
    goto yyerrlab;

  if (hyd->yyn == hytbl->yyfinal)
    {
      ExprPtr e;
      
      emptyParseStack (hyd);
      (void) KEEP (yyval);
      e = Kdata (0, optttm, yyval);
      QUIT;
      return e;
      /*YYACCEPT;*/
    }

  /* Shift the lookahead token.  */

  if (yydebug)
    {
      hfprintf(hstderr, "Shifting token %d (", hyd->yychar);
      prtTname (hytbl, yychar1);
      hfprintf (hstderr, "), ");
    }

  /* Discard the token being shifted unless it is eof.  */
  if (hyd->yychar != YYEOF)
    hyd->yychar = YYEMPTY;

  *++hyd->yyvsp = yylval;
  HITBARRIER (hydx);
  YYPUSHLOC();

  /* count tokens shifted since error; after three, turn off error status.  */
  if (hyd->yyerrstatus)
    hyd->yyerrstatus--;

  hyd->yystate = hyd->yyn;
  goto yynewstate;

/* Do the default action for the current state.  */
yydefault:

  hyd->yyn = hytbl->yydefact[hyd->yystate];
  if (hyd->yyn == 0)
    goto yyerrlab;

/* Do a reduction.  yyn is the number of a rule to reduce with.  */
yyreduce:
  yylen = hytbl->yyr2[hyd->yyn];
#ifdef OBSOLETE
  if (yylen > 0)
    yyval = hyd->yyvsp[1-yylen]; /* implement default value of the action */
#endif

  if (yydebug)
    {
      int i;

      hfprintf (hstderr, "Reducing via rule %d (line %d), ",
		hyd->yyn, hytbl->yyrline[hyd->yyn]);

      /* Print the symbols being reduced, and their result.  */
      for (i = hytbl->yyprhs[hyd->yyn]; hytbl->yyrhs[i] > 0; i++)
	{
	  prtTname (hytbl, hytbl->yyrhs[i]);
	  hfputc (' ', hstderr);
	}
      hfputs (" -> ", hstderr);
      prtTname (hytbl, hytbl->yyr1[hyd->yyn]);
      hfputs ("\n", hstderr);
      hfflush (hstderr);
    }

  {
    ExprPtr action;

    action = AVAL(hytbl->yyactiontable)->expr[hyd->yyn];
    yyval = (*ACTION (action)) (ENV_CONS(QENV (action),1), hydx);
    if (yyval->magic!=0) /* NONE */
      YYERROR;
    /* SOME yyval */
    yyval = CDR(yyval);
    (void) KEEPONLY (yyval);
  }

  hyd->yyvsp -= yylen;
  hyd->yyssp -= yylen;
  hyd->yylsp -= yylen;

  if (yydebug)
    {
      SHORT *ssp1 = hyd->yyss - 1;
      hfprintf (hstderr, "state stack now");
      while (ssp1 != hyd->yyssp)
	hfprintf (hstderr, " %d", *++ssp1);
      hfprintf (hstderr, "\n");
      hfflush (hstderr);
    }

  *++hyd->yyvsp = yyval;
  HITBARRIER (hydx);

  hyd->yylsp++;
  {
    INT *ip;
    COUNTER n;

    ip = (INT *)EXTDATA (hyd->pos);
    n = *ip++;
    if (n>2)
      n = 2; /* we only use the first 2 entries (first_line, first_column). */
    if (yylen == 0)
      {
	COUNTER i;
	YYLTYPE *yyls0 = hyd->yylsp;

	for (i=0; i<n; i++)
	  yyls0->pos[i] = ip[i];
	for (; i<2; i++)
	  yyls0->pos[i] = 0; /* fill in with default value. */
	yyls0->pos[2] = (yyls0-1)->pos[2];
	yyls0->pos[3] = (yyls0-1)->pos[3];
	/*
	  yyls0->first_line = yylloc.first_line;
	  yyls0->first_column = yylloc.first_column;
	  yyls0->last_line = (yyls0-1)->last_line;
	  yyls0->last_column = (yyls0-1)->last_column;
	*/
      }
    else
      {
	YYLTYPE *yyls0 = hyd->yylsp;

	yyls0->pos[2] = (yyls0+yylen-1)->pos[2];
	yyls0->pos[3] = (yyls0+yylen-1)->pos[3];
	/*
	  yyls0->last_line = (yyls0+yylen-1)->last_line;
	  yyls0->last_column = (yyls0+yylen-1)->last_column;
	*/
      }
  }

  /* Now "shift" the result of the reduction.
     Determine what state that goes to,
     based on the state we popped back to
     and the rule number reduced by.  */

  hyd->yyn = hytbl->yyr1[hyd->yyn];

  hyd->yystate = hytbl->yypgoto[hyd->yyn - hytbl->yyntbase] + *hyd->yyssp;
  if (hyd->yystate >= 0 && hyd->yystate <= hytbl->yylast && hytbl->yycheck[hyd->yystate] == *hyd->yyssp)
    hyd->yystate = hytbl->yytable[hyd->yystate];
  else
    hyd->yystate = hytbl->yydefgoto[hyd->yyn - hytbl->yyntbase];

  goto yynewstate;

yyerrlab:   /* here on detecting error */

  if (!hyd->yyerrstatus)
    /* If not already recovering from an error, report this error.  */
    {
      ExprPtr expected = NIL;

      ++hyd->yynerrs;
      hyd->yyn = hytbl->yypact[hyd->yystate];
      if (hyd->yyn>hytbl->yyflag && hyd->yyn < hytbl->yylast)
	{
	  COUNTER x, n;

	  n = AVAL(hytbl->yytname)->n;
	  /* Start X at -yyn if negative to avoid negative indexes in hytbl->yycheck.  */
	  for (x = (hyd->yyn < 0 ? -hyd->yyn : 0); x<n; x++)
	    if (hytbl->yycheck[x + hyd->yyn] == x)
	      {
		BEGINQUICKKEEP();
		QUICKKEEP(expected);
		expected = Kons (AVAL(hytbl->yytname)->expr[x], expected);
		ENDQUICKKEEP();
	      }
	}
      /*
      BEGINQUICKKEEP();
      QUICKKEEP(expected);
      expected = MakePair (E_CONS, parseErrExc, expected);
      ENDQUICKKEEP();
      */
      (void) KEEP (expected);
      (void) (*ACTION (hyd->yyerror)) (ENV_CONS(QENV (hyd->yyerror),1), expected);
      /*
      (void) Raise (expected);
      */
      /* parse error */
    }
  goto yyerrlab1;
yyerrlab1:   /* here on error raised explicitly by an action */

  if (hyd->yyerrstatus == 3)
    {
      /* if just tried and failed to reuse lookahead token after an error, discard it.  */

      /* return failure if at end of input */
      if (hyd->yychar == YYEOF)
	{
	  emptyParseStack (hyd);
	  QUIT;
	  return none;
	  /*YYABORT;*/
	}

      if (yydebug)
	{
	  hfprintf(hstderr, "Discarding token %d (", hyd->yychar);
	  prtTname (hytbl, yychar1);
	  hfputs (").\n", hstderr);
	}

      hyd->yychar = YYEMPTY;
    }

  /* Else will try to reuse lookahead token
     after shifting the error token.  */

  hyd->yyerrstatus = 3;		/* Each real token shifted decrements this */

  goto yyerrhandle;

yyerrdefault:  /* current state does not do anything special for the error token. */
yyerrpop:   /* pop the current state because it cannot handle the error token */

  if (hyd->yyssp == hyd->yyss)
    {
      emptyParseStack (hyd);
      QUIT;
      return none;
      /*YYABORT;*/
    }
  hyd->yyvsp--;
  hyd->yystate = *--hyd->yyssp;
  hyd->yylsp--;

  if (yydebug)
    {
      SHORT *ssp1 = hyd->yyss - 1;
      hfprintf (hstderr, "Error: state stack now");
      while (ssp1 != hyd->yyssp)
	hfprintf (hstderr, " %d", *++ssp1);
      hfprintf (hstderr, "\n");
      hfflush (hstderr);
    }

yyerrhandle:

  hyd->yyn = hytbl->yypact[hyd->yystate];
  if (hyd->yyn == hytbl->yyflag)
    goto yyerrdefault;

  hyd->yyn += YYTERROR;
  if (hyd->yyn < 0 || hyd->yyn > hytbl->yylast || hytbl->yycheck[hyd->yyn] != YYTERROR)
    goto yyerrdefault;

  hyd->yyn = hytbl->yytable[hyd->yyn];
  if (hyd->yyn < 0)
    {
      if (hyd->yyn == hytbl->yyflag)
	goto yyerrpop;
      hyd->yyn = -hyd->yyn;
      goto yyreduce;
    }
  else if (hyd->yyn == 0)
    goto yyerrpop;

  if (hyd->yyn == hytbl->yyfinal)
    {
      ExprPtr e;
      
      emptyParseStack (hyd);
      (void) KEEP (yyval);
      e = Kdata (0, optttm, yyval);
      QUIT;
      return e;
      /*YYACCEPT;*/
    }

  if (yydebug)
    hfprintf(hstderr, "Shifting error token, ");

  *++hyd->yyvsp = yylval;
  HITBARRIER (hydx);
  YYPUSHLOC();

  hyd->yystate = hyd->yyn;
  goto yynewstate;
}

DEF_PRIVATE_ACTION_1ARG(q_hyparse)
{
  return hyparse ((struct hytables *)EXTDATA(ENV_UNDERLYING(env)), arg);
}
ACTION_END

DEF_ACTION_1ARG(hyparse) /* : 'val hytables -> (int, 'val) hydata -> 'val option,
			    where 'val will be YYSTYPE
			    (we take ('token, 'val) hydata with 'token = int:
			    the lexer has to return integer tokens.) */
{
  return MakeCompiled (E_COMPILED_CLOSURE, ACTION_NAME (q_hyparse), arg);
}
ACTION_END

PRIVATE void hytblSweep (extdata, doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  struct hytables *hytbl;

  hytbl = (struct hytables *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      outInt ((INT) hytbl->yyfinal);
      if (errno) return;
      outInt ((INT) hytbl->yyflag);
      if (errno) return;
      outInt ((INT) hytbl->yyntbase);
      if (errno) return;
      outInt ((INT) hytbl->max_user_token);
      if (errno) return;
      outInt ((INT) hytbl->nsyms);
      if (errno) return;
      outInt ((INT) hytbl->yylast);
      if (errno) return;
      outInt ((INT) hytbl->yyr1_n);
      if (errno) return;
      outInt ((INT) hytbl->yyr2_n);
      if (errno) return;
      outInt ((INT) hytbl->yydefact_n);
      if (errno) return;
      outInt ((INT) hytbl->yydefgoto_n);
      if (errno) return;
      outInt ((INT) hytbl->yypact_n);
      if (errno) return;
      outInt ((INT) hytbl->yypgoto_n);
      if (errno) return;
      outInt ((INT) hytbl->yytable_n);
      if (errno) return;
      outInt ((INT) hytbl->yycheck_n);
      if (errno) return;
      outInt ((INT) hytbl->yytranslate_n);
      if (errno) return;
      outInt ((INT) hytbl->yyprhs_n);
      if (errno) return;
      outInt ((INT) hytbl->yyrhs_n);
      if (errno) return;
      outInt ((INT) hytbl->yyrline_n);
      if (errno) return;
      outShorts (hytbl->yyr1,
		 hytbl->yyr1_n + hytbl->yyr2_n + hytbl->yydefact_n
		 + hytbl->yydefgoto_n + hytbl->yypact_n + hytbl->yypgoto_n
		 + hytbl->yytable_n + hytbl->yycheck_n + hytbl->yytranslate_n
		 + hytbl->yyprhs_n + hytbl->yyrhs_n + hytbl->yyrline_n);
      if (errno) return;
      /*FALLTHROUGH*/
    case X_MARK:
    case X_MARKDATA:
      (*doext->action) (hytbl->yyactiontable);
      (*doext->action) (hytbl->yytname);
      break;
    case X_FREE:
      MARKFREE(extdata);
      break;
    }
}

PRIVATE ExprPtr hytblLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  struct hytables *hytbl;
  INT final, flag, ntbase, maxtok, nsyms, last, nprhs, nrhs, nrline, nr1, nr2, ndefact,
    ndefgoto, npact, npgoto, ntable, ncheck, ntrans, n;
  SHORT *p;
  ExprPtr e;

  final = inInt();
  if (errno) return NIL;
  flag = inInt();
  if (errno) return NIL;
  ntbase = inInt();
  if (errno) return NIL;
  maxtok = inInt();
  if (errno) return NIL;
  nsyms = inInt();
  if (errno) return NIL;
  last = inInt();
  if (errno) return NIL;
  n = nr1 = inInt();
  if (errno) return NIL;
  n += nr2 = inInt();
  if (errno) return NIL;
  n += ndefact = inInt();
  if (errno) return NIL;
  n += ndefgoto = inInt();
  if (errno) return NIL;
  n += npact = inInt();
  if (errno) return NIL;
  n += npgoto = inInt();
  if (errno) return NIL;
  n += ntable = inInt();
  if (errno) return NIL;
  n += ncheck = inInt();
  if (errno) return NIL;
  n += ntrans = inInt();
  if (errno) return NIL;
  n += nprhs = inInt();
  if (errno) return NIL;
  n += nrhs = inInt();
  if (errno) return NIL;
  n += nrline = inInt();
  if (errno) return NIL;
  hytbl = (struct hytables *)MALLOC (sizeof(struct hytables) + n * sizeof(SHORT));
  hytbl->yyfinal = final;
  hytbl->yyflag = flag;
  hytbl->yyntbase = ntbase;
  hytbl->max_user_token = maxtok;
  hytbl->nsyms = nsyms;
  hytbl->yylast = last;
  hytbl->yyr1_n = nr1;
  hytbl->yyr2_n = nr2;
  hytbl->yydefact_n = ndefact;
  hytbl->yydefgoto_n = ndefgoto;
  hytbl->yypact_n = npact;
  hytbl->yypgoto_n = npgoto;
  hytbl->yytable_n = ntable;
  hytbl->yycheck_n = ncheck;
  hytbl->yytranslate_n = ntrans;
  hytbl->yyprhs_n = nprhs;
  hytbl->yyrhs_n = nrhs;
  hytbl->yyrline_n = nrline;
  p = (SHORT *)(((char *)hytbl) + sizeof (struct hytables));
  inShorts (p, n);
  if (errno)
    {
    err:
      MARKFREE (hytbl);
      return NIL;
    }
  hytbl->yyr1 = p;
  p += nr1;
  hytbl->yyr2 = p;
  p += nr2;
  hytbl->yydefact = p;
  p += ndefact;
  hytbl->yydefgoto = p;
  p += ndefgoto;
  hytbl->yypact = p;
  p += npact;
  hytbl->yypgoto = p;
  p += npgoto;
  hytbl->yytable = p;
  p += ntable;
  hytbl->yycheck = p;
  p += ncheck;
  hytbl->yytranslate = p;
  p += ntrans;
  hytbl->yyprhs = p;
  p += nprhs;
  hytbl->yyrhs = p;
  p += nrhs;
  hytbl->yyrline = p;
  /*p += nrline;*/
  hytbl->yyactiontable = (*load) (sharedref);
  if (errno) goto err;
  hytbl->yytname = (*load) (sharedref);
  if (errno) goto err;
  e = MakeExternal (desc, (char *)hytbl);
  (void) KEEP (e);
  return e;
}

PUBLIC extDesc hytblExtDesc = {
  (CHAR *)"hyacc-tables",
  { 'y', 'a', 'k', 't' }, NULL,
  NULL, /* no hash function, hytables are not shared */
  NULL, /* no comparison function, hytables are not shared */
  NULL, /* no print function */
  hytblSweep,
  hytblLoad
};

PRIVATE SHORT *list_to_shortp (l, p)
IN CONST ExprPtr l;
IN CONST SHORT *p;
{
  while (ISCONS (l))
    {
      *p++ = (SHORT)(IVAL (CAR (l)));
      l = CDR (l);
    }
  return p;
}

PRIVATE ExprPtr hy_build_tables (arg)
IN CONST ExprPtr arg;
/* arg : (('token, 'val) hydata -> 'val option) list * string array
         (0) yyactiontable                    (1) yytname
	 * int * int * int * int * int * int
	 (2-7) yyfinal, yyflag, yyntbase, max_user_token, nsyms, yylast
	 * int list * int list * int list * int list * int list * int list * int list * int list * int list * int list * int list * int list
	 (8-16) yyr1, yyr2, yydefact, yydefgoto, yypact, yypgoto, yytable, yycheck, yytranslate,
	 (17-19) yyprhs, yyrhs, yyrline
	 -> 'val hytables
*/
{
  VecPtr vp;
  struct hytables *hytbl;
  ExprPtr prhs, rhs, rline, r1, r2, defact, defgoto, pact, pgoto, table, check, trans;
  INT nprhs, nrhs, nrline, nr1, nr2, ndefact, ndefgoto, npact, npgoto, ntable, ncheck, ntrans, n;
  SHORT *p;
  ExprPtr e, actions;
  ENTER;

  vp = AVAL (arg);
  actions = ListToArray0 (vp->expr[0]);
  (void) KEEP (actions);
  vp = AVAL (arg);
  n = nr1 = ListLength (r1 = vp->expr[8 /*11*/]);
  n += nr2 = ListLength (r2 = vp->expr[9 /*12*/]);
  n += ndefact = ListLength (defact = vp->expr[10 /*13*/]);
  n += ndefgoto = ListLength (defgoto = vp->expr[11 /*14*/]);
  n += npact = ListLength (pact = vp->expr[12 /*15*/]);
  n += npgoto = ListLength (pgoto = vp->expr[13 /*16*/]);
  n += ntable = ListLength (table = vp->expr[14 /*17*/]);
  n += ncheck = ListLength (check = vp->expr[15 /*18*/]);
  n += ntrans = ListLength (trans = vp->expr[16 /*19*/]);
  n += nprhs = ListLength (prhs = vp->expr[17]);
  n += nrhs = ListLength (rhs = vp->expr[18]);
  n += nrline = ListLength (rline = vp->expr[19]);
  hytbl = (struct hytables *)MALLOC (sizeof (struct hytables) + n*sizeof(SHORT));
  p = (SHORT *) (((char *)hytbl) + sizeof (struct hytables));
  hytbl->yyr1_n = nr1;
  p = list_to_shortp (r1, hytbl->yyr1 = p);
  hytbl->yyr2_n = nr2;
  p = list_to_shortp (r2, hytbl->yyr2 = p);
  hytbl->yydefact_n = ndefact;
  p = list_to_shortp (defact, hytbl->yydefact = p);
  hytbl->yydefgoto_n = ndefgoto;
  p = list_to_shortp (defgoto, hytbl->yydefgoto = p);
  hytbl->yypact_n = npact;
  p = list_to_shortp (pact, hytbl->yypact = p);
  hytbl->yypgoto_n = npgoto;
  p = list_to_shortp (pgoto, hytbl->yypgoto = p);
  hytbl->yytable_n = ntable;
  p = list_to_shortp (table, hytbl->yytable = p);
  hytbl->yycheck_n = ncheck;
  p = list_to_shortp (check, hytbl->yycheck = p);
  hytbl->yytranslate_n = ntrans;
  p = list_to_shortp (trans, hytbl->yytranslate = p);
  hytbl->yyprhs_n = nprhs;
  p = list_to_shortp (prhs, hytbl->yyprhs = p);
  hytbl->yyrhs_n = nrhs;
  p = list_to_shortp (rhs, hytbl->yyrhs = p);
  hytbl->yyrline_n = nrline;
  (void) list_to_shortp (rline, hytbl->yyrline = p);
  vp = AVAL (arg);
  hytbl->yyfinal = IVAL (vp->expr[2]);
  hytbl->yyflag = IVAL (vp->expr[3]);
  hytbl->yyntbase = IVAL (vp->expr[4]);
  hytbl->max_user_token = IVAL (vp->expr[5]);
  hytbl->nsyms = IVAL (vp->expr[6]);
  hytbl->yylast = IVAL (vp->expr[7]);
  hytbl->yyactiontable = actions;
  hytbl->yytname = vp->expr[1];
  e = MakeExternal (&hytblExtDesc, (char *)hytbl);
  QUIT;
  return e;
}

DEF_ACTION(hbuildhytbl)
#define N_ARGS 20 /* 20 arguments */
{
  va_list _args;
  ExprPtr arg;

  VA_START;
  if (ENV_ARITY(env)==N_ARGS) {
    ExprPtr args[N_ARGS];
    COUNTER i;
    ENTER;

    for (i=0; i<N_ARGS; i++)
      args[i] = va_arg(_args,ExprPtr);
    va_end(_args);
    arg = MakeRecord(E_TUPLE_GNRL, NIL, N_ARGS, args);
    (void) KEEP (arg);
    arg = hy_build_tables (arg);
    QUIT;
  }
  else
    {
      arg = va_arg(_args,ExprPtr);
      va_end(_args);
      arg = hy_build_tables (arg);
    }
  return arg;
}
#undef N_ARGS

PRIVATE void hydataSweep (extdata, doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  struct hydata *hyd;

  hyd = (struct hydata *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      outInt ((INT)hyd->initdepth);
      if (errno) return;
      outInt ((INT)hyd->maxdepth);
      if (errno) return;
      outInt ((INT) hyd->yyerrstatus);
      if (errno) return;
      outInt ((INT) hyd->yychar);
      if (errno) return;
      outInt ((INT) hyd->yystacksize);
      if (errno) return;
      outInt ((INT) hyd->yystate);
      if (errno) return;
      outInt ((INT) hyd->yyn);
      if (errno) return;
      outInt ((INT) hyd->yynerrs);
      if (errno) return;
      outChars (&hyd->yydebug, 1);
      if (errno) return;
      if (hyd->yyss==NULL) /* then yyvs, yyvsp, yyls and yylsp should be NULL as well */
	outInt ((INT) -1);
      else /* then none of yyss, yyvs, yyvsp, yyls, yylsp should be NULL. */
	{
	  INT i;
	  INT m = hyd->yyssp + 1 - hyd->yyss;
	  INT n = hyd->yyvsp + 1 - hyd->yyvs; /* should be <=yystacksize */
	  YYLTYPE *yylp;

	  outInt (m);
	  if (errno) return;
	  if (m>0)
	    {
	      outShorts (hyd->yyss, m);
	      if (errno) return;
	    }
	  outInt (n);
	  if (errno) return;
	  for (i=0, yylp = hyd->yyls; i<n; i++, yylp++)
	    {
	      outInt ((INT) yylp->pos[0]);
	      if (errno) return;
	      outInt ((INT) yylp->pos[1]);
	      if (errno) return;
	      outInt ((INT) yylp->pos[2]);
	      if (errno) return;
	      outInt ((INT) yylp->pos[3]);
	      if (errno) return;
	    }
	}
      /*FALLTHROUGH*/
    case X_MARK:
    case X_MARKDATA:
      if (hyd->yyss!=NULL)
	{
	  INT i;
	  INT n = hyd->yyvsp + 1 - hyd->yyvs; /* should be <=yystacksize */
	  ExprPtr *ep;

	  for (i=0, ep = hyd->yyvs; i<n; i++)
	    {
	      (*doext->action) (*ep++);
	      if (errno) return;
	    }
	}
      (*doext->action) (hyd->pos);
      (*doext->action) (hyd->lexarg);
      (*doext->action) (hyd->lex);
      (*doext->action) (hyd->default_value);
      (*doext->action) (hyd->yylval);
      (*doext->action) (hyd->yyerror);
      break;
    case X_FREE:
      MARKFREE (extdata);
      break;
    }
}

PRIVATE ExprPtr hydataLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  struct hydata *hyd;
  INT stksize, m, n;
  ExprPtr e;

  hyd = (struct hydata *)MALLOC (sizeof (struct hydata));
  hyd->initdepth = (COUNTER) inInt ();
  if (errno)
    {
    err:
      MARKFREE (hyd);
      return NIL;
    }
  hyd->maxdepth = (COUNTER) inInt ();
  if (errno) goto err;
  hyd->yyerrstatus = (COUNTER) inInt ();
  if (errno) goto err;
  hyd->yychar = (COUNTER) inInt ();
  if (errno) goto err;
  hyd->yystacksize = stksize = inInt ();
  if (errno) goto err;
  hyd->yystate = inInt ();
  if (errno) goto err;
  hyd->yyn = inInt ();
  if (errno) goto err;
  hyd->yynerrs = inInt ();
  if (errno) goto err;
  inChars (&hyd->yydebug, 1);
  if (errno) goto err;
  m = inInt ();
  if (errno) goto err;
  if (m<0)
    {
      hyd->yyss = hyd->yyssp = NULL;
      hyd->yyvs = hyd->yyvsp = NULL;
      hyd->yyls = hyd->yylsp = NULL;
    }
  else
    {
      hyd->yyss = (SHORT *) MALLOC (stksize * sizeof (SHORT));
      if (m>0)
	{
	  inShorts (hyd->yyss, m);
	  if (errno)
	    {
	    err1:
	      if (stksize) { MARKFREE (hyd->yyss); }
	      goto err;
	    }
	}
      hyd->yyssp = hyd->yyss + m - 1;
      n = inInt ();
      if (errno) goto err1;
      hyd->yyls = (YYLTYPE *) MALLOC (stksize * sizeof (YYLTYPE));
      {
	INT i;
	YYLTYPE *yylp;

	for (i=0, yylp = hyd->yyls; i<n; i++, yylp++)
	  {
	    yylp->pos[0] = inInt ();
	    if (errno)
	      {
	      err2:
		if (n>=0) { MARKFREE (hyd->yyls); }
		goto err1;
	      }
	    yylp->pos[1] = inInt ();
	    if (errno) goto err2;
	    yylp->pos[2] = inInt ();
	    if (errno) goto err2;
	    yylp->pos[3] = inInt ();
	    if (errno) goto err2;
	  }
      }
      hyd->yylsp = hyd->yyls + n - 1;
      hyd->yyvs = (YYSTYPE *)MALLOC (stksize * sizeof (YYSTYPE));
      {
	INT i;
	YYSTYPE *ep;

	for (i=0, ep = hyd->yyvs; i<n; i++)
	  {
	    *ep++ = (*load) (sharedref);
	    if (errno)
	      {
	      err3:
		if (n>=0) { MARKFREE (hyd->yyvs); }
		goto err2;
	      }
	  }
      }
      hyd->yyvsp = hyd->yyvs + n - 1;
    }
  hyd->pos = (*load) (sharedref);
  if (errno) goto err3;
  hyd->lexarg = (*load) (sharedref);
  if (errno) goto err3;
  hyd->lex = (*load) (sharedref);
  if (errno) goto err3;
  hyd->default_value = (*load) (sharedref);
  if (errno) goto err3;
  hyd->yylval = (*load) (sharedref);
  if (errno) goto err3;
  hyd->yyerror = (*load) (sharedref);
  if (errno) goto err3;
  e = MakeExternal (desc, (char *)hyd);
  (void) KEEP (e);
  return e;
}

PUBLIC extDesc hydataExtDesc = {
  (CHAR *)"hyacc-data",
  { 'y', 'a', 'k', 'd' }, NULL,
  NULL, /* no hash function, hydata are not shared */
  NULL, /* no comparison function, hydata are not shared */
  NULL, /* no print function */
  hydataSweep,
  hydataLoad
};

DEF_ACTION_6ARGS(hynewdata) /* : 'a * ('a -> 'token) * 'val * 'val ref * intarray * (string list -> unit)
			       -> ('token, 'val) hydata */
{
  struct hydata *hyd;
  ExprPtr e;

  hyd = (struct hydata *) MALLOC (sizeof (struct hydata));
  hyd->pos = arg5;
  hyd->lexarg = arg1;
  hyd->lex = arg2;
  hyd->default_value = arg3;
  hyd->yylval = arg4;
  hyd->yyerror = arg6;
  hyd->initdepth = YYINITDEPTH;
  hyd->maxdepth = YYMAXDEPTH;
  hyd->yyerrstatus = 0;
  hyd->yychar = YYEMPTY;
  hyd->yystacksize = YYINITDEPTH;
  hyd->yystate = 0;
  hyd->yyn = 0;
  hyd->yynerrs = 0;
  hyd->yyss = hyd->yyssp = NULL;
  hyd->yyvs = hyd->yyvsp = NULL;
  hyd->yyls = hyd->yylsp = NULL;
  hyd->yydebug = 0;
  e = MakeExternal (&hydataExtDesc, (char *)hyd);
  return e;
}
ACTION_END

DEF_ACTION_1ARG(hyerrmsg) /* int -> string */
{
  INT i = IVAL (arg);
  CHAR *s;

  if (i<0 || i>8)
    return voidString;
  s = (CHAR *)message (721+i);
  return Kstr (s);
}
ACTION_END

