%{
/* Patch parser file as returned by bison 1.22, so as to produce
   parser files that can be compiled on the Macintosh. (Cures the
   problem of tables that are >32K by putting them in a resource
   file and reading the latter at initialization time.)
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "system.h"

typedef enum { lngs, ints, chrs, strgs } type;
char *mactype[] = {"lngs", "ints", "chrs", "str#"};
/* First entry may be patched in case of 16 bit ints; see main() */
char *ctype[] = {"int", "short", "char", "char * "};

#define NTABLES 32
#define RES_START 256
#define MAXBUF 80

char name[MAXBUF];
int n = 0;
int size;
int nentries;

struct resource {
  char name[MAXBUF];
  char size[MAXBUF];
  type type;
} resources[NTABLES];

FILE *macf;

void put(s)
char *s;
{
  fputs(s,stdout);
  fputs(s,macf);
}

%}

SPACE [\ \t]
DIGIT [0-9]
LETTER [A-Za-z_]

%x table
%x keyword
%option noyywrap

%%

<INITIAL>static{SPACE}const	{ /* this is the start of a table */
  nentries = 0;
  BEGIN(table);
}
<INITIAL>\"[^\"\n]*\"	|
<INITIAL>.	|
<INITIAL>\n	{ ECHO; }

<table>short({SPACE}+int)?	{ resources[n].type = ints; }
<table>int			|
<table>long({SPACE}+int)?	{ resources[n].type = lngs; }
<table>(unsigned{SPACE}+)?char	{ resources[n].type = chrs; }
<table>char{SPACE}*\*{SPACE}*(const)?	{ resources[n].type = strgs; }

<table>{LETTER}+{DIGIT}*\[{DIGIT}*\]	{
  char *s,*t;

  for (s = resources[n].name, t = yytext; (*s++ = *t++)!='['; );
  s[-1] = '\0';
  for (s = resources[n].size; (*s++ = *t++)!=']'; );
  s[-1] = '\0';
}

<table>=		{ /* time to output the tables! */
  /* first, correct a bug: the table yytname[] is inside an "#if YYDEBUG!=0",
     together with yyrline[]. However, we also need yytname[] when YYDEBUG==0
     but YYERROR_VERBOSE is defined. */
  if (strcmp(resources[n].name,"yytname")==0)
    {
      /* so, instead we close the "#if YYDEBUG!=0", and insert what's needed: */
      fputs("\n#define WANT_YYTNAME\n#endif\n\n#ifndef WANT_YYTNAME\n#ifdef YYERROR_VERBOSE\n#define WANT_YYTNAME\n#endif\n#endif\n\n#ifdef WANT_YYTNAME\n",stdout);
    }
  /* now, back to printing the tables: */
  fprintf(stdout,"\n#if OS==MAC\nstatic %s *%s = NULL;\n",
	  ctype[resources[n].type],resources[n].name);
  fprintf(stdout,"#else\nstatic %s %s[%s] =\n",
	  ctype[resources[n].type],resources[n].name,resources[n].size);
  fprintf(macf,"\nresource '%s' (%d,preload,purgeable) { /* %s[%s] */",
	  mactype[resources[n].type],n+RES_START,resources[n].name,resources[n].size);
}


<table>\};	{ /* this is the end of a table */
  ECHO; fputs ("}\n};\n",macf);
  fputs("\n#endif",stdout);
  fprintf(stdout,"\n#define %s_SIZE %d\n",resources[n].name,nentries);
  n++;
  BEGIN(INITIAL);
}
<table>\"	{ nentries++; put(yytext); BEGIN(keyword); }
<table>{DIGIT}+	{ nentries++; put(yytext); }
<table>.	|
<table>\n	{ put(yytext); }

<keyword>kw_	{ put("\177"); /* signals a description rather than a proper keyword */ }
<keyword>_	{ put(" "); }
<keyword>kw_left	{ put("left associative"); }
<keyword>kw_right	{ put("right associative"); }
<keyword>{DIGIT}/\"	{ put("precedence "); yytext[1] = 0; put(yytext); put(" operator"); }
<keyword>{DIGIT}	{ put("precedence "); put(yytext); }
<keyword>open_paren	{ put("("); }
<keyword>close_paren	{ put(")"); }
<keyword>open_bracket	{ put("["); }
<keyword>close_bracket	{ put("]"); }
<keyword>open_brace	{ put("{"); }
<keyword>close_brace	{ put("}"); }
<keyword>open_brace_under	{ put("<{"); }
<keyword>open_record		{ put("|["); }
<keyword>close_record	{ put("]|"); }
<keyword>comma		{ put(","); }
<keyword>colon		{ put(":"); }
<keyword>semicolon	{ put(";"); }
<keyword>ellipsis	{ put("..."); }
<keyword>underscore	{ put("_"); }
<keyword>vbar		{ put("|"); }
<keyword>vbar_vbar	{ put("||"); }
<keyword>equal		{ put("="); }
<keyword>big_arrow	{ put("=>"); }
<keyword>arrow		{ put("->"); }
<keyword>star		{ put("*"); }
<keyword>sharp		{ put("#"); }
<keyword>scale_star	{ put("`"); }
<keyword>scale_power	{ put("^"); }
<keyword>union		{ put("U"); }
<keyword>map_arrow	{ put("-m>"); }
<keyword>open_mu	{ put("++|["); }
<keyword>pinf		{ put("+inf"); }
<keyword>minf		{ put("~inf"); }
<keyword>nan{DIGIT}+	{ put("NAN("); put(yytext+3); put(")"); }
<keyword>mnan{DIGIT}+	{ put("~NAN("); put(yytext+4); put(")"); }
<keyword>memo_fn	{ put("memofn"); }
<keyword>memo_fun	{ put("memofun"); }
<keyword>kw_cons_id	{ put("label"); }
<keyword>kw_identifier		|
<keyword>kw_type_variable	{ put(yytext+3); }
<keyword>_l/\"		{ put(" list"); }
<keyword>\"		{ put(yytext); BEGIN(table); }
<keyword>@{DIGIT}+	|
<keyword>.		|
<keyword>\n		{ put(yytext); }

%%

#ifdef HAS_PROTOTYPES
void outmacinit (void)
#else
void outmacinit()
#endif
{
  int i;

  fputs("\n#if OS==MAC\n",stdout);
  for (i=0; i<n; i++)
    {
      if (resources[i].type==strgs)
	; /* special case: no handle allocated */
      else
	fprintf(stdout,"static Handle %s_H = (Handle)0L;\n",resources[i].name);
    }
  fputs("\nvoid bison_mac_end()\n{\n",stdout);
  for (i=0; i<n; i++)
    {
      if (resources[i].type==strgs)
	; /* special case: no handle allocated */
      else
	fprintf(stdout,"  if (%s_H!=(Handle)0L) HUnlock(%s_H);\n",
		resources[i].name, resources[i].name);
    }
  fputs("}\n\nvoid bison_mac_init()\n{\n",stdout);
  for (i=0; i<n; i++)
    {
      if (resources[i].type==strgs)
	{ /* special case, must convert from linear array of 0-delimited chars to
	     array of pointers to strings. */
	  fprintf(stdout,"  if (%s==NULL)\n    {\n    %s = load_mac_str(%d);\n",
		  resources[i].name, resources[i].name, i+RES_START);
	  fprintf(stdout,"    if (%s==NULL) goto error;\n    }\n",resources[i].name);
	}
      else
	{
	  fprintf(stdout,"  if (%s_H==(Handle)0L || %s==NULL)\n    {\n      %s_H = GetResource('%s',%d);\n",
		  resources[i].name, resources[i].name, resources[i].name,
		  mactype[resources[i].type], i+RES_START);
	  fprintf(stdout,"      if (%s_H==(Handle)0L) goto error;\n",resources[i].name);
	  fprintf(stdout,"      MoveHHi(%s_H);\n",resources[i].name);
	  fprintf(stdout,"      %s = (%s *)*%s_H;\n",
		  resources[i].name, ctype[resources[i].type], resources[i].name);
	  fprintf(stdout,"    }\n  HLock(%s_H);\n",resources[i].name);
	}
    }
  fputs("  return;\nerror:\n  bison_mac_end();\n",stdout);
  fputs("  yyerror(\"Not enough memory to load grammar tables\");\n}\n",stdout);
  fputs("#endif\n",stdout);
}

main()
{
  macf = fopen("tab.r","w");
  if (macf==NULL)
    {
      fputs("Error: Cannot open tab.r for writing.\n",stderr);
      exit(10);
    }
#if OS==MAC
  if (sizeof(int)==sizeof(short))
    mactype[lngs] = mactype[ints];
  else if (sizeof(int)!=sizeof(long))
    {
      fputs("Sorry, the Mac port needs ints of the same size as shorts or longs.\n",stderr);
      fprintf(stderr,"This is not the case (sizeof(int)=%d, but sizeof(short)=%d and sizeof(long)=%d),\n",sizeof(int),sizeof(short),sizeof(long));
      fputs("the only way to correct this is to hack makemactab.l.\n",stderr);
      exit(10);
    }
#endif
  fputs("/* Flex tables for the Mac, as resources; this is generated by makemactab. */\n\n",
	macf);
  fputs("type 'lngs' {\n  wide array { longint; };\n};\n",macf);
  fputs("type 'ints' {\n  wide array { integer; };\n};\n",macf);
  fputs("type 'chrs' {\n  wide array { byte; };\n};\n\n",macf);
  fputs("type 'str#' {\n  integer = $$Countof(StringArray);\n  array StringArray {\n    cstring;\n  };\n};\n\n", macf);
  yylex();
  outmacinit();
  fclose(macf);
  exit(0);
}
