/* Determine system and compiler dependencies.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005,2011 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#ifdef THINK_C
#include <unix.h>
extern char *malloc();
#endif
#include "general.h"

#ifdef __STDC__
#include <stdlib.h>
#endif
#if OS==AMIGA_OS
#include <time.h>

struct tms {
  time_t tms_utime;
  time_t tms_stime;
};

#define HERTZ CLOCKS_PER_SEC
void times (struct tms *tms);
#define random rand
#else
#include <sys/times.h>
#endif

FILE *out;
long maxTwoNLong;
int nBitsLong;
#define MAX_TWO_TO_THE_N_LONG maxTwoNLong
#define NBITS_IN_LONG nBitsLong

int ndigs;

#define CHAR unsigned char
#define SHORT short

#define OTHER 0

#ifndef OS
#ifdef THINK_C
# define OS MAC
# define OSNAME "Mac"
# define EXTENDED double
# define REAL short double
# define EXTENDED_NAME "double"
# define REAL_NAME "short double"
#else
#ifdef MPW
# define OS MAC
# define OSNAME "Mac"
# define EXTENDED extended
# define REAL double
# define EXTENDED_NAME "extended"
# define REAL_NAME "double"
#else
# define OS OTHER
# define EXTENDED double
# define REAL   double
# define EXTENDED_NAME "double"
# define REAL_NAME "double"
#endif
#endif
#endif

#if OS==AMIGA_OS
# define OSNAME "AmigaOS"
# define EXTENDED double
# define REAL double
# define EXTENDED_NAME "double"
# define REAL_NAME "double"
#endif

#ifdef __STDC__
#define P(list) list
#else
#define P(list) ()
#endif

void fatal P((char *msg));
     void done P((int n));
     void InitRandom P((int os));
     int descendingstack P((void));
     int memequal P((char *p1,int len1,char *p2,int len2));
     void putifloor P((FILE *out));
     void prepalloc P((void));
     void prepdiv P((FILE *out));
     
#define DEFPROTO(name) \
       typedef struct name { \
	 CHAR kind; \
	 CHAR mark; \
	 SHORT magic; \
	 char *link; \
	 union { \
	   struct { \
	     long len; \
	     char *text; \
	     } str; \
	   struct { \
	     struct name *car,*cdr; \
	     } pair; \
	   REAL x; \
	   } what; \
	 } name

DEFPROTO(protoExpr);
     
typedef struct MBlock
{
  unsigned long MB_size;
  struct  MBlock *MB_next;  /* or MB_contents for allocated ptrs */
}   MBlock,*MBPtr;

typedef struct  MZone
{
  MBlock MZ_block;
  char *(*MZ_hook)(/* unsigned long size */);
} MZone,*MZPtr;

typedef struct block {
  struct block *next;
  struct block *prev;
  char *b1;
  char *end;
  struct bcontents {
    char data[32];
  } b;
} block;

#define NREALS 7

REAL xreal[] = {
  0.0, 1.0, -1.0, 2.0, -2.0, 3.0, -3.0
  };

unsigned short xshort[][4] = {
  { 0x0000, 0x0000, 0x0000, 0x0000 },
  { 0x3ff0, 0x0000, 0x0000, 0x0000 },
  { 0xbff0, 0x0000, 0x0000, 0x0000 },
  { 0x4000, 0x0000, 0x0000, 0x0000 },
  { 0xc000, 0x0000, 0x0000, 0x0000 },
  { 0x4008, 0x0000, 0x0000, 0x0000 },
  { 0xc008, 0x0000, 0x0000, 0x0000 },
};

unsigned short xrevshort[][4] = {
  { 0x0000, 0x0000, 0x0000, 0x0000 },
  { 0xf03f, 0x0000, 0x0000, 0x0000 },
  { 0xf0bf, 0x0000, 0x0000, 0x0000 },
  { 0x0040, 0x0000, 0x0000, 0x0000 },
  { 0x00c0, 0x0000, 0x0000, 0x0000 },
  { 0x0840, 0x0000, 0x0000, 0x0000 },
  { 0x08c0, 0x0000, 0x0000, 0x0000 },
};

unsigned short xchar[][8] = {
  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
  { 0x3f, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
  { 0xbf, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
  { 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
  { 0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
  { 0x40, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
  { 0xc0, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
};

unsigned short xintelshort[][4] = {
  { 0x0000, 0x0000, 0x0000, 0x0000 },
  { 0x0000, 0x0000, 0x0000, 0x3ff0 },
  { 0x0000, 0x0000, 0x0000, 0xbff0 },
  { 0x0000, 0x0000, 0x0000, 0x4000 },
  { 0x0000, 0x0000, 0x0000, 0xc000 },
  { 0x0000, 0x0000, 0x0000, 0x4008 },
  { 0x0000, 0x0000, 0x0000, 0xc008 },
};

struct Page {
  struct Page *next;
  char *end,*top;
  char dummy;
};

#define BUCKET_SIZE 512

struct Bucket {
  struct Bucket *next;
  short firstNewCell;
  short occupancy[MAX_GENERATIONS];
  protoExpr cell[BUCKET_SIZE];
  protoExpr *newcells[BUCKET_SIZE];
} Bucket,*BucketPtr;

typedef struct TZone {
  unsigned short line1,line2;
  unsigned short pos1,pos2;
} TZone;

struct build_header {
  CHAR kind; /* = BUILD_HEADER_KIND */
  CHAR mark;
  short type;
  TZone z;
  struct Expression *args;
};

#define BH_BUCKET_SIZE 32

typedef struct bhBucket {
  struct bhBucket *next;
  short firstNewCell;
  short occupancy[MAX_GENERATIONS];
  struct build_header bh[BH_BUCKET_SIZE];
  struct build_header *bhcells[BH_BUCKET_SIZE];
} bhBucket;

long alignment P((void));
long pgcd P((long a,long b));
long ppcm P((long a,long b));
long align;
void byteordering P((void));
void byteordering_short P((void));

#if OS==AMIGA_OS
void times(tms)
struct tms *tms;
{
  (tms)->tms_stime = 0L;
  /* unfortunately, clock() is buggy under SAS/C 6.3,
     and returns sometimes random values. We then use
     timer() instead, unless it is not available. */
#ifndef _STRICT_ANSI
  {
    unsigned int clock[2];

    (void) timer(clock);
    tms->tms_utime = CLOCKS_PER_SEC*clock[0]+clock[1]/(1000000/CLOCKS_PER_SEC);
  }
#else
  tms->tms_utime = clock();
#endif
}
#endif

int markbit(out,nlowbits)
     FILE *out;
     int nlowbits;
{
  protoExpr *e;
  unsigned long l;
  enum encoding { E_NONE, E_LOWBIT_LONG } enc;

  enc = E_NONE;
  e = malloc(sizeof(protoExpr));
  l = (unsigned long) e;
  if ((protoExpr *)l == e)
    {
      if ((l & 1)==0 && nlowbits>0)
	enc = E_LOWBIT_LONG;
    }
  free(e);
  switch (enc)
    {
    case E_LOWBIT_LONG:
      fprintf (out, "#define M_TYPE unsigned long\n");
      fprintf (out, "#define M_BITSET(e) (((M_TYPE)(e)) & 0x1L)\n");
      fprintf (out, "#define M_SETBIT(type,e) (type)(((M_TYPE)(e)) | 0x1L)\n");
      fprintf (out, "#define M_CLRBIT(type,e) (type)(((M_TYPE)(e)) & ~0x1L)\n");
      fprintf (out, "#define M_VARCALL(x,f,env,n) \\\n");
      fprintf (out, "  {\\\n    switch (n) { \\\n");
      {
	int i, j;

	for (i=0; i<(1<<nlowbits); i++)
	  {
	    if (i==0)
	      fprintf (out, "      case 0: {");
	    else 
	      fprintf (out, "      case %d: VA_START; {", i);
	    for (j=0; j<i; j++)
	      fprintf (out, " ExprPtr _x%d = va_arg(_args,ExprPtr);", j+1);
	    if (i!=0)
	      fprintf (out, " va_end(_args);");
	    fprintf (out, " x = f (ENV_CONS(env,%d)", i, i);
	    for (j=0; j<i; j++)
	      fprintf (out, ", _x%d", j+1);
	    fprintf (out, "); } break; \\\n");
	  }
      }
      fprintf (out, "      default: abort (); /*NOTREACHED*/ break;\\\n    }\\\n  }\n");
      fprintf (out, "#define M_VARCALL2(x,f,bc,env,n) \\\n");
      fprintf (out, "  {\\\n    switch (n) { \\\n");
      {
	int i, j;

	for (i=0; i<(1<<nlowbits); i++)
	  {
	    if (i==0)
	      fprintf (out, "      case 0: {");
	    else 
	      fprintf (out, "      case %d: VA_START; {", i);
	    for (j=0; j<i; j++)
	      fprintf (out, " ExprPtr _x%d = va_arg(_args,ExprPtr);", j+1);
	    if (i!=0)
	      fprintf (out, " va_end(_args);");
	    fprintf (out, " x = f (bc,ENV_CONS(env,%d)", i, i);
	    for (j=0; j<i; j++)
	      fprintf (out, ", _x%d", j+1);
	    fprintf (out, "); } break; \\\n");
	  }
      }
      fprintf (out, "      default: abort (); /*NOTREACHED*/ break;\\\n    }\\\n  }\n");
      fprintf (out, "#define M_VARCALL_CC(x,f,env,n,data) \\\n");
      fprintf (out, "  {\\\n    ExprPtr _env = env;\\\n    switch (n) { \\\n");
      {
	int i, j;

	for (i=0; i<(1<<nlowbits); i++)
	  {
	    fprintf (out, "      case %d: ", i);
	    fprintf (out, " x = f (_env", i, i);
	    for (j=0; j<i; j++)
	      fprintf (out, ", data[%d]", j+1);
	    fprintf (out, "); break; \\\n");
	  }
      }
      fprintf (out, "      default: abort (); /*NOTREACHED*/ break;\\\n    }\\\n  }\n");
      fprintf (out, "#define M_VARCALL2_CC(x,f,bc,env,n,data) \\\n");
      fprintf (out, "  {\\\n    ExprPtr _env = env;\\\n    switch (n) { \\\n");
      {
	int i, j;

	for (i=0; i<(1<<nlowbits); i++)
	  {
	    fprintf (out, "      case %d: ", i);
	    fprintf (out, " x = f (bc,_env", i, i);
	    for (j=0; j<i; j++)
	      fprintf (out, ", data[%d]", j+1);
	    fprintf (out, "); break; \\\n");
	  }
      }
      fprintf (out, "      default: abort (); /*NOTREACHED*/ break;\\\n    }\\\n  }\n");
      break;
    default:
      return 1;
    }
  return 0;
}

main()
{
  CHAR c;
  int ieee754,chars8bits;
  int extp,informat;
  int i,k,nba,nlowbits;
  long m,n,al;
  char *addr;
  /*
  struct protoExprAlign {
    protoExpr expr[2];
  } peal;
  */
  protoExpr pe;
  struct Page page;
  struct Bucket bucket;
  bhBucket bhb;
  FILE *file;
  /*long align2;*/
  int gap;
  unsigned long ul,lastul;
  block blk;
  int lipzamp = 1;
  int lipzdiv = 1;
  int nbits_in_char;
  
  out = fopen("system.h","w");
  if (out==NULL)
    fatal("could not create file 'system.h'");
  fputs("/* System configuration file, automatically generated by mksys and mksyscc/mksys.cli */\n\n",out);
  /* searching for permanent brain damages: */
  c = 0xff;
  if (((unsigned)c)!=((unsigned)0xff))
    fatal("major lossage: characters are less than 8 bits");
  for (i=0; i<1000; i++)
    {
      if (-i != 1+(~i))
	fatal("fatal misfeature: not a two's complement machine");
    }
  /* size of characters */
  for (i=0, c = (CHAR)1; c!=0; i++, c<<=1);
  nbits_in_char = i;
  fprintf(out,"#define NBITS_IN_CHAR %d\n", i);
  chars8bits = (i==8);
  fprintf(out,"#define MAX_UCHAR (%u)\n",(CHAR)(-1));
  fprintf(out,"#define SIZEOFLONG %ld\n", sizeof(long));
  {
    int bitsoflong = i * sizeof(long);
    int nbits;
    int nbitsh;
    long radix, radixroot;
    EXTENDED log10rad;
    int size;

    fprintf (out,"#define BITSOFLONG %d\n",bitsoflong);
    nbits = bitsoflong;
    if (i&1)
      nbits--;
    else nbits -= 2;
    while (1)
      { /* copied from zstart(): */
	EXTENDED local_one = (EXTENDED) 1;
	EXTENDED local_half = 1 / (EXTENDED) 2;
	EXTENDED epsilon = local_one;
	EXTENDED fudge = local_one + epsilon;

	radix = (long)(1L<<nbits);
	/* the following loop is sick, but we have to: */
	/* comparing one to one + epsilon does not have */
	/* the desired effect on some machines, because */
	/* one + epsilon is Not cast to EXTENDED (as it should) */
	while (local_one != fudge)
	  {
	    epsilon = epsilon * local_half;
	    fudge = local_one + epsilon;
	  }
	epsilon += epsilon;
	if ((epsilon * radix) > local_one)
	  {
	    nbits -= 2;
	  }
	else break;
      }

    fprintf (out,"#define NBITS_IN_LIP_LONG %d\n",nbits);
#define NBITS nbits
    fprintf (out,"#define NBITSH %d\n", nbitsh = nbits>>1);
#define NBITSH nbitsh
    fprintf (out,"#define RADIX %ld\n", radix);
#define RADIX radix
#define RADIXM (radix-1)
    fprintf (out,"#define RADIXROOT %ld\n", radixroot = (long)(1L<<nbitsh));
#define RADIXROOT radixroot
#define RADIXROOTM (radixroot-1)
    fprintf (out,"#define LOG10RAD %.20f\n", log10rad = log10 ((EXTENDED)radix));
    fprintf (out,"#define LOG10RAD_BYTES ");
    for (i=0; i<sizeof(EXTENDED); i++)
      {
	if (i!=0)
	  fputs (", ", out);
	fprintf (out, "0x%x", ((unsigned char *)&log10rad)[i]);
      }
    fputs ("\n", out);
    size = (bitsoflong+nbits-1)/nbits;
    fprintf (out, "#define LIP_SIZE %d\n", size);
    {
#define N_LIPZAMP_CHECKS 1000000
      long a, a2, b, d, t, t2;

      {
#define PRIVATE
#include "lipzamp1.h"
#undef PRIVATE

	for (i=0; i<N_LIPZAMP_CHECKS; i++)
	  {
	    a = a2 = random () & RADIXM;
	    b = random () & RADIXM;
	    d = random () & RADIXM;
	    t = t2 = random () & RADIXM;
	    zaddmulp(&a,b,d,&t);
#undef zaddmulp
#undef zaddmulpsq
#include "lipzamp2.h"
	    zaddmulp(&a2,b,d,&t2);
	    if (a2!=a || t2!=t)
	      {
		lipzamp = 2;
		fprintf (out,
			 "/* Use lipzamp2.h, since lipzamp1.h gives incorrect results. */\n");
		break;
	      }
#undef zaddmulp
#undef zaddmulpsq
	  }
      }
      if (lipzamp==1)
	{
	  struct tms start1, end1, start2, end2;
	  long sumt1=0, suma1=0, sumt2=0, suma2=0;
#define PRIVATE
#include "lipzamp1.h"
#undef PRIVATE

	  (void) times (&start1);
	  for (i=0; i<N_LIPZAMP_CHECKS; i++)
	    {
	      a = random () & RADIXM;
	      b = random () & RADIXM;
	      d = random () & RADIXM;
	      t = random () & RADIXM;
	      zaddmulp(&a,b,d,&t);
	      sumt1 += t;
	      suma1 += a;
#undef zaddmulp
#undef zaddmulpsq
	    }
	  (void) times (&end1);
#include "lipzamp2.h"
	  (void) times (&start2);
	  for (i=0; i<N_LIPZAMP_CHECKS; i++)
	    {
	      a = random ();
	      b = random ();
	      d = random ();
	      t = random ();
	      zaddmulp(&a,b,d,&t);
	      sumt2 += t;
	      suma2 += a;
#undef zaddmulp
#undef zaddmulpsq
	    }
	  (void) times (&end2);
	  {
	    long t1 = end1.tms_utime - start1.tms_utime,
	      t2 = end2.tms_utime - start2.tms_utime;
	    fprintf (out, "/* lipzamp1.h: %ld clock ticks; lipzamp2.h: %ld clock ticks. */\n",
		     t1, t2);
	    if (t1>t2)
	      lipzamp = 2;
	    if (sumt1!=sumt2 || suma1!=suma2)
	      fprintf (out, "/* */\n"); /* so that sum[at][12] is used by compiler. */
	  }
	}
      fprintf (out, "#define LIPZAMP %d\n", lipzamp);
    }
    {
#define N_LIPZDIV_CHECKS 1000000
      long numhi, numlo, denom, quot, rem, quot2, rem2;

      {
	EXTENDED fradix = (EXTENDED) RADIX;

#include "lipzdiv1.h"
	for (i=0; i<N_LIPZDIV_CHECKS; i++)
	  {
	    numhi = random () & RADIXM;
	    denom = random () & RADIXM;
	    if (numhi==denom)
	      continue;
	    if (numhi>denom)
	      {
		numlo = numhi;
		numhi = denom;
		denom = numlo;
	      }
	    numlo = random () & RADIXM;
	    { zdiv21 (numhi, numlo, denom, &quot, &rem); }
#undef zdiv21
#include "lipzdiv2.h"
#include "lipzamp2.h"
	    { zdiv21 (numhi, numlo, denom, &quot2, &rem2); }
#undef zdiv21
#undef zaddmulp
#undef zaddmulpsq
	    if (quot!=quot2 || rem!=rem2)
	      {
		lipzdiv = 2;
		fprintf (out,
			 "/* Use lipzdiv2.h, since lipzdiv1.h gives incorrect results. */\n");
		break;
	      }
	  }
	if (lipzdiv==1)
	  {
	    struct tms start1, end1, start2, end2;
	    long sumt1=0, suma1=0, sumt2=0, suma2=0;
#include "lipzdiv1.h"

	    (void) times (&start1);
	    for (i=0; i<N_LIPZDIV_CHECKS; i++)
	      {
		numhi = random () & RADIXM;
		denom = random () & RADIXM;
		if (numhi==denom)
		  continue;
		if (numhi>denom)
		  {
		    numlo = numhi;
		    numhi = denom;
		    denom = numlo;
		  }
		numlo = random () & RADIXM;
		{ zdiv21 (numhi, numlo, denom, &quot, &rem); }
		sumt1 += quot;
		suma1 += rem;
	      }
	    (void) times (&end1);
#undef zdiv21
#include "lipzamp2.h"
#include "lipzdiv2.h"
	    (void) times (&start2);
	    for (i=0; i<N_LIPZDIV_CHECKS; i++)
	      {
		numhi = random () & RADIXM;
		denom = random () & RADIXM;
		if (numhi==denom)
		  continue;
		if (numhi>denom)
		  {
		    numlo = numhi;
		    numhi = denom;
		    denom = numlo;
		  }
		numlo = random () & RADIXM;
		{ zdiv21 (numhi, numlo, denom, &quot, &rem); }
		sumt2 += quot;
		suma2 += rem;
	      }
	    (void) times (&end2);
#undef zdiv21
#undef zaddmulp
#undef zaddmulpsq
	    {
	      long t1 = end1.tms_utime - start1.tms_utime,
		t2 = end2.tms_utime - start2.tms_utime;
	      fprintf (out, "/* lipzdiv1.h: %ld clock ticks; lipzdiv2.h: %ld clock ticks. */\n",
		       t1, t2);
	      if (t1>t2)
		lipzdiv = 2;
	      if (sumt1!=sumt2 || suma1!=suma2)
		fprintf (out, "/* */\n"); /* so that sum[at][12] is used by compiler. */
	    }
	  }
      }
      fprintf (out, "#define LIPZDIV %d\n", lipzdiv);
    }
    {
#define N_LIPZMM_CHECKS 100000
      long a, b, n, r, r2;
      int lipzmm = 1;

      for (i=0; i<N_LIPZMM_CHECKS; i++)
	{
	  a = random ();
	  b = random ();
	  n = random ();
	  {
#include "lipzmm1.h"
	    r = lr;
	  }
	  {
#include "lipzamp2.h"
#include "lipzmm2.h"
	    r2 = lr;
#undef zaddmulp
#undef zaddmulpsq
	  }
	  if (r!=r2)
	    {
	      lipzmm = 2;
	      break;
	    }
	}
      fprintf (out, "#define LIPZMM %d\n", lipzmm);
    }
#undef RADIXROOTM
#undef RADIXROOT
#undef RADIXM
#undef RADIX
#undef NBITSH
#undef NBITS
  }
  {
    FILE *f;
    CHAR c;

    f = fopen ("n1bits.h", "w");
    if (f==NULL)
      fatal ("could not open file 'n1bits.h'");
    fprintf (f, "%d, ", 0);
    for (c=0; (++c)!=0; )
      {
	int i = 0;
	CHAR cc;

	for (cc=c; cc!=0; cc>>=1)
	  if (cc&1)
	    i++;
	fprintf (f, "%d, ", i);
	if ((c & 7)==7)
	  fputc ('\n', f);
      }
    fclose (f);
  }
  {
    FILE *f;
    CHAR c;

    f = fopen ("bitrev.h", "w");
    if (f==NULL)
      fatal ("could not open file 'bitrev.h'");
    fprintf (f, "%d, ", 0);
    for (c=0; (++c)!=0; )
      {
	int j;
	CHAR cc, cin;

	for (cc=0, cin=c, j=0; j<nbits_in_char; j++)
	  {
	    cc <<= 1;
	    if (cin & 1)
	      cc |= 1;
	    cin >>= 1;
	  }
	fprintf (f, "%d, ", (int)cc);
	if ((c & 7)==7)
	  fputc ('\n', f);
      }
    fclose (f);
  }
  /* defining the floating point format: */
  ieee754 = chars8bits;
  extp = 0;
  if (!ieee754) goto outfloat;
  if (sizeof(REAL)!=8)
    {
      ieee754 = 0;
      goto outfloat;
    }
  switch (sizeof(EXTENDED))
    {
    case 8:
      extp = 0;
      break;
    case 10:
      extp = 1;
      break;
    case 12:
      extp = 2;
      break;
    default:
      {
	char buf[128];
	
	sprintf(buf,"strange size for an EXTENDED: %ld",sizeof(EXTENDED));
	fatal(buf);
      }
    }
  informat = 1;
  for (i=0; i<NREALS; i++)
    {
      if (!memequal((char *)&xreal[i],sizeof(xreal[i]),
		    (char *)xshort[i],sizeof(xshort[i])))
	break;
    }
  if (i<NREALS)
    {
      informat = 2;
      for (i=0; i<NREALS; i++)
	{
	  if (!memequal((char *)&xreal[i],sizeof(xreal[i]),
			(char *)xrevshort[i],sizeof(xrevshort[i])))
	    break;
	}
    }
  if (i<NREALS)
    {
      informat = 3;
      for (i=0; i<NREALS; i++)
	{
	  if (!memequal((char *)&xreal[i],sizeof(xreal[i]),
			(char *)xchar[i],sizeof(xchar[i])))
	    break;
	}
    }
  if (i<NREALS)
    {
      informat = 4;
      for (i=0; i<NREALS; i++)
	{
	  if (!memequal((char *)&xreal[i],sizeof(xreal[i]),
			(char *)xintelshort[i],sizeof(xintelshort[i])))
	    break;
	}
    }
  if (i<NREALS)
    {
      ieee754 = 0;
      goto outfloat;
    }
 outfloat:
  fprintf(out,"#define FLOAT_FORMAT %s\n",ieee754?"1 /* IEEE754 */":"2 /* OTHER_FLOAT */");
  fprintf(out,"#define EXTP %d\n",extp);
  fprintf(out,"#define FLOAT_LAYOUT %d\n",informat);
  fprintf(out,"#define EXTENDED %s\n",EXTENDED_NAME);
  fprintf(out,"#define REAL %s\n",REAL_NAME);
  fprintf(out,"#define EXT_SIZE %ld\n",sizeof(EXTENDED));
  fprintf(out,"#define REAL_SIZE %ld\n",sizeof(REAL));
  fprintf(out,"#define LONG_SIZE %ld\n",sizeof(long));
  putifloor(out);
  fprintf(out,"\n");
  /* sizeof longs, shorts, etc. */
  align = alignment();
  for (i=0, n=1; n!=0; i++, n<<=1);
  nBitsLong = i;
  fprintf(out,"#define NBITS_IN_LONG (%d)\n",i);
  i /= 2;
  fprintf(out,"#define N_BITS_IN_HALF_LONG (%d)\n",i);
  for (n=0; i!=0; i--, n = 2L*n+1L);
  fprintf(out,"#define MASK_HALF_LONG ((long)0x%lx)\n",n);
  for (i=1; k = 2*i; i=k);
  fprintf(out,"#define MAX_TWO_TO_THE_N_INT (0x%x)\n",i);
  for (m=1; n = 2L*m; m=n);
  maxTwoNLong = m;
  fprintf(out,"#define MAX_TWO_TO_THE_N_LONG ((long)0x%lx)\n",m);
  for (i=1; k = 2*i+1, k>=0; i=k);
  fprintf(out,"#define MAX_SIGNED_INT (0x%x)\n",i);
  for (m=1; n = 2L*m+1L, n>=0; m=n);
  fprintf(out,"#define MAX_SIGNED_LONG ((long)0x%lx)\n",m);
  for (m=-1; (n = (2*m))<0;)
    m = n;
  fprintf(out,"#define MIN_SIGNED_LONG ((long)0x%lx)\n",(unsigned long)m);
  for (i=0, addr=(char *)1; addr!=0;
       i++, addr = (char *)(((unsigned long)addr)<<1));
  fprintf(out,"#define NBITS_IN_ADDR (%d)\n",i);
  nba = i;
  for (ndigs=0, ul=1L; lastul = ul, ul *= 10L, (ul/10L)==lastul;)
    ndigs++;
  fprintf(out,"#define NDIGS_IN_ULONG (%d)\n",ndigs);
  fprintf(out,"#define DECL_POW10(name) EXTENDED name[]={ \\\n 1E-10");
  for (i=-9; i<=11 || i<=ndigs; i++)
    fprintf(out,", 1E%d",i);
  fprintf(out," \\\n};\n");
  
  prepdiv(out);

  fprintf(out,"#%s MALLOC_0_0\n",(malloc(0L)==NULL)?"define":"undef");

  /*align2 = (long)(((char *)&peal.expr[1])-((char *)&peal.expr[0]));*/
  gap = (int)(((char *)&pe.link)-(((char *)&pe.magic)+sizeof(SHORT)));
  /*align2 = pgcd(align,align2);*/
  for (i=0, al=align; (al & 1L)==0L; i++, al >>= 1);
  fprintf(out,"#define NLOWBITS (%d)\n",i);
  nlowbits = i;
  fprintf(out,"#define MAP_CHAIN_LENGTH %d\n",nba-i);
  fprintf(out,"#define MAP_STK0");
  for (k=i; k<nba; k++)
    fprintf(out," NOEXPR,");
  fputc('\n',out);
  fprintf(out,"#define EXPR_GAP %d\n",gap);
  gap = (((unsigned long)((char *)(&pe+1) - (char *)&pe)) % align);
  if (gap!=0)
    gap = align-gap;
  fprintf(out,"#define EXPR_PADDING %d\n",gap);
  
  gap = ((unsigned long)(((char *)blk.b.data)-(char *)&blk)) % align;
  if (gap!=0)
    gap = align-gap;
  fprintf(out,"#define BLOCK_GAP %d\n",gap);

  gap = align-(((char *)&bucket.occupancy[MAX_GENERATIONS])-((char *)&bucket));
  while (gap<0)
    gap += align;
  fprintf(out,"#define BUCKET_GAP %d\n",gap);
  gap = align-(((char *)&bhb.occupancy[MAX_GENERATIONS])-((char *)&bhb));
  while (gap<0)
    gap += align;
  fprintf(out,"#define BH_BUCKET_GAP %d\n",gap);
  gap = (((unsigned long)((char *)&bhb.bh[1] - (char *)&bhb.bh[0])) % align);
  if (gap!=0)
    gap = align-gap;
  fprintf(out,"#define BH_PADDING %d\n",gap);
  gap = align-(((char *)&page.top)-((char *)&page)+sizeof(char *));
  while (gap<0)
    gap += align;
  fprintf(out,"#define PAGE_GAP %d\n",gap);
  fprintf(out,"#%s ASCENDING_STACK\n",
	  descendingstack()?"undef":"define");
  prepalloc();
  byteordering();
  byteordering_short();
  fputc('\n',out);
  if (markbit (out, nlowbits))
    {
      fclose (out);
      fatal ("Cannot determine how to hack a mark bit into ExprPtrs");
    }
  file = fopen("/etc/passwd","r");
  if (file!=NULL)
    {
      fclose(file);
      file = fopen("/usr/include/netinet/ip.h","r");
      if (file!=NULL)
	{
	  fclose(file);
	  fprintf(out,"#define OS UNIX_BSD\n");
	}
      else
	fprintf(out,"#define OS UNIX_SYS_V\n");
      fclose(out);
      InitRandom(0);
    }
  else
    {
      file = fopen("s:startup-sequence","r");
      if (file!=NULL)
	{
	  fclose(file);
	  fprintf(out,"#define OS AMIGA_OS\n");
	  /* other options are set by the mksys.cli shell script */
	  fprintf(out,"#undef HAS_ALLOCA\n");
	  fprintf(out,"#define SETJMP(jb) setjmp(jb)\n");
	  fprintf(out,"#define LONGJMP(jb,n) longjmp(jb,n)\n");
	  /* even if it had alloca(), we should not use it, because
	     it uses precious stack space */
	  fclose(out);
	  InitRandom(0);
	}
      else
	{
	  file = fopen(":Utils:init.ml","r");
	  if (file!=NULL)
	    {
	      fclose(file);
	      fprintf(out,"#define OS MAC\n");
	      fprintf(out,"#ifdef THINK_C\n");
	      fprintf(out,"#define HAS_PROTOTYPES\n");
	      fprintf(out,"#define HAS_HI_PROTOS\n");
	      fprintf(out,"#define MERGE(a,b) a\\\nb\n");
	      fprintf(out,"#undef HAS_VOLATILE\n");
	      fprintf(out,"#else\n");
	      fprintf(out,"#define HAS_PROTOTYPES\n");
	      fprintf(out,"#define MERGE(a,b) a\\\nb\n");
	      fprintf(out,"#define HAS_VOLATILE\n");
	      fprintf(out,"#endif\n");
	      fprintf(out,"#undef BRAIN_DAMAGED_LONGJMP\n");
	      fprintf(out,"#undef HAS_ALLOCA\n");
	      /* even if it had alloca(), we should not use it, because
		 it uses precious stack space */
	      fprintf(out,"#undef HAS_STDARG\n");
	      /* some compilers on the Mac have stdarg.h (Think C>=4.0,
		 MPW C>=3.0), but not all (MPW C 2.0), so we currently
		 undefine HAS_STDARG */
	      fprintf(out,"#define HAS_VARARGS\n");
	      fprintf(out,"#define HAS_STDLIB\n");
	      fprintf(out,"#define SETJMP(jb) setjmp(jb)\n");
	      fprintf(out,"#define LONGJMP(jb,n) longjmp(jb,n)\n");
	      /* on the Mac, string functions are found in <string.h>,
		 not <strings.h> */
	      fprintf(out,"#define HAS_STRING_H\n");
	      fclose(out);
	      InitRandom(1);
	    }
	  else
	    {
	      fclose(out);
	      InitRandom(0);
	      fatal("Unknown OS");
	    }
	}
    }
  done(0);
}

int descendingstack1(cp)
     char *cp;
{
  char c;
  
  return (&c)<cp;
}

int descendingstack()
{
  char c;
  
  return descendingstack1(&c);
}

typedef void (*divfun) P((int a,int b,int *q,int *r));
     
     void divplain P((int,int,int *,int *));
     void divplain(a,b,q,r)
     int a,b,*q,*r;
{
  *q = a/b;
  *r = a%b;
}

void divb P((int,int,int *,int *));
     void divb(a,b,q,r)
     int a,b,*q,*r;
{ /* correct divplain if it is only wrong when b<0 */
  if (b<0)
    {
      divplain(a,-b,q,r);
      /* now, a = -bq+r, sign(r)=-sign(b), 0<=|r|<|b| */
      *q = -*q;
      /* now, a = bq+r, sign(r)=-sign(b), 0<=|r|<|b| */
      if (*r!=0)
	{
	  --*q;
	  *r += b;
	}
      /* now, a = -bq+r, sign(r)=sign(b), 0<=|r|<|b| */
    }
  else
    divplain(a,b,q,r);
}

void diva P((int,int,int *,int *));
     void diva(a,b,q,r)
     int a,b,*q,*r;
{ /* correct divplain if it is only wrong when a<0 */
  if (a<0)
    {
      divplain(-a,b,q,r);
      /* now, a = -bq-r, sign(r)=sign(b), 0<=|r|<|b| */
      *q = -*q;
      /* now, a = bq-r, sign(r)=sign(b), 0<=|r|<|b| */
      if (*r!=0)
	{
	  --*q;
	  *r = b-*r;
	}
      /* now, a = bq-r, sign(r)=sign(b), 0<=|r|<|b| */
    }
  else divplain(a,b,q,r);
}

void divab P((int,int,int *,int *));
     void divab(a,b,q,r)
     int a,b,*q,*r;
{ /* correct divplain if it is wrong when a<0 or b<0 */
  if (b<0)
    {
      if (a<0)
	{
	  divplain(-a,-b,q,r);
	  /* now, a=bq-r, 0<=r<-b */
	  *r = -*r;
	}
      else
	{ /* taken from divb */
	  divplain(a,-b,q,r);
	  *q = -*q;
	  if (*r!=0)
	    {
	      --*q;
	      *r += b;
	    }
	}
    }
  else
    {
      if (a<0)
	{ /* taken from diva */
	  divplain(-a,b,q,r);
	  *q = -*q;
	  if (*r!=0)
	    {
	      --*q;
	      *r = b-*r;
	    }
	}
      else divplain(a,b,q,r);
    }
}

struct { int a,b,q,r; } test[3] = {
  {7, -3, -3, -2},
  {-7, 3, -3, 2},
  {-7, -3, 2, -1}};

int testdiv(div)
     divfun div;
{
  int q,r,i;
  
  /* the quotient should have the smallest possible
     absolute value (the remainder has the same sign as b). */
  for (i=0; i<3; i++)
    {
      (*div)(test[i].a,test[i].b,&q,&r);
      if (q!=test[i].q || r!=test[i].r)
	return 0;
    }
  return 1;
}

struct alt { char *name; divfun div; } alt[5] = {
  {"DIV_CORRECT", divplain},
  {"DIV_CORRECT_B", divb},
  {"DIV_CORRECT_A", diva},
  {"DIV_CORRECT_AB", divab},
  {NULL, (void (*)())0}};

void prepdiv(out)
     FILE *out;
{
  struct alt *altp;
  
  for (altp = alt; altp->name!=NULL; altp++)
    if (testdiv(altp->div))
      {
	fprintf(out,"#define %s\n",altp->name);
	return;
      }
}

int memequal(b1,len1,b2,len2)
     char *b1,*b2;
     int len1,len2;
{
  if (len1!=len2)
    return 0;
  while (len1--!=0)
    if (*b1++ != *b2++)
      return 0;
  return 1;
}

long ifloor0(x)
     double x;
{
  return (long)(double)x;
}

long ifloor1(x)
     double x;
{
  return (long)floor(x);
}

struct tstifloor {
  long (*ifloorfun) P((double x));
  char *ifloordef;
} tstifloor[] = {
  { ifloor0,
      "(long)(EXTENDED)(x)" },
  { ifloor1,
      "(long)floor(x)" },
  { NULL,
      NULL }
};

int ifloortst(fun)
     long (*fun) P((double x));
{
  double x;
  
  x = 3.0;
  if ((*fun)(x)!=3)
    return 0;
  x = 3.7;
  if ((*fun)(x)!=3)
    return 0;
  x=-0.2;
  if ((*fun)(x)!=-1)
    return 0;
  x = -1.0;
  if ((*fun)(x)!=-1)
    return 0;
  x = 0.0;
  if ((*fun)(x)!=0)
    return 0;
  return 1;
}

void putifloor(out)
     FILE *out;
{
  struct tstifloor *tst;
  
  for (tst=tstifloor; tst->ifloordef!=NULL; tst++)
    {
      if (ifloortst(tst->ifloorfun))
	{
	  fprintf(out,"#define DEFAULT_IFLOOR(x) %s\n",tst->ifloordef);
	  return;
	}
    }
  fatal("brain-damage: no way of computing floor of double, giving int");
}

#define NALIGNTESTS 2000

long pgcd(a,b)
     long a,b;
{
  long r;
  
  if (a<b)
    {
      r = a; a = b; b = r;
    }
  while (r = a%b, r!=0)
    {
      a = b;
      b = r;
    }
  return b;
}

long ppcm(a,b)
     long a,b;
{
  return (a*b)/pgcd(a,b);
}

long alignment()
{
  int i;
  long align,d;
  char *old,*new;
  
  old = (char *)malloc(1L);
  new = (char *)malloc(1L);
  align = new-old;
  if (align<0)
    align = -align;
  for (i = 0; i<NALIGNTESTS; i++)
    {
      old = new;
      new = (char *)malloc(rand () & 0x1ff);
      d = new-old;
      if (d<0)
	d = -d;
      d = pgcd(d,align);
      if (d!=align)
	{
	  align = d;
	  i = 0;
	}
    }
  return align;
}

void prepalloc()
{
  long mbstride,stride,hdsize,adder;
  MBlock mb[2];
  
  /*align = alignment();*/
  mbstride = (long)(((char *)&mb[1])-((char *)&mb[0]));
  stride = ppcm(align,mbstride);
  fprintf(out,"\n#define ALIGNMENT (%ld)\n",align);
  fprintf(out,"#define MBSTRIDE (%ld)\n",stride);
  /* the real size of a block must be the least multiple of align
     that is at least size + sizeof(MB_size) */
  hdsize = (long)(((char *)&mb->MB_next)-((char *)mb));
  fprintf(out,"#define HDSIZE (%ld)\n",hdsize);
  if ((align & -align)==align) /* if align is a power of two, modulo is an and */
    fprintf(out,"#define MODALIGN(x) ((x) & (-%ldL))\n",align);
  else
    fprintf(out,"#define MODALIGN(x) ((x) %% (%ldL))\n",align);
  if ((stride & -stride)==stride) /* if stride is a power of two, modulo is an and */
    fprintf(out,"#define MODMBSTRIDE(x) ((x) & (-%ldL))\n",stride);
  else
    fprintf(out,"#define MODMBSTRIDE(x) ((x) %% (%ldL))\n",stride);
  adder = hdsize+stride-1;
  fprintf(out,"#define MBREALSIZE(size) MODMBSTRIDE(size+%ldL)\n",adder);
  /* define the offset from the beginning of a MZone to the first
     MBlock */
  adder = ((sizeof(MZone)+hdsize+align-1)/align)*align-hdsize;
  fprintf(out,"#define MZHDSIZE (%ld)\n",adder);
}

void byteordering()
{
  long l;
  char *s;
  int i,j;
  int scrambled;

  fputc('\n',out);
  s = (char *)&l;
  for (i=0, l=0; i<sizeof(long); i++)
    l = (l<<8) | (char)i;
  for (scrambled=0, i=0; i<sizeof(long); i++)
    if (scrambled=(s[i]!=(char)i))
      break;
  if (scrambled)
    {
      fprintf(out,"#define BEGIN_FROM_INT(l) CHAR MERGE(_ca_,l)[%ld] = { ",
	      sizeof(long));
      for (i=0; i<sizeof(long); i++)
	for (j=0; j<sizeof(long); j++)
	  if (s[j]==(char)i)
	    {
	      fprintf(out,"((CHAR *)&l)[%d], ",j);
	      break;
	    }
      fprintf(out,"}\n");
      fprintf(out,"#define FROM_INT(l) MERGE(_ca_,l)\n");
      fprintf(out,"#define END_FROM_INT(l)\n");
      fprintf(out,"#define BEGIN_TO_INT(l) CHAR MERGE(_cb_,l)[%ld]\n",
	      sizeof(long));
      fprintf(out,"#define TO_INT(l) MERGE(_cb_,l)\n");
      fprintf(out,"#define END_TO_INT(l)");
      for (i=0; i<sizeof(long); i++)
	fprintf(out," ((CHAR *)&l)[%d] = MERGE(_cb_,l)[%d];",i,(int)s[i]);
      fprintf(out,"\n");
    }
  else
    {
      fprintf(out,"#define BEGIN_FROM_INT(l)\n");
      fprintf(out,"#define FROM_INT(l) ((CHAR *)&l)\n");
      fprintf(out,"#define END_FROM_INT(l)\n");
      fprintf(out,"#define BEGIN_TO_INT(l)\n");
      fprintf(out,"#define TO_INT(l) ((CHAR *)&l)\n");
      fprintf(out,"#define END_TO_INT(l)\n");
    }
}

void byteordering_short()
{
  short l;
  char *s;
  int i,j;
  int scrambled;

  fputc('\n',out);
  s = (char *)&l;
  for (i=0, l=0; i<sizeof(short); i++)
    l = (l<<8) | (char)i;
  for (scrambled=0, i=0; i<sizeof(short); i++)
    if (scrambled=(s[i]!=(char)i))
      break;
  if (scrambled)
    {
      fprintf(out,"#define BEGIN_FROM_SHORT(l) CHAR MERGE(_ca_,l)[%ld] = { ",
	      sizeof(short));
      for (i=0; i<sizeof(short); i++)
	for (j=0; j<sizeof(short); j++)
	  if (s[j]==(char)i)
	    {
	      fprintf(out,"((CHAR *)&l)[%d], ",j);
	      break;
	    }
      fprintf(out,"}\n");
      fprintf(out,"#define FROM_SHORT(l) MERGE(_ca_,l)\n");
      fprintf(out,"#define END_FROM_SHORT(l)\n");
      fprintf(out,"#define BEGIN_TO_SHORT(l) CHAR MERGE(_cb_,l)[%ld]\n",
	      sizeof(short));
      fprintf(out,"#define TO_SHORT(l) MERGE(_cb_,l)\n");
      fprintf(out,"#define END_TO_SHORT(l)");
      for (i=0; i<sizeof(short); i++)
	fprintf(out," ((CHAR *)&l)[%d] = MERGE(_cb_,l)[%d];",i,(int)s[i]);
      fprintf(out,"\n");
    }
  else
    {
      fprintf(out,"#define BEGIN_FROM_SHORT(l)\n");
      fprintf(out,"#define FROM_SHORT(l) ((CHAR *)&l)\n");
      fprintf(out,"#define END_FROM_SHORT(l)\n");
      fprintf(out,"#define BEGIN_TO_SHORT(l)\n");
      fprintf(out,"#define TO_SHORT(l) ((CHAR *)&l)\n");
      fprintf(out,"#define END_TO_SHORT(l)\n");
    }
}

#define INIT_RANDOM
#include "random.src"

void fatal(msg)
     char *msg;
{
  fputs(msg,stderr);
  fputc('\n',stderr);
  done(10);
}

void done(n)
     int n;
{
  exit(n);
}
