/* Support functions for Standard ML modules (not used), plus a few utilities, himmlrun version.
   Copyright (C) 2003-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "eval.h"
#include "type.h"
#include "expr.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rmodules.h"
#ifdef HAS_PROTOTYPES
#include "smodules.h"
#endif

#undef DEBUG

/* a TDG is either true, false
   or KONS(cond,(l,h)), meaning (if cond then h else l),
   or KTUPLE2(cons,(l,h)), meaning not(if cond then h else l)
*/

EXTERN ExprPtr tdgAlt; /* = (false,true) */

PUBLIC Boolean tdgVarLess(x,y)
IN CONST ExprPtr x,y;
{ /* for module saving purposes, we need to have an order on tdg variables
     that is independent of, say, the system ordering <. */
  REGISTER ExprPtr xx,yy;

  xx = (x->kind==(CHAR)E_REF)?REFVAL(x):x;
  yy = (y->kind==(CHAR)E_REF)?REFVAL(y):y;
  switch (CompareStrings(xx,yy))
    {
    case -1:
      return true;
    case 0:
      return x->kind<y->kind;
    default:
      return false;
    }
}

PUBLIC ExprPtr tdgNot(t)
IN CONST ExprPtr t;
{
  if (t==trueValue)
    return falseValue;
  if (t==falseValue)
    return trueValue;
  if (t->kind==(CHAR)E_CONS)
    {
      t = Ktuple2(CAR(t),CDR(t));
      return KEEP(t);
    }
  return KONS(CAR(t),CDR(t));
}

#define tdgChs(sign,t) ((sign)?(t):tdgNot(t))

PRIVATE ExprPtr makeTdg(pos,x,l,h)
IN Boolean pos;
IN CONST ExprPtr x,l,h;
	/* normalize a TDG by ensuring the h part is positive */
{
  if (l==h)
    return tdgChs(pos,l);
  if (h==trueValue || h->kind==(CHAR)E_CONS)
    {
      l = KONS (l,h);
      if (pos)
	l = Kons (x, l);
      else l = Ktuple2 (x, l);
      /*l = Kpair((pos?E_CONS:E_TUPLE_2),x,l);*/
    }
  else
    {
      l = KONS(tdgNot(l),tdgNot(h));
      if (pos)
	l = Ktuple2 (x,l);
      else l = Kons (x,l);
      /*l = Kpair((pos?E_TUPLE_2:E_CONS),x,l);*/
    }
  return KEEP(l);
}

PRIVATE ExprPtr tdgOr1 P((IN CONST ExprPtr t1,IN CONST ExprPtr t2,
 			  IN Boolean sign1,IN Boolean sign2));

DEF_PRIVATE_ACTION_1ARG(tdgor)
{
  struct tdgor *data;

  data = (struct tdgor *)ENV_UNDERLYING(env);
  return tdgOr1(data->t1,data->t2,data->sign1,data->sign2);
}
ACTION_END

PRIVATE ExprPtr tdgOrk(t1,t2,sign1,sign2)
IN CONST ExprPtr t1,t2;
IN Boolean sign1,sign2;
{
  struct tdgor *data = (struct tdgor *)alloctmp(sizeof(struct tdgor));

  data->t1 = t1;
  data->t2 = t2;
  data->sign1 = sign1;
  data->sign2 = sign2;
  return CallCC(ACTION_NAME(tdgor),(ExprPtr)data,true);
}

PRIVATE ExprPtr tdgOr1(t1,t2,sign1,sign2)
IN CONST ExprPtr t1,t2;
IN Boolean sign1,sign2;
/* if we write [sign]t the quantity computed by tdgChs(sign,t),
   that is, t is pos==true, and tdgNot(t) otherwise,
   then this function computes [sign1]t1 or [sign2]t2
   */
{
  ExprPtr x1,x2,v1,v2;
  ExprPtr *ep;
  COUNTER i;

  if (STKFULL(i))
    return tdgOrk(t1,t2,sign1,sign2);
  if (t1==t2)
    return (sign1==sign2)?tdgChs(sign1,t1):trueValue;
  if (t1==(sign1?trueValue:falseValue) ||
      t2==(sign2?falseValue:trueValue))
    return tdgChs(sign1,t1);
  if (t2==(sign2?trueValue:falseValue) ||
      t1==(sign1?falseValue:trueValue))
    return tdgChs(sign2,t2);
  if (t1>t2)
    {
      Boolean auxs;

      x1 = t1; t1 = t2; t2 = x1;
      auxs = sign1; sign1 = sign2; sign2 = auxs;
    }
  i = (((((unsigned long)t1)+((unsigned long)t2))%TDG_OR_HASH_SIZE) << 2)
	+(sign1<<1)+sign2;
  i += (i<<1); /* multiply i by 3 */
  ep = &AVAL(tdgOrHashTable)->expr[i];
  if (ep[0]==t1 && ep[1]==t2)
    return ep[2];
  x1 = CAR(t1);
  v1 = CDR(t1);
  x2 = CAR(t2);
  v2 = CDR(t2);
  if (x1==x2)
    {
      sign1 = ((t1->kind==(CHAR)E_CONS)==sign1);
      sign2 = ((t2->kind==(CHAR)E_CONS)==sign2);
      x1 = makeTdg(true,x1,tdgOr1(CAR(v1),CAR(v2),sign1,sign2),
			   tdgOr1(CDR(v1),CDR(v2),sign1,sign2));
    }
  else if (tdgVarLess(x1,x2))
    {
      sign1 = ((t1->kind==(CHAR)E_CONS)==sign1);
      x1 = makeTdg(true,x1,tdgOr1(CAR(v1),t2,sign1,sign2),
			   tdgOr1(CDR(v1),t2,sign1,sign2));
    }
  else
    {
      sign2 = ((t2->kind==(CHAR)E_CONS)==sign2);
      x1 = makeTdg(true,x2,tdgOr1(CAR(v2),t1,sign2,sign1),
			   tdgOr1(CDR(v2),t1,sign2,sign1));
    }
#if STOP_AND_COPY_VECTORS
  ep = &AVAL(tdgOrHashTable)->expr[i];
#endif
  ep[0] = t1;
  ep[1] = t2;
  ep[2] = x1;
  HITBARRIER(tdgOrHashTable);
  return x1;
}

PUBLIC ExprPtr tdgOr(t1,t2)
IN CONST ExprPtr t1,t2;
{
  ENTER;

  t1 = tdgOr1(t1,t2,true,true);
  KEEPONLY(t1);
  return t1;
}

PUBLIC ExprPtr tdgAnd(t1,t2)
IN CONST ExprPtr t1,t2;
{
  ENTER;

  t1 = tdgNot(tdgOr1(t1,t2,false,false));
  KEEPONLY(t1);
  return t1;
}

PUBLIC ExprPtr tdgImply(t1,t2)
IN CONST ExprPtr t1,t2;
{
  ENTER;

  t1 = tdgOr1(t1,t2,false,true);
  KEEPONLY(t1);
  return t1;
}

DEF_PRIVATE_ACTION_1ARG(tdginst)
{
  struct tdginst *data;

  data = (struct tdginst *)ENV_UNDERLYING(env);
  return tdgInst(data->t,data->subst,data->defselfp);
}
ACTION_END

PRIVATE ExprPtr tdgInst1(t,subst,defselfp)
IN CONST ExprPtr t;
IN CONST ExprPtr subst;
IN Boolean defselfp;
{
  struct tdginst *data = (struct tdginst *)alloctmp(sizeof(struct tdginst));

  data->t = t;
  data->subst = subst;
  data->defselfp = defselfp;
  return CallCC(ACTION_NAME(tdginst),(ExprPtr)data,true);
}

PUBLIC ExprPtr tdgInst(t,subst,defselfp)
IN CONST ExprPtr t;
IN CONST ExprPtr subst;
IN Boolean defselfp;
{ /* subst is a map from names to tdgs;
     by default, names are mapped to themselves
     if defselfp==true, otherwise to true */
  ExprPtr x,y,v,l,h;
  Boolean found;
  ENTER;

  if (STKFULL(found))
    return tdgInst1(t,subst,defselfp);
again:
  if (t==trueValue || t==falseValue)
    return t;
  v = CDR(t);
  x = CAR(t);
  y = MapGet(x,subst,&found);
  if (!found)
    {
      if (defselfp)
	{
	  l = tdgInst(CAR(v),subst,defselfp);
	  h = tdgInst(CDR(v),subst,defselfp);
	  if (l==CAR(v) && h==CDR(v))
	    ;
	  else
	    t = makeTdg((t->kind==(CHAR)E_CONS),x,l,h);
	}
      else
	{
	  t = CDR(v);
	  goto again;
	}
    }
  else if (y==trueValue)
    {
      t = CDR(v);
      goto again;
    }
  else if (y==falseValue)
    {
      t = CAR(v);
      goto again;
    }
  else
    {
      l = tdgInst(CAR(v),subst,defselfp);
      h = tdgInst(CDR(v),subst,defselfp);
      if (y->kind==(CHAR)E_CONS && CAR(y)==x && CDR(y)==tdgAlt)
	{ /* no substitution: x is unknown */
	  if (l==CAR(v) && h==CDR(v))
	    ;
	  else
	    t = makeTdg((t->kind==(CHAR)E_CONS),x,l,h);
	}
      else
	{
	  l = tdgAnd(tdgImply(y,h),tdgOr(y,l));
	  t = tdgChs((t->kind==(CHAR)E_CONS),l);
	}
    }
  KEEPONLY(t);
  return t;
}

PUBLIC ExprPtr funEqTdg(f,state)
IN CONST SymPtr f;
IN CONST ExprPtr state;
{
  ExprPtr t,tt;
  Boolean found;

  t = MapGet(f,S_NONEQS(state),&found);
  if (found)
    {
      ENTER;
      
      tt = tdgInst(t,S_NONEQS(state),false);
      if (t!=tt)
	{
	  if (tt==trueValue)
	    setMapRemove(f,&S_NONEQS(state));
	  else
	    setMapPut(f,tt,&S_NONEQS(state));
	  t = tt;
	}
      QUIT; /* t is kept in the state */
    }
  else
    t = trueValue;
  return t;
}

PUBLIC ExprPtr funNotRefTdg(f,state)
IN CONST SymPtr f;
IN CONST ExprPtr state;
{
  ExprPtr var,t;
  ENTER;
  
  if (f==refSym || f==arraySym || f==tableSym)
    return falseValue;
  var = Kref(E_REF,f);
  t = funEqTdg(KEEP(var),state);
  QUIT; /* t is kept in the state or is true */
  return t;
}

