/* Static structures for Hyacc parser engine.
   Copyright (C) 2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

typedef struct yyltype
{
  COUNTER pos[4];
  /*
    int first_line;
    int first_column;
    int last_line;
    int last_column;
  */
} yyltype;

#define YYLTYPE yyltype

typedef struct hydata {
  ExprPtr pos; /* : intarray; of length 4, should contain first_line, first_column, last_line, last_column;
		*/
  ExprPtr lexarg; /* : 'a? (typically yydata); argument to lex below. */
  ExprPtr lex; /* : 'a? -> int; typically yydata -> int, in case lex is hlex. */
  ExprPtr default_value; /* : 'val? */
  ExprPtr yylval; /* : 'val? ref, used to return values from lex; initialized by hyparse to default_value. */
  ExprPtr yyerror; /* : string list -> unit */
  COUNTER initdepth; /* by default, YYINITDEPTH, e.g., 200 */
  COUNTER maxdepth; /* by default, YYMAXDEPTH, e.g., 10000 */
  COUNTER yyerrstatus; /* initially, 0: number of tokens to shift before error messages enabled */
  COUNTER yychar;
  COUNTER yystacksize; /* initialized to initdepth on entry to yyparse(). */
  COUNTER yystate;
  COUNTER yyn;
  COUNTER yynerrs;
  SHORT *yyss; /* state stack, initialized to NULL */
  SHORT *yyssp; /* pointer to last element in yyss, as for yyvsp and yyvs. */
  ExprPtr *yyvs; /* value stack, initialized to NULL */
  ExprPtr *yyvsp; /* pointer to last element in yyvs (not just after the last element, rather the last
		     element itself), if yyvs!=NULL. */
  YYLTYPE *yyls; /* location stack, initialized to NULL */
  YYLTYPE *yylsp; /* pointer to last element in yyls, as for yyvsp and yyvs. */
  CHAR yydebug;
} hydata;

typedef struct hytables {
  ExprPtr yyactiontable; /* vector of actions, i.e. closures of types hyaccdata -> YYSTYPE option;
			    returning NONE means we want to trigger a parse error. */
  ExprPtr yytname; /* : string array */
  COUNTER yyfinal;
  COUNTER yyflag; /* usually -32768 */
  COUNTER yyntbase;
  COUNTER max_user_token;
  COUNTER nsyms;
  COUNTER yylast;
  COUNTER yyr1_n;
  COUNTER yyr2_n;
  COUNTER yydefact_n;
  COUNTER yydefgoto_n;
  COUNTER yypact_n;
  COUNTER yypgoto_n;
  COUNTER yytable_n;
  COUNTER yycheck_n;
  COUNTER yytranslate_n;
  COUNTER yyprhs_n;
  COUNTER yyrhs_n;
  COUNTER yyrline_n;
  SHORT *yyr1; /* inlined, i.e. points to inside this very record. */
  SHORT *yyr2; /* same */
  SHORT *yydefact; /* same */
  SHORT *yydefgoto; /* same */
  SHORT *yypact; /* same */
  SHORT *yypgoto; /* same */
  SHORT *yytable; /* same */
  SHORT *yycheck; /* same */
  SHORT *yytranslate; /* same */
  SHORT *yyprhs; /* same */
  SHORT *yyrhs; /* same */
  SHORT *yyrline; /* same */
} hytables;
