/* Definitions for Union-Find structures (unused).
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
type (''a,'b) union_find_tree =
    |[link : ''a -m> ''a, (* links going up towards roots (aka colors) *)
      info : ''a -m> num * 'b (* maps from colors to data: the number of elements
				in this color, and other data *)
      ]|
*/
/* implemented as an updatable map (functions NGet1, etc.), where
   maplets point either to other maplets (link field) or to a COUNTED
   reference (the info field).
*/

typedef ExprPtr uFindPtr,colorPtr;
/* uFindPtr : union-find trees,
   colorPtr : maplet representing the root of an inverted tree for some x
*/
typedef ExprPtr (*infoMergeFunc) P((IN CONST colorPtr x1,IN CONST colorPtr x2,
				    INOUT ExprPtr info1,IN CONST ExprPtr info2,
				    char *data));
	/* the infos being the *uncounted* infos */
#define INFOMERGEFUNC(fun) \
PRIVATE ExprPtr fun P((IN CONST colorPtr x1,IN CONST colorPtr x2, \
		       INOUT ExprPtr info1,IN CONST ExprPtr info2, \
		       char *_data)); \
PRIVATE ExprPtr fun(x1,x2,info1,info2,_data) \
IN CONST colorPtr x1,x2; \
INOUT ExprPtr info1,info2; \
INOUT char *_data;
#define INFOMERGEDATA(type) ((type)_data)

#define newUFind() NIL
IMPORT colorPtr uFind1 P((IN CONST uFindPtr utree,IN CONST ExprPtr e,OUT Boolean *foundp));
#define colorOne(color) MAPLETX(color)
#define colorInfo(color) MAPLETY(color)
#define uinfoCard(info) COUNTEDN(info)
#define uinfoData(info) COUNTEDVAL(info)
IMPORT void uMerge P((IN CONST uFindPtr utree,
		      IN CONST colorPtr x,IN CONST colorPtr y,
		      IN CONST infoMergeFunc doinfo,IN char *data));
IMPORT Boolean uColored P((IN CONST uFindPtr utree,IN CONST ExprPtr x));
IMPORT void uNewColor P((INOUT uFindPtr utree,
			     IN CONST ExprPtr x,IN CONST ExprPtr info));

	/* congruence closure */

	/* we need a notion of sets, for which uniqueness is not crucial,
	  since it will be handled automatically. Instead of using lists,
	  for which union (concatenation) is slow, we use heaps. */
/*
datatype ''a heap =
    H0
  | H1 of ''a
  | H2 of ''a heap * ''a heap

	H0 will be represented by NIL,
	H1(x) will be E_TUPLE_1(x), and
	H2(h1,h2) will be E_TUPLE_2.

	We do not need heaps to admit equality, so sharing is not used.
*/
#define H0 NIL
#define H1 E_TUPLE_1
#define H2 E_TUPLE_2

IMPORT ExprPtr heapConcat P((IN CONST ExprPtr h1,IN CONST ExprPtr h2));
typedef CONST Boolean (*heapSweepFunc) P((IN CONST ExprPtr elt,
					  INOUT CHAR *data));
#define HEAPSWEEPFUNC(fun) \
PRIVATE Boolean fun P((IN CONST ExprPtr x,INOUT CHAR *_data)); \
PRIVATE Boolean fun(x,_data) \
IN CONST ExprPtr x; \
INOUT CHAR *_data;
#define HEAPSWEEPDATA(type) ((type)_data)

IMPORT Boolean heapSweep P((IN CONST ExprPtr h,heapSweepFunc f,CHAR *data));
/*
  the type of objects we consider in congruence closure are heaps, too.
  H0 is a constant, H1 x does not propagate colors, only H2 (x,y) do.
  The type of signature tables is color * color -m> color,
  which maps (color(x),color(y)) to color(H2 (x,y)).
*/
typedef ExprPtr sigTablePtr;
 /* updatable map */
typedef ExprPtr congPtr;
 /* couple (uFindPtr,sigTablePtr) */
#define CONGUTREE(cong) CAR(cong)
#define CONGSIGS(cong) CDR(cong)
typedef ExprPtr (*newCongInfoFunc) P((colorPtr x,char *data));
IMPORT congPtr newCong P((newCongInfoFunc newinfo,char *data));

/* the info field for congruence closure is:
  type (''a,'other) conginfo = (int * ''a heap) * 'ext
   where (int * ''a heap) is an COUNTED reference, and 'ext is another data.
   The couple is an updatable TUPLE_2.
*/
IMPORT colorPtr congColorize P((INOUT congPtr cong,
				IN CONST ExprPtr h,
				IN newCongInfoFunc newinfo,IN char *newinfodata,
				IN infoMergeFunc mergeinfo,IN char *mergeinfodata));
	/* returns the color of h, creating it if necessary */
IMPORT void congEqualize P((INOUT congPtr cong,
			    IN CONST MapPtr eqns,
			    IN newCongInfoFunc newinfo,IN char *newinfodata,
			    IN infoMergeFunc mergeinfo,IN char *mergeinfodata));
	/* merges colors (and propagates equalities) from equalities in eqns */
